/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper Server
 
    Comments: Registration of endpoints at the gatekeeper

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/GkRegistration.h          $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 6.01.06 11:14                                                $
   $Modtime:: 6.01.06 10:09                                                $
  $Revision:: 8                                                            $
  $Workfile:: GkRegistration.h                                             $
$NoKeywords: $
----------------------------------------------------------------------------*/

#ifndef _H323GkServerRegistration_h_
#define _H323GkServerRegistration_h_

#include "transports.h"
#include "h323pdu.h"

// forwards
class H323GkServer;

typedef std::list<H323GloballyUniqueID> H323CallIdentifierList;

/**Class for global stuff.
 */
class H323GkSrv
{
public:
    enum Err
    {
        /// Successful completion
        Success,
        /// Something failed
        Failure,
        /// Function is not yet implemented
        NotImpl,
        /// An argument given to the function is invalid
        InvalidArg,
        /// The specified user could be unambiguously determined
        UserAmbiguous,
        /// User access denied
        AccessDenied,
        /// User already registered
        UserAlreadyRegistered,
        /// User account disabled
        AccountDisabled,
        /// No license available
        LicenseLimitReached,
        MaxErr,
    };
};
 
/**Class to store aliases.
 */
class H323GkAliases : public Tools::SObject
{
  public:
    H323GkAliases(const Tools::SStringVector & rE164Aliases,
                  const Tools::SStringVector & rH323Aliases);

    virtual void PrintOn(std::ostream & rStream) const;

    // Vectors of aliases received from client (in RRQ)
    Tools::SStringVector        m_E164Aliases; 
    Tools::SStringVector        m_H323Aliases;
    
    // Aliases successfully registered at server (in RCF)
    Tools::SStringVector        m_RegisteredE164Aliases;
    Tools::SString              m_RegisteredH323Alias;
    Tools::SStringVector        m_RegisteredMixedAliases;
};

class H323RegisteredEPList;
/**Information about the remote registered terminal.
 */
#define NoOfSupportedCodecs 7

class H323RemoteEndpoint
 : public Tools::SObject,
   public Tools::SRefCnt
{
public:
    // Registration timer states
    enum SRegTimerStates { 
        RegistrationTimerStopped, 
        RegistrationTimerStarted, 
        RegistrationTimerExpired
    };

    // enum to build a bit field to determine codecs supported by H.323 endpoint
    enum SupportedCodecs
    {
        CodecNone  = 0x00,
        CodecG711a = 0x01,
        CodecG711u = 0x02,
        CodecSpeex = 0x04,
        CodecG723  = 0x08,
        CodecG729  = 0x10,
        CodecGSM   = 0x20,
        CodecT38   = 0x40,
        CodecAll   = 0x4f
    };
	
	// On a timeout of a timer, a new thread from the treadpool will
    // be started and execute OnTimeoutThread which calls STimer::OnTimeout().
    // This is done to prevent other timeouts being blocked by long running
    // OnTimeout functions.
    SDECLARE_NOTIFIER( Tools::SThreadPool, H323RemoteEndpoint, OnRegistrationExpired );

    H323RemoteEndpoint(H323GkServer &rGKServer,
                       const Tools::SIpAddress & rIpAddress,
                       const H323TransportAddress & rCallSignalAddress,
                       const H323TransportAddress & rRasAddress,
                       const Tools::SStringVector & rE164Aliases,
                       const Tools::SStringVector & rH323Aliases,
                       Tools::STimeInterval TimeToLive,
                       const H323EndPoint::TerminalTypes & rTerminalType,
                       const Tools::SWString &rszVersion,
                       int   iSupportedCodecs = CodecNone,
                       bool  bRequiresDelayedDisconnect = false,
                       int   iLineId = 0,
					   int   iNumOfChannels = 0
);
    
    ~H323RemoteEndpoint();

    virtual void PrintOn(std::ostream & rStream) const;

    bool FindMixedAlias(const Tools::SString & rAlias);

    H323GkServer              & m_rGKServer;

    Tools::SIpAddress           m_IpAddress;

    H323TransportAddress        m_CallSignalAddress;
    H323TransportAddress        m_RasAddress;

    H323GkAliases               m_H323GkAliases;

    H323EndPoint::TerminalTypes m_TerminalType;
    Tools::SWString             m_szVersion;
    bool                        m_bRequiresDelayedDisconnect; 
    int                         m_iSupportedCodecs;
    int                         m_iLineId; 
    int                         m_iNumOfChannels; 

    Tools::SString              m_EndpointIdentifier;
    H323CallIdentifierList      m_H323CallIdentifier;

    H323RasPDU                  m_RasPdu;
    unsigned                    m_uSequenceNumber;
    unsigned                    m_uRequestRetryNumber;
    unsigned                    m_uTimeToLiveInSeconds;

    Tools::STimeInterval        m_TimeToLive;
    SRegTimerStates             m_RegistrationStatus;
    SSemaphore                  m_ProcessingSync;           // blocks as long as timer's timeout function is executed but not returned
    bool                        m_bToBeDeleted;             // true if timer deletes itself with 'delete this' after the next timeout

    Tools::STime                m_TimeOfRegistration;             // absolute time of first registration
    Tools::STime                m_TimeOfRegistrationExpiration;   // absolute time when registration will expire

	Tools::SObject            * m_pUserData;
};

typedef stdext::hash_map<Tools::SString, H323RemoteEndpoint*> H323RemoteEndpointHashmap;
typedef std::vector<H323RemoteEndpoint*>                      H323RemoteEndpointVector;

// Defines order of registration timeout in the list of registered remote endpoints.
// Note: This mechanism is the same as in the SwyxLib timer list implementation!!!
struct lessthanH323RemoteEndpoint
{
    // Remote endpoints are ordered by their registration timeout time first and by their pointer value then.
    // The pointer value is involved for unique order.
    bool operator()(const H323RemoteEndpoint* ep1, const H323RemoteEndpoint* ep2) const;

};
typedef std::set<H323RemoteEndpoint*,lessthanH323RemoteEndpoint> H323RemoteEndpointRegTimerList;
typedef std::vector<H323RemoteEndpoint*> H323RemoteEndpointVector;


/**Class to store all registered H.323-endpoints. All entries
   are hashed to ensure fast retrival of the data.
   The key for hashing is the endpointIdentifier of the registered terminal.
 */
class H323RegisteredEPList : public Tools::SObject
{
public:

    H323RegisteredEPList(H323GkServer & rGatekeeper,
                     bool AutoDelete = false);
    virtual ~H323RegisteredEPList();

    virtual void PrintOn(std::ostream & rStream) const;
    
    H323GkSrv::Err Insert(const Tools::SIpAddress & rIpAddress,
                          H323RemoteEndpoint      * pEndpoint);
    
    bool Erase(const Tools::SString & rEndpointIdentifier);

    H323RemoteEndpoint * FindEndpointByTsap(const Tools::SIpAddress & rIpAddress);
    H323RemoteEndpoint * FindEndpointByAlias(const Tools::SString    & rAlias,
                                             const Tools::SIpAddress & rIpAddress,
                                             bool bMatchIpAndPort = false,
                                             bool bSkipExpiredEPs = false);
    H323RemoteEndpointVector FindEndpointsByAlias(const Tools::SString    & rAlias);
    H323RemoteEndpoint * FindEndpointByAliasAndLineId(const Tools::SString & rAlias,
                                                      int iLineId,
                                                      const Tools::SIpAddress & rIpAddress,
                                                      bool bMatchIpAndPort = false);
    H323RemoteEndpoint * FindEndpointByAliases(const Tools::SStringVector & rAliases,
                                               const Tools::SIpAddress    & rIpAddress,
                                               bool bMatchIpAndPort = false,
                                               bool bSkipExpiredEPs = false);
    H323RemoteEndpoint * FindEndpointByEndpointId(const Tools::SString & rEndpointIdentifier);
    WORD                 FindPortByCallSignalingAddr(const Tools::SIpAddress & rIpAddress);
    WORD                 FindPortByH323Alias(const Tools::SString    & rAlias,
                                             const Tools::SIpAddress & rIpAddress);
    WORD                 FindPortByH323AliasAndLineId(const Tools::SString    & rAlias,
                                                      int iLineId,
                                                      const Tools::SIpAddress & rIpAddress);
    bool FindIsRequiresDelayedDisconnectEPByH323Alias(const Tools::SString    & rH323Alias,
                                                      const Tools::SIpAddress & rIpAddress);
    Tools::SString       FindH323AliasByCallIdentifier(const H323GloballyUniqueID & rCallIdentifier,
                                                       const Tools::SStringVector & rE164Aliases);

    H323RemoteEndpointHashmap & GetEndpoints() { return m_Endpoints; }
    
    void SetAutoDelete(bool d = true) { m_AutoDelete = d; }

    /// Stuff to handle the timer list for registered EPs
    bool AddEPToRegTimerList     (H323RemoteEndpoint * pEndpoint);  // EP to add to in timer list, note: time to live value already set!

    bool UpdateEPRegTimer        (H323RemoteEndpoint * pEndpoint,   // EP to update in timer list
                                  Tools::STimeInterval TimeToLive); // new time to live value for updating

    bool RemoveEPFromRegTimerList(H323RemoteEndpoint * pEndpoint,   // EP to remove from timer list
                                  bool bSignalListChange = true );  // true = process timer list to determine new timeout
    
    SMutex& GetRegTimerListMutex() { return m_EPRegTimerListMutex; };

    /* Decrement all the created registration timers and dispatch to their callback
       functions if they have expired. The #STimer::Tick()# function
       value is used to determine the time elapsed since the last call to
       Process().

       The return value is the number of milliseconds until the next timer
       needs to be despatched. The function need not be called again for this
       amount of time, though it can (and usually is).
       
       @return
       maximum time interval before function should be called again.
    */
    STimeInterval ProcessRegTimerList();

    // Handle registration timeouts
    bool HandleRasRegistrationTimeout(H323RemoteEndpoint * pEndpoint);
    
protected:

    H323GkServer                       & m_rGatekeeper;

    SCriticalSection                     m_EndpointsCritSect;
    H323RemoteEndpointHashmap            m_Endpoints;
    
    // Mutual exclusion and sync. point for multi tasking
    SMutex                               m_EPRegTimerListMutex;
    H323RemoteEndpointRegTimerList       m_EndpointsRegTimerList;
    // timer thread pool; any OnRegistrationExpired-function will get its own thread
    // to prevent blocking in GK registration thread thread caused by long running
    // OnRegistrationExpired-function calls
    SThreadPool                        * m_pRegTimeoutThreadPool;
    // flag to set for auto deletion of endpoints after erasing from container
    bool                                 m_AutoDelete;
};


#endif // _H323GkServerRegistration_h_
