/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Codecs
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/SwyxSrv.root/H323/include/codecs.h                   $
    $Author:: Martin.hueser                                                $
      $Date:: 6.08.04 16:29                                                $
   $Modtime:: 3.08.04 9:45                                                 $
  $Revision:: 18                                                           $
  $Workfile:: codecs.h                                                     $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __CODECS_H
#define __CODECS_H


/* The following classes have forward references to avoid including the VERY
   large header files for H225 and H245. If an application requires access
   to the protocol classes they can include them, but for simple usage their
   inclusion can be avoided.
 */
class H245_MiscellaneousCommand_type;
class H245_MiscellaneousIndication_type;

class H323Connection;
class H323VideoDevice;

class RTP_DataFrame;



/**This class embodies the implementation of a specific codec instance
   used to transfer data via the logical channels opened and managed by the
   H323 control channel.

   An application may create a descendent off this class and override
   functions as required for descibing a codec.
 */
class H323Codec : public Tools::SObject
{
  public:
    enum Direction {
      Encoder,
      Decoder
    };

    H323Codec(
      Direction direction   /// Direction in which this instance runs
    );

    enum CodecType
    {
        Undefined,
        G711a,
        G711u,
        G723,
        G729,
        G722,
        GSM_FullRate,
        T38,
        MaxCodec
    };
    
    /**Open the codec.
       This will open the codec for encoding or decoding, it is called after
       the logical channel have been established and the background threads to
       drive them have been started. This is primarily used to delay allocation
       of resources until the last millisecond.

       A descendent class may be created by the application and it may cast
       the connection parameter to the application defined descendent of 
       H323Connection to obtain information needed to open the codec.

       The default behaviour does nothing.
      */
    virtual bool Open(
      H323Connection & connection /// Connection between the endpoints
    );

    /**Close the codec.
      */
    virtual void Close() = 0;

    /**Encode the data from the appropriate device.
       This will encode a frame of data for transmission. The exact size and
       description of the data placed in the buffer is codec dependent but
       should be less than H323Capability::GetTxPacketSize() in length.

       The length parameter is filled with the actual length of the encoded
       data, often this will be the same as the size parameter.

       This function is called every GetFrameRate() milliseconds, so MUST
       take less than (or equal to) that amount of time to complete!

       Note that a returned length of zero indicates that time has passed but
       there is no data encoded. This is typically used for silence detection
       in an audio codec.
     */
    virtual bool Read(
      BYTE * buffer,    /// Buffer of encoded data
      unsigned & length /// Actual length of encoded data buffer
    ) = 0;

    /**Decode the data and output it to appropriate device.
       This will decode a single frame of received data. The exact size and
       description of the data required in the buffer is codec dependent but
       should be less than H323Capability::GetRxPacketSize() in length.

       It is expected this function anunciates the data. That is, for example
       with audio data, the sound is output on a speaker.

       This function is called every GetFrameRate() milliseconds, so MUST
       take less than that amount of time to complete!
     */
    virtual bool Write(
      const BYTE * buffer,          /// Buffer of encoded data
      unsigned length,              /// Length of encoded data buffer
      const RTP_DataFrame & frame,  /// Entire RTP frame
      unsigned & written            /// Number of bytes used from data buffer
    ) = 0;

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const = 0;

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const = 0;

    /**Get the frame rate in RTP timestamp units.
      */
    virtual unsigned GetFrameRate() const = 0;

    /**Get the number of RTP timestamp units per millisecond.
      */
    virtual unsigned GetTimeUnits() const = 0;

    /**Get the bandwidth used by the channel in 100's of bits/sec.
       The default behaviour is pure.
     */
    virtual unsigned GetBandwidth() const = 0;

    /**Indicate whether the codec requires jitter buffering.
      */
    virtual bool NeedsJitterBuffer() const = 0;

    /**Limit bit flow for the logical channel.
       The default behaviour does nothing.
     */
    virtual void OnFlowControl(
      long bitRateRestriction   /// Bit rate limitation
    );

    /**Process a miscellaneous command on the logical channel.
       The default behaviour does nothing.
     */
    virtual void OnMiscellaneousCommand(
      const H245_MiscellaneousCommand_type & type  /// Command to process
    );

    /**Process a miscellaneous indication on the logical channel.
       The default behaviour does nothing.
     */
    virtual void OnMiscellaneousIndication(
      const H245_MiscellaneousIndication_type & type  /// Indication to process
    );

    Direction GetDirection()   const { return direction; }
    BYTE GetRTPPayloadType()   const { return (BYTE)rtpPayloadType; }


  protected:

    static const char * const CodecNames[MaxCodec];
    friend std::ostream & operator<<(std::ostream & o, CodecType type) { return o << CodecNames[type]; }

    Direction direction;
    unsigned  rtpPayloadType;
};


/**This class defines a codec class that will use the standard platform PCM
   output device.

   An application may create a descendent off this class and override
   functions as required for descibing a specific codec.
 */
class H323AudioCodec : public H323Codec
{
  public:
    /** Create a new audio codec.
        This opens the standard PCM audio output device, for input and output
        and allows descendent codec classes to do audio I/O after
        decoding/encoding.
      */
    H323AudioCodec(
      Direction direction,      /// Direction in which this instance runs
      unsigned samplesPerFrame  /// Number of samples in a frame
    );

    ~H323AudioCodec();

    /**Open the codec.
       This will open the codec for encoding or decoding. This is primarily
       used to delay allocation of resources until the last minute.

       The default behaviour calls the H323EndPoint::OpenAudioChannel()
       function and assigns the result of that function to the raw data
       channel in the H323Codec class.
      */
    virtual bool Open(
      H323Connection & connection /// Connection between the endpoints
    );

    /**Close down the codec.
       This will close the codec breaking any block on the Read() or Write()
       functions.

       The default behaviour will close the rawDataChannel if it is not NULL
       and then delete it if deleteChannel is true.
      */
    virtual void Close();

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const = 0;

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const = 0;

    /**Get the frame rate in milliseconds.
      */
    virtual unsigned GetFrameRate() const;

    /**Get the number of RTP timestamp units per millisecond.
      */
    virtual unsigned GetTimeUnits() const;

    /**Indicate whether the codec requires jitter buffering.
       Default behaviour returns true, audio codecs require jitter buffering.
      */
    bool NeedsJitterBuffer() const;

    /**Attach the raw data channel for use by codec.
       Note the channel provided will be deleted on destruction of the codec.
      */
    bool AttachChannel(
      Tools::SChannel * channel,  /// Channel to read/write raw codec data
      bool autoDelete = true      /// Channel is to be automatically deleted
    );


    /**Enable/Disable silence detection.
      */
    void EnableSilenceDetection(
      bool enable = true   /// Flag for enable/disable
    );

    /**Disable silence detection.
      */
    void DisableSilenceDetection() { EnableSilenceDetection(true); }

    /**Check energy in frame for a talk burst.
       This does the deadband calculations on the provided energy value based
       on the energyThreshold, signalDeadbandFrames and silenceDeadbandFrames
       member variables.
      */
    bool DetectSilence();

    /**Calculate the energy in the audio stream.
       This is called from within DetectSilence() to calculate the average
       energy since the last call to DetectSilence().
      */
    virtual unsigned CalculateEnergy();


  protected:
    unsigned          samplesPerFrame;
    Tools::SChannel * rawDataChannel;
    bool              deleteChannel;

    unsigned energyThreshold;       // In 16 bit PCM values (0=disabled)
    unsigned signalDeadbandFrames;  // Frames of signal before talk burst starts
    unsigned silenceDeadbandFrames; // Frames of silence before talk burst ends
    unsigned framesReceived;        // Signal/Silence frames so far
    bool     inTalkBurst;           // Currently sending RTP data
};


/**This class defines a codec class that will use the standard platform PCM
   output device, and the encoding/decoding has fixed blocks. That is each
   input block of n samples is encoded to exactly the same sized compressed
   data, eg G.711, GSM etc.

   An application may create a descendent off this class and override
   functions as required for descibing a specific codec.
 */
class H323FramedAudioCodec : public H323AudioCodec
{
  public:
    /** Create a new audio codec.
        This opens the standard PCM audio output device, for input and output
        and allows descendent codec classes to do audio I/O after
        decoding/encoding.
      */
    H323FramedAudioCodec(
      Direction direction,      /// Direction in which this instance runs
      unsigned samplesPerFrame, /// Number of samples in a frame
      unsigned bytesPerFrame    /// Number of bytes in a frame
    );

    /**Encode the data from the appropriate device.
       This will encode a frame of data for transmission. The exact size and
       description of the data placed in the buffer is codec dependent but
       should be less than H323Capability::GetTxPacketSize() in length.

       The length parameter is filled with the actual length of the encoded
       data, often this will be the same as the size parameter.

       This function is called every GetFrameRate() milliseconds, so MUST
       take less than (or equal to) that amount of time to complete!

       Note that a returned length of zero indicates that time has passed but
       there is no data encoded. This is typically used for silence detection
       in an audio codec.
     */
    virtual bool Read(
      BYTE * buffer,    /// Buffer of encoded data
      unsigned & length /// Actual length of encoded data buffer
    );

    /**Decode the data and output it to appropriate device.
       This will decode a single frame of received data. The exact size and
       description of the data required in the buffer is codec dependent but
       should be less than H323Capability::GetRxPacketSize() in length.

       It is expected this function anunciates the data. That is, for example
       with audio data, the sound is output on a speaker.

       This function is called every GetFrameRate() milliseconds, so MUST
       take less than that amount of time to complete!
     */
    virtual bool Write(
      const BYTE * buffer,            /// Buffer of encoded data
      unsigned length,                /// Length of encoded data buffer
      const RTP_DataFrame & rtpFrame, /// RTP data frame
      unsigned & written              /// Number of bytes used from data buffer
    );

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const = 0;

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const = 0;

    /**Get the bandwidth used by the channel in 100's of bits/sec.

       The default behaviour calculates the bandwidth used from the
       bytesPerFrame and frame rate parameters.
     */
    virtual unsigned GetBandwidth() const;


    /**Calculate the energy in the audio stream.
       This is called from within DetectSilence() to calculate the average
       energy since the last call to DetectSilence().
      */
    virtual unsigned CalculateEnergy();


    /**Encode a sample block into the buffer specified.
       The samples have been read and are waiting in the readBuffer member
       variable. it is expected this function will encode exactly
       bytesPerFrame bytes.
     */
    virtual bool EncodeFrame(
      BYTE * buffer     /// Buffer into which encoded bytes are placed
    ) = 0;

    /**Decode a sample block from the buffer specified.
       The samples must be placed into the writeBuffer member variable. It is
       expected that exactly samplesPerFrame samples is decoded.
     */
    virtual bool DecodeFrame(
      const BYTE * buffer   /// Buffer from which encoded data is found
    ) = 0;

  protected:
    Tools::SShortVector sampleBuffer;
    unsigned            bytesPerFrame;
    unsigned            framesPerPacket;
};


/**This class defines a codec class that will use the standard platform PCM
   output device, and the encoding/decoding is streamed. That is each input
   16 bit PCM sample is encoded to 8 bits or less of encoded data and no
   blocking of PCM data is required, eg G.711, G.721 etc.

   An application may create a descendent off this class and override
   functions as required for descibing a specific codec.
 */
class H323StreamedAudioCodec : public H323FramedAudioCodec
{
  public:
    /** Create a new audio codec.
        This opens the standard PCM audio output device, for input and output
        and allows descendent codec classes to do audio I/O after
        decoding/encoding.
      */
    H323StreamedAudioCodec(
      Direction direction,      /// Direction in which this instance runs
      unsigned samplesPerFrame, /// Number of samples in a frame
      unsigned bits             /// Bits per sample
    );

    /**Encode a sample block into the buffer specified.
       The samples have been read and are waiting in the readBuffer member
       variable. it is expected this function will encode exactly
       encodedBlockSize bytes.
     */
    virtual bool EncodeFrame(
      BYTE * buffer     /// Buffer into which encoded bytes are placed
    );

    /**Decode a sample block from the buffer specified.
       The samples must be placed into the writeBuffer member variable. It is
       expected that no more than frameSamples is decoded. The return value
       is the number of samples decoded. Zero indicates an error.
     */
    virtual bool DecodeFrame(
      const BYTE * buffer   /// Buffer from which encoded data is found
    );

    /**Encode a single sample value.
     */
    virtual int Encode(short sample) const = 0;

    /**Decode a single sample value.
     */
    virtual short Decode(int sample) const = 0;

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const = 0;

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const = 0;

  protected:
    unsigned bitsPerSample;
};


/**This class defines a codec class that will use the standard platform image
   output device.

   An application may create a descendent off this class and override
   functions as required for descibing a specific codec.
 */
class H323VideoCodec : public H323Codec
{
  public:
    /** Create a new video codec.
        This opens the standard image output device, for input and output
        and allows descendent codec classes to do video I/O after
        decoding/encoding.
      */
    H323VideoCodec(
      Direction direction      /// Direction in which this instance runs
    );

    ~H323VideoCodec();

    /**Open the codec.
       This will open the codec for encoding or decoding. This is primarily
       used to delay allocation of resources until the last minute.

       The default behaviour calls the H323EndPoint::OpenVideoDevice()
       function and assigns the result of that function to the raw data
       channel in the H323Codec class.
      */
    virtual bool Open(
      H323Connection & connection /// Connection between the endpoints
    );

    /**Close down the codec.
       This will close the codec breaking any block on the Read() or Write()
       functions.

       The default behaviour will close the rawDataChannel if it is not NULL
       and then delete it if deleteChannel is true.
      */
    virtual void Close();

    /**Get the number of RTP timestamp units per millisecond.
      */
    virtual unsigned GetTimeUnits() const;

    /**Indicate whether the codec requires jitter buffering.
       Default behaviour returns false, video codecs no NOT require jitter buffering.
      */
    bool NeedsJitterBuffer() const;

    /**Process a miscellaneous command on the logical channel.
       The default behaviour does nothing.
     */
    virtual void OnMiscellaneousCommand(
      const H245_MiscellaneousCommand_type & type  /// Command to process
    );

    /**Process a miscellaneous indication on the logical channel.
       The default behaviour does nothing.
     */
    virtual void OnMiscellaneousIndication(
      const H245_MiscellaneousIndication_type & type  /// Indication to process
    );

    /**Attach the raw data channel for use by codec.
       Note the channel provided will be deleted on destruction of the codec.
      */
    bool AttachDevice(
      H323VideoDevice * device, /// Device to read/write data
      bool autoDelete = true    /// Device is to be automatically deleted
    );

    /**Process a FreezePicture command from remote endpoint.
       The default behaviour does nothing.
     */
    virtual void OnFreezePicture();

    /**Process a FastUpdatePicture command from remote endpoint.
       The default behaviour does nothing.
     */
    virtual void OnFastUpdatePicture();

    /**Process a FastUpdateGOB command from remote endpoint.
       The default behaviour does nothing.
     */
    virtual void OnFastUpdateGOB(unsigned firstGOB, unsigned numberOfGOBs);

    /**Process a FastUpdateMB command from remote endpoint.
       The default behaviour does nothing.
     */
    virtual void OnFastUpdateMB(int firstGOB, int firstMB, unsigned numberOfMBs);

    /**Process a OnVideoIndicateReadyToActivate indication from remote endpoint.
       The default behaviour does nothing.
     */
    virtual void OnVideoIndicateReadyToActivate();

    /**Process a OnVideoTemporalSpatialTradeOff indication from remote endpoint.
       The default behaviour does nothing.
     */
    virtual void OnVideoTemporalSpatialTradeOff();

    /**Process a OnVideoNotDecodedMBs indication from remote endpoint.
       The default behaviour does nothing.
     */
    virtual void OnVideoNotDecodedMBs(unsigned firstMB, unsigned numberOfMBs, unsigned temporalReference);

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const = 0;

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const = 0;

  protected:
    H323VideoDevice * rawDevice;
    bool              deleteDevice;
};

//----------------------------------------------------------------------------------
// The simplest codec is the G.711 PCM codec.

/**This class is a G711 ALaw codec.
 */
class H323_ALawCodec : public H323StreamedAudioCodec
{
  public:
  /**@name Construction */
  //@{
    /**Create a new G.711 codec for ALaw.
     */
    H323_ALawCodec(
      Direction direction,  /// Direction in which this instance runs
      bool at56kbps,        /// Encoding bit rate.
      unsigned frameSize    /// Size of frame in bytes
    );
  //@}

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const { return G711a; }

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const { return (GetFrameRate()/GetTimeUnits()); }

    virtual int   Encode(short sample) const;
    virtual short Decode(int   sample) const;

  protected:
    bool sevenBit;
};


/**This class is a G711 uLaw codec.
 */
class H323_muLawCodec : public H323StreamedAudioCodec
{
  public:
  /**@name Construction */
  //@{
    /**Create a new G.711 codec for muLaw.
     */
    H323_muLawCodec(
      Direction direction,  /// Direction in which this instance runs
      bool at56kbps,        /// Encoding bit rate.
      unsigned frameSize    /// Size of frame in bytes
    );
  //@}

    virtual int   Encode(short sample) const;
    virtual short Decode(int   sample) const;

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const { return G711u; }

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const { return (GetFrameRate()/GetTimeUnits()); }

  protected:
    bool sevenBit;
};


//-----------------------------------------------------------------------------
// G.723.1 Codec (dummy!!!)
//-----------------------------------------------------------------------------

class H323_G7231Codec : public H323FramedAudioCodec
{
  public:
  /**@name Construction */
  //@{
    /**Create a new G.723 dummy codec.
     */
    H323_G7231Codec(
      Direction direction  /// Direction in which this instance runs
    );
  //@}

    /**Encode a sample block into the buffer specified.
       The samples have been read and are waiting in the readBuffer member
       variable. it is expected this function will encode exactly
       bytesPerFrame bytes.
     */
    virtual bool EncodeFrame(
      BYTE * buffer     /// Buffer into which encoded bytes are placed
    );

    /**Decode a sample block from the buffer specified.
       The samples must be placed into the writeBuffer member variable. It is
       expected that exactly samplesPerFrame samples is decoded.
     */
    virtual bool DecodeFrame(
      const BYTE * buffer   /// Buffer from which encoded data is found
    );

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const { return G723; }

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const { return 30; }
};  


//-----------------------------------------------------------------------------
// G.729 Codec (dummy!!!)
//-----------------------------------------------------------------------------

class H323_G729Codec : public H323FramedAudioCodec
{
  public:
  /**@name Construction */
  //@{
    /**Create a new G.729 dummy codec.
     */
    H323_G729Codec(
      Direction direction  /// Direction in which this instance runs
    );
  //@}

    /**Encode a sample block into the buffer specified.
       The samples have been read and are waiting in the readBuffer member
       variable. it is expected this function will encode exactly
       bytesPerFrame bytes.
     */
    virtual bool EncodeFrame(
      BYTE * buffer     /// Buffer into which encoded bytes are placed
    );

    /**Decode a sample block from the buffer specified.
       The samples must be placed into the writeBuffer member variable. It is
       expected that exactly samplesPerFrame samples is decoded.
     */
    virtual bool DecodeFrame(
      const BYTE * buffer   /// Buffer from which encoded data is found
    );

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const { return G729; }

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const { return 20; }
};  

//-----------------------------------------------------------------------------
// G.722 Codec (dummy!!!)
//-----------------------------------------------------------------------------

class H323_G722Codec : public H323FramedAudioCodec
{
  public:
  /**@name Construction */
  //@{
    /**Create a new G.722 dummy codec.
     */
    H323_G722Codec(
      Direction direction  /// Direction in which this instance runs
    );
  //@}

    /**Encode a sample block into the buffer specified.
       The samples have been read and are waiting in the readBuffer member
       variable. it is expected this function will encode exactly
       bytesPerFrame bytes.
     */
    virtual bool EncodeFrame(
      BYTE * buffer     /// Buffer into which encoded bytes are placed
    );

    /**Decode a sample block from the buffer specified.
       The samples must be placed into the writeBuffer member variable. It is
       expected that exactly samplesPerFrame samples is decoded.
     */
    virtual bool DecodeFrame(
      const BYTE * buffer   /// Buffer from which encoded data is found
    );

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const { return G722; }

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const { return 20; }
};  


#endif // __CODECS_H
