/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper client
 
    Comments: Gatekeeper client protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxGate.root/H323/include/gkclient.h               $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 10.08.05 16:42                                               $
   $Modtime:: 10.08.05 15:49                                               $
  $Revision:: 3                                                            $
  $Workfile:: gkclient.h                                                   $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __GKCLIENT_H
#define __GKCLIENT_H


#include "SwyxExtHelper.h"
#include "transports.h"


class H225_GatekeeperRequest;
class H225_GatekeeperConfirm;
class H225_GatekeeperReject;
class H225_RegistrationRequest;
class H225_RegistrationConfirm;
class H225_RegistrationReject;
class H225_UnregistrationRequest;
class H225_UnregistrationConfirm;
class H225_UnregistrationReject;
class H225_AdmissionRequest;
class H225_AdmissionConfirm;
class H225_AdmissionReject;
class H225_BandwidthRequest;
class H225_BandwidthConfirm;
class H225_BandwidthReject;
class H225_DisengageRequest;
class H225_DisengageConfirm;
class H225_DisengageReject;
class H225_LocationRequest;
class H225_LocationConfirm;
class H225_LocationReject;
class H225_InfoRequest;
class H225_InfoRequestResponse;
class H225_NonStandardMessage;
class H225_UnknownMessageResponse;
class H225_RequestInProgress;
class H225_ResourcesAvailableIndicate;
class H225_ResourcesAvailableConfirm;
class H225_InfoRequestAck;
class H225_InfoRequestNak;


class H323Gatekeeper : public Tools::SObject, public Tools::SRefCnt
{
  public:

  /**@name Construction */
  //@{
    /**Create a new gatekeeper.
     */
    H323Gatekeeper(
      H323EndPoint & endpoint,  /// Endpoint gatekeeper is associated with.
      const Tools::SIpAddress & transportAddress /// Transport address over which gatekeepers communicates.
    );

    /**Destroy gatekeeper.
     */
    ~H323Gatekeeper();

    /**Background thread handler.
      */
    virtual void HandleRasChannel();

    /**Handle and dispatch a RAS PDU..
      */
    virtual bool HandleRasPDU(
      const H323RasPDU & response
    );

    /**Unregister all clients.

       @param reason                   Reason for unregistration (-1 = no reason)
      */
    bool UnregisterAllClients(int reason = -1);

    /**Shutdown gatekeeper.
     */
    void Shutdown();
//@}

  /**@name Overrides from PObject */
  //@{
    void PrintOn(std::ostream & strm) const;
  //@}

  /**@name Protocol operations */
  //@{
    /**Discover a gatekeeper on the local network.
       If the address string is empty then it is assumed that the transport
       is already connected to the gatekeeper via the Discover() function.
     */
    bool Discover(
      const H323TransportAddress & address /// Address of gatekeeper.
    );

    /**Register with gatekeeper.

       @param szH323Alias              H323 alias of client performing this RRQ
       @param bIsLightWeightRRQ        True to perform light weight RRQ for keep alive

     */
    bool RegistrationRequest(const Tools::SString & szH323Alias,
                             bool                   bIsLightWeightRRQ = false,
                             int                    iLineId = 0,                // used by SwyxGate
                             int                    iNumOfChannels = 0,         // used by SwyxFax >= v2.x
                             bool                   bCleanupCaseOfFailure = false);

    /**Unregister with gatekeeper.

       @param szH323Alias              H323 alias of client performing this RRQ
       @param reason                   Reason for unregistration
     */
    bool UnregistrationRequest(const Tools::SString & szH323Alias,
                               int                    iLineId = 0,                // used by SwyxGate
                               int                    reason = -1);

    /**Location request to gatekeeper.
     */
    bool LocationRequest(
      const Tools::SString & alias,          /// Alias name we wish to find.
      H323TransportAddress & address  /// Resultant transport address.
    );

    /**Location request to gatekeeper.
     */
    bool LocationRequest(
      const Tools::SStringList & aliases,    /// Alias names we wish to find.
      H323TransportAddress & address  /// Resultant transport address.
    );

    /**Admission request to gatekeeper.
     */
    bool AdmissionRequest(
      H323Connection & connection  /// Connection we wish to change.
    );


    /**Admission request to gatekeeper.
     */
    bool AdmissionRequest(
      H323Connection & connection,    /// Connection we wish to change.
      H323TransportAddress & address  /// Gatekeeper routed transport address.
    );

    /**Disengage request to gatekeeper.
     */
    bool DisengageRequest(
      const H323Connection & connection,  /// Connection we wish admitted.
      unsigned reason                     /// Reason code for disengage
    );

    /**Bandwidth request to gatekeeper.
     */
    bool BandwidthRequest(
      H323Connection & connection,    /// Connection we wish to change.
      unsigned requestedBandwidth     /// New bandwidth wanted in 0.1kbps
    );

    /**Send an info response to the gatekeeper.
     */
    bool InfoRequestResponse(
      const H323Connection * connection,  /// Connection to send info about
      unsigned seqNum = 1
    );
  //@}

  /**@name Gatekeeper handling */
  //@{    
    /**Try to reregister client at gatekeeper again.

       @param szH323Alias               H323 alias of client performing this reregister
       @param bIsLightWeightRRQ         True to perform light weight RRQ for keep alive
       @param iLineId                   used by SwyxGate to indicate while line to register
       @param bSetStatusToUnregistered  True will set the registration status to 'unregistered'
                                        (this will force the gateway to reject incoming calls during
                                         reregistration attempt to server!)
     */
    bool RegisterClientAgain(const Tools::SString & szH323Alias,
                             bool                   bIsLightWeightRRQ = true,
                             int                    iLineId = 0,
                             bool                   bSetStatusToUnregistered = false );

    /**Restart RAS thread and try registering at gatekeeper again.
     */
    void StartGkRegisterThread();

    /**Check if all clients are registered.

       @return     If true is returned all clients are registered.
     */
    bool AreAllClientsRegistered();
  //@}

  /**@name Protocol callbacks */
  //@{
#define DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(name) \
    virtual void OnSend##name(H225_##name &); \
    virtual bool OnReceive##name(const H225_##name &);

    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestResponse);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(NonStandardMessage);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnknownMessageResponse);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RequestInProgress);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(ResourcesAvailableIndicate);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(ResourcesAvailableConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestAck);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestNak);

#undef DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS

    /**Handle unknown PDU type.
      */
    virtual bool OnReceiveUnkown(
      const H323RasPDU & pdu  /// PDU that was not handled.
    );
  //@}

    /** Add the H225 RRQ nonStandardData part to a RRQ signal PDU.

        @param rrq            H.225 RRQ signal PDU
        @param strm           Non standard data stream to attach to RRQ
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool AttachRRQNonStandardData(const Tools::SByteVector & data,
                                  H225_RegistrationRequest & rrq);

    /** Add the H225 URQ nonStandardData part to a URQ signal PDU.

        @param urq            H.225 RRQ signal PDU
        @param strm           Non standard data stream to attach to URQ
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool AttachURQNonStandardData(const Tools::SByteVector   & data,
                                  H225_UnregistrationRequest & urq);

    /** Get the H225 RRQ nonStandardData part out of a RCF signal PDU.

        @param rcf            H.225 RCF signal PDU
        @param data           Non standard data stream from RCF
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool GetRCFNonStandardData(const H225_RegistrationConfirm & rcf,
                               Tools::SByteVector             & data);

    /** Get the H225 RRQ nonStandardData part out of a RRJ signal PDU.

        @param rrj            H.225 RRJ signal PDU
        @param data           Non standard data stream from RRJ
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool GetRRJNonStandardData(const H225_RegistrationReject & rrj,
                               Tools::SByteVector            & data);

    /** Get the H225 URQ nonStandardData part out of a URQ signal PDU.

        @param urq            H.225 URQ signal PDU
        @param data           Non standard data stream from URQ
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool GetURQNonStandardData(const H225_UnregistrationRequest & urq,
                               Tools::SByteVector               & data);


    const H323TransportAddress & GetLastGKAddress() const { return m_LastGKAddress;       }
    void  SetH323Transport( H323Transport * newTransport) { transport = newTransport;     }
    const Tools::SString & GetGatekeeperName() const      { return m_szGatekeeperName;    }
    const Tools::SIpAddress & GetTransportAddress() const { return m_TransportAddress;    }
    void  SetRASChannelInvalid()                          { m_bRASChannelInvalid = true;  }
    void  SetRASChannelValid()                            { m_bRASChannelInvalid = false; }
    bool  IsRASChannelInvalid()                     const { return m_bRASChannelInvalid;  }
    void  SetAlternativeGatekeeper()                      { m_bIsAlternativeGK = true;    }
    bool  IsAlternativeGatekeeper()                 const { return m_bIsAlternativeGK;    }

    enum GKRegStatus
    { gk_RegistrationStatusUnknown,
      gk_RegistrationStatusUnregistered,
      gk_RegistrationStatusOK,
      gk_RegistrationStatusInKeepAlive,
      gk_RegistrationStatusNotPermitted,
      gk_RegistrationStatusRegistrationLost,
    };

    /** Class that contains client (H323 endpoint) specific data.
     */
    class Client : public Tools::SObject, public Tools::SRefCnt
    {
    public:

	  enum ClientType
	  {
		  Unknown = 0,
		  Gateway,
		  Link,
		  FaxServer,
	  };

	  typedef struct
	  {
          H323Transport  * transport;
          Tools::SString   szH323Alias;
		  ClientType       clientType;
		  union
		  {
			  struct
			  {
				  unsigned lineId;
			  } GatewayData;
			  struct
			  {
				  unsigned numOfChannels;
			  }  FaxServerData;
		  } Data;
	  }  ClientDataStruct;

      /**@name Construction */
      //@{
        /**Create a client specific object.
         */
        Client( H323Gatekeeper    * gatekeeper,
                H323EndPoint      & endpoint,
			    ClientDataStruct  & rClientData );

        /**Destroy a client specific object.
         */
        ~Client();
      //@}

      /**@name Member variable access */
      //@{
        /**Determine if the endpoint is registered with the gatekeeper.
          */
        bool IsRegistered() const { return m_GKRegStatus == gk_RegistrationStatusOK || m_GKRegStatus == gk_RegistrationStatusInKeepAlive; }
        bool IsRegisteredToMainGK() const { return IsRegistered() && m_bRegisteredToMainGK; }
        bool IsRegisteredToAlternativGK() const { return IsRegistered() && !m_bRegisteredToMainGK; }
      //@}


        // Declare timer for the timeToLive parameter in the RCF msg.
        // This will implement a keep alive mechanism
        SDECLARE_NOTIFIER( Tools::STimer, Client, OnRegistrationExpiry );

        H323Gatekeeper  * gatekeeper;
        H323EndPoint    & endpoint;
        H323Transport   * transport;

        ClientDataStruct  m_ClientData;

        Tools::SString    m_szEndpointIdentifier;

        Tools::STimer     m_RegTimer;
        GKRegStatus       m_GKRegStatus;
        bool              m_bRegisteredToMainGK;

        unsigned          m_uLastReqSequenceNumber;
        unsigned          m_uNumberOfRRQAttempts;
    };

    Client * FindClientByAlias(const Tools::SString & rAlias);
    Client * FindClientByAliasAndLineId(const Tools::SString & rAlias, int iLineId = 0);
    Client * FindClientBySeqNum(unsigned seqNum);
    Client * FindClientByEndpointId(const Tools::SString & rEndpointId);

protected:

    /** The object ClientsMap contains the data of all clients (H.323 endpoints)
        which are used in one H323Gatekeeper object. This is necessary to mangage
        multiple clients (H.323 endpoints) in one gatekeeper object, e.g. on a
        gateway you can have one client/H.323 endpoint for each controller.
        All entries are in a map. The key for the client is the H323 alias
        of the registered terminal.
     */
    typedef std::map<Tools::SString, Client *>  ClientsMap;

    /** GK register thread.
     */
    class GKRegisterThread : public Tools::SThread
    {
      public:
        GKRegisterThread( H323Gatekeeper & gatekeeper,
                          H323EndPoint   & endpoint,
                          H323Transport  * transport,
                          const Tools::SStringVector & H323Aliases);

        ~GKRegisterThread();

      protected:
        void Main();

        H323Gatekeeper       & gatekeeper;
        H323EndPoint         & endpoint;
        H323Transport        * transport;
        bool                   m_bStopReRegistration;
        Tools::SSyncPoint      m_ReRegistrationSyncPoint;
        Tools::SStringVector   m_H323Aliases;
    };

    class RASReqInProgress : public Tools::SObject
    {
    public :
        
        enum RASRequestResult {
          RequestInProgress,
          ConfirmReceived,
          RejectReceived,
          RequestTimeout
        } ;

        RASReqInProgress::RASReqInProgress(const H323RasPDU & rPendingReq)
            : m_rPendingReq(rPendingReq), m_RASReqResult(RequestInProgress), m_NumberOfReqRetry(0) {}
        RASReqInProgress::~RASReqInProgress() {};

        Tools::SSyncPoint    m_PendingReqSync;
        const H323RasPDU   & m_rPendingReq;
        RASRequestResult     m_RASReqResult;
        Tools::STimeInterval m_ReqTimeout;
        unsigned             m_NumberOfReqRetry;
        unsigned             m_MaxNumberOfReqRetries;
        Tools::STimeInterval m_LastReqAttempt;
    };
    typedef std::map< unsigned, RASReqInProgress *>  RASReqInProgressMap;

    // This will check for RAS rquest timeouts
    SDECLARE_NOTIFIER( Tools::STimer, H323Gatekeeper, OnRASReqExpiryCheck );
    Tools::STimer   m_RASReqExpiryCheckTimer;

    unsigned GetNextSequenceNumber();
    
    RASReqInProgress * MakeRequest(const H323RasPDU & request, const unsigned uSeqNumber);
    void CleanupReqInProgress(RASReqInProgress * pRASReqInProgress);

    RASReqInProgress * CheckForReqInProgress(unsigned, unsigned);
    bool CheckForResponse(unsigned, unsigned);
    void CheckForRequestTimeout();

    H323EndPoint         & endpoint;
    H323Transport        * transport;

    // Master RAS req timeout value
    Tools::STimeInterval   m_MaxMakeReqestTimeout;
    // Map to handle all RAS request which are ongoing 
    RASReqInProgressMap    m_RASReqInProgressMap;
    // Main mutex to protect H323Gatekeeper members
    Tools::SMutex          m_GkMutex;
    // master member to determine next RAS message sequence number to use
    unsigned               m_uNextRASReqSequenceNumber;

    // helper variables for GRQ/GCF handling
    unsigned               m_uLastRASReqSequenceNumber;
    unsigned               m_uLastRASReqTag;


    bool                   m_bGatekeeperDiscoveryComplete;
    bool                   m_bShutdownGatekeeper;
    bool                   m_bRASChannelInvalid;
    bool                   m_bIsAlternativeGK;
    
    Tools::SString         m_szGatekeeperName;
    Tools::SString         m_szGatekeeperIdentifier;
    Tools::SString         m_szObjectIdentifier;

    GKRegisterThread     * m_pGKRegisterThread;
    H323TransportAddress   m_LastGKAddress;

    unsigned               m_uAllocatedBandwidth;
    H323TransportAddress   m_LocatedAddress;

    Tools::SIpAddress      m_TransportAddress;
    Tools::SIpAddress      m_LastReceiveIpAddress;
    int                    m_LastTransportError;

    ClientsMap             m_Clients;
};


#endif // __GKCLIENT_H
