/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper functionality for links between server
 
    Comments: Gatekeeper/RAS protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/LinkMgr.root/H323/include/gklink.h                   $
    $Author:: Ulrich.findeisen                                             $
      $Date:: 9.08.05 21:25                                                $
   $Modtime:: 9.08.05 18:46                                                $
  $Revision:: 2                                                            $
  $Workfile:: gklink.h                                                     $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __GKLINK_H
#define __GKLINK_H

#include "transports.h"


class H225_GatekeeperRequest;
class H225_GatekeeperConfirm;
class H225_GatekeeperReject;
class H225_RegistrationRequest;
class H225_RegistrationConfirm;
class H225_RegistrationReject;
class H225_UnregistrationRequest;
class H225_UnregistrationConfirm;
class H225_UnregistrationReject;
class H225_AdmissionRequest;
class H225_AdmissionConfirm;
class H225_AdmissionReject;
class H225_BandwidthRequest;
class H225_BandwidthConfirm;
class H225_BandwidthReject;
class H225_DisengageRequest;
class H225_DisengageConfirm;
class H225_DisengageReject;
class H225_LocationRequest;
class H225_LocationConfirm;
class H225_LocationReject;
class H225_InfoRequest;
class H225_InfoRequestResponse;
class H225_NonStandardMessage;
class H225_UnknownMessageResponse;
class H225_RequestInProgress;
class H225_ResourcesAvailableIndicate;
class H225_ResourcesAvailableConfirm;
class H225_InfoRequestAck;
class H225_InfoRequestNak;



class H323GkLink : public Tools::SObject
{
  public:
  /**@name Construction */
  //@{
    /**Create a new gatekeeper link object.
     */
    H323GkLink(
      H323EndPoint & endpoint,  /// Endpoint gatekeeper is associated with.
      const Tools::SIpAddress & transportAddress /// Transport address over which gatekeepers communicates.
    );

    /**Destroy gatekeeper link object.
     */
    ~H323GkLink();

    /**Background thread handler.
      */
    virtual void HandleRasChannel();

    /**Handle and dispatch a RAS PDU..
      */
    virtual bool HandleRasPDU(
      const H323RasPDU & response
    );

    /**Unregister all clients.

       @param reason                   Reason for unregistration (-1 = no reason)
      */
    bool UnregisterAllClients(int reason = -1);
  //@}

  /**@name Overrides from SObject */
  //@{
    void PrintOn(std::ostream & strm) const;
  //@}

  /**@name Protocol operations */
  //@{
    /**Discover a gatekeeper on the local network.
       If the address string is empty then it is assumed that the transport
       is already connected to the gatekeeper via the Discover() function.
     */
    bool Discover(
      const H323TransportAddress & address /// Address of gatekeeper.
    );

    /**Register with gatekeeper.

       @param szH323Alias              H323 alias of client performing this RRQ
       @param bIsLightWeightRRQ        True to perform light weight RRQ for keep alive

     */
    bool RegistrationRequest(const Tools::SString & szH323Alias,
                             bool                   bIsLightWeightRRQ = false);

    /**Unregister with gatekeeper.

       @param szH323Alias              H323 alias of client performing this RRQ
       @param reason                   Reason for unregistration
     */
    bool UnregistrationRequest(const Tools::SString & szH323Alias,
                               int                    reason = -1);

    /**Location request to gatekeeper.
     */
    bool LocationRequest(
      const Tools::SString & alias,          /// Alias name we wish to find.
      H323TransportAddress & address  /// Resultant transport address.
    );

    /**Location request to gatekeeper.
     */
    bool LocationRequest(
      const Tools::SStringList & aliases,    /// Alias names we wish to find.
      H323TransportAddress & address  /// Resultant transport address.
    );

    /**Admission request to gatekeeper.
     */
    bool AdmissionRequest(
      H323Connection & connection  /// Connection we wish to change.
    );


    /**Admission request to gatekeeper.
     */
    bool AdmissionRequest(
      H323Connection & connection,    /// Connection we wish to change.
      H323TransportAddress & address  /// Gatekeeper routed transport address.
    );

    /**Disengage request to gatekeeper.
     */
    bool DisengageRequest(
      const H323Connection & connection,  /// Connection we wish admitted.
      unsigned reason                     /// Reason code for disengage
    );

    /**Bandwidth request to gatekeeper.
     */
    bool BandwidthRequest(
      H323Connection & connection,    /// Connection we wish to change.
      unsigned requestedBandwidth     /// New bandwidth wanted in 0.1kbps
    );

    /**Send an info response to the gatekeeper.
     */
    bool InfoRequestResponse(
      const H323Connection * connection,  /// Connection to send info about
      unsigned seqNum = 1
    );
  //@}

  /**@name Gatekeeper handling */
  //@{    
    /**Try to reregister client at gatekeeper again.

       @param szH323Alias              H323 alias of client performing this reregister
       @param bIsLightWeightRRQ        True to perform light weight RRQ for keep alive
     */
    bool RegisterClientAgain(const Tools::SString & szH323Alias,
                             bool                   bIsLightWeightRRQ);

    /**Method called if a registration failure occurs, eg no response on RRQ.
     */
    virtual bool OnRegisterFailure();

    /**Restart RAS thread and try registering at gatekeeper again.
     */
    void StartGkRegisterThread();

    /**Check if all clients are registered.

       @return     If true is returned all clients are registered.
     */
    bool AreAllClientsRegistered();
  //@}

  /**@name Protocol callbacks */
  //@{
#define DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(name) \
    virtual void OnSend##name(H225_##name &); \
    virtual bool OnReceive##name(const H225_##name &);

    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(GatekeeperReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RegistrationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnregistrationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(AdmissionReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(BandwidthReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(DisengageReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(LocationReject);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequest);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestResponse);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(NonStandardMessage);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(UnknownMessageResponse);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(RequestInProgress);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(ResourcesAvailableIndicate);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(ResourcesAvailableConfirm);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestAck);
    DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS(InfoRequestNak);

#undef DECLARE_GATEKEEPER_PROTOCOL_CALLBACKS

    /**Handle unknown PDU type.
      */
    virtual bool OnReceiveUnkown(
      const H323RasPDU & pdu  /// PDU that was not handled.
    );
  //@}

    /** Get the H225 RRQ nonStandardData part out of a RCF signal PDU.

        @param rcf            H.225 RCF signal PDU
        @param data           Non standard data stream from RCF
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool GetRCFNonStandardData(const H225_RegistrationConfirm & rcf,
                               Tools::SByteVector             & data);


    const H323TransportAddress & GetLastGKAddress() const { return m_LastGKAddress;       }
    void  SetH323Transport( H323Transport * newTransport) { transport = newTransport;     }
    const Tools::SString & GetGatekeeperName() const      { return m_szGatekeeperName;    }
    const Tools::SIpAddress & GetTransportAddress() const { return m_TransportAddress;    }
    void  SetRASChannelInvalid()                          { m_bRASChannelInvalid = true;  }
    void  SetRASChannelValid()                            { m_bRASChannelInvalid = false; }
    bool  IsRASChannelInvalid()                     const { return m_bRASChannelInvalid;  }

    /** Get local interface address.
    */
    Tools::SIPSocket::Address GetLocalInterfaceAddress() const { return m_LocalInterfaceAddress.m_Address; }

  protected:
    /** Class that contains client (H323 endpoint) specific data.
     */
    class Client : public Tools::SObject
    {
      public:
      /**@name Construction */
      //@{
        /**Create a client specific object.
         */
        Client(H323GkLink     & gatekeeper,
               H323EndPoint   & endpoint,
               H323Transport  * transport,
               Tools::SString   m_szH323Alias);

        /**Destroy a client specific object.
         */
        ~Client();
      //@}

      /**@name Member variable access */
      //@{
        /**Determine if the endpoint is registered with the gatekeeper.
          */
        bool IsRegistered() const { return m_bIsRegistered; }
      //@}

        // Declare timer for the timeToLive parameter in the RCF msg.
        // This will implement a keep alive mechanism
        SDECLARE_NOTIFIER( Tools::STimer, Client, OnRegistrationExpiry );

        // Declare timer for the timeToLive RRQ message (not receiving RCF).
        SDECLARE_NOTIFIER( Tools::STimer, Client, OnTTLRegRequestTimeout );

        H323GkLink      & m_rGkLink;
        H323EndPoint    & endpoint;
        H323Transport   * transport;

        bool              m_bIsRegistered;
        bool              m_bLightweightRRQInProgress;
        unsigned          m_uLastReqSequenceNumber;
        unsigned          m_uTTLRegRequestRetries;
        Tools::SString    m_szH323Alias;
        Tools::SString    m_szEndpointIdentifier;

        Tools::STimer     m_RegTimer;
        Tools::STimer     m_TTLRegRequestTimer;
    };

    /** The object ClientsMap contains the data of all clients (H.323 endpoints)
        which are used in one H323Gatekeeper object. This is necessary to mangage
        multiple clients (H.323 endpoints) in one gatekeeper object, e.g. on a
        gateway you can have one client/H.323 endpoint for each controller.
        All entries are in a map. The key for the client is the H323 alias
        of the registered terminal.
     */
    typedef std::map<Tools::SString, Client *>  ClientsMap;
 
    /** GK link register thread.
     */
    class GkLinkRegisterThread : public Tools::SThread
    {
      public:
        GkLinkRegisterThread(H323GkLink                 & rGkLink,
                             H323EndPoint               & rEndpoint,
                             H323Transport              * rTransport,
                             const Tools::SStringVector & rH323Aliases);

        ~GkLinkRegisterThread();

      protected:
        void Main();

        H323GkLink           & m_rGkLink;
        H323EndPoint         & m_rEndpoint;
        H323Transport        * m_pTransport;
        bool                   m_bStopReRegistration;
        Tools::SSyncPoint      m_ReRegistrationSyncPoint;
        Tools::SStringVector   m_H323Aliases;
    };

    class RASReqInProgress : public Tools::SObject
    {
    public :
        
        enum RASRequestResult {
          RequestInProgress,
          ConfirmReceived,
          RejectReceived,
          RequestTimeout
        } ;

        RASReqInProgress::RASReqInProgress(const H323RasPDU & rPendingReq)
            : m_rPendingReq(rPendingReq), m_RASReqResult(RequestInProgress), m_NumberOfReqRetry(0) {}
        RASReqInProgress::~RASReqInProgress() {};

        Tools::SSyncPoint    m_PendingReqSync;
        const H323RasPDU   & m_rPendingReq;
        RASRequestResult     m_RASReqResult;
        Tools::STimeInterval m_ReqTimeout;
        unsigned             m_NumberOfReqRetry;
        unsigned             m_MaxNumberOfReqRetries;
        Tools::STimeInterval m_LastReqAttempt;
    };
    typedef std::map< unsigned, RASReqInProgress *>  RASReqInProgressMap;

    // This will check for RAS rquest timeouts
    SDECLARE_NOTIFIER( Tools::STimer, H323GkLink, OnRASReqExpiryCheck );
    Tools::STimer   m_RASReqExpiryCheckTimer;

    unsigned GetNextSequenceNumber();
    
    RASReqInProgress * MakeRequest(const H323RasPDU & request, const unsigned uSeqNumber);
    void CleanupReqInProgress(RASReqInProgress * pRASReqInProgress);

    RASReqInProgress * CheckForReqInProgress(unsigned, unsigned);
    bool CheckForResponse(unsigned, unsigned);
    void CheckForRequestTimeout();

    Client * FindClientByAlias(const Tools::SString & rAlias);
    Client * FindClientBySeqNum(unsigned seqNum);
    Client * FindClientByEndpointId(const Tools::SString & rEndpointId);

    H323EndPoint         & endpoint;
    H323Transport        * transport;

    // Master RAS req timeout value
    Tools::STimeInterval   m_MaxMakeReqestTimeout;
    // Map to handle all RAS request which are ongoing 
    RASReqInProgressMap    m_RASReqInProgressMap;
    // Main mutex to protect H323Gatekeeper members
    Tools::SMutex          m_GkMutex;
    // master member to determine next RAS message sequence number to use
    unsigned               m_uNextRASReqSequenceNumber;

    // helper variables for GRQ/GCF handling
    unsigned               m_uLastRASReqSequenceNumber;
    unsigned               m_uLastRASReqTag;

    bool                   m_bGatekeeperDiscoveryComplete;
    bool                   m_bRASChannelInvalid;

    Tools::SString         m_szGatekeeperName;
    Tools::SString         m_szGatekeeperIdentifier;
    Tools::SString         m_szObjectIdentifier;

    GkLinkRegisterThread * m_pGKLinkRegisterThread;
    H323TransportAddress   m_LastGKAddress;

    unsigned               m_uAllocatedBandwidth;
    H323TransportAddress   m_LocatedAddress;

    Tools::SIpAddress      m_TransportAddress;
    Tools::SIpAddress      m_LastReceiveIpAddress;
    int                    m_LastTransportError;

    Tools::SIpAddress      m_LocalInterfaceAddress;

    ClientsMap             m_Clients;
};


#endif // __GKLINK_H
