/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper server
 
    Comments: Gatekeeper server protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/gkserver.h                $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 23.12.05 15:26                                               $
   $Modtime:: 23.12.05 15:13                                               $
  $Revision:: 4                                                            $
  $Workfile:: gkserver.h                                                   $
$NoKeywords: $
----------------------------------------------------------------------------*/

#ifndef _GkServer_h_
#define _GkServer_h_

#include "transports.h" // H323Transport(UDP)
#include "h323.h" // H323Connection, H323ConnectionDict

#include "GkRegistration.h"


// forward declarations
class H323GkTimeoutThread;

/**
   This class implements the server functionality of a H.323 gatekeeper.
   There are two ports the server has to listen for RAS messages on.
   On udp port 1718 in multicast network 224.0.1.41 we have to detect
   incoming GRQ messages, this port is called the discovery port.
   On udp port 1719 we have to listen for registration and status
   messages. Therefore it is called the registration and status port of
   the gatekeeper. Version 1 gatekeepers will send GRQ messages to
   this port, too.
*/
class H323GkServer : public Tools::SObject
{
    enum BandwidthAmount
    {
        DefaultEndpointBandwidth = 256
    };

public:

    /** Class that contains RASPdu specific data.
     */
    class RasPDU : public Tools::SObject, public Tools::SRefCnt
    {
    public:


      /**@name Construction */
      //@{
        /**Create a RASPdu specific object.
         */
        //RasPDU();

        /**Destroy a RASPdu specific object.
         */
        //~RasPDU();
      //@}

      /**@name Member variable access */
      //@{
      //@}

        H323RasPDU             m_H323RasPDU;
        H323TransportAddress   m_ReceiveTransportAddress;
        Tools::SIpAddress      m_ReceiveIpAddress;
    };
    /// dequeue to store RasPDU objects for async processing
    typedef std::deque< RasPDU * > RasPDUQueue;

	enum GkServerThreadType
    {
        RASMulticastThread,
        RASThread,
        TimeoutThread
    };

    /**@name Construction */
    //@{
    /**Create a new gatekeeper server.
     */
    H323GkServer( H323EndPoint            & rEndpoint,                  /// Gatekeepers endpoint
                  const Tools::SIpAddress & rTransportAddress,          /// Transport address for the RAS messages
                  const Tools::SIpAddress & rMulticastTransportAddress, /// Transport address for the multicast RAS messages
                  bool GkRoutedCallModel = true );                      /// Gk routed call model

    /**Destroy the gatekeeper server.
     */
    ~H323GkServer();

    /**Force erasure of a client.
     */
    void ForceEraseClient(const Tools::SString    & szH323Alias,
                          const Tools::SIpAddress & rIpAddress);
    /**Force erasure of all devices with a specific alias.
     */
    void ForceEraseClient(const Tools::SString & szH323Alias);

    /**Unregister all clients
     */
    void UnregisterAllClients();
    
    /**Shutdown gatekeeper.
     */
    void Shutdown();

    /**RAS registration timeout handler.
     */
    virtual bool HandleRasRegistrationTimeout(H323RemoteEndpoint * pEndpoint);
    
    /**Background thread handler.
     */
    virtual bool HandleRasChannel();

    /**Background thread handler for multicast messages.
     */
    virtual bool HandleRasMulticastChannel();

    /**Handle and dispatch the RAS PDUs
     */
    virtual bool HandleRasPDU(H323GkServer::RasPDU * pRasPDU);
    //@}

    /**@name Overrides from SObject */
    //@{
    void PrintOn(std::ostream & rStream) const;
    //@}

    /**@name Protcol operations */
    //@{

    /**Tell the client, that we are his gatekeeper.
     */
    bool GatekeeperConfirm();

    /**We do not want to serve the client
     */
    bool GatekeeperReject();

    /**Confirm the registration of the client
     */
    bool RegistrationConfirm();

    /**Reject the registration of the client
     */
    bool RegistrationReject();

    /**Unregister an endpoint
     */
    bool UnregistrationRequest(H323RemoteEndpoint & rEndpoint);

    /**Confirm unregistration of an endpoint
     */
    bool UnregistrationConfirm();

    /**Reject unregistration of an endpoint
     */
    bool UnregistrationReject();

    /**Confirm admission of endpoint's connection
     */
    bool AdmissionConfirm();

    /**Reject admission of endpoint's connection
     */
    bool AdmissionReject();

    /**Request bandwidth
     */
    bool BandwidthRequest();

    /**Confirm bandwidth request
     */
    bool BandwidthConfirm();

    /**Reject bandwidth request
     */
    bool BandwidthReject();

    /**Request info
     */
    bool InfoRequest();

    /**Info request acknowledge
     */
    bool InfoRequestAck();

    /**Info request not acknowledge
     */
    bool InfoRequestNack();

    /**Request disengage
     */
    bool DisengageRequest();

    /**Confirm disengage request
     */
    bool DisengageConfirm();

    /**Reject disengage request
     */
    bool DisengageReject();

    /**Locate an endpoint
     */
    bool LocationRequest();

    /**Confirm a location request; endpoint resolved
     */
    bool LocationConfirm();

    /**Reject a location request; endpoint not resolved
     */
    bool LocationReject();

    /**Send a NonStandardMessage to an endpoint
     */
    bool NonStandardMessage();

    /**Tell remote endpoint, that we did not understand his message
     */
    bool UnknownMessageResponse();

    /**Request is still in progress
     */
    bool RequestInProgress();

    /**Confirm requested resources
     */
    bool RequestAvailableConfirm();
    
    //@}

    /**@name Protocol callbacks */
    //@{

    virtual bool OnReceiveGatekeeperRequest(RasPDU *pRasPDU);
    virtual bool OnReceiveRegistrationRequest(RasPDU *pRasPDU);
    virtual bool OnReceiveUnregistrationRequest(RasPDU *pRasPDU);
    virtual bool OnReceiveUnregistrationConfirm(RasPDU *pRasPDU);
    virtual bool OnReceiveUnregistrationReject(RasPDU *pRasPDU);
    virtual bool OnReceiveAdmissionRequest(RasPDU *pRasPDU);
    virtual bool OnReceiveBandwidthRequest(RasPDU *pRasPDU);
    virtual bool OnReceiveBandwidthConfirm(RasPDU *pRasPDU);
    virtual bool OnReceiveBandwidthReject(RasPDU *pRasPDU);
    virtual bool OnReceiveInfoRequestResponse(RasPDU *pRasPDU);
    virtual bool OnReceiveDisengageRequest(RasPDU *pRasPDU);
    virtual bool OnReceiveDisengageConfirm(RasPDU *pRasPDU);
    virtual bool OnReceiveDisengageReject(RasPDU *pRasPDU);
    virtual bool OnReceiveLocationRequest(RasPDU *pRasPDU);
    virtual bool OnReceiveLocationConfirm(RasPDU *pRasPDU);
    virtual bool OnReceiveLocationReject(RasPDU *pRasPDU);
    virtual bool OnReceiveNonStandardMessage(RasPDU *pRasPDU);
    virtual bool OnReceiveRequestInProgress(RasPDU *pRasPDU);
    virtual bool OnReceiveResourcesAvailableIndicate(RasPDU *pRasPDU);

    virtual bool OnReceiveUnknown(RasPDU *pRasPDU); /// PDU type UNKNOWN.

    virtual void OnSendGatekeeperConfirm(H225_GatekeeperConfirm &);
    virtual void OnSendGatekeeperReject(H225_GatekeeperReject &);
    virtual void OnSendRegistrationConfirm(H225_RegistrationConfirm &);
    virtual void OnSendRegistrationReject(H225_RegistrationReject &);
    virtual void OnSendUnregistrationRequest(H225_UnregistrationRequest &);
    virtual void OnSendUnregistrationConfirm(H225_UnregistrationConfirm &);
    virtual void OnSendUnregistrationReject(H225_UnregistrationReject &);
    virtual void OnSendAdmissionConfirm(H225_AdmissionConfirm &);
    virtual void OnSendAdmissionReject(H225_AdmissionReject &);
    virtual void OnSendBandwidthRequest(H225_BandwidthRequest &);
    virtual void OnSendBandwidthConfirm(H225_BandwidthConfirm &);
    virtual void OnSendBandwidthReject(H225_BandwidthReject &);
    virtual void OnSendInfoRequest(H225_InfoRequest &);
    virtual void OnSendInfoRequestAck(H225_InfoRequestAck &);
    virtual void OnSendInfoRequestNak(H225_InfoRequestNak &);
    virtual void OnSendDisengageRequest(H225_DisengageRequest &);
    virtual void OnSendDisengageConfirm(H225_DisengageConfirm &);
    virtual void OnSendDisengageReject(H225_DisengageReject &);
    virtual void OnSendLocationRequest(H225_LocationRequest &);
    virtual void OnSendLocationConfirm(H225_LocationConfirm &);
    virtual void OnSendLocationReject(H225_LocationReject &);
    virtual void OnSendNonStandardMessage(H225_NonStandardMessage &);
    virtual void OnSendUnknownMessageResponse(H225_UnknownMessageResponse &);
    virtual void OnSendRequestInProgress(H225_RequestInProgress &);
    virtual void OnSendResourcesAvailableConfirm(H225_ResourcesAvailableConfirm &);
    
    /**@name Abstract callbacks */
    //@{
    /**Callback to register remote endpoint with higher layer protocols.

       @param rAddress          Address of remote endpoint to register
       @param rAliases          H323 and E164 aliases of remote endpoint to register
       @param ppUserData        Pointer to registered remote endpoint of higher
                                layer protocol.
       @param rTerminalType     Terminal type of remote endpoint.
       @param SupportedCodecs   int, bit field to determine supported codecs by endpoint
       @param rszVersion        endpoint version information as textual string
	   @param iLineId           line id in case of SwyxGate registration, oherwise = 0!
	   @param iNumOfChannels    number of channels in case of SwyxFax server >= 2.0 registration, oherwise = 0!
    */
    virtual H323GkSrv::Err OnRegisterRemoteEndpoint(const Tools::SIpAddress           &  rAddress,
                                                    H323GkAliases                     &  rAliases,
                                                    Tools::SObject                    ** ppUserData,
                                                    const H323EndPoint::TerminalTypes &  rTerminalType,
                                                    int                                  iSupportedCodecs,
                                                    Tools::SWString                   &  rszVersion,
                                                    int                                  iLineId = 0,
													int                                  iNumOfChannels = 0);

    /**Callback to release remote endpoint at higher layer protocols.

       @param rAddress      Address of remote endpoint to release
       @param rAliases      H323 and E164 aliases of remote endpoint to release
       @param pUserData     Pointer to registered remote endpoint of higher
                            layer protocol.
       @param rTerminalType Terminal type of remote endpoint.
    */
    virtual bool OnReleaseRemoteEndpoint(const Tools::SIpAddress           & rAddress,
                                         H323GkAliases                     & rAliases,
                                         Tools::SObject                   *& pUserData,
                                         const H323EndPoint::TerminalTypes & rTerminalType);

    /**Callback to resolve a remote endpoint with the help of an IP address

       @param rAddress      Address of endpoint to resolve
       @param pUserData     Resolved endpoint for given address
    */
    virtual bool OnResolveAddress(const Tools::SIpAddress & rAddress,
                                  Tools::SObject ** ppUserData);

    /**Callback to resolve a remote endpoint with the help of an alias.

       @param rAlias        E164-alias of endpoint to resolve
       @param rIpAddress    Resolved IpAddress of endpoint for given alias
    */
    virtual bool OnResolveAddress(const Tools::SString & rAlias,
                                  Tools::SIpAddress & rIpAddress);

    /**Callback for a call to ask for permission at higher layer protocols.

       @param rAliases      H323 and E164 aliases of remote endpoint
       @param pUserData     Pointer to registered remote endpoint of higher
                            layer protocol.
       @param rTerminalType Terminal type of remote endpoint.
    */
    virtual bool OnAdmissionRequest(H323GkAliases                       & rAliases,
                                    Tools::SObject                    * & pUserData,
                                    const H323EndPoint::TerminalTypes   & rTerminalType);

    /**Callback for a call to disengage at higher layer protocols.

       @param rAliases      H323 and E164 aliases of remote endpoint
       @param pUserData     Pointer to registered remote endpoint of higher
                            layer protocol.
       @param rTerminalType Terminal type of remote endpoint.
    */
    virtual bool OnDisengageRequest(H323GkAliases                       & rAliases,
                                    Tools::SObject                    * & pUserData,
                                    const H323EndPoint::TerminalTypes   & rTerminalType);
    //@}

    /**@name Miscellaneous methods */
    //@{
    Tools::SIPSocket::Address GetLocalAddress() const;
    bool SetUpOwnCallSignalTransportPDU(H225_TransportAddress & pdu) const;
    //@}

    /**@name Member variable access */
    //@{
    
    /**Query if we are running in gatekeeper routed call model.
     */
    bool IsGkRoutedCallModel() const { return m_GkRoutedCallModel; }
    
    /**Get pointer to object that stores the registered endpoints.
     */
    H323RegisteredEPList & GetRegisteredEndpoints() { return m_RegisteredEPList; }

    /**Get H323 alias for registered endpoint by call identifier.
     */
    Tools::SString GetH323Alias(const H323GloballyUniqueID & rCallIdentifier,
                                const Tools::SStringVector & rE164Aliases);
    //@}
    
    void SignalShutdown( GkServerThreadType threadType );

    bool GatekeeperShutdownInProgress() { return m_ShutdownGatekeeper; }

    // EP added / removed to / from regiration timer list -> process list to determine new timeout
    void SignalRegistrationTimerListChange();
    
protected:

    /** Add the H225 RRQ nonStandardData part to a RCF signal PDU.

        @param rrq            H.225 RCF signal PDU
        @param strm           Non standard data stream to attach to RCF
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool AttachRCFNonStandardData(const Tools::SByteVector & data,
                                  H225_RegistrationConfirm & rcf);

    /** Get the H225 RRQ nonStandardData part out of a RRQ signal PDU.

        @param rcf            H.225 RRQ signal PDU
        @param data           Non standard data stream from RRQ
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    bool GetRRQNonStandardData(const H225_RegistrationRequest & rrq,
                               Tools::SByteVector             & data);

    unsigned GetNextSequenceNumber();

    bool MakeRequest(H323RemoteEndpoint & rEndpoint,
                     const H323RasPDU & rRequest);
    
    bool CheckForResponse(H323RemoteEndpoint & rEndpoint,
                          unsigned Tag,
                          unsigned SequenceNumber);

    bool SendRasPdu(const H323TransportAddress & rDest,
                    const H323RasPDU & rPdu);

    // called after reception of RAS message
    SDECLARE_NOTIFIER( Tools::SThreadPool, H323GkServer, OnRASRequest );
    // called at last line of RAS/timeout thread before stopping. This is to signal that a new thread must be created
    SDECLARE_NOTIFIER( Tools::SThreadPool, H323GkServer, OnThreadShutdown );

    // thread pool; every terminating RAS thread will queue a work item to signal termination.
    // within the OnTimeout-function it's decided to restart the thread if neccessary
    Tools::SThreadPool   * m_pThreadPool;

	// basic configuration information
    Tools::STimeInterval   m_RasTimeout;
    unsigned               m_RasRetries;
    
    bool                   m_GkRoutedCallModel;
    bool                   m_ShutdownGatekeeper;

    Tools::SString         m_GatekeeperIdentifier;
    Tools::SString         m_szObjectIdentifier;

    // communication related stuff
    H323EndPoint         & m_rEndpoint;

    H323Transport        * m_pMulticastTransport;
    Tools::SIpAddress      m_MulticastTransportAddress;

    H323Transport        * m_pTransport;
    Tools::SIpAddress      m_TransportAddress;
    int                    m_LastTransportError;

    H323GkTimeoutThread  * m_pTimeoutThread;
    
    Tools::SMutex          m_Mutex;
    H323RegisteredEPList   m_RegisteredEPList;

    SCriticalSection       m_LastSequenceNumberCritSec;
    unsigned               m_LastSequenceNumber;
};


#endif // _GkServer_h_
