/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: GSM Codec
 
    Comments: H.323 protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/SwyxSrv.root/H323/include/gsmcodec.h                 $
    $Author:: Martin.hueser                                                $
      $Date:: 6.08.04 16:29                                                $
   $Modtime:: 3.08.04 9:45                                                 $
  $Revision:: 8                                                            $
  $Workfile:: gsmcodec.h                                                   $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __GSM_CODEC_H
#define __GSM_CODEC_H

#include "h323caps.h"


/**This class describes the GSM 06.10 codec capability.
 */
class H323_GSM0610Capability : public H323AudioCapability
{
  public:
  /**@name Construction */
  //@{
    /**Create a new GSM 06.10 capability.
     */
    H323_GSM0610Capability();
  //@}

  /**@name Overrides from class SObject */
  //@{
    /**Create a copy of the object.
      */
    virtual Tools::SObject * Clone() const;

    /**Print to stream.
       Standard output function from SObject implements stream << operator.
      */
    virtual void PrintOn(
      std::ostream & strm    /// Stream to output to.
    ) const;
  //@}

  /**@name Operations */
  //@{
    /**Create the codec instance, allocating resources as required.
     */
    virtual H323Codec * CreateCodec(
      H323Codec::Direction direction  /// Direction in which this instance runs
    ) const;
  //@}

  /**@name Identification functions */
  //@{
    /**Get the sub-type of the capability. This is a code dependent on the
       main type of the capability.

       This returns one of the four possible combinations of mode and speed
       using the enum values of the protocol ASN H245_AudioCapability class.
     */
    virtual unsigned GetSubType() const;
  //@}

  /**@name Protocol manipulation */
  //@{
    /**This function is called whenever and outgoing TerminalCapabilitySet
       or OpenLogicalChannel PDU is being constructed for the control channel.
       It allows the capability to set the PDU fields from information in
       members specific to the class.

       The default behaviour sets the data rate field in the PDU.
     */
    virtual bool OnSendingPDU(
      H245_AudioCapability & pdu,  /// PDU to set information on
      unsigned packetSize          /// Packet size to use in capability
    ) const;

    /**This function is called whenever and incoming TerminalCapabilitySet
       or OpenLogicalChannel PDU has been used to construct the control
       channel. It allows the capability to set from the PDU fields,
       information in members specific to the class.

       The default behaviour gets the data rate field from the PDU.
     */
    virtual bool OnReceivedPDU(
      const H245_AudioCapability & pdu,  /// PDU to get information from
      unsigned & packetSize              /// Packet size to use in capability
    );
  //@}
};


struct gsm_state;

/**This class is a GSM 06.10 codec.
 */
class H323_GSM0610Codec : public H323FramedAudioCodec
{
  public:
    enum {
      SamplesPerFrame = 160,  // 20 milliseconds
      BytesPerFrame = 33      // Encoded size
    };

  /**@name Construction */
  //@{
    /**Create a new GSM 06.10 codec for ALaw.
     */
    H323_GSM0610Codec(
      Direction direction         /// Direction in which this instance runs
    );
    ~H323_GSM0610Codec();
  //@}

    /**Encode a sample block into the buffer specified.
       The samples have been read and are waiting in the readBuffer member
       variable. it is expected this function will encode exactly
       encodedBlockSize bytes.
     */
    virtual bool EncodeFrame(
      BYTE * buffer     /// Buffer into which encoded bytes are placed
    );

    /**Decode a sample block from the buffer specified.
       The samples must be placed into the writeBuffer member variable. It is
       expected that no more than frameSamples is decoded. The return value
       is the number of samples decoded. Zero indicates an error.
     */
    virtual bool DecodeFrame(
      const BYTE * buffer   /// Buffer from which encoded data is found
    );

    /**Get type of codec.
      */
    virtual CodecType GetCodecType() const { return GSM_FullRate; }

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const { return 65; } // Microsoft GSM (?)

  protected:
    struct gsm_state * gsm;
};


#endif // __GSM_CODEC_H

