/*---------------------------------------------------------------------------

   Swyx Solutions AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) Swyx Solutions AG
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/h323.h                    $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 28.06.05 12:36                                               $
   $Modtime:: 28.06.05 11:03                                               $
  $Revision:: 2                                                            $
  $Workfile:: h323.h                                                       $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __H323_H
#define __H323_H

//#define ROUND_TRIP_DELAY

#include "h323lib.h"
#include "h323caps.h"
#include "transports.h"
#include "gkclient.h"
#include "gklink.h"
#include "channels.h"
#include "codecs.h"
#include "../include/rtp.h" // prevent conflicts with platform SDK
#include "q931.h"
#include "h450pdu.h"
#include "STlsConnection.h"


// Module including Swyx H.323 specific stuff, e.g. tracing
extern H323Module *g_pH323Module;


// Defines
#define MAX_H245_ACCEPT_WAITING 120 // Seconds until H.245 Accept timeout occurs.


/* The following classes have forward references to avoid including the VERY
   large header files for H225 and H245. If an application requires access
   to the protocol classes they can include them, but for simple usage their
   inclusion can be avoided.
 */
class SPerStream;
class SAsnOctetString;

class H225_EndpointType;
class H225_VendorIdentifier;
class H225_H221NonStandard;
class H225_TransportAddress;
class H225_GatekeeperRequest;
class H225_GatekeeperConfirm;
class H225_ArrayOf_SAsnOctetString;

class H245_TerminalCapabilitySet;
class H245_TerminalCapabilitySetReject;
class H245_OpenLogicalChannel;
class H245_OpenLogicalChannelAck;
class H245_TransportAddress;
class H245_UserInputIndication;
class H245_RequestMode;
class H245_RequestModeAck;
class H245_RequestModeReject;
class H245_SendTerminalCapabilitySet;
class H245_MultiplexCapability;
class H245_FlowControlCommand;
class H245_EndSessionCommand;
class H245_MiscellaneousCommand;
class H245_MiscellaneousIndication;
class H245_JitterIndication;

class H323SignalPDU;
class H323ControlPDU;
class H323VideoDevice;

class H245NegMasterSlaveDetermination;
class H245NegTerminalCapabilitySet;
class H245NegLogicalChannels;
class H245NegRequestMode;
class H245NegRoundTripDelay;


/** H225 call thread.
 */
class H225CallThread : public Tools::SThread, public Tools::SConnectionObject
{
  public:
    H225CallThread(H323EndPoint & endpoint,
                   H323Connection & connection,
                   H323Transport & transport,
                   const Tools::SString & remoteParty);

  protected:
    void Main();

    H323Connection & connection;
    H323Transport  & transport;
    Tools::SString   remoteParty;
};


/**Globally unique ID definition.
 */
class H323GloballyUniqueID : public std::vector<BYTE>
{
  public:
    H323GloballyUniqueID();
    H323GloballyUniqueID(const SAsnOctetString &);

    Tools::SString AsString() const;
    void PrintOn(std::ostream & strm) const;
};


/**This class represents a particular H323 connection between two endpoints.
   There are at least two threads in use, this one to look after the
   signalling channel, and another to look after the control channel. There
   would then be additional threads created for each data channel created by
   the control channel protocol thread.
 */
class H323Connection : public Tools::SConnectionObject, public Tools::SRefCnt
{
  public:
  /**@name Construction */
  //@{
    /**Create a new connection.
     */
    H323Connection(
      H323EndPoint & endpoint,          /// H323 End Point object
      unsigned callReference,           /// Call reference
      const Tools::SString & token,     /// Token to identify connection in active connections container
      const unsigned long    ulConnId   /// Connection ID for connection releated trace output
    );

    /**Destroy the connection
     */
    ~H323Connection();

    /** Initialize several objects.
       (Moved from constructor to check pointers).
     */
    bool Init();

    /**Delete several objects.
       (Moved from destructor to this method because of reference counting).
     */
    void Delete();

    /**Lock connection.
       When the H323EndPoint::FindConnectionWithLock() function is used to gain
       access to a connection object, this is called to prevent it from being
       closed and deleted by the background threads.
     */
    void Lock() { inUseFlag.Wait(); }

    /**Unlock connection.
       If the H323EndPoint::FindConnectionWithLock() function is used to gain
       access to a connection object, this MUST be called to allow it to
       subsequently be closed and disposed of.
     */
    void Unlock() { inUseFlag.Signal(); }

    enum TerminationReason {
      LocalTermination,         /// Local endpoint application terminated call
      NoAcceptTermination,      /// Local endpoint did not accept call OnIncomingCall()=FALSE
      AnswerDeniedTermination,  /// Local endpoint declined to answer call
      RemoteTermination,        /// Remote endpoint application terminated call
      RefusalTermination,       /// Remote endpoint refused call
      NoAnswerTermination,      /// Remote endpoint did not answer in required time
      CallAbortTermination,     /// Remote endpoint stopped calling
      AbnormalTermination,      /// Transport error terminated call
      ConnectFailTermination,   /// Transport connection failed to establish call
      GatekeeperForcedTermination,  /// Gatekeeper has termianted call
      EndedByCapabilityExchange,/// Could not find common capabilities
      NumTerminationReasons
    };

    /**Get the termination reason for this connection shutting down.
       Note that this function is only generally useful in the
       H323EndPoint::OnConnectionCleared() function. This is due to the
       connection not being terminated before that, and the object not even
       exiting after that.
      */
    TerminationReason GetTerminationReason() const { return terminationReason; }

    /**Set the termination reason.
       An application should have no cause to use this function. It is present
       for the H323EndPoint::ClearCall() function to set the termination reason.
      */
    void SetTerminationReason(
      TerminationReason reason   /// Reason for termination of connection.
    );

    /** Get the Cause for disconnecting call.

        @return   Cause.
    */
    Q931::Cause GetCause() const { return m_Cause; }

    /** Set the Cause for disconnecting call.

        @param Cause   Cause for disconnecting call.
      */
    void SetCause(Q931::Cause cause) { m_Cause = cause; }

    /** Get the Release Complete reason.

        @return  ReleaseCompleteReason (type and reason).
    */
    H225_ReleaseCompleteReason GetReleaseCompleteReason() const { return m_ReleaseCompleteReason; }

    /** Set the Release Complete reason.

        @param Reason   ReleaseCompleteReason (type and reason).
      */
    void SetReleaseCompleteReason(H225_ReleaseCompleteReason reason) { m_ReleaseCompleteReason = reason; }

    /**Clear a current connection.
       This hangs up the connection to a remote endpoint. It actually just
       calls the endpoint version of the ClearCall() function to avoid
       possible multithreading race conditions.
      */
    virtual bool ClearCall(
      H323Connection::TerminationReason terminationReason =
                  H323Connection::LocalTermination /// Reason for call clearing
    );

    /**Clean up the termination of the connection.
       This function will do any internal cleaning up and waiting on background
       threads that may be using the connection object. After this returns it
       is then safe to delete the object.

       An application will not typically use this function as it is used by the
       H323EndPoint during a clear call.
      */
    virtual void CleanUpOnTermination();
  //@}


  /**@name Signalling Channel */
  //@{
    /**Attach a transport to this connection as the signalling channel.
      */
    void AttachSignalChannel(
      H323Transport * channel  /// Transport for the PDU's
    );

    /**Write a PDU to the signalling channel.
      */
    bool WriteSignalPDU(
      H323SignalPDU & pdu       /// PDU to write.
    );

    /**Handle reading PDU's from the signalling channel.
     */
    void HandleSignallingChannel();

    /**Handle PDU from the signalling channel.
     */
    virtual bool HandleSignalPDU(
      H323SignalPDU & pdu       /// PDU to handle.
    );

    /**Handle Control PDU tunnelled in the signalling channel.
     */
    virtual void HandleTunnelPDU(
      const H323SignalPDU & pdu,    /// PDU to handle.
      H323SignalPDU * reply         /// Optional PDU to piggy back replies
    );

    /**Build list of fast start entries.
    */
    bool BuildFastStartList(const H323Channel & channel,
                            H225_ArrayOf_SAsnOctetString & array,
                            H323Channel::Directions reverseDirection);

    /**Handle an incoming Q931 setup PDU.
       The default behaviour is to do the handshaking operation calling a few
       virtuals at certain moments in the sequence.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.
     */
    virtual bool OnReceivedSignalSetup(
      const H323SignalPDU & pdu   /// Received setup PDU
    );

    /**Handle an incoming Q931 Setup Acknowledge PDU.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.
     */
    virtual bool OnReceivedSetupAck(
      const H323SignalPDU & pdu   /// Received Setup Acknowledge PDU
    );

    /**Handle an incoming Q931 Information PDU.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.
     */
    virtual bool OnReceivedInformation(
      const H323SignalPDU & pdu   /// Received Information PDU
    );

    /**Handle an incoming Q931 call proceeding PDU.
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour checks for hH245Address field and if present
       starts the separate H245 channel, if successful or not present it
       returns true.
     */
    virtual bool OnReceivedCallProceeding(
      const H323SignalPDU & pdu   /// Received call proceeding PDU
    );

    /**Handle an incoming Q931 alerting PDU.
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour obtains the display name and calls OnAlerting().
     */
    virtual bool OnReceivedAlerting(
      const H323SignalPDU & pdu   /// Received Alerting PDU
    );

    /**Handle an incoming Q931 connect PDU.
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour checks for hH245Address field and if present
       starts the separate H245 channel, if successful it returns true.
       If not present and there is no H245Tunneling then it returns false.
     */
    virtual bool OnReceivedSignalConnect(
      const H323SignalPDU & pdu   /// Received connect PDU
    );

    /**Handle an incoming Q931 progress PDU.
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour checks for progress indicator IE, fast start
       channels and h245Address field and if present it starts that channel.
       If successful or not present it returns true.
     */
    virtual bool OnReceivedProgress(
      const H323SignalPDU & pdu
    );

    /** Handle an incoming Q931 Notify PDU.

        The default behaviour checks for notification indicator IE.
        If successful or not present it returns true.

        @param pdu       Received Notify PDU
        @return          If false is returned the connection is aborted and a Release Complete
                         PDU is sent.
                         The default behaviour returns true.
     */
    virtual bool OnReceivedNotify(
      const H323SignalPDU & pdu
    );

    /**Handle an incoming Q931 facility PDU.
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour checks for hH245Address field and if present
       starts the separate H245 channel, if successful or not present it
       returns true.
     */
    virtual bool OnReceivedFacility(
      const H323SignalPDU & pdu   /// Received Facility PDU
    );

    /**Handle an incoming Q931 Status PDU.
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.
     */
    virtual bool OnReceivedStatus(
      const H323SignalPDU & pdu   /// Received Status PDU
    );

    /**Handle an incoming Q931 Status Enquiry PDU.
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour sends a Q931 Status PDU back.
     */
    virtual bool OnReceivedStatusEnquiry(
      const H323SignalPDU & pdu   /// Received Status Enquiry PDU
    );

    /**Handle an incoming Q931 Release Complete PDU. (SwTCL extension)
       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour calls OnReleaseComplete().
     */
    virtual bool OnReceivedReleaseComplete(
      const H323SignalPDU & pdu   /// Received release complete PDU
    );

    /**This function is called from the HandleSignallingChannel() function
       for unhandled PDU types.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent. The default behaviour returns true.
     */
    virtual bool OnUnknownSignalPDU(
      const H323SignalPDU & pdu  /// Received PDU
    );

    /**Call back for incoming call.
       This function is called from the OnReceivedSignalSetup() function
       before it sends the Alerting PDU. It gives an opportunity for an
       application to alter the reply before transmission to the other
       endpoint.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour simply returns true.
     */
    virtual bool OnIncomingCall(
      const H323SignalPDU & setupPDU,   /// Received setup PDU
      H323SignalPDU & alertingPDU       /// Alerting PDU to send
    );

    /**Initiate the transfer of an existing call (connection) to a new remote party
       using H.450.2.  This sends a Call Transfer Initiate Invoke message from the
       A-Party (transferring endpoint) to the B-Party (transferred endpoint).
     */
    void TransferCall(
      const Tools::SString & remoteParty,   /// Remote party to transfer the existing call to
      const Tools::SString & callIdentity = Tools::SString()
                                    /// Call Identity of secondary call if present
    );

    /**Transfer the call through consultation so the remote party in the primary call is connected to
       the called party in the second call using H.450.2.  This sends a Call Transfer Identify Invoke 
       message from the A-Party (transferring endpoint) to the C-Party (transferred-to endpoint).
     */
    void ConsultationTransfer(
      const Tools::SString & primaryCallToken  /// Primary call
    );

    /**Handle the reception of a callTransferSetupInvoke APDU whilst a secondary call exists.  This 
       method checks whether the secondary call is still waiting for a callTransferSetupInvoke APDU and 
       proceeds to clear the call if the call identies match.
       This is an internal function and it is not expected the user will call
       it directly.
     */
    virtual void HandleConsultationTransfer(
      const Tools::SString & callIdentity, /**Call Identity of secondary call 
                                       received in SETUP Message. */
      H323Connection & incoming     /// Connection upon which SETUP PDU was received.
    );

    /**Determine whether this connection is being transferred.
     */
    bool IsTransferringCall() const;

    /**Determine whether this connection is the result of a transferred call.
     */
    bool IsTransferredCall() const;

    /**Handle the transfer of an existing connection to a new remote.
       This is an internal function and it is not expected the user will call
       it directly.
     */
    virtual void HandleTransferCall(
        const Tools::SString & token,
        const Tools::SString & identity
    );

    /**Get transfer invoke ID dureing trasfer.
       This is an internal function and it is not expected the user will call
       it directly.
      */
    int GetCallTransferInvokeId();

    /**Handle the failure of a call transfer operation at the Transferred Endpoint.  This method is
       used to handle the following transfer failure cases that can occur at the Transferred Endpoint. 
       The cases are:
       Reception of an Admission Reject
       Reception of a callTransferSetup return error APDU.
       Expiry of Call Transfer timer CT-T4.
     */
    virtual void HandleCallTransferFailure(
      const int returnError    /// Failure reason code
    );

    /**Store the passed token on the current connection's H4502Handler.
       This is an internal function and it is not expected the user will call
       it directly.
     */
    void SetAssociatedCallToken(
      const Tools::SString & token  /// Associated token
    );

    /**Callback to indicate a successful transfer through consultation.  The paramter passed is a
       reference to the existing connection between the Transferring endpoint and Transferred-to 
       endpoint.
     */
    virtual void OnConsultationTransferSuccess(
      H323Connection & secondaryCall  /// Secondary call for consultation
    );

    enum AnswerCallResponse {
      AnswerCallNow,
      AnswerCallDenied,
      AnswerCallPending,
      NumAnswerCallResponses
    };

    /**Call back for answering an incoming call.
       This function is called from the OnReceivedSignalSetup() function
       before it sends the Connect PDU. It gives an opportunity for an
       application to alter the reply before transmission to the other
       endpoint.

       It also gives an application time to wait for some event before
       signalling to the endpoint that the connection is to proceed. For
       example the user pressing an "Answer call" button.

       If AnswerCallDenied is returned the connection is aborted and a Release
       Complete PDU is sent. If AnswerCallNow is returned then the H.323
       protocol proceeds. Finally if AnswerCallPending is returned then the
       protocol negotiations are paused until the AnsweringCall() function is
       called.

       The default behaviour simply returns AnswerNow.
     */
    virtual AnswerCallResponse OnAnswerCall(
      const Tools::SString & callerName, /// Name of caller
      const H323SignalPDU & setupPDU,   /// Received setup PDU
      H323SignalPDU & connectPDU        /// Connect PDU to send. 
    );

    /**Indicate the result of answering an incoming call.
       This should only be called if the OnAnswerCall() callback function has
       returned a AnswerCallPending response.
      */
    void AnsweringCall(
      AnswerCallResponse response /// Answer response to incoming call
    );

    /**Send first PDU in signalling channel.
       This function does the signalling handshaking for establishing a
       connection to a remote endpoint. The transport (TCP/IP) connection
       is assumed to have already been made.
     */
    virtual bool SendSignalSetup(
      const Tools::SString & remoteParty   /// Name of remote party (or host)
    );

    /**Adjust setup PDU being sent on initialisation of signal channel.
       This function is called from the SendSignalSetup() function before it
       sends the Setup PDU. It gives an opportunity for an application to
       alter the request before transmission to the other endpoint.

       The default behaviour simply returns true. Note that this is usually
       overridden by the transport dependent descendent class, eg the
       H323ConnectionTCP descendent fills in the destCallSignalAddress field
       with the TCP/IP data. Therefore if you override this in your
       application make sure you call the ancestor function.
     */
    virtual bool OnSendSignalSetup(
      H323SignalPDU & setupPDU   /// Setup PDU to send
    );

    /**Call back for remote party being alerted.
       This function is called from the SendSignalSetup() function after it
       receives the optional Alerting PDU from the remote endpoint. That is
       when the remote "phone" is "ringing".

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour simply returns true.
     */
    virtual bool OnAlerting(
      const H323SignalPDU & alertingPDU,  /// Received Alerting PDU
      const Tools::SString & user         /// Username of remote endpoint
    );

    /**This function is called from the SendSignalSetup() function after it
       receives the Connect PDU from the remote endpoint, but before it
       attempts to open the control channel.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour simply returns true.
     */
    virtual bool OnOutgoingCall(
      const H323SignalPDU & connectPDU   /// Received Connect PDU
    );

    /** This function could be called if a Release Complete PDU is sent to the remote endpoint.
        (SwTCL extension)

        @param ReleaseCompletePDU       Release Complete PDU to send
        @param callReference            Call reference value
        @param reason                   Release Complete reason (see H.225 for reference)
        @param cause                    Cause for disconnecting the call (mutally exclusive to release complete reason!)
        @return                         The default behaviour returns true.
     */
    virtual bool OnSendReleaseComplete(const H323SignalPDU               & ReleaseCompletePDU,
                                       int                                 callReference,
                                       const H225_ReleaseCompleteReason  & reason,
                                       const Q931::Cause                 & cause);

    /**This function is called after a Release Complete PDU is received from the remote endpoint.
       (SwTCL extension)

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour simply returns true.
     */
    virtual bool OnReleaseComplete(
      const H323SignalPDU & releasePDU   /// Received Release Complete PDU
    );

    /**Handle the acknowldege of a fast start.
       This function is called from one of a number of functions after it
       receives a PDU from the remote endpoint that has a fastStart field. It
       is in response to a request for a fast strart from the local endpoint.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour parses .
     */
    virtual bool HandleFastStartAcknowledge(
      const H225_ArrayOf_SAsnOctetString & array   /// Array of H245_OpenLogicalChannel
    );

    /**Start a separate H245 channel.
       This function is called from one of a number of functions after it
       receives a PDU from the remote endpoint that has a h245Address field.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent.

       The default behaviour checks to see if it is a known transport and
       creates a corresponding H323Transport decendent for the control
       channel.
     */
    virtual bool StartControlChannel(
      const H225_TransportAddress & h245Address   /// H245 address
    );

    /** Attach the H225 nonStandardData part to a signal PDU.

        @param pdu            H323 signal PDU
        @param data           Non standard data
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    virtual bool AttachH225NonStandardData(H323SignalPDU            & pdu,
                                           const Tools::SByteVector & data);

    /** Get and handle the H225 nonStandardData part of a signal PDU.

        @param pdu            H323 signal PDU
        @param data           Non standard data
        @return               If false is returned an error occured.
                              The default behaviour returns true.
    */
    virtual bool HandleH225NonStandardData(const H323SignalPDU & pdu,
                                           Tools::SByteVector  & data);

    /// Class including NonStandardParameter data.
    class NonStandardParameter
    {
      public:
        NonStandardParameter(const SAsnObjectId & rObjectId,
                             const SByteVector  & rData)
         :  m_bObjectIdUsed(true),
            m_ObjectId(rObjectId),
            m_Data(rData) {};

        NonStandardParameter(const H225_H221NonStandard & rH221NonStandard,
                             const SByteVector          & rData)
         :  m_bObjectIdUsed(false),
            m_H221NonStandard(rH221NonStandard),
            m_Data(rData) {};

        virtual ~NonStandardParameter() {};

        bool                    m_bObjectIdUsed;
        SAsnObjectId            m_ObjectId;
        H225_H221NonStandard    m_H221NonStandard;
        SByteVector             m_Data;
    };

    /** Attach the H225 nonStandardControl part to a signal PDU.

        @param pdu                  H323 signal PDU
        @param rNonStdParamVector   Vector including NonStandardParameter data.
        @return                     If false is returned an error occured.
                                    The default behaviour returns true.
    */
    virtual bool AttachH225NonStandardControl(H323SignalPDU                           & pdu,
                                              const std::vector<NonStandardParameter> & rNonStdParamVector);

    /** Get and handle the H225 nonStandardControl part of a signal PDU.

        @param pdu                  H323 signal PDU
        @param rNonStdParamVector   Vector including NonStandardParameter data.
        @return                     If false is returned an error occured.
                                    The default behaviour returns true.
    */
    virtual bool HandleH225NonStandardControl(const H323SignalPDU               & pdu,
                                              std::vector<NonStandardParameter> & rNonStdParamVector);
  //@}

  /**@name Control Channel */
  //@{
    /**Write a PDU to the control channel.
       If there is no control channel open then this will tunnel the PDU
       into the signalling channel.
      */
    virtual bool WriteControlPDU(
      const H323ControlPDU & pdu
    );

    /**Start control channel negotiations.
      */
    bool StartControlNegotiations();

    /**Handle reading data on the control channel.
     */
    virtual void HandleControlChannel();

    /**Handle incoming data on the control channel.
       This decodes the data stream into a PDU and calls HandleControlPDU().

       If false is returned the connection is aborted. The default behaviour
       returns true.
     */
    virtual bool HandleControlData(
      SPerStream & strm
    );

    /**Handle incoming PDU's on the control channel. Dispatches them to the
       various virtuals off this class.

       If false is returned the connection is aborted. The default behaviour
       returns true.
     */
    virtual bool HandleControlPDU(
      const H323ControlPDU & pdu
    );

    /**This function is called from the HandleControlPDU() function
       for unhandled PDU types.

       If false is returned the connection is aborted and a Release Complete
       PDU is sent. The default behaviour returns true.

       The default behaviour send a FunctioNotUnderstood indication back to
       the sender, and returns true to continue operation.
     */
    virtual bool OnUnknownControlPDU(
      const H323ControlPDU & pdu  /// Received PDU
    );

    /**Handle incoming request PDU's on the control channel.
       Dispatches them to the various virtuals off this class.
     */
    virtual bool OnH245Request(
      const H323ControlPDU & pdu  /// Received PDU
    );

    /**Handle incoming response PDU's on the control channel.
       Dispatches them to the various virtuals off this class.
     */
    virtual bool OnH245Response(
      const H323ControlPDU & pdu  /// Received PDU
    );

    /**Handle incoming command PDU's on the control channel.
       Dispatches them to the various virtuals off this class.
     */
    virtual bool OnH245Command(
      const H323ControlPDU & pdu  /// Received PDU
    );

    /**Handle incoming indication PDU's on the control channel.
       Dispatches them to the various virtuals off this class.
     */
    virtual bool OnH245Indication(
      const H323ControlPDU & pdu  /// Received PDU
    );

    /**Handle H245 command to send terminal capability set.
     */
    virtual bool OnH245_SendTerminalCapabilitySet(
      const H245_SendTerminalCapabilitySet & pdu  /// Received PDU
    );

    /**Handle H245 command to control flow control.
       This function calls OnLogicalChannelFlowControl() with the channel and
       bit rate restriction.
     */
    virtual bool OnH245_FlowControlCommand(
      const H245_FlowControlCommand & pdu  /// Received PDU
    );

    /**Handle H245 miscellaneous command.
       This function passes the miscellaneous command on to the channel
       defined by the pdu.
     */
    virtual bool OnH245_MiscellaneousCommand(
      const H245_MiscellaneousCommand & pdu  /// Received PDU
    );

    /**Handle H245 end session command. (SwTCL extension)
       This function passes the end session command on to the channel
       defined by the pdu.
     */
    virtual bool OnH245_EndSessionCommand(
      const H245_EndSessionCommand & pdu  /// Received PDU
    );

    /**Handle H245 miscellaneous indication.
       This function passes the miscellaneous indication on to the channel
       defined by the pdu.
     */
    virtual bool OnH245_MiscellaneousIndication(
      const H245_MiscellaneousIndication & pdu  /// Received PDU
    );

    /**Handle H245 indication of received jitter.
       This function calls OnLogicalChannelJitter() with the channel and
       estimated jitter.
     */
    virtual bool OnH245_JitterIndication(
      const H245_JitterIndication & pdu  /// Received PDU
    );

    /**Error discriminator for the OnControlProtocolError() function.
      */
    enum ControlProtocolErrors {
      e_MasterSlaveDetermination,
      e_CapabilityExchange,
      e_LogicalChannel,
      e_ModeRequest,
      e_RoundTripDelay
    };

    /**This function is called from the HandleControlPDU() function or
       any of its sub-functions for protocol errors, eg unhandled PDU types.

       The errorData field may be a string or PDU or some other data depending
       on the value of the errorSource parameter. These are:
          e_UnhandledPDU                    &H323ControlPDU
          e_MasterSlaveDetermination        const char *

       If false is returned the connection is aborted. The default behaviour
       returns true.
     */
    virtual bool OnControlProtocolError(
      ControlProtocolErrors errorSource,  // Source of the proptoerror
      const void * errorData = NULL       // Data associated with error
    );

    /**This function is called from the HandleControlPDU() function when
       it is about to send the Capabilities Set to the remote endpoint. This
       gives the application an oppurtunity to alter the PDU to be sent.

       The default behaviour does nothing.
     */
    virtual void OnSendCapabilitySet(
      H245_TerminalCapabilitySet & pdu  /// PDU to send
    );

    /**This function is called when the remote endpoint sends its capability
       set. This gives the application an opportunity to determine what codecs
       are available and if it supports any of the combinations of codecs.

       Note any codec types that are the remote system supports that are not in
       the codecs list member variable for the endpoint are ignored and not
       included in the remoteCodecs list.

       The default behaviour assigns the table and set to member variables and
       returns true if the remoteCodecs list is not empty.
     */
    virtual bool OnReceivedCapabilitySet(
      const H323Capabilities & remoteCaps,      /// Capability combinations remote supports
      const H245_MultiplexCapability * muxCap,  /// Transport capability, if present
      H245_TerminalCapabilitySetReject & reject /// Rejection PDU (if return false)
    );

    /**This function sends an empty TerminalCapabilitySet messages to initiate pause
       and re-routing handling.
     */
    virtual bool SendEmptyCapabilitySet();

    /**This function sends a full TerminalCapabilitySet and MasterSlaveDetermination messages
       to make peer side opening its logical channel again after pause and re-routing handling.
     */
    virtual bool SendFullCapabilitySet();

    /**This function is called when the remote endpoint sends an empty capability
       set to initiate pause and re-routing handling.
     */
    virtual bool OnReceivedEmptyCapabilitySet();

    /**This function is called when the remote endpoint sends 
	   H.450 4 remote call hold invoke
     */
	virtual bool OnReceivedHoldRequest();

    /**This function is called when the remote endpoint sends 
	   H.450 4 remote call retrieve invoke
     */
	virtual bool OnReceivedActivateRequest();

    /**This function is called when the remote endpoint sends 
	   H.450 2 CT initiate invoke
     */
	virtual bool OnReceivedCallTransferRequest(H323Connection *pTransferredToCall);

    /**This function is resets the capability table of this connection.
       It is used when the remote endpoint has send an empty capability
       set to initiate pause and re-routing handling.
     */
    virtual void ResetCapabilities();

    /**Return if this H245 connection is a master or slave
     */
    bool IsH245Master() const;

#ifdef ROUND_TRIP_DELAY
    /**Start the round trip delay calculation over the control channel.
     */
    void StartRoundTripDelay();

    /**Get the round trip delay over the control channel.
     */
    Tools::STimeInterval GetRoundTripDelay() const;
#endif // ROUND_TRIP_DELAY

    /**Call back for control channel open.
       This is called when the control channel has negotiated the master/slave
       relationship and the remote endpoint has acknowledged receipt of the
       capability set.

       If false is returned the connection is cleared.

       The default behaviour of this function is to call 
       OnSelectLogicalChannels() and return true.
    */
    virtual bool OnControlChannelOpen();
  //@}

  /**@name Logical Channel Management */
  //@{
    /**Call back to select logical channels to start.

       This function must be defined by the descendent class. It is used
       to select the logical channels to be opened between the two endpoints.
       There are three ways in which this may be called: when a "fast start"
       has been initiated by the local endpoint (via SendSignalSetup()
       function), when a "fast start" has been requested from the remote
       endpoint (via the OnReceivedSignalSetup() function) or when the H245
       capability set (and master/slave) negotiations have completed (via the
       OnControlChannelOpen() function.

       The function would typically examine several member variable to decide
       which mode it is being called in and what to do. If fastStartState is
       FastStartDisabled then non-fast start semantics should be used. The
       H245 capabilities in the remoteCapabilities members should be
       examined, and appropriate transmit channels started using
       OpenLogicalChannel().

       If fastStartState is FastStartInitiate, then the local endpoint has
       initiated a call and is asking the application if fast start semantics
       are to be used. If so it is expected that the function call 
       OpenLogicalChannel() for all the channels that it wishes to be able to
       be use. A subset (possibly none!) of these would actually be started
       when the remote endpoint replies.

       If fastStartState is FastStartResponse, then this indicates the remote
       endpoint is attempting a fast start. The fastStartChannels member
       contains a list of possible channels from the remote that the local
       endpoint is to select which to accept. For each accepted channel it
       simply necessary to call the Start() function on that channel eg
       fastStartChannels[0].Start();

       The default behaviour selects the first codec of each session number
       that is available. This is according to the order of the capabilities
       in the remoteCapabilities, the local capability table or of the
       fastStartChannels list respectively for each of the above scenarios.

       @param bStartDataSession    Boolean to possible disable starting data session.
    */
    virtual void OnSelectLogicalChannels(bool bStartDataSession = true);

    /**Select default logical channel for normal start.
      */
    virtual void SelectDefaultLogicalChannel(
      unsigned sessionID    /// Session ID to find default logical channel.
    );

    /**Select default logical channel for fast start.
      */
    virtual void SelectFastStartChannels(
      unsigned sessionID    /// Session ID to find default logical channel.
    );

    /**Restart logical channels (select logical channels to start) after
       state "transmitter side paused" is leaved
       (see H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing).

       The default behaviour selects the first codec of each session number
       that is available. This is according to the order of the capabilities
       in the remoteCapabilityTable or the local capability table, respectively.
      */
    virtual void RestartSelectLogicalChannels();

    /**Open a new logical channel.
       This function will open a channel between the endpoints for the
       specified capability.

       If this function is called while there is not yet a conenction
       established, eg from the OnFastStartLogicalChannels() function, then
       a "trial" receiver/transmitter channel is created. This channel is not
       started until the remote enpoint has confirmed that they are to start.
       Any channels not confirmed are deleted.

       If this function is called later in the call sequence, eg from
       OnSelectLogicalChannels(), then it may only establish a transmit
       channel, ie fromRemote must be false.
      */
    virtual bool OpenLogicalChannel(
      const H323Capability & capability,  /// Capability to open channel with
      unsigned sessionID,                 /// Session for the channel
      H323Channel::Directions dir         /// Direction of channel
    );

    /**This function is called when the remote endpoint want's to open
       a new channel.

       If the return value is false then the open is rejected using the
       errorCode as the cause, this would be a value from the enum
       H245_OpenLogicalChannelReject_cause::Choices.

       The default behaviour simply returns true.
     */
    virtual bool OnOpenLogicalChannel(
      const H245_OpenLogicalChannel & openPDU,  /// Received PDU for the channel open
      H245_OpenLogicalChannelAck & ackPDU,      /// PDU to send for acknowledgement
      unsigned & errorCode                      /// Error to return if refused
    );

    /**This function is called after creation of a new logical channel object.

        @param pChannel         Channel that has been started or NULL.
        @param channelNumber    Channel number of OLC.
        @param uErrorCode       Error code if OLC was refused.
     */
    virtual void OnOLCAfterCreation(H323Channel             * pChannel,
                                    const H323ChannelNumber & channelNumber,
                                    unsigned                  uErrorCode);

    /**Create a new logical channel object.
       This is in response to a request from the remote endpoint to open a
       logical channel.
      */
    virtual H323Channel * CreateLogicalChannel(
      const H245_OpenLogicalChannel & open, /// Parameters for opening channel
      unsigned & errorCode                  /// Reason for create failure
    );

    /**This is called after an OpenLogicalChannelAck was received (SwTCL extension).
     */
    virtual bool OnReceivedLogicalChannelAckPDU(
      const H245_OpenLogicalChannelAck & ack  /// Acknowledgement PDU
    );

    /**Call back function when a logical channel thread begins.

       The default behaviour does nothing and returns true.
      */
    virtual bool OnStartLogicalChannel(
      H323Channel & channel    /// Channel that has been started.
    );

    /**Open a channel for use by an audio codec.
       The H323AudioCodec class will use this function to open the channel to
       read/write PCM data.

       The default behaviour calls the equivalent function on the endpoint.
      */
    virtual bool OpenAudioChannel(
      bool isEncoding,       /// Direction of data flow
      unsigned bufferSize,   /// Size of each sound buffer
      H323AudioCodec & codec /// codec that is doing the opening
    );

    /**Open a channel for use by an video codec.
       The H323VideoCodec class will use this function to open the channel to
       read/write image data.

       The default behaviour calls the equivalent function on the endpoint.
      */
    virtual bool OpenVideoDevice(
      bool isEncoding,       /// Direction of data flow
      H323VideoCodec & codec /// codec doing the opening
    );

    /**Close a logical channel.
      */
    virtual void CloseLogicalChannel(
      unsigned number,    /// Channel number to close.
      bool fromRemote     /// Indicates close request of remote channel
    );

    /**Close all logical channels (Tx if fromRemote is false, Rx if fromRemote is true).
      */
    virtual void CloseAllLogicalChannels(
      bool fromRemote     /// Indicates close request of remote channel
    );

    /**This function is called when the remote endpoint has closed down
       a logical channel.

       The default behaviour does nothing.
     */
    virtual void OnClosedLogicalChannel(
      const H323Channel & channel   /// Channel that was closed
    );

    /**This function is called when the remote endpoint request the close of
       a logical channel.

       The application may get an opportunity to refuse to close the channel by
       returning false from this function.

       The default behaviour returns true.
     */
    virtual bool OnClosingLogicalChannel(
      H323Channel & channel   /// Channel that is to be closed
    );

    /**This function is called when the remote endpoint wishes to limit the
       bit rate being sent on a channel.

       If channel is NULL, then the bit rate limit applies to all channels.

       The default behaviour does nothing if channel is NULL, otherwise calls
       H323Channel::OnFlowControl() on the specific channel.
     */
    virtual void OnLogicalChannelFlowControl(
      H323Channel * channel,   /// Channel that is to be limited
      long bitRateRestriction  /// Limit for channel
    );

    /**This function is called when the remote endpoint indicates the level
       of jitter estimated by the receiver.

       If channel is NULL, then the jitter applies to all channels.

       The default behaviour does nothing if channel is NULL, otherwise calls
       H323Channel::OnJitter() on the specific channel.
     */
    virtual void OnLogicalChannelJitter(
      H323Channel * channel,   /// Channel that is to be limited
      DWORD jitter,            /// Estimated received jitter in microseconds
      int skippedFrameCount,   /// Frames skipped by decodec
      int additionalBuffer     /// Additional size of video decoder buffer
    );

    /**Find a logical channel.
       Locates a channel give a RTP session ID. Each session would usually
       have two logical channels associated with it, so the fromRemote flag
       may be used to distinguish which channel to return.
      */
    H323Channel * FindChannel(
      unsigned sessionId,   /// Session ID to search for.
      bool fromRemote       /// Indicates the direction of RTP data.
    ) const;
    
    /**This method deletes all unused fast start channel objects which have not
       been moved to the logical channels map.
      */
    bool DeleteUnusedFastStartChannels();
  //@}

  /**@name Bandwidth Management */
  //@{
    /**Get the bandwidth currently used.
       This totals the open channels and returns the total bandwidth used in
       100's of bits/sec
      */
    unsigned GetBandwidthUsed() const;

    /**Request use the available bandwidth in 100's of bits/sec.
       If there is insufficient bandwidth available, false is returned. If
       sufficient bandwidth is available, then true is returned and the amount
       of available bandwidth is reduced by the specified amount.
      */
    bool UseBandwidth(
      unsigned bandwidth,     /// Bandwidth required
      bool removing           /// Flag for adding/removing bandwidth usage
    );

    /**Get the available bandwidth in 100's of bits/sec.
      */
    unsigned GetBandwidthAvailable() const { return bandwidthAvailable; }

    /**Set the available bandwidth in 100's of bits/sec.
       Note if the force parameter is true this function will close down
       active logical channels to meet the new bandwidth requirement.
      */
    bool SetBandwidthAvailable(
      unsigned newBandwidth,    /// New bandwidth limit
      bool force = false        /// Force bandwidth limit
    );
  //@}

  /**@name Indications */
  //@{
    /**Send a user input indication to the remote endpoint.
       The two forms are for basic user input of a simple string using the
       SendUserInput() function or a full DTMF emulation user input using the
       SendUserInputTone() function.

       An application could do more sophisticated usage by filling in the 
       H245_UserInputIndication structure directly and using this function.
      */
    virtual void SendUserInputIndication(
      const H245_UserInputIndication & pdu    /// Full user indication PDU
    );

    /**Send a user input indication to the remote endpoint.
       This is for sending arbitrary strings as user indications. Simple DTMF
       tones can be sent this way with the tone value being a single
       character string.
      */
    virtual void SendUserInput(
      const Tools::SString & value            /// String value of indication
    );

    /**Send a user input indication to the remote endpoint.
       This sends DTMF emulation user input. If something more sophisticated
       than the simple tones that can be sent using the SendUserInput()
       function.

       A duration of zero indicates that no duration is to be indicated.
       A non-zero logical channel indicates that the tone is to be syncronised
       with the logical channel at the rtpTimestamp value specified.

       The tone parameter must be one of "0123456789#*ABCD!" where '!'
       indicates a hook flash. If tone is a ' ' character then a
       signalUpdate PDU is sent that updates the last tone indication
       sent. See the H.245 specifcation for more details on this.
      */
    virtual void SendUserInputTone(
      const Tools::SString & tone, /// DTMF tone code
      unsigned duration = 0,       /// Duration of tone in milliseconds
      unsigned logicalChannel = 0, /// Logical channel number for RTP sync.
      unsigned rtpTimestamp = 0    /// RTP timestamp in logical channel sync.
    );

    /**Send a user input indication to the remote endpoint.
       This sends a Hook Flash emulation user input.
      */
    void SendUserInputHookFlash(
      int duration = 500  /// Duration of tone in milliseconds
    ) { SendUserInputTone(SString('!'), duration); }

    /**Call back for remote enpoint has sent user input.
       The default behaviour calls OnUserInputString() if the PDU is of the
       alphanumeric type, or OnUserInputTone() if of a tone type.
      */
    virtual void OnUserInputIndication(
      const H245_UserInputIndication & pdu  /// Full user indication PDU
    );

    /**Call back for remote enpoint has sent user input as a string.

       The default behaviour calls the endpoint function of the same name.
      */
    virtual void OnUserInputString(
      const Tools::SString & value   /// String value of indication
    );

    /**Call back for remote enpoint has sent user input.

       The default behaviour calls the endpoint function of the same name.
      */
    virtual void OnUserInputTone(
      const Tools::SString & tone, /// DTMF tone code
      unsigned duration,           /// Duration of tone in milliseconds
      unsigned logicalChannel,     /// Logical channel number for RTP sync.
      unsigned rtpTimestamp        /// RTP timestamp in logical channel sync.
    );
  //@}

  /**@name RTP Session Management */
  //@{
    /**Get an RTP session for the specified ID.
       If there is no session of the specified ID, NULL is returned.
      */
    RTP_Session * GetSession(
      unsigned sessionID
    ) const;

    /**Get the names of the codecs in use for the RTP session.
       If there is no session of the specified ID, an empty string is returned.
      */
    Tools::SString GetSessionCodecNames(
      unsigned sessionID
    ) const;

    /**Use an RTP session for the specified ID.
       If there is no session of the specified ID, a new one is created using
       the information provided in the H245_TransportAddress PDU. If the system
       does not support the specified transport, NULL is returned.

       If this function is used, then the ReleaseSession() function MUST be
       called or the session is never deleted for the lifetime of the H323
       connection.
      */
    RTP_Session * UseSession(
      unsigned sessionID,
      const H245_TransportAddress & pdu
    );

    /**Release the session. If the session ID is not being used any more any
       clients via the UseSession() function, then the session is deleted.
     */
    void ReleaseSession(
      unsigned sessionID
    );
  //@}

  /**@name Request Mode Changes */
  //@{
    /**Make a request to mode change to remote.
      */
    void RequestModeChange();

    /**Received request for mode change from remote.
      */
    virtual bool OnRequestModeChange(
      const H245_RequestMode & pdu,     /// Received PDU
      H245_RequestModeAck & ack,        /// Ack PDU to send
      H245_RequestModeReject & reject,  /// Reject PDU to send
      int & selectedMode                /// Which mode was selected
    );

    /**Completed request for mode change from remote.
       This is a call back that occurs after the ack has been sent to the
       remote as indicated by the OnRequestModeChange() return result. This
       function is intended to actually implement the mode change after it
       had been accepted.
      */
    virtual void OnModeChanged(
      const H245_ModeDescription & newMode
    );

    /**Received acceptance of last mode change request from remote.
      */
    virtual void OnAcceptModeChange(
      const H245_RequestModeAck & pdu  /// Received PDU
    );

    /**Received reject of last mode change request from remote.
      */
    void OnRefusedModeChange(
      const H245_RequestModeReject * pdu  /// Received PDU, if NULL is a timeout
    );
  //@}

  /**@name Member variable access */
  //@{
    /**Get the owner endpoint for this connection.
     */
    H323EndPoint & GetEndPoint() const { return endpoint; }

    /**Get the call direction for this connection.
     */
    bool HadAnsweredCall() const { return callAnswered; }

    /**Get the internal OpenH323 call token for this connection.
     */
    const Tools::SString & GetCallToken() const { return callToken; }

    /**Get the call reference for this connection.
     */
    unsigned GetCallReference() const { return callReference; }

    /**Get the call identifier for this connection.
     */
    const H323GloballyUniqueID & GetCallIdentifier() const { return callIdentifier; }

    /**Get the conference identifier for this connection.
     */
    const H323GloballyUniqueID & GetConferenceIdentifier() const { return conferenceIdentifier; }

    /**Get the caller name/alias.
      */
    const Tools::SString & GetRemotePartyName() const { return remotePartyName; }

    /**Set the caller name/alias from information in the PDU.
      */
    void SetRemotePartyName(const H323SignalPDU & pdu);

    /**Get the remotes capability table for this connection.
     */
    const H323Capabilities & GetLocalCapabilities() const { return localCapabilities; }

    /**Get the remotes capability table for this connection.
     */
    const H323Capabilities & GetRemoteCapabilities() const { return remoteCapabilities; }

    /**Get the maximum audio jitter delay.
     */
    unsigned GetRemoteMaxAudioDelayJitter() const { return remoteMaxAudioDelayJitter; }

    /**Get the signalling channel being used.
      */
    const H323Transport * GetSignallingChannel() const { return signallingChannel; }

    /**Get the control channel being used (may return signalling channel).
      */
    const H323Transport & GetControlChannel() const;

    /**Get the time at which the connection was established
      */
    Tools::STime GetConnectionStartTime() const { return connectionStartTime; }

    /** Get the ProgressIndicator information.

        @return  ProgressIndicator  Class containing ProgressIndicator information.
    */
    Q931::ProgressIndicator GetProgressIndicator() const { return m_ProgressIndicator; }
    
    /**Set the source media TSAP (RTP IP address and port of caller) (SwTCL extension).
      */
    void SetSourceMediaTSAP(const Tools::SIpAddress & srcMediaTSAP) { m_srcMediaTSAP = srcMediaTSAP; }

    /**Get the source media TSAP (RTP IP address and port of caller) (SwTCL extension).
      */
    virtual Tools::SIpAddress GetSourceMediaTSAP(const H323Codec::CodecType type) const { return m_srcMediaTSAP; }

    /**Set the source media control TSAP (RTCP IP address and port of caller) (SwTCL extension).
      */
    void SetSourceMediaControlTSAP(const Tools::SIpAddress & srcMediaControlTSAP) { m_srcMediaControlTSAP = srcMediaControlTSAP; }

    /**Get the source media control TSAP (RTCP IP address and port of caller) (SwTCL extension).
      */
    virtual Tools::SIpAddress GetSourceMediaControlTSAP(const H323Codec::CodecType type) const { return m_srcMediaControlTSAP; }

    /**Get IP address of local interface.
      */
    virtual Tools::SIPSocket::Address GetLocalInterfaceAddress() const { return Tools::SIPSocket::Address(); }

    /**Set the address of the local H.245 interface (SwTCL extension).
      */
    void SetH245LocalAddress(const Tools::SIPSocket::Address & H245LocalAddress) { m_H245LocalAddress = H245LocalAddress; }

    /**Get the address of the local H.245 interface (SwTCL extension).
      */
    Tools::SIPSocket::Address GetH245LocalAddress() const { return m_H245LocalAddress; }

    /**Set the port of the local H.245 interface (SwTCL extension)
      */
    void SetH245LocalPort(WORD H245LocalPort) { m_H245LocalPort = H245LocalPort; }

    /**Get the port of the local H.245 interface (SwTCL extension)
      */
    WORD GetH245LocalPort() const { return m_H245LocalPort; }

    /**Get the H233 alias to be used for gatekeeper registration.
     */
    const Tools::SString & GetH323Alias() const { return m_szH323Alias; }

    /**Set H245 tunneling flag.
    */
    void SetH245Tunneling(bool bH245Tunneling) { h245Tunneling = bH245Tunneling; }

    /**Get information if connection is already connected (already received Connect).
     */
    bool IsConnected() const { return (   connectionState == HasExecutedSignalConnect
                                       || connectionState == EstablishedConnection
                                       || connectionState == TransmitterSidePaused); }

    /** Get refmutex stored within the transport object*/
    Tools::SAutoRefPtr<SRefCriticalSection> GetRefMutex()
    {
        H323TransportTCP* pTransportTcp = dynamic_cast<H323TransportTCP*>(signallingChannel);
        return ( pTransportTcp!=NULL ) ? pTransportTcp->m_pRefMutex : NULL;
    }

  //@}

  protected:
    H323EndPoint       &endpoint;
    Tools::SSemaphore  inUseFlag;
    Tools::STime       connectionStartTime;
    bool               isConsultationTransfer;

    bool                 callAnswered;
    Tools::SString       callToken;
    unsigned             callReference;
    H323GloballyUniqueID callIdentifier;
    H323GloballyUniqueID conferenceIdentifier;
    Tools::SString       objectIdentifier;

    H323Capabilities   localCapabilities; // Capabilities local system supports
    Tools::SString     remotePartyName;
    H323Capabilities   remoteCapabilities; // Capabilities remote system supports
    unsigned           remoteMaxAudioDelayJitter;
#ifdef ROUND_TRIP_DELAY
    Tools::STimer      roundTripDelayTimer;
#endif // ROUND_TRIP_DELAY

    unsigned bandwidthAvailable;

    H323Transport * signallingChannel;
    H323Transport * controlChannel;
    bool            h245Tunneling;
    H323SignalPDU * h245TunnelPDU;
    H323SignalPDU * h245ParallelPDU;

    enum ConnectionStates {
      NoConnectionActive,
      AwaitingGatekeeperAdmission,
      AwaitingTransportConnect,
      AwaitingSignalConnect,
      HasExecutedSignalConnect,
      EstablishedConnection,
      TransmitterSidePaused,  // H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing
      ShuttingDownConnection,
      NumConnectionStates
    } connectionState;

    Q931::ProgressIndicator      m_ProgressIndicator;
    Q931::Cause                  m_Cause;
    H225_ReleaseCompleteReason   m_ReleaseCompleteReason;
    
    TerminationReason  terminationReason;
    AnswerCallResponse answerResponse;
    Tools::SSyncPoint  answerWaitFlag;

    RTPSessionDict rtpSessions;
    Tools::SMutex  rtpSessionMutex;

    enum FastStartStates {
      FastStartDisabled,
      FastStartInitiate,
      FastStartResponse,
      FastStartAcknowledged,
      NumFastStartStates
    };
    FastStartStates        fastStartState;
    H323LogicalChannelList fastStartChannels;

    // SwTCL extensions
    Tools::SIpAddress         m_srcMediaTSAP;        // RTP  IP address and port of caller
    Tools::SIpAddress         m_srcMediaControlTSAP; // RTCP IP address and port of caller
    Tools::SIPSocket::Address m_H245LocalAddress;    // Address of the local H.245 interface
    WORD                      m_H245LocalPort;       // Port of the local H.245 interface
    Tools::SString            m_szH323Alias;         // H323 alias (registered at GK)

    bool                      m_bControlDataHandled;


//#if STRACING
    static const char * const ConnectionStatesNames[NumConnectionStates];
    friend std::ostream & operator<<(std::ostream & o, ConnectionStates s) { return o << ConnectionStatesNames[s]; }
    static const char * const FastStartStateNames[NumFastStartStates];
    friend std::ostream & operator<<(std::ostream & o, FastStartStates s) { return o << FastStartStateNames[s]; }
    static const char * const AnswerCallResponseNames[NumAnswerCallResponses];
    friend std::ostream & operator<<(std::ostream & o, AnswerCallResponse s) { return o << AnswerCallResponseNames[s]; }
    static const char * const TerminationReasonNames[NumTerminationReasons];
    friend std::ostream & operator<<(std::ostream & o, TerminationReason s) { return o << TerminationReasonNames[s]; }
//#endif


    // The following pointers are to protocol procedures, they are pointers to
    // hide their complexity from the H323Connection classes users.
    H245NegMasterSlaveDetermination  * masterSlaveDeterminationProcedure;
    H245NegTerminalCapabilitySet     * capabilityExchangeProcedure;
    H245NegLogicalChannels           * logicalChannels;
    H245NegRequestMode               * requestModeProcedure;
    H245NegRoundTripDelay            * roundTripDelayProcedure;
    H450xDispatcher                  * h450dispatcher;
    H4502Handler                     * h4502handler;
    H4504Handler                     * h4504handler;
};

//PLIST(H323ConnectionList, H323Connection);
typedef std::list<H323Connection *> H323ConnectionList;
//typedef std::vector<H323Connection *> H323ConnectionList;

//PDICTIONARY(H323ConnectionDict, PString, H323Connection);
typedef stdext::hash_map<Tools::SString, H323Connection* > H323ConnectionDict;

class H323EndPoint;

class SH323TcpConnection : public Tools::STcpConnection
{
    H323TransportTCP* m_pTransport;
    H323EndPoint&     m_endpoint;
    bool              m_ClosedFromApp;

public:
    SH323TcpConnection( H323EndPoint& endpoint,
                        unsigned int uTraceId = 0, 
                        unsigned long ulConnectionId = 0)
    : STcpConnection( NULL, uTraceId, ulConnectionId ),
      m_pTransport(NULL), m_endpoint(endpoint), m_ClosedFromApp(false)
      {}

    virtual ~SH323TcpConnection()
    { if( m_pTransport ) m_pTransport->ClosedByWire(); }



    virtual bool ProcessRead();

    void SetTransport(H323TransportTCP* pTransport) 
    { m_pTransport=pTransport; 
      if(pTransport==NULL) 
        m_ClosedFromApp=true; 
    }

    void RemoveTransport() 
    { m_pTransport=NULL; m_ClosedFromApp=true; }


};


class SH323TcpConnectionFactory : public Tools::STcpConnectionFactory
{
public:
    SH323TcpConnectionFactory() : STcpConnectionFactory() {}

    virtual Tools::STcpConnection * CreateNew(Tools::STcpListener * pListener, unsigned int uTraceId = 0, unsigned long ulConnectionId = 0) const
    { return new STcpConnection(pListener, uTraceId, ulConnectionId); }

    virtual Tools::STcpConnection * CreateNew(H323EndPoint& endpoint, unsigned int uTraceId = 0, unsigned long ulConnectionId = 0) const
    { return new SH323TcpConnection( endpoint, uTraceId, ulConnectionId); }
};

class SH323TlsConnection : public Tools::STlsConnection
{
    H323TransportTls* m_pTransport;
    H323TransportTls* m_pProvisionalTransport;
    H323EndPoint&     m_endpoint;

public:
    SH323TlsConnection( H323EndPoint& endpoint,
                        const SSimpleArray<char>& certificate, 
                        const SSimpleArray<char>& privateKey,
                        unsigned int uTraceId = 0, 
                        unsigned long ulConnectionId = 0)
    : STlsConnection( NULL, certificate, privateKey, uTraceId, ulConnectionId ),
      m_pTransport(NULL), m_pProvisionalTransport(NULL), m_endpoint(endpoint)
      { }

    virtual ~SH323TlsConnection()
    {
        if( m_pTransport )              m_pTransport->ClosedByWire();
        if( m_pProvisionalTransport )   m_pProvisionalTransport->ClosedByWire();
    }

    virtual bool ProcessRead();

    void SetTransport(H323TransportTls* pTransport) 
    {   
        // a new H323 connection re-using this connection for its transport 
        //   should trigger an still pending connection to be aborted and removed
        if( m_pTransport!=NULL && m_pTransport!=pTransport) m_pTransport->AbortH323Connection(false/*IsFromWire*/);
        m_pTransport=pTransport;
    }

    void RemoveTransport(H323TransportTls* pTransport) 
    { 
      if( pTransport==m_pTransport ) 
        m_pTransport=NULL; 
      if( pTransport==m_pProvisionalTransport )
        m_pProvisionalTransport=NULL;
    }

    void ConfirmTransport()
    { m_pTransport=m_pProvisionalTransport; m_pProvisionalTransport=NULL; }

    void ResetProvisionalTransport()
    { m_pProvisionalTransport=NULL; }
};

class SH323TlsConnectionFactory : public STlsConnectionFactory
{
public:
    SH323TlsConnectionFactory(const SSimpleArray<char>& certificate, const SSimpleArray<char>& privateKey) 
        : STlsConnectionFactory(certificate,privateKey) {}

    virtual STcpConnection * CreateNew(H323EndPoint& endpoint, unsigned int uTraceId = 0, unsigned long ulConnectionId = 0) const
    { return new SH323TlsConnection( endpoint, m_certificate, m_privateKey, uTraceId, ulConnectionId); }
};

/**This class manages the H323 endpoint.
   An endpoint may have zero or more listeners to create incoming connections
   or zero or more outgoing conenctions initiated via the MakeCall() function.
   Once a conection exists it is managed by this class instance.

   The main thing this class embodies is the capabilities of the application,
   that is the codecs and protocols it is capable of.

   An application may create a descendent off this class and overide the
   CreateConnection() function, if they require a descendent of H323Connection
   to be created. This would be quite likely in most applications.
 */
class H323EndPoint : public Tools::SObject
{
  public:

    /**Endpoint types.
     */
    enum TerminalTypes {
      e_TerminalOnly = 50,
      e_TerminalAndMC = 70,
      e_GatewayOnly = 60,
      e_GatewayAndMC = 80,
      e_GatewayAndMCWithDataMP = 90,
      e_GatewayAndMCWithAudioMP = 100,
      e_GatewayAndMCWithAVMP = 110,
      e_GatekeeperOnly = 120,
      e_GatekeeperWithDataMP = 130,
      e_GatekeeperWithAudioMP = 140,
      e_GatekeeperWithAVMP = 150,
      e_MCUOnly = 160,
      e_MCUWithDataMP = 170,
      e_MCUWithAudioMP = 180,
      e_MCUWithAVMP = 190
    };

  /**@name Construction */
  //@{
    /**Create a new endpoint.
     */
    H323EndPoint(SH323TcpConnectionFactory& tcpConFab, SH323TlsConnectionFactory& tlsConFab);

    /**Destroy endpoint.
     */
    ~H323EndPoint();

    /**Set the endpoint information in H225 PDU's.
      */
    virtual void SetEndpointTypeInfo(
      H225_EndpointType & info, const SString sLinkName=""
    ) const;

    /**Set the vendor information in H225 PDU's.
      */
    virtual void SetVendorIdentifierInfo(
      H225_VendorIdentifier & info
    ) const;

    /**Set the tunnelled prototocol in H225 PDU's.
      */
///*hk: this is to add supported tunnelled protocols 
    virtual void SetSupportedProtocolsToTunnelledProtocol(
      H225_ArrayOf_TunnelledProtocol & info
    ) const;

    /**Set the H221NonStandard information in H225 PDU's.
      */
    virtual void SetH221NonStandardInfo(
      H225_H221NonStandard & info
    ) const;
  //@}


  /**@name Capabilities */
  //@{
    /**Add a codec to the capabilities table. This will assure that the
       assignedCapabilityNumber field in the codec is unique for all codecs
       installed on this endpoint.

       If the specific instnace of the capability is already in the table, it
       is not added again. Ther can be multiple instances of the same
       capability class however.
     */
    void AddCapability(
      H323Capability * capability   /// New codec specification
    );

    /**Set the capability descriptor lists. This is three tier set of
       codecs. The top most level is a list of particular capabilities. Each
       of these consists of a list of alternatives that can operate
       simultaneously. The lowest level is a list of codecs that cannot
       operate together. See H323 section 6.2.8.1 and H245 section 7.2 for
       details.

       Note that the capability specified here is automatically added to the
       capability table using the AddCapability() function.
     */
    void SetCapability(
      int descriptorNum,    /// The member of the capabilityDescriptor to add
      int simultaneous,     /// The member of the SimultaneousCapabilitySet to add
      H323Capability * cap  /// New capability specification
    );

    /**Find a capability that has been registered.
     */
    H323Capability * FindCapability(
      const H245_Capability & cap  /// H245 capability table entry
    ) const;

    /**Find a capability that has been registered.
     */
    H323Capability * FindCapability(
      const H245_DataType & dataType  /// H245 data type of codec
    ) const;

    /**Find a capability that has been registered.
     */
    H323Capability * FindCapability(
      H323Capability::MainTypes mainType,   /// Main type of codec
      unsigned subType,                     /// Subtype of codec
      bool useEncryption = false
    ) const;
  //@}

  /**@name Gatekeeper management */
  //@{
    /**Select and register with an explicit gatekeeper.
       This will use the specified transport and a string giving a transport
       dependent address to locate a specific gatekeeper. The endpoint will
       register with that gatekeeper and, if successful, set it as the current
       gatekeeper used by this endpoint.

       Note the transport being passed in will be deleted by this function or
       the H323Gatekeeper object it becomes associated with.
     */
    bool SetGatekeeper(
      const Tools::SString &    address, /// Address of gatekeeper to add.
      const Tools::SString &    alias,   /// H323 alias of endpoint to register at gatekeeper.
	  const Tools::SIpAddress & transportAddress, /// Transport address over which to talk to gatekeeper.
      const bool                bIsAlternativeGatekeeper = false, // flag to indicate if GK address is alternative GK
      const int                 iLineId = 0, // line Id to register SwyxGate
      const int                 iNumOfChannels = 0 // number of channels to register by SwyxFax
    );

    /**Discover and select gatekeeper.
       This function will use the automatic gatekeeper discovery methods to
       locate a gatekeeper on a particular transport. The
       OnReceiveGatekeeperConfirmPDU() function is called for each gatekeeper
       suitable.
       oneOnly parameter determines if the endpoint is to register with only
       the first gatekeeper discovered, or all of the them. Each one that is
       successfully registered is added to the gatekeeper list used by this
       endpoint.

       Note the transport being passed in will be deleted becomes owned by the
       endpoint and will be deleted by it.
     */
    bool DiscoverGatekeeper(
      const Tools::SIpAddress & transportAddress  /// Transport address over which to find gatekeepers.
    );

    /**Create a gatekeeper.
       This allows the application writer to have the gatekeeper as a
       descendent of the H323Gatekeeper in order to add functionality to the
       base capabilities in the library.

       The default creates an instance of the H323Gatekeeper class.
     */
    virtual H323Gatekeeper * CreateGatekeeper(
      const Tools::SIpAddress & transportAddress  /// Transport address over which gatekeepers communicates.
    );

    /**Get the gatekeeper we are currently registered with.
     */
    H323Gatekeeper * GetGatekeeper();

    /**Get the main gatekeeper.
     */
    H323Gatekeeper * GetMainGatekeeper() const { return gatekeeper; }

    /**Get the alternative gatekeeper.
     */
    H323Gatekeeper * GetAlternativeGatekeeper() const { return alternativeGatekeeper; }

    /**Unregister and delete the gatekeeper we are registered with.
     */
    bool RemoveGatekeeper(
      int reason = -1    /// Reason for gatekeeper removal
    );

    /**Unregister and delete the alternative gatekeeper we are registered with.
     */
    bool RemoveAlternativeGatekeeper(
      int reason = -1    /// Reason for gatekeeper removal
    );

    /**Called by gatekeeper registration thread if a gatekeeper registration is lost.

        @param iClientId    ID of the H.323 client which lost registration
    */
    virtual bool OnGatekeeperRegistrationLost(int iClientId);

    /**Called by gatekeeper registration thread if a gatekeeper registration was successful.

        @param iClientId    ID of the H.323 client which established registration
    */
    virtual bool OnGatekeeperRegistrationDone(int iClientId);
  //@}

  /**@name Connection management */
  //@{
    /**Add a listener to the endoint.
       This allows for the automatic creating of incoming call connections. An
       application should use OnConnectionEstablished() to monitor when calls
       have arrived and been successfully negotiated.
      */
    bool StartListener(
      H323Listener * listener /// Transport dependent listener.
    );

    /**Search for appropriate signalling channel via remote address/port to reuse it for
       sending call signalling/control messages (TLS socket should be reused with SwyxPhones)
      */
    virtual Tools::SChannel * GetSignallingChannelForReuseByRemote(const Tools::SIPSocket::Address & remoteAddress, DWORD remotePort)
    { return 0; }

    /**Search for appropriate signalling channel via local address/port to reuse it for
       sending call signalling/control messages (TLS socket should be reused with SwyxPhones)
      */
    virtual Tools::SChannel* GetSignallingChannelForReuseByLocal(const Tools::SIPSocket::Address& localAddress, DWORD localPort)
    { return 0; }

    virtual bool GetIsTlsServer()
    { return false; }

    /**Search for appropriate signalling channel via remote address/port to reuse it for
       sending call signalling/control messages (TLS socket should be reused with SwyxPhones)
      */
    bool PutSignallingChannelToListener(Tools::SChannel *socket) 
    { if(!listeners.empty()) return listeners[0]->AddConnectedConnection(socket); 
      return false;}

    virtual void OnSignallingChannelAccepted(SIPSocket::Address remoteAddr, WORD remotePort) {}
    virtual void OnSignallingChannelClosed(SIPSocket::Address remoteAddr, WORD remotePort) {}

    /**Make a call to a remote party using the specified transport.
       The remoteParty may be a hostname or other user name that is to be
       passed to a gateway, if present.

       This function returns almost immediately with the call occurring in a
       new background thread.

       Note, the returned pointer to the connection is not locked and may be
       deleted at any time. This is extremely unlikely immediately after the
       function is called, but you should not keep this pointer beyond that
       brief time. The the FindConnectionWithLock() function for future
       references to the connection object.
     */
    H323Connection * MakeCall(
      const Tools::SString & remoteParty,  /// Remote party to call
      H323Transport * transport,           /// Transport to use for call.
      Tools::SString & token,              /// String to receive token for connection
      const unsigned long ulConnId         /// Connection ID for connection releated trace output
    );

    /**Clear a current connection.
       This hangs up the connection to a remote endpoint.
      */
    bool ClearCall(
      const Tools::SString & token,  /// Token for identifying connection
      H323Connection::TerminationReason terminationReason =
                  H323Connection::LocalTermination /// Reason for call clearing
    );

    /**Clear a current connection via different thread context
     */
    bool ClearCall( H323Connection * connection );

    /**Clear all current connections.
       This hangs up all the connections to remote endpoints.
      */
    void ClearAllCalls(
      H323Connection::TerminationReason terminationReason =
                  H323Connection::LocalTermination /// Reason for call clearing
    );

    /**Setup the transfer of an existing call (connection) to a new remote party
       using H.450.2.  This sends a Call Transfer Setup Invoke message from the
       B-Party (transferred endpoint) to the C-Party (transferred-to endpoint).

       If the transport parameter is NULL the transport is determined from the
       remoteParty description. The general form for this parameter is
       [alias@][transport$]host[:port] where the default alias is the same as
       the host, the default transport is "ip" and the default port is 1720.

       This function returns almost immediately with the transfer occurring in a
       new background thread.

       Note, the returned pointer to the connection is not locked and may be
       deleted at any time. This is extremely unlikely immediately after the
       function is called, but you should not keep this pointer beyond that
       brief time. The the FindConnectionWithLock() function for future
       references to the connection object.

       This function is declared virtual to allow an application to override
       the function and get the new call token of the forwarded call.
     */
    virtual H323Connection * SetupTransfer(
      const Tools::SString & token,        /// Existing connection to be transferred
      const Tools::SString & callIdentity, /// Call identity of the secondary call (if it exists)
      const Tools::SString & remoteParty,  /// Remote party to transfer the existing call to
      Tools::SString & newToken,           /// String to receive token for the new connection
      void * userData = NULL        /// user data to pass to CreateConnection
    );

    /**Initiate the transfer of an existing call (connection) to a new remote
       party using H.450.2.  This sends a Call Transfer Initiate Invoke
       message from the A-Party (transferring endpoint) to the B-Party
       (transferred endpoint).
     */
    void TransferCall(
      const Tools::SString & token,        /// Existing connection to be transferred
      const Tools::SString & remoteParty,  /// Remote party to transfer the existing call to
      const Tools::SString & callIdentity = Tools::SString()
                                    /// Call Identity of secondary call if present
    );

    /**Transfer the call through consultation so the remote party in the
       primary call is connected to the called party in the second call
       using H.450.2.  This sends a Call Transfer Identify Invoke message
       from the A-Party (transferring endpoint) to the C-Party
       (transferred-to endpoint).
     */
    void ConsultationTransfer(
      const Tools::SString & primaryCallToken,   /// Token of primary call
      const Tools::SString & secondaryCallToken  /// Token of secondary call
    );

    /**A call back function whenever a connection is established.
       This indicates that a connection to a remote endpoint was established
       with a control channel and zero or more logical channels.

       The default behaviour does nothing.
      */
    virtual void OnConnectionEstablished(
      H323Connection & connection,    /// Connection that was established
      const Tools::SString & token    /// Token for identifying connection
    );

    /**A call back function whenever a connection is broken.
       This indicates that a connection to a remote endpoint is no longer
       available.

       The default behaviour does nothing.
      */
    virtual void OnConnectionCleared(
      H323Connection & connection,    /// Connection that was established
      const Tools::SString & token    /// Token for identifying connection
    );

    /**Find a connection that uses the specified token.
       This searches the endpoint for the connection that contains the token
       as provided by functions such as MakeCall() (as built by the
       BuildConnectionToken() function). if not found it will then search for
       the string representation of the CallIdentifier for the connection, and
       finally try for the string representation of the ConferenceIdentifier.

       Note the caller of this function MUST call the H323Connection::Unlock()
       function if this function returns a non-NULL pointer. If it does not
       then a deadlock can occur.
      */
    H323Connection * FindConnectionWithLock(
      const Tools::SString & token     /// Token to identify connection
    );

    /**Build a unique token for the connection.
       This identifies the call using the Q931 transport host name and the
       Q931 call reference number.
      */
    static Tools::SString BuildConnectionToken(
      const H323Transport & transport,  /// Transport for connection
      unsigned callReference,           /// Call reference of Q.931 link
      bool fromRemote                   /// Call reference is from remote endpoint
    );

    /**Get a connection that uses the specified call reference.
      */
    H323Connection * GetConnection(
      const H323Transport & transport,  /// Transport for connection
      unsigned callReference,           /// Call reference of Q.931 link
      bool fromRemote                   /// Call reference is from remote endpoint
    );

    /**Create a connection that uses the specified call reference.
      */
    virtual H323Connection * CreateConnection(
      unsigned callReference,           /// Call reference to identify connection
      const Tools::SString & token,     /// Token to identify connection in active connections container
      const unsigned long ulConnId      /// Connection ID for connection releated trace output
      ) = 0;

    /** Destroy a connection.

        This function deletes the H323Connection.
        It is a virtual method that could be overloaded by a derived
        class of H323EndPoint but the implementation in the base
        class H323EndPoint should exist because e.g. CleanUpConnections
        may be called when the derived instance does not exist
        any longer. (Destructor of a derived class is called first!)
      */
    virtual void DestroyConnection(
      H323Connection * pConnection /// Connection to destroy
    );

    /**Clean up connections.
       This function is called from a background thread and checks for closed
       connections to clean up.

       This would not normally be called by an application.
      */
    void CleanUpConnections();

    /**Inform endpoint that gatekeeper connection is brokenn.
       On clients this will trigger a gatekeeper chnage to alternative gatekeeper (if name available).
      */
    virtual bool GatekeeperConnectionIsBroken(int iClientID);
  //@}

  /**@name Logical Channels management */
  //@{
    /**Call back for opening a logical channel.

       The default behaviour simply returns true.
      */
    virtual bool OnStartLogicalChannel(
      H323Connection & connection,    /// Connection for the channel
      H323Channel & channel           /// Channel being started
    );

    /**Call back for closed a logical channel.

       The default behaviour does nothing.
      */
    virtual void OnClosedLogicalChannel(
      H323Connection & connection,    /// Connection for the channel
      const H323Channel & channel     /// Channel being started
    );

    /**Open a channel for use by an audio codec.
       The H323AudioCodec class will use this function to open the channel to
       read/write PCM data.

       The default function creates a SSoundChannel using the member variables
       soundChannelPlayDevice or soundChannelRecordDevice.
      */
    virtual bool OpenAudioChannel(
      H323Connection & connection,  /// Connection for the channel
      bool isEncoding,              /// Direction of data flow
      unsigned bufferSize,          /// Size of each sound buffer
      H323AudioCodec & codec        /// codec that is doing the opening
    );

    /**Open a channel for use by an video codec.
       The H323VideoCodec class will use this function to open the channel to
       read/write image data.

       The default function returns NULL at present.
      */
    virtual bool OpenVideoDevice(
      H323Connection & connection,  /// Connection for the channel
      bool isEncoding,              /// Direction of data flow
      H323VideoCodec & codec        /// codec doing the opening
    );
  //@}

  /**@name Indications */
  //@{
    /**Call back for remote enpoint has sent user input as a string.

       The default behaviour does nothing.
      */
    virtual void OnUserInputString(
      H323Connection & connection,  /// Connection for the input
      const Tools::SString & value         /// String value of indication
    );

    /**Call back for remote enpoint has sent user input.

       The default behaviour calls H323Connection::OnUserInputTone().
      */
    virtual void OnUserInputTone(
      H323Connection & connection,  /// Connection for the input
      const Tools::SString & tone,  /// DTMF tone code
      unsigned duration,            /// Duration of tone in milliseconds
      unsigned logicalChannel,      /// Logical channel number for RTP sync.
      unsigned rtpTimestamp         /// RTP timestamp in logical channel sync.
    );
  //@}

  /**@name Member variable access */
  //@{
    /**Set the user name to be used for the local end of any connections. This
       defaults to the logged in user as obtained from the
       SProcess::GetUserName() function.
     */
    void SetLocalUserName(const Tools::SString & name);

    /**Get the user name to be used for the local end of any connections. This
       defaults to the logged in user as obtained from the
       SProcess::GetUserName() function.
     */
    const Tools::SString & GetLocalUserName() const { return localUserName; }

    /**Set calling party number to be used e.g. in Setup.
     */
    void SetCallingPartyNumber(const Tools::SString & number) { m_CallingPartyNumber = number; }

    /**Get calling party number to be used e.g. in Setup.
     */
    const Tools::SString & GetCallingPartyNumber() const { return m_CallingPartyNumber; }

    /**Set the E.164 terminal alias to be used for the gatekeeper registration request.
     */
    void SetE164TerminalAlias(const Tools::SString & alias) { m_E164TerminalAlias = alias; }

    /**Set the H323-Id terminal alias to be used for the gatekeeper registration request.
     */
    void SetH323IdTerminalAlias(const Tools::SString & alias) { m_H323IdTerminalAlias = alias; }

    /**Get the E.164 terminal alias to be used for the gatekeeper registration request.
     */
    const Tools::SString & GetE164TerminalAlias() const { return m_E164TerminalAlias; }

    /**Get the H233-Id terminal alias to be used for the gatekeeper registration request.
     */
    const Tools::SString & GetH323IdTerminalAlias() const { return m_H323IdTerminalAlias; }

    /**Set the name for the sound channel to be used for output.
       If the name is not suitable for use with the SSoundChannel class then
       the function will return false and not change the device.

       This defaults to the value of the SSoundChannel::GetDefaultDevice()
       function.
     */
    virtual bool SetSoundChannelPlayDevice(const Tools::SString & name);

    /**Get the name for the sound channel to be used for output.
       This defaults to the value of the SSoundChannel::GetDefaultDevice()
       function.
     */
    const Tools::SString & GetSoundChannelPlayDevice() const { return soundChannelPlayDevice; }

    /**Set the name for the sound channel to be used for input.
       If the name is not suitable for use with the SSoundChannel class then
       the function will return false and not change the device.

       This defaults to the value of the SSoundChannel::GetDefaultDevice()
       function.
     */
    virtual bool SetSoundChannelRecordDevice(const Tools::SString & name);

    /**Get the name for the sound channel to be used for input.
       This defaults to the value of the SSoundChannel::GetDefaultDevice()
       function.
     */
    const Tools::SString & GetSoundChannelRecordDevice() const { return soundChannelRecordDevice; }

    /**Get the set of listeners (incoming call transports) for this endpoint.
     */
    const H323ListenerList & GetListeners() const { return listeners; }

    /**Get the current capability table for this endpoint.
     */
    const H323Capabilities & GetCapabilities() const { return capabilities; }

    /**Get the endpoint terminal type.
     */
    TerminalTypes GetTerminalType() const { return terminalType; }

    /**Determine if endpoint is terminal type.
     */
    bool IsTerminal() const;

    /**Determine if endpoint is gateway type.
     */
    bool IsGateway() const;

    /**Determine if endpoint is gatekeeper type.
     */
    bool IsGatekeeper() const;

    /**Determine if endpoint is gatekeeper type.
     */
    bool IsMCU() const;

    /** Checks whether a string is a E164 number or not.
        That means the number contains only characters from the set "1234567890*#".

        @param   number  String containing a number
        @return  Boolean (true if it is a E164 number)
    */
    bool IsE164(const Tools::SString & number) const;

    /**Get the default maximum audio delay jitter parameter.
     */
    WORD GetMaxAudioDelayJitter() const { return maxAudioDelayJitter; }

    /**Set the default maximum audio delay jitter parameter.
     */
    void SetMaxAudioDelayJitter(unsigned jitter) { maxAudioDelayJitter = (WORD)jitter; }

    /**Get the initial bandwidth parameter.
     */
    unsigned GetInitialBandwidth() const { return initialBandwidth; }

    /**Get the initial bandwidth parameter.
     */
    void SetInitialBandwidth(unsigned bandwidth) { initialBandwidth = bandwidth; }

    /**Get the UDP port number base for RTP channels.
     */
    WORD GetRtpPortBase() const { return rtpPortBase; }

    /**Get the default timeout for calling another endpoint.
     */
    Tools::STimeInterval GetSignallingChannelCallTimeout() const { return signallingChannelCallTimeout; }

    /**Get the default timeout for master slave negotiations.
     */
    Tools::STimeInterval GetMasterSlaveDeterminationTimeout() const { return masterSlaveDeterminationTimeout; }

    /**Get the default retries for H245 master slave negotiations.
     */
    unsigned GetMasterSlaveDeterminationRetries() const { return masterSlaveDeterminationRetries; }

    /**Get the default timeout for H245 capability exchange negotiations.
     */
    Tools::STimeInterval GetCapabilityExchangeTimeout() const { return capabilityExchangeTimeout; }

    /**Get the default timeout for H245 logical channel negotiations.
     */
    Tools::STimeInterval GetLogicalChannelTimeout() const { return logicalChannelTimeout; }

    /**Get the default timeout for H245 request mode negotiations.
     */
    Tools::STimeInterval GetRequestModeTimeout() const { return logicalChannelTimeout; }

    /**Get the default timeout for H245 round trip delay negotiations.
     */
    Tools::STimeInterval GetRoundTripDelayTimeout() const { return roundTripDelayTimeout; }

#ifdef ROUND_TRIP_DELAY
    /**Get the default rate H245 round trip delay is calculated by connection.
     */
    Tools::STimeInterval GetRoundTripDelayRate() const { return roundTripDelayRate; }

    /**Get the default rate H245 round trip delay is calculated by connection.
     */
    bool ShouldClearCallOnRoundTripFail() const { return clearCallOnRoundTripFail; }

    /**Get bool flag to check if sending H245 round trip delay requests is enabled.
     */
    bool IsSendRoundTripDelay() const { return m_bSendRoundTripDelay; }
#endif // ROUND_TRIP_DELAY

    /**Get the default timeout for Call Transfer Timer CT-T1.
     */
    const STimeInterval & GetCallTransferT1() const { return callTransferT1; }

    /**Get the default timeout for Call Transfer Timer CT-T2.
     */
    const STimeInterval & GetCallTransferT2() const { return callTransferT2; }

    /**Get the default timeout for Call Transfer Timer CT-T3.
     */
    const STimeInterval & GetCallTransferT3() const { return callTransferT3; }

    /**Get the default timeout for Call Transfer Timer CT-T4.
     */
    const STimeInterval & GetCallTransferT4() const { return callTransferT4; }

    /**Get bool flag to check if overlapped sending is enabled.
     */
    bool IsOverlappedSendingEnabled() const { return m_bIsOverlappedSendingEnabled; }

    /**Get bool flag to check if callIdentifier should be used to resolve H.323 client.
     */
    bool UseCallIdentifier() const { return m_bUseCallIdentifier; }

    /**Get bool flag to check if registration workaround for Siemens TA 1100 should be used.
     */
    bool UseTA1100RegWorkaround() const { return (m_uTA1100RegisterTTLSeconds == 0 ? false : true); }

    /**Get seconds for registration time to live workaround for Siemens TA 1100.
     */
    unsigned GetTA1100RegisterTTLSeconds() const { return m_uTA1100RegisterTTLSeconds; }

    /**Get the default timeout for GatekeeperRequest and Gatekeeper discovery.
     */
    Tools::STimeInterval GetGatekeeperRequestTimeout() const { return gatekeeperRequestTimeout; }

    /**Get the default retries for GatekeeperRequest and Gatekeeper discovery.
     */
    unsigned GetGatekeeperRequestRetries() const { return gatekeeperRequestRetries; }

    /**Get the default timeout for RAS protocol transactions.
     */
    Tools::STimeInterval GetRasRequestTimeout() const { return rasRequestTimeout; }

    /**Get the default retries for RAS protocol transations.
     */
    unsigned GetRasRequestRetries() const { return rasRequestRetries; }

    /**Set the default retries for RAS protocol transations.
     */
    void SetRasRequestRetries(unsigned rasReqRetries) { rasRequestRetries = rasReqRetries; }

    /**Get the maximum retries for RAS protocol transations
       before concering the gatekeeper connection as broken.
     */
    unsigned GetMaxRasRequestRetries() const { return rasMaxRequestRetries; }

    /**Set the maximum retries for RAS protocol transations
       before concering the gatekeeper connection as broken.
     */
    void SetMaxRasRequestRetries(unsigned rasMaxReqRetries) { rasMaxRequestRetries = rasMaxReqRetries; }

    /**Get the default registration time to live in seconds.
     */
    unsigned GetRegistrationTTL() const { return rasRegistrationTTL; }

    /**Get the default reregistration timeout in seconds.
     */
    unsigned GetReRegisterTimeout() const { return rasReRegisterTimeout; }

    /**Get the default stack size of listener threads.
     */
    int GetListenerThreadStackSize() const { return listenerThreadStackSize; }

    /**Get the default stack size of signalling channel threads.
     */
    int GetSignallingThreadStackSize() const { return signallingThreadStackSize; }

    /**Get the default stack size of control channel threads.
     */
    int GetControlThreadStackSize() const { return controlThreadStackSize; }

    /**Get the default stack size of logical channel threads.
     */
    int GetChannelThreadStackSize() const { return logicalThreadStackSize; }

    /**Get the default stack size of RAS channel threads.
     */
    int GetRasThreadStackSize() const { return rasThreadStackSize; }

    /**Get the dictionary of <callIdentities, connections>
     */
    H323ConnectionDict& GetCallIdentityDictionary() { return secondaryConenctionsActive; }
  //@}

  /** The following methods must be overloaded by Link Manager implementation
   */
  //@{
    /** Indicates if this endpoint is the H323 endpoint for the remote side 
      */
    virtual bool IsRemoteSideEndpoint() { return false; }

    /** Reads the setting from configuration. 
        Used by LinkManager to trigger the special handling for SwyxLinks to third party gatekeepers
        in comparison to the previous method this one has the link name has parameter and only applies
        to the link with the specified link name.

        @return   Boolean if true the remote server of this link is third party gatekeeper
                  If the given LinkName is empty or not matching to the configuration -> false is returned.
     */
    virtual bool IsRemoteSideLinkToThirdParty(Tools::SString sLinkName) const { return false; }
    
    /** Reads the setting from configuration. 
        Used in LinkManager third party case to decide if the special handling with T.38 call 
        should be used or not.
        Default is yes, if IsRemoteSideLinkToThirdParty() returns true. 
        Otherwise the default is false.

        @return   Boolean according to the configuration.
                  If there is no configuration for the link with the given linkname the default is as follows:
                  yes if IsRemoteSideLinkToThirdParty() for the same link name returns true
                  false if IsRemoteSideLinkToThirdParty() returns false.
     */
    virtual bool UseSpecialT38Handling(Tools::SString sLinkName) const { return false; }

    /** Reads the setting from configuration.
        Used in LinkManager third party case to decide if the workaround for certain cisco gatekeepers should be
        applied or not.
        The workaround is that for overlapped sending the gatekeeper does not respond the setup message 
        with a setup acknowledge. 
     */
    virtual bool WaitForSetupAck(Tools::SString sLinkName) const { return false; }


    /** Reads the setting from configuration. Used in LinkManager third party case to decide 
     */
    virtual bool SuppressUsingE164asH323alias() const { return false; }

  //@}
  
    virtual void WakeUpListener() { }

  protected:

    H323Connection * FindConnectionWithoutLocks(
        const Tools::SString & token     /// Token to identify connection
    );

    // Configuration variables, commonly changed
    Tools::SString     localUserName;
    Tools::SString     m_CallingPartyNumber;
    Tools::SString     m_E164TerminalAlias;
    Tools::SString     m_H323IdTerminalAlias;
    Tools::SString     soundChannelPlayDevice;
    Tools::SString     soundChannelRecordDevice;

    BYTE               t35CountryCode;
    BYTE               t35Extension;
    WORD               manufacturerCode;
    TerminalTypes      terminalType;

    /* Protect against absence of a response to the ctIdentify reqest
       (Transferring Endpoint - Call Transfer with a secondary Call) */
    Tools::STimeInterval      callTransferT1;
    /* Protect against failure of completion of the call transfer operation
       involving a secondary Call (Transferred-to Endpoint) */
    Tools::STimeInterval      callTransferT2;
    /* Protect against failure of the Transferred Endpoint not responding
       within sufficient time to the ctInitiate APDU (Transferring Endpoint) */
    Tools::STimeInterval       callTransferT3;
    /* May optionally operate - protects against absence of a response to the
       ctSetup request (Transferred Endpoint) */
    Tools::STimeInterval       callTransferT4;

    WORD               maxAudioDelayJitter;
    unsigned           initialBandwidth;  // in 100s of bits/sec
#ifdef ROUND_TRIP_DELAY
    bool               clearCallOnRoundTripFail;
    bool               m_bSendRoundTripDelay;
#endif // ROUND_TRIP_DELAY

    bool               m_bIsOverlappedSendingEnabled;
    bool               m_bUseCallIdentifier;
    unsigned           m_uTA1100RegisterTTLSeconds;

    // Some more configuration variables, rarely changed.
    WORD                 rtpPortBase;
    Tools::STimeInterval signallingChannelCallTimeout;
    Tools::STimeInterval masterSlaveDeterminationTimeout;
    unsigned             masterSlaveDeterminationRetries;
    Tools::STimeInterval capabilityExchangeTimeout;
    Tools::STimeInterval logicalChannelTimeout;
    Tools::STimeInterval requestModeTimeout;
    Tools::STimeInterval roundTripDelayTimeout;
#ifdef ROUND_TRIP_DELAY
    Tools::STimeInterval roundTripDelayRate;
#endif // ROUND_TRIP_DELAY
    Tools::STimeInterval gatekeeperRequestTimeout;
    unsigned             gatekeeperRequestRetries;
    WORD                 rasPort;
    Tools::STimeInterval rasRequestTimeout;
    unsigned             rasRequestRetries;
    unsigned             rasMaxRequestRetries;
    unsigned             rasRegistrationTTL;
    unsigned             rasReRegisterTimeout;

    int listenerThreadStackSize;
    int signallingThreadStackSize;
    int controlThreadStackSize;
    int logicalThreadStackSize;
    int rasThreadStackSize;

    // delete connection in the context of the connections cleaner thread
    void OnConnectionDeletion(H323Connection * pConnection);

    // Dynamic variables
    H323ListenerList         listeners;

    H323Capabilities         capabilities;

    Tools::SCriticalSection  m_EndpointGkCritSec;
    H323Gatekeeper         * gatekeeper;
    H323Gatekeeper         * alternativeGatekeeper;
    bool                     m_bInGKDiscovery;

    H323ConnectionDict       connectionsActive;
    Tools::SMutex            connectionsMutex;
    H323ConnectionDict       secondaryConenctionsActive;

    H323ConnectionList       connectionsToBeCleaned;
    H323ConnectionList       connectionsToBeDeleted;
public:
    SH323TcpConnectionFactory&   m_tcpConFactory;
    SH323TlsConnectionFactory&   m_tlsConFactory;

}; // H323EndPoint


#endif // __H323_H