/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/SwyxSrv.root/H323/include/h323neg.h                  $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 3.11.04 13:10                                                $
   $Modtime:: 2.11.04 16:07                                                $
  $Revision:: 23                                                           $
  $Workfile:: h323neg.h                                                    $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __H323NEG_H
#define __H323NEG_H

#include "h323pdu.h"


/**Base class for doing H245 negotiations
 */
class H245Negotiator : public Tools::SConnectionObject, public Tools::SRefCnt
{
  public:
    H245Negotiator(H323EndPoint & endpoint, H323Connection & connection);
    ~H245Negotiator();

    virtual void Delete();

    Tools::SMutex & GetMutex() { return mutex; };

  protected:
    SDECLARE_NOTIFIER(Tools::STimer, H245Negotiator, HandleTimeout);

    H323EndPoint   & endpoint;
    H323Connection & connection;
    Tools::STimer    replyTimer;
    Tools::SMutex    mutex;
};


/**Determine the master and slave on a H245 connection as per H245 section 8.2
 */
class H245NegMasterSlaveDetermination : public H245Negotiator
{
  public:
    H245NegMasterSlaveDetermination(H323EndPoint & endpoint, H323Connection & connection);

    bool Start();
    bool HandleIncoming(const H245_MasterSlaveDetermination & pdu);
    bool HandleAck(const H245_MasterSlaveDeterminationAck & pdu);
    bool HandleReject(const H245_MasterSlaveDeterminationReject & pdu);
    bool HandleRelease(const H245_MasterSlaveDeterminationRelease & pdu);
    void HandleTimeout(Tools::STimer &, INT);

    bool IsMaster() const     { return status == e_DeterminedMaster; }
    bool IsDetermined() const { return state == e_Idle && status != e_Indeterminate; }

  protected:
    bool Restart();

    enum States {
      e_Idle, e_Outgoing, e_Incoming,
      e_NumStates
    } state;
//#if STRACING
    static const char * const StateNames[e_NumStates];
    friend std::ostream & operator<<(std::ostream & o, States s) { return o << StateNames[s]; }
//#endif

    DWORD    determinationNumber;
    unsigned retryCount;

    enum MasterSlaveStatus {
      e_Indeterminate, e_DeterminedMaster, e_DeterminedSlave,
      e_NumStatuses
    } status;
//#if STRACING
    static const char * const StatusNames[e_NumStatuses];
    friend std::ostream & operator<<(std::ostream & o , MasterSlaveStatus s) { return o << StatusNames[s]; }
//#endif
};


/**Exchange capabilities on a H245 connection as per H245 section 8.3
 */
class H245NegTerminalCapabilitySet : public H245Negotiator
{
  public:
    H245NegTerminalCapabilitySet(H323EndPoint & endpoint, H323Connection & connection);

    bool Start();
    bool HandleIncoming(const H245_TerminalCapabilitySet & pdu);
    bool HandleAck(const H245_TerminalCapabilitySetAck & pdu);
    bool HandleReject(const H245_TerminalCapabilitySetReject & pdu, bool bIsError=true);
    bool HandleRelease(const H245_TerminalCapabilitySetRelease & pdu);
    void HandleTimeout(Tools::STimer &, INT);
    bool SendEmptyTerminalCapabilitySet();
    void Reset();

    bool HasSentCapabilities() const { return state == e_Sent; }
    bool HasReceivedCapabilities() const { return receivedCapabilites; }

  protected:
    enum States {
      e_Idle, e_InProgress, e_Sent,
      e_NumStates
    } state;
//#if STRACING
    static const char * const StateNames[e_NumStates];
    friend std::ostream & operator<<(std::ostream & o, States s) { return o << StateNames[s]; }
//#endif

    unsigned inSequenceNumber;
    unsigned outSequenceNumber;

    bool receivedCapabilites;
};


/**Logical Channel signalling on a H245 connection as per H245 section 8.4
 */
class H245NegLogicalChannel : public H245Negotiator
{
  public:
    H245NegLogicalChannel(H323EndPoint & endpoint,
                          H323Connection & connection,
                          const H323ChannelNumber & channelNumber);
    H245NegLogicalChannel(H323EndPoint & endpoint,
                          H323Connection & connection,
                          H323Channel & channel);
    ~H245NegLogicalChannel();

    virtual void Delete();

    bool Open(const H323Capability & capability, unsigned sessionID);
    bool Close();
    bool HandleOpen(const H245_OpenLogicalChannel & pdu);
    bool HandleOpenAck(const H245_OpenLogicalChannelAck & pdu);
    bool HandleOpenConfirm(const H245_OpenLogicalChannelConfirm & pdu);
    bool HandleReject(const H245_OpenLogicalChannelReject & pdu);
    bool HandleClose(const H245_CloseLogicalChannel & pdu);
    bool HandleCloseAck(const H245_CloseLogicalChannelAck & pdu);
    bool HandleRequestClose(const H245_RequestChannelClose & pdu);
    bool HandleRequestCloseAck(const H245_RequestChannelCloseAck & pdu);
    bool HandleRequestCloseReject(const H245_RequestChannelCloseReject & pdu);
    bool HandleRequestCloseRelease(const H245_RequestChannelCloseRelease & pdu);
    void HandleTimeout(Tools::STimer &, INT);

    H323Channel * GetChannel();

    bool IsReleased() const { return state == e_Released; }

  protected:
    H323Channel * channel;

    H323ChannelNumber channelNumber;

    enum States {
      e_Released,
      e_AwaitingEstablishment,
      e_Established,
      e_AwaitingRelease,
      e_AwatingConfirmation,
      e_AwaitingResponse,
      e_NumStates
    } state;
//#if STRACING
    static const char * const StateNames[e_NumStates];
    friend std::ostream & operator<<(std::ostream & o, States s) { return o << StateNames[s]; }
//#endif

  private:
    void Release();
};


//PDICTIONARY(H245LogicalChannelDict, H323ChannelNumber, H245NegLogicalChannel);
typedef stdext::hash_map<H323ChannelNumber, H245NegLogicalChannel*> H245LogicalChannelDict;


/**Dictionary of all Logical Channels
 */
class H245NegLogicalChannels : public H245Negotiator
{
  public:
    H245NegLogicalChannels(H323EndPoint & endpoint, H323Connection & connection);
    virtual ~H245NegLogicalChannels();

    void Add(H323Channel & channel);

    bool Open(const H323Capability & capability, unsigned sessionID, H323ChannelNumber & rNumber);
    bool Close(unsigned channelNumber, bool fromRemote);
    bool Erase(H323Channel * channel);
    void CloseAllChannels();

    bool HandleOpen(const H245_OpenLogicalChannel & pdu);
    bool HandleOpenAck(const H245_OpenLogicalChannelAck & pdu);
    bool HandleOpenConfirm(const H245_OpenLogicalChannelConfirm & pdu);
    bool HandleReject(const H245_OpenLogicalChannelReject & pdu);
    bool HandleClose(const H245_CloseLogicalChannel & pdu);
    bool HandleCloseAck(const H245_CloseLogicalChannelAck & pdu);
    bool HandleRequestClose(const H245_RequestChannelClose & pdu);
    bool HandleRequestCloseAck(const H245_RequestChannelCloseAck & pdu);
    bool HandleRequestCloseReject(const H245_RequestChannelCloseReject & pdu);
    bool HandleRequestCloseRelease(const H245_RequestChannelCloseRelease & pdu);

    H323ChannelNumber GetNextChannelNumber();
    int GetSize() const { return channels.size(); }
//  H323Channel * GetChannelAt(int i) const { return channels.GetDataAt(i).GetChannel(); }
    H323Channel * GetChannelAt(int i); // Swyx version using STL hash_map !
    H323Channel * FindChannel(unsigned channelNumber, bool fromRemote);
    H245NegLogicalChannel * FindNegLogicalChannel(unsigned channelNumber, bool fromRemote);
    H323Channel * FindChannelBySession(unsigned rtpSessionId, bool fromRemote);
    bool IsChannelInMap(H323Channel * pChannel);

  protected:
    H323ChannelNumber      lastChannelNumber;
    H245LogicalChannelDict channels;
};


/**Request mode change as per H245 section 8.9
 */
class H245NegRequestMode : public H245Negotiator
{
  public:
    H245NegRequestMode(H323EndPoint & endpoint, H323Connection & connection);

    bool StartRequest(const H323Capability & rCapability);
    bool HandleRequest(const H245_RequestMode & pdu);
    bool HandleAck(const H245_RequestModeAck & pdu);
    bool HandleReject(const H245_RequestModeReject & pdu);
    bool HandleRelease(const H245_RequestModeRelease & pdu);
    void HandleTimeout(Tools::STimer &, INT);

  protected:
    bool awaitingResponse;
    unsigned inSequenceNumber;
    unsigned outSequenceNumber;
};


/**Request mode change as per H245 section 8.9
 */
class H245NegRoundTripDelay : public H245Negotiator
{
  public:
    H245NegRoundTripDelay(H323EndPoint & endpoint, H323Connection & connection);

    bool StartRequest();
    bool HandleRequest(const H245_RoundTripDelayRequest & pdu);
    bool HandleResponse(const H245_RoundTripDelayResponse & pdu);
    void HandleTimeout(Tools::STimer &, INT);

    Tools::STimeInterval GetRoundTripDelay() const { return roundTripTime; }
#ifdef ROUND_TRIP_DELAY
    bool IsRemoteOffline() const { return retryCount == 0; }
#endif // ROUND_TRIP_DELAY

  protected:
    bool                 awaitingResponse;
    unsigned             sequenceNumber;
    Tools::STimeInterval tripStartTime;
    Tools::STimeInterval roundTripTime;
#ifdef ROUND_TRIP_DELAY
    unsigned             retryCount;
#endif // ROUND_TRIP_DELAY
};


#endif // __H323NEG_H
