/*---------------------------------------------------------------------------

   Swyx Solutions GmbH
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Solutions GmbH are
  Copyright (C) 2002-2005 Swyx Solutions GmbH.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323 T.38
 
    Comments: H.323 T.38 logical channel establishment

  ---------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/h323t38.h                 $
    $Author:: Ulrich.findeisen                                             $
      $Date:: 14.06.05 17:48                                               $
   $Modtime:: 8.06.05 18:36                                                $
  $Revision:: 2                                                            $
  $Workfile:: h323t38.h                                                    $
$NoKeywords: $
----------------------------------------------------------------------------*/

#ifndef __H323T38_H
#define __H323T38_H


//-----------------------------------------------------------------------------

/**This class includes the (default) parameters of a T38 standard capability.
 */
class H323_T38CapParams : public Tools::SObject
{
  public:

    /// T38 rate management
    enum T38RateManagement
    {
        T38LocalTCF = 0,    /// recommended for TCP
        T38TransferredTCF   /// recommended for UDP
    };

    /// T38 UDP option: Error correction
    enum T38UDPErrCorrection
	{
        T38UDPFEC = 0,
        T38UDPRedundancy
    };

    H323_T38CapParams();
    virtual ~H323_T38CapParams();

    H323_T38CapParams & operator=(const H323_T38CapParams & params);

    /// Print parameters of T38 capability.
    virtual void PrintOn(std::ostream & strm) const;

    /// Version of T38 protocol (Version 0, the default, refers to T.38 (1998))
    unsigned                m_uVersion;

    /// Rate Management
    T38RateManagement       m_RateManagement;

    /// Maximum bit rate, units 100 bit/s (usually 144 = 14400 bit/s)
    unsigned                m_uMaxBitRate;

    /// Bit option: Fill Bit Removal
    bool                    m_bFillBitRemoval;

    /// Bit option: Transcoding JBIG
    bool                    m_bTranscodingJBIG;

    /// Bit option: Transcoding MMR
    bool                    m_bTranscodingMMR;

    /// UDP option: Maximum buffer size
    unsigned                m_uUDPMaxBufferSize;

    /// UDP option: Maximum datagram size
    unsigned                m_uUDPMaxDatagramSize;

    /// UDP option: Error correction
    T38UDPErrCorrection     m_UDPErrCorrection;
};


/**This class describes the T.38 standard capability.
 */
class H323_T38Capability : public H323DataCapability
{
  public:
  /**@name Construction */
  //@{
    enum TransportMode {
      e_UDP,
      e_DualTCP,        // two unidirectional channels
      e_SingleTCP,      // one bi-directional channel
      NumTransportModes
    };

    /**Create a new capability.
     */
    H323_T38Capability(
      TransportMode mode
    );
  //@}

  /**@name Overrides from class SObject */
  //@{
    /**Compare two capability instances. This compares the main and sub-types
       of the capability.
     */
    Comparison Compare(const Tools::SObject & obj) const;

    /**Create a copy of the object.
      */
    virtual Tools::SObject * Clone() const;

    /**Print to stream.
       Standard output function from SObject implements stream << operator.
      */
    virtual void PrintOn(
      std::ostream & strm    /// Stream to output to.
    ) const;
  //@}

  /**@name Identification functions */
  //@{
    /**Get the sub-type of the capability. This is a code dependent on the
       main type of the capability.

       This returns the e_t38fax enum value from the protocol ASN
       H245_DataApplicationCapability_application class.
     */
    virtual unsigned GetSubType() const;

    /**Check if mode of capability is bidirectional.
     */
    virtual bool IsBidirectional() const { return mode == e_SingleTCP; };

    /**Get the name of the media data format this class represents.
     */
    virtual Tools::SString GetFormatName() const;
  //@}

  /**@name Operations */
  //@{
    /**Create the channel instance, allocating resources as required.
     */
    virtual H323Channel * CreateChannel(
      H323Connection & connection,    /// Owner connection for channel
      H323Channel::Directions dir,    /// Direction of channel
      unsigned sessionID,             /// Session ID for RTP channel
      const H245_H2250LogicalChannelParameters * param
                                      /// Parameters for channel
    ) const;
  //@}

  /**@name Protocol manipulation */
  //@{
    /**This function is called whenever and outgoing TerminalCapabilitySet
       or OpenLogicalChannel PDU is being constructed for the control channel.
       It allows the capability to set the PDU fields from information in
       members specific to the class.

       The default behaviour sets the pdu and calls OnSendingPDU with a
       H245_DataProtocolCapability parameter.
     */
    virtual bool OnSendingPDU(
      H245_DataApplicationCapability & pdu
    ) const;

    /**This function is called whenever and outgoing RequestMode
       PDU is being constructed for the control channel. It allows the
       capability to set the PDU fields from information in members specific
       to the class.

       The default behaviour sets the pdu and calls OnSendingPDU with a
       H245_DataProtocolCapability parameter.
     */
    virtual bool OnSendingPDU(
      H245_DataMode & pdu  /// PDU to set information on
    ) const;

    /**This function is called whenever and outgoing PDU is being constructed
       for the control channel. It allows the capability to set the PDU fields
       from information in members specific to the class.

       The default behaviour sets tcp or udp as required.
     */
    virtual bool OnSendingPDU(
      H245_DataProtocolCapability & proto,  /// PDU to set information on
      H245_T38FaxProfile & profile          /// PDU to set information on
    ) const;

    /**This function is called whenever and incoming TerminalCapabilitySet
       or OpenLogicalChannel PDU has been used to construct the control
       channel. It allows the capability to set from the PDU fields,
       information in members specific to the class.

       The default behaviour gets the data rate field from the PDU.
     */
    virtual bool OnReceivedPDU(
      const H245_DataApplicationCapability & pdu  /// PDU to set information on
    );

    /** Set parameters of T38 capability. */
    void SetT38CapParams(const H323_T38CapParams & T38CapParams) { m_T38CapParams = T38CapParams; }

    /** Get parameters of T38 capability. */
    H323_T38CapParams GetT38CapParams() const { return m_T38CapParams; }
  //@}

    TransportMode GetTransportMode() const { return mode; }

  protected:
    TransportMode mode;

    /// Parameters of a T38 standard capability
    H323_T38CapParams   m_T38CapParams;
};

//-----------------------------------------------------------------------------

/**This class describes the T.38 logical channel.
 */
class H323_T38Channel : public H323DataChannel
{
  public:
  /**@name Construction */
  //@{
    /**Create a new channel.
     */
    H323_T38Channel(
      H323Connection & connection,            /// Connection to endpoint for channel
      const H323Capability & capability,      /// Capability channel is using
      Directions direction,                   /// Direction of channel
      unsigned sessionID,                     /// Session ID for channel
      H323_T38Capability::TransportMode mode  /// Transport mode
    );
    ~H323_T38Channel();
  //@}

  /**@name Overrides from class H323Channel */
  //@{
    /**Fill out the OpenLogicalChannel PDU for the particular channel type.
     */
    virtual bool OnSendingPDU(
      H245_OpenLogicalChannel & openPDU  /// Open PDU to send. 
    ) const;

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.

       The default makes sure the parameters are compatible and passes on
       the PDU to the rtp session.
     */
    virtual bool OnReceivedPDU(
      const H245_OpenLogicalChannel & pdu,    /// Open PDU
      unsigned & errorCode                    /// Error code on failure
    );

    /**This is called to clean up any threads on connection termination.
     */
    virtual void CleanUpOnTermination();

    /**Handle channel data reception.

       This is called by the thread started by the Start() function and is
       a loop reading from the transport and calling HandlePacket() for each
       PDU read.
      */
    virtual void Receive();

    /**Handle channel data transmission.

       This is called by the thread started by the Start() function and is
       typically a loop reading from the codec and writing to the transport
       (eg an RTP_session).
      */
    virtual void Transmit();

    /** Get type of codec.
     */
    virtual H323Codec::CodecType GetCodecType() const { return H323Codec::T38; }

    /** Get framesize in milliseconds.
     */
    virtual unsigned long GetFramesize() const { return 0; }
  //@}

  protected:
    bool              usesTCP;
};

//-----------------------------------------------------------------------------

#endif // __H323T38_H
