/*---------------------------------------------------------------------------

   Swyx Solutions GmbH
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Solutions GmbH are
  Copyright (C) 2000-2005 Swyx Solutions GmbH.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Q.931
 
    Comments: Q.931 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/q931.h                    $
    $Author:: Martin.hueser                                                $
      $Date:: 20.12.05 18:34                                               $
   $Modtime:: 16.12.05 16:35                                               $
  $Revision:: 3                                                            $
  $Workfile:: q931.h                                                       $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __Q931_H
#define __Q931_H

#include <stlib.h>
#include "h225.h"


// Q.931 PDU

class Q931 : public Tools::SObject
{
  public:
    enum MsgTypes {
      NationalEscapeMsg  = 0x00,
      AlertingMsg        = 0x01,
      CallProceedingMsg  = 0x02,
      ProgressMsg        = 0x03,
      SetupMsg           = 0x05,
      ConnectMsg         = 0x07,
      SetupAckMsg        = 0x0d,
      ConnectAckMsg      = 0x0f,
      ReleaseCompleteMsg = 0x5a,
      FacilityMsg        = 0x62,
      NotifyMsg          = 0x6e,
      StatusEnquiryMsg   = 0x75,
      InformationMsg     = 0x7b,
      StatusMsg          = 0x7d
    };

    class ProgressIndicator
    {
      public:
        enum Location  // Q.931, table 4-20
        {
            e_User                            = 0x0, // User
            e_PrivateNetworkServingLocalUser  = 0x1, // Private network serving the local user
            e_PublicNetworkServingLocalUser   = 0x2, // Public network serving the local user
            e_TransitNetwork                  = 0x3, // Transit network (Note 2)
            e_PublicNetworkServingRemoteUser  = 0x4, // Public network serving the remote user
            e_PrivateNetworkServingRemoteUser = 0x5, // Private network serving the remote user
            e_NetworkBeyondInterworkingPoint  = 0x6  // Network beyond the interworking point 
        };

        enum ProgressDescription  // Q.931, table 4-20
        {
            e_CallIsNotEndToEndIsdn       = 1, // Call is not end-to-end ISDN; further call progress information may be available in-band
            e_DestinationAddressIsNonIsdn = 2, // Destination address is non ISDN
            e_OriginationAddressIsNonIsdn = 3, // Origination address is non ISDN
            e_CallHasReturnedToTheIsdn    = 4, // Call has returned to the ISDN
            e_InterworkingHasOccurred     = 5, // Interworking has occurred and has resulted in a telecommunication service change (Note 5)
            e_InBandInformationAvailable  = 8  // In-band information or an appropriate pattern is now available.
        };

        ProgressIndicator::ProgressIndicator() { bIndicatorIsSet = false; }

        bool                  bIndicatorIsSet;
        Location              location;
        ProgressDescription   description;
    };

    class NotificationIndicator
    {
      public:
        enum NotificationDescription
        {
            // Q.931, table 4-19
            e_UserSuspended            = 0x0,  // User suspended
            e_UserResumed              = 0x1,  // User resumed
            e_BearerServiceChange      = 0x2,  // Bearer service change

            // Swyx proprietary values
            e_SwyxStopDtmfDetection    = 0x7e, // Stop DTMF detection
            e_SwyxStartDtmfDetection   = 0x7f  // Start DTMF detection
        };

        NotificationIndicator::NotificationIndicator() { }

        NotificationDescription   description;
    };

    class Cause
    {
    public:
        enum Causes  // ETSI 300 102-1, table 4.13
        {
            e_UnknownCause                     = 0x0000,
            
            // class 0        
            e_UnallocatedNumber                = 0x0001,
            e_NoRouteToSpecifiedTransitNetwork = 0x0002,
            e_NoRouteToDestination             = 0x0003,
            e_ChannelUnacceptable              = 0x0006,
            e_CallAwarded                      = 0x0007,

            // class 1
            e_NormalCallClearing               = 0x0010,
            e_UserBusy                         = 0x0011,
            e_NoUserResponding                 = 0x0012,
            e_NoAnswerFromUser                 = 0x0013,
            e_CallRejected                     = 0x0015,
            e_NumberChanged                    = 0x0016,
            e_NonSelectedUserClearing          = 0x001A,
            e_DestinationOutOfOrder            = 0x001B,
            e_InvalidNumberFormat              = 0x001C,
            e_FacilityRejected                 = 0x001D,
            e_ResponsetoStatusEnquiry          = 0x001E,
            e_Normal                           = 0x001F,

            // class 2
            e_NoChannelAvailable               = 0x0022,
            e_NetworkOutOfOrder                = 0x0026,
            e_TemporaryFailure                 = 0x0029,
            e_NetworkCongestion                = 0x002A,
            e_AccessInformationDiscarded       = 0x002B,
            e_RequestedChannelNotAvailable     = 0x002C,
            e_ResourcesUnavailable             = 0x002F,

            // class 3
            e_QualityOfServiceUnavailable      = 0x0031,
            e_RequestedFacilityNotSubscribed   = 0x0032,
            e_BearerCapabilityNotAuthorized    = 0x0039,
            e_BearerCapabilityNotPresentlyAvailable = 0x003A,
            e_ServiceOrOptionNotAvailable      = 0x003F,

            // class 4
            e_BearerCapabilityNotImplemented   = 0x0041,
            e_ChannelTypeNotImplemented        = 0x0042,
            e_RequestedFacilityNotImplemented  = 0x0045,
            e_OnlyRestrictedDigitalInformationBearerCapabilityIsAvailable = 0x0046,
            e_ServiceOrOptionNotImplemented    = 0x004F,

            // class 5
            e_InvalidCallReferenceValue        = 0x0051,
            e_IdentifiedChannelDoesNotExist    = 0x0052,
            e_SuspendedCallExists              = 0x0053,
            e_CallIdentityInUse                = 0x0054,
            e_NoCallSuspended                  = 0x0055,
            e_CallHavingTheRequestedCallIdentityHasBeenCleared = 0x0056,
            e_IncompatibleDestination          = 0x0058,
            e_InvalidTransitNetworkSelection   = 0x005B,
            e_InvalidMessage                   = 0x005F,

            // class 6
            e_MandatoryInformationElementIsMissing           = 0x0060,
            e_MessageTypeNonExistentOrNotImplemented         = 0x0061,
            e_MessageNotCompatibleWithCallStateOrNonExistent = 0x0062,
            e_InformationElementNonExistentOrNotImplemented  = 0x0063,
            e_InvalidInformationElementContents              = 0x0064,
            e_MessageNotCompatibleWithCallState              = 0x0065,
            e_RecoveryOnTimerExpiry                          = 0x0066,
            e_ProtocolError                                  = 0x006F,

            // class 7
            e_Interworking                                   = 0x007F,
        };
        
        Cause::Cause()          : cause(e_UnknownCause) {}
        Cause::Cause(Causes c)  : cause(c) {}

        Tools::SString GetCauseString() const;
        
        Causes   cause;
    };
    
    Q931();
    ~Q931();
    Q931 & operator=(const Q931 & other);

    void BuildFacility(int callRef, bool fromDest);
    void BuildCallProceeding(int callRef);
    void BuildAlerting(int callRef, const Q931::ProgressIndicator & ProgressInd);
    void BuildSetup(int callRef, const Tools::SString & name, bool bSendingComplete);
    void BuildSetupAck(int callRef);
    void BuildProgress(int callRef, bool fromDest, const Q931::ProgressIndicator & ProgressInd, const Q931::Cause & cause);
    void BuildNotify(int callRef, bool fromDest, const Q931::NotificationIndicator & NotificationInd);
    void BuildConnect(int callRef, const Tools::SString & name);
    void BuildInformation(int callRef, bool fromDest);
    void BuildStatus(int callRef, bool fromDest);
    void BuildReleaseComplete(int callRef, bool fromDest, const Q931::Cause & cause);

    bool Decode(const Tools::SByteVector & data);
    bool Encode(Tools::SByteVector & data) const;

    void PrintOn(std::ostream & strm) const;
    Tools::SString GetMessageTypeName() const;

    static unsigned GenerateCallReference();
    unsigned GetCallReference() const { return callReference; }
    bool IsFromDestination() const { return fromDestination; }
    MsgTypes GetMessageType() const { return messageType; }

    enum InformationElementCodes
    {
        BearerCapabilityIE                      = 0x04,
        CauseIE                                 = 0x08,
        CallIdentityIE                          = 0x10,
        CallStateIE                             = 0x14,
        ChannelIdentificationIE                 = 0x18,
        FacilityIE                              = 0x1c,
        ProgressIndicatorIE                     = 0x1e,
        NetworkSpecificFacilitiesIE             = 0x20,
        NotificationIndicatorIE                 = 0x27,
        DisplayIE                               = 0x28,
        DateTimeIE                              = 0x29,
        KeypadFacilityIE                        = 0x2c,
        SignalIE                                = 0x34,
        InformationRateIE                       = 0x40,
        EndToEndTransitDelayIE                  = 0x42,
        TransitDelaySelectionAndIndicationIE    = 0x43,
        PacketLayerBinaryParametersIE           = 0x44,
        PacketLayerWindowSizeIE                 = 0x45,
        PacketSizeIE                            = 0x46,
        ClosedUserGroupIE                       = 0x47,
        ReverseChargeIndicationIE               = 0x4a,
        ConnectedNumberIE                       = 0x4c,
        CallingPartyNumberIE                    = 0x6c,
        CallingPartySubaddressIE                = 0x6d,
        CalledPartyNumberIE                     = 0x70,
        CalledPartySubaddressIE                 = 0x71,
        RedirectingNumberIE                     = 0x74,
        TransitNetworkSelectionIE               = 0x78,
        RestartIndicatorIE                      = 0x79,
        LowLayerCompatibilityIE                 = 0x7c,
        HighLayerCompatibilityIE                = 0x7d,
        UserUserIE                              = 0x7e,
        SendingCompleteIE                       = 0xa1
    };
    bool HasIE(InformationElementCodes ie) const;
    Tools::SByteVector GetIE(InformationElementCodes ie) const;
    void SetIE(InformationElementCodes ie, const Tools::SByteVector & userData);

    void SetProgressIndicator(const Q931::ProgressIndicator & ProgressInd);
    bool GetProgressIndicator(Q931::ProgressIndicator & ProgressInd) const;
    Tools::SString GetProgressIndicatorInfo(Q931::ProgressIndicator & ProgressInd) const;

    void SetNotificationIndicator(const Q931::NotificationIndicator & NotificationInd);
    bool GetNotificationIndicator(Q931::NotificationIndicator & NotificationInd) const;
    Tools::SString GetNotificationIndicatorInfo(Q931::NotificationIndicator & NotificationInd) const;

    enum SignalInfo {
      SignalDialToneOn,
      SignalRingBackToneOn,
      SignalInterceptToneOn,
      SignalNetworkCongestionToneOn,
      SignalBusyToneOn,
      SignalConfirmToneOn,
      SignalAnswerToneOn,
      SignalCallWaitingTone,
      SignalOffhookWarningTone,
      SignalPreemptionToneOn,
      SignalTonesOff = 0x3f,
      SignalAlertingPattern0 = 0x40,
      SignalAlertingPattern1,
      SignalAlertingPattern2,
      SignalAlertingPattern3,
      SignalAlertingPattern4,
      SignalAlertingPattern5,
      SignalAlertingPattern6,
      SignalAlertingPattern7,
      SignalAlertingOff = 0x4f,
      SignalErrorInIE = 0x100
    };
    void SetSignalInfo(SignalInfo value);
    SignalInfo GetSignalInfo() const;
    Tools::SString GetSignalInfoName(SignalInfo value) const;

    void SetDisplayName(const Tools::SString & name);
    Tools::SString GetDisplayName() const;

    void SetKeypadFacility(const Tools::SString & number);
    Tools::SString GetKeypadFacility() const;

    void SetCallingPartyNumber(
      const Tools::SString & number, // Number string
      unsigned plan = 1,             // 1 = ISDN/Telephony numbering system
      unsigned type = 0,             // 0 = Unknown number type
      int presentation = -1,         // 0 = presentation allowed, -1 = no octet3a
      int screening = -1             // 0 = user provided, not screened
    );
    bool GetCallingPartyNumber(
      Tools::SString & number,        // Number string
      unsigned * plan = NULL,         // ISDN/Telephony numbering system
      unsigned * type = NULL,         // Number type
      unsigned * presentation = NULL, // Presentation indicator
      unsigned * screening = NULL,    // Screening indicator
      unsigned defPresentation = 0,   // Default value if octet3a not present
      unsigned defScreening = 0       // Default value if octet3a not present
    ) const;

    void SetCalledPartyNumber(
      const Tools::SString & number, // Number string
      unsigned plan = 1,             // 1 = ISDN/Telephony numbering system
      unsigned type = 0              // 0 = Unknown number type
    );
    bool GetCalledPartyNumber(
      Tools::SString & number,  // Number string
      unsigned * plan = NULL,   // ISDN/Telephony numbering system
      unsigned * type = NULL    // Number type
    ) const;

    void SetCause(const Q931::Cause & cause);
    bool GetCause(Cause & cause) const;

    void SetFacilityIE(
      const Tools::SByteVector & data // Data of Facility info element
    );
    bool GetFacilityIE(
      Tools::SByteVector & data       // Data of Facility info element
    ) const;

  protected:
    void ClearIEs();

    unsigned callReference;
    bool fromDestination;
    unsigned protocolDiscriminator;
    MsgTypes messageType;
    bool     m_bCopy; // Little hack!!! -> avoid clearing the 'IEs' if the copy constructor was called

//  PDICTIONARY(InternalInformationElements, POrdinalKey, PBYTEArray);
    typedef stdext::hash_map<int, Tools::SByteVector *>  InternalInformationElements;
    mutable InternalInformationElements informationElements;
};


#endif // __Q931_H
