/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: RTP
 
    Comments: RTP protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/rtp.h                     $
    $Author:: Martin.hueser                                                $
      $Date:: 20.12.05 18:34                                               $
   $Modtime:: 16.12.05 16:35                                               $
  $Revision:: 2                                                            $
  $Workfile:: rtp.h                                                        $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __RTP_H
#define __RTP_H


#include <stlib.h>

#include "codecs.h"

class H225_RTPSession;

class H245_H2250LogicalChannelParameters;
class H245_H2250LogicalChannelAckParameters;

class H323Connection;


//-----------------------------------------------------------------------------
// Real Time Protocol - IETF RFC1889 and RFC1890

/**This class represents a RTP data frame.
 */
class RTP_DataFrame : public Tools::SSimpleArray<BYTE>
{
  public:
    RTP_DataFrame(int payloadSize = 2048);

    enum {
      ProtocolVersion = 2,
      MinHeaderSize = 12
    };

    enum PayloadTypes {
      PCMU,
      FS1016,
      G721,
      GSM,
      G7231,
      DVI4_8k,
      DVI4_16k,
      LPC,
      PCMA,
      G722,
      L16_Stereo,
      L16_Mono,
      G723,
      CN,
      MPA,
      G728,
      DVI4_11k,
      DVI4_22k,
      G729,
      CelB = 25,
      JPEG,
      H261 = 31,
      MPV,
      MP2T,
      H263,
      DynamicBase = 96,
      MaxPayloadType = 127
    };

    unsigned GetVersion() const { return (m_pArray[0]>>6)&3; }

    bool GetExtension() const   { return (m_pArray[0]&0x10) != 0; }
    void SetExtension(bool ext);

    bool GetMarker() const { return (m_pArray[1]&0x80) != 0; }
    void SetMarker(bool m);

    unsigned GetPayloadType() const { return  m_pArray[1]&0x7f; }
    void     SetPayloadType(unsigned t);

    WORD GetSequenceNumber() const { return *(Tools::SUInt16b *)&m_pArray[2]; }
    void SetSequenceNumber(WORD n) { *(Tools::SUInt16b *)&m_pArray[2] = n; }

    DWORD GetTimestamp() const  { return *(Tools::SUInt32b *)&m_pArray[4]; }
    void  SetTimestamp(DWORD t) { *(Tools::SUInt32b *)&m_pArray[4] = t; }

    DWORD GetSyncSource() const  { return *(Tools::SUInt32b *)&m_pArray[8]; }
    void  SetSyncSource(DWORD s) { *(Tools::SUInt32b *)&m_pArray[8] = s; }

    int    GetContribSrcCount() const { return m_pArray[0]&0xf; }
    DWORD  GetContribSource(int idx) const;
    void   SetContribSource(int idx, DWORD src);

    int    GetHeaderSize()     const { return 12 + 4*GetContribSrcCount(); }

    int    GetPayloadSize() const { return payloadSize; }
    bool   SetPayloadSize(int sz);
    BYTE * GetPayloadPtr() const { return (BYTE *)(m_pArray+GetHeaderSize()); }

  protected:
    int payloadSize;
};


class RTP_ControlFrame : public Tools::SSimpleArray<BYTE>
{
  public:
    RTP_ControlFrame(int payloadSize = 2048);

    unsigned GetVersion() const { return (BYTE)m_pArray[0]>>6; }
    unsigned GetCount()   const { return (BYTE)m_pArray[0]&0x1f; }

    enum PayloadTypes {
      e_SenderReport = 200,
      e_ReceiverReport,
      e_SourceDescription,
      e_Goodbye,
      e_ApplDefined
    };

    unsigned GetPayloadType() const { return (BYTE)m_pArray[1]; }
    void     SetPayloadType(unsigned t);

    int    GetPayloadSize() const { return 4*(*(Tools::SUInt16b *)&m_pArray[2]); }
    void   SetPayloadSize(int sz);

    BYTE * GetPayloadPtr() const { return (BYTE *)(m_pArray+4); }

#pragma pack(1)
    struct ReceptionReport {
      Tools::SUInt32b ssrc;      /* data source being reported */
      BYTE            fraction;  /* fraction lost since last SR/RR */
      BYTE            lost[3];   /* cumulative number of packets lost (signed!) */
      Tools::SUInt32b last_seq;  /* extended last sequence number received */
      Tools::SUInt32b jitter;    /* interarrival jitter */
      Tools::SUInt32b lsr;       /* last SR packet from this source */
      Tools::SUInt32b dlsr;      /* delay since last SR packet */

      unsigned GetLostPackets() const { return (lost[0]<<16U)+(lost[1]<<8U)+lost[2]; }
    };

    struct SenderReport {
      Tools::SUInt32b ssrc;      /* source this RTCP packet refers to */
      Tools::SUInt32b ntp_sec;   /* NTP timestamp */
      Tools::SUInt32b ntp_frac;
      Tools::SUInt32b rtp_ts;    /* RTP timestamp */
      Tools::SUInt32b psent;     /* packets sent */
      Tools::SUInt32b osent;     /* octets sent */ 
    };

    struct SourceDescription {
      Tools::SUInt32b src;       /* first SSRC/CSRC */
      struct Item {
        BYTE type;        /* type of SDES item (rtcp_sdes_type_t) */
        BYTE length;      /* length of SDES item (in octets) */
        char data[1];     /* text, not zero-terminated */
      } item[1];          /* list of SDES items */
    };
#pragma pack()
};


/**This class is for encpsulating the IETF Real Time Protocol interface.
 */
class RTP_Session : public Tools::SObject
{
  public:
    enum {
      DefaultAudioSessionID = 1,
      DefaultVideoSessionID = 2,
      DefaultDataSessionID = 3      
    };

  /**@name Construction */
  //@{
    /**Create a new channel.
     */
    RTP_Session(
      const H323Connection & connection, /// Owner of the RTP session
      unsigned id                        /// Session ID for RTP channel
    );
  //@}

  /**@name Operations */
  //@{
    /**Read a data frame from the RTP channel.
       Any control frames received are dispatched to callbacks and are not
       returned by this function. It will block until a data frame is
       available or an error occurs.
      */
    virtual bool ReadData(RTP_DataFrame & frame) = 0;

    /**Write a data frame from the RTP channel.
      */
    virtual bool WriteData(RTP_DataFrame & frame) = 0;

    /**Write a control frame from the RTP channel.
      */
    virtual bool WriteControl(RTP_ControlFrame & frame) = 0;

    /**Close down the RTP session.
      */
    virtual void Close(
      bool reading    /// Closing the read side of the session
    ) = 0;

    /**Fill out the OpenLogicalChannel PDU for the particular channel type.
     */
    virtual bool OnSendingPDU(
      H245_H2250LogicalChannelParameters & param, /// Open PDU to send.
      bool includeMediaChannel,                   /// Include media channel field
      H323Codec::CodecType type                   /// Type of codec
    ) const = 0;

    /**This is called when request to create a channel is received from a
       remote machine and is about to be acknowledged.
     */
    virtual void OnSendingAckPDU(
      H245_H2250LogicalChannelAckParameters & param, /// Acknowledgement PDU
      H323Codec::CodecType type                      /// Type of codec
    ) const = 0;

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.
     */
    virtual bool OnReceivedPDU(
      const H245_H2250LogicalChannelParameters & param, /// Acknowledgement PDU
      unsigned & errorCode                              /// Error on failure
    ) = 0;

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.
     */
    virtual bool OnReceivedAckPDU(
      const H245_H2250LogicalChannelAckParameters & param  /// Acknowledgement PDU
    ) = 0;

    /**This is called when a gatekeeper wants to get status information from
       the endpoint.

       The default behaviour fills in the session ID's and SSRC parameters
       but does not do anything with the transport fields.
     */
    virtual void OnSendRasInfo(
      H225_RTPSession & info  /// RTP session info PDU
    );
  //@}

  /**@name Call back functions */
  //@{
    enum SendReceiveStatus {
      e_ProcessPacket,
      e_IgnorePacket,
      e_AbortTransport
    };
    virtual SendReceiveStatus OnSendData(RTP_DataFrame & frame);
    virtual SendReceiveStatus OnReceiveData(const RTP_DataFrame & frame);
    virtual SendReceiveStatus OnReceiveControl(const RTP_ControlFrame & frame);

    virtual void OnRxSenderReport(const RTP_ControlFrame::SenderReport & sender,
                                  const RTP_ControlFrame::ReceptionReport * reports,
                                  int count);
    virtual void OnRxReceiverReport(DWORD src,
                                    const RTP_ControlFrame::ReceptionReport * reports,
                                    int count);
    virtual void OnRxSourceDescription(const RTP_ControlFrame::SourceDescription & description);
    virtual void OnRxGoodbye(const Tools::SUInt32b * ssrc, int count,
                             const Tools::SString & reason);
    virtual void OnRxApplDefined(const Tools::SString & type, unsigned subtype, DWORD src,
                                 const BYTE * data, int size);
  //@}

  /**@name Member variable access */
  //@{
    /**Get the ID for the RTP session.
      */
    unsigned GetSessionID() const { return sessionID; }

    /**Increment reference count for RTP session.
      */
    void IncrementReference() { referenceCount++; }

    /**Decrement reference count for RTP session.
      */
    bool DecrementReference() { return --referenceCount == 0; }

    /**Get total packets sent.
      */
    DWORD GetPacketsSent() const { return packetsSent; }

    /**Get total octets sent.
      */
    DWORD GetOctetsSent() const { return octetsSent; }

    /**Get total packets received.
      */
    DWORD GetPacketsReceived() const { return packetsReceived; }

    /**Get total octets received.
      */
    DWORD GetOctetsReceived() const { return octetsReceived; }

    /**Get total received packets lost.
      */
    DWORD GetPacketsLost() const { return packetsLost; }

    /** Get remote RTP IP address and port (TSAP). */
    virtual Tools::SIpAddress GetRemoteRtpTsap() const { return Tools::SIpAddress(); } 

    /** Get remote RTCP IP address and port (TSAP). */
    virtual Tools::SIpAddress GetRemoteRtcpTsap() const { return Tools::SIpAddress(); }
  //@}

  protected:
    const H323Connection & connection;

    BYTE     sessionID;
    unsigned referenceCount;
    bool     shutdownRead;
    bool     shutdownWrite;

    DWORD    syncSourceOut;
    unsigned senderReportInterval;

    WORD     lastSequenceNumber;
    WORD     expectedSequenceNumber;

    DWORD    packetsSent;
    DWORD    octetsSent;

    DWORD    packetsReceived;
    DWORD    octetsReceived;
    DWORD    packetsLost;
    DWORD    jitterLevel;
};

//PDICTIONARY(RTPSessionDict, POrdinalKey, RTP_Session);
typedef stdext::hash_map<int, RTP_Session*> RTPSessionDict;


/**This class is for the IETF Real Time Protocol interface on UDP/IP.
 */
class RTP_UDP : public RTP_Session
{
  public:
  /**@name Construction */
  //@{
    /**Create a new RTP channel.
     */
    RTP_UDP(
      const H323Connection & connection, /// Owner of the RTP session
      unsigned id,  /// Session ID for RTP channel
      WORD portBase /// Base port number for socket pair
    );

    /// Destroy the RTP
    ~RTP_UDP();
  //@}

  /**@name Overrides from class RTP_Session */
  //@{
    /**Read a data frame from the RTP channel.
       Any control frames received are dispatched to callbacks and are not
       returned by this function. It will block until a data frame is
       available or an error occurs.
      */
    virtual bool ReadData(RTP_DataFrame & frame);

    /**Write a data frame from the RTP channel.
      */
    virtual bool WriteData(RTP_DataFrame & frame);

    /**Write a control frame from the RTP channel.
      */
    virtual bool WriteControl(RTP_ControlFrame & frame);

    /**Close down the RTP session.
      */
    virtual void Close(
      bool reading    /// Closing the read side of the session
    );

    /**Fill out the OpenLogicalChannel PDU for the particular channel type.
     */
    virtual bool OnSendingPDU(
      H245_H2250LogicalChannelParameters & param, /// Open PDU to send.
      bool includeMediaChannel,                   /// Include media channel field
      H323Codec::CodecType type                   /// Type of codec
    ) const;

    /**This is called when request to create a channel is received from a
       remote machine and is about to be acknowledged.
     */
    virtual void OnSendingAckPDU(
      H245_H2250LogicalChannelAckParameters & param,  /// Acknowledgement PDU
      H323Codec::CodecType type                       /// Type of codec
    ) const;

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.

       The default behaviour sets the remote ports to send UDP packets to.
     */
    virtual bool OnReceivedPDU(
      const H245_H2250LogicalChannelParameters & param, /// Acknowledgement PDU
      unsigned & errorCode                              /// Error on failure
    );

    /**This is called after a request to create a channel occurs from the
       local machine via the H245LogicalChannelDict::Open() function, and
       the request has been acknowledged by the remote endpoint.

       The default behaviour sets the remote ports to send UDP packets to.
     */
    virtual bool OnReceivedAckPDU(
      const H245_H2250LogicalChannelAckParameters & param  /// Acknowledgement PDU
    );

    /**This is called when a gatekeeper wants to get status information from
       the endpoint.

       The default behaviour calls the ancestor functon and then fills in the 
       transport fields.
     */
    virtual void OnSendRasInfo(
      H225_RTPSession & info  /// RTP session info PDU
    );

    /** Get remote RTP IP address and port (TSAP). */
    virtual Tools::SIpAddress GetRemoteRtpTsap() const; 

    /** Get remote RTCP IP address and port (TSAP). */
    virtual Tools::SIpAddress GetRemoteRtcpTsap() const; 
  //@}

  protected:
    SendReceiveStatus ReadDataPDU(RTP_DataFrame & frame);
    SendReceiveStatus ReadControlPDU();

    Tools::SIPSocket::Address remoteAddress;
    WORD                      remoteDataPort;
    WORD                      remoteControlPort;

    Tools::SUDPSocket dataSocket;
    Tools::SUDPSocket controlSocket;
};


#endif // __RTP_H
