/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Transports
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/include/transports.h              $
    $Author:: Martin.hueser                                                $
      $Date:: 20.12.05 18:34                                               $
   $Modtime:: 16.12.05 16:35                                               $
  $Revision:: 3                                                            $
  $Workfile:: transports.h                                                 $
$NoKeywords: $
----------------------------------------------------------------------------*/


#ifndef __TRANSPORTS_H
#define __TRANSPORTS_H


#include <stlib.h>

class SRefCriticalSection : public Tools::SCriticalSection, public Tools::SRefCnt
{
    virtual ~SRefCriticalSection() {}
};


class H225_Setup_UUIE;
class H225_TransportAddress;
class H225_ArrayOf_TransportAddress;
class H225_TransportAddress_ipAddress;

class H245_TransportAddress;

class H323SignalPDU;
class H323RasPDU;
class H323EndPoint;
class H323Connection;
class H323Transport;
class H323Gatekeeper;
class H323GkLink;
class H323GloballyUniqueID;


#define RAS_MULTICAST_ADDRESS      "224.0.1.41"


struct lessthanSTcpConnection
{
    // Timers are ordered by their timeout time first and by their pointer value then.
    // The pointer value is involved for unique order.
    bool operator()(const Tools::STcpConnection* s1, const Tools::STcpConnection* s2) const;

};


/**String representation of a transport address.
 */

class H323TransportAddress : public Tools::SString
{
  public:
    H323TransportAddress() { }
    H323TransportAddress(const char *);
    H323TransportAddress(const Tools::SString &);
    H323TransportAddress(const H225_TransportAddress &);
    H323TransportAddress(const H245_TransportAddress &);
    H323TransportAddress(const Tools::SIPSocket::Address &, WORD);

    bool SetPDU(H225_TransportAddress & pdu) const;
    bool SetPDU(H245_TransportAddress & pdu) const;

    /**Extract the hostname/ip address and port number.
       Returns false, if the address is not an IP transport.
      */
    bool GetIpAndPort(
      Tools::SIPSocket::Address & ip,
      WORD & port,
      const char * proto = "tcp"
    ) const;

  protected:
    void Validate();
};


/**This class describes a "listener" on a transport protocol.
   A "listener" is an object that listens for incoming connections on the
   particular transport. It is executed as a separate thread.

   The Main() function is used to handle incoming H.323 connections and
   dispatch them in new threads based on the actual H323Transport class. This
   is defined in the descendent class that knows what the low level transport
   is, eg H323ListenerIP for the TCP/IP protocol.

   An application may create a descendent off this class and override
   functions as required for operating the channel protocol.
 */
class H323Listener : public Tools::SThread
{
  public:
  /**@name Construction */
  //@{
    /**Create a new listener.
     */
    H323Listener(
      H323EndPoint & endpoint      /// Endpoint instance for channel
    );
  //@}

  /**@name Operations */
  //@{
    /** add a further socket to the listener to receive and process messages from wire
    */
    virtual bool AddConnectedConnection(Tools::SChannel* socket) {return false;}

    /** Open the listener.
      */
    virtual bool Open() = 0;

    /**Stop the listener thread and no longer accept incoming connections.
     */
    virtual bool Close() = 0;


    /**Set up a transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_ArrayOf_TransportAddress & pdu,  /// List of transport addresses listening on
      const H225_TransportAddress & first   /// Transport address to take precedence
    ) = 0;

    /**Set up a only one transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_ArrayOf_TransportAddress & pdu,   /// List of transport addresses listening on
      const Tools::SIPSocket::Address & addr /// Transport address
    ) = 0;
  //@}

  protected:
    H323EndPoint & endpoint;  /// Endpoint that owns the listener.

    bool    m_bTraceFullIP;
};


typedef std::vector<H323Listener *> H323ListenerList;



/**This class describes a I/O transport protocol..
   A "transport" is an object that listens for incoming connections on the
   particular transport.
 */
class H323Transport : public Tools::SIndirectChannel
{
  public:
  /**@name Construction */
  //@{
    /**Create a new transport channel.
     */
    H323Transport(H323EndPoint & endpoint);
    ~H323Transport();
  //@}

  /**@name Operations */
  //@{
    virtual void SetLocalAddress(Tools::SIPSocket::Address addr) =0;

    /**Get the transport dependent name of the local endpoint.
      */
    virtual H323TransportAddress GetLocalAddress() const = 0;

    /**Get the transport dependent name of the remote endpoint.
      */
    virtual H323TransportAddress GetRemoteAddress() const = 0;

    /**Get the transport dependent name as string of the remote endpoint.
      */
    virtual Tools::SString GetRemoteAddressString() const = 0;

    /**Set remote address to connect to.
       Note that this does not necessarily initiate a transport level
       connection, but only indicates where to connect to. The actual
       connection is made by the Connect() function.
      */
    virtual bool SetRemoteAddress(
      const H323TransportAddress & address
    ) = 0;


    virtual void SetConnection(H323Connection* pConnection) {};

    /**Connect to the remote address.
      */
    virtual bool Connect() = 0;

    /**Close the channel.
      */
    virtual bool Close();

    /**Check that the transport address PDU is compatible with transport.
      */
    virtual bool IsCompatibleTransport(
      const H225_TransportAddress & pdu
    ) const;

    /**Set up a transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_TransportAddress & pdu,
      bool bIsLocalTsap
    ) const;

    /**Set up a transport address PDU for logical channel.
      */
    virtual void SetUpTransportPDU(
      H245_TransportAddress & pdu,
      unsigned tsap
    ) const;

    /**Set up a transport address PDU for logical channel. (SwTCL extension)
      */
    virtual void SetUpTransportPDU(
      H245_TransportAddress & pdu,
      Tools::SIpAddress tsap
    ) const;

    /**Read a protocol data unit from the transport.
       This will read using the transports mechanism for PDU boundaries, for
       example UDP is a single Read() call, while for TCP there is a TPKT
       header that indicates the size of the PDU.
      */
    virtual bool ReadPDU(
      Tools::SByteVector & pdu   /// PDU read from transport
    ) = 0;

    /**Write a protocol data unit from the transport.
       This will write using the transports mechanism for PDU boundaries, for
       example UDP is a single Write() call, while for TCP there is a TPKT
       header that indicates the size of the PDU.
      */
    virtual bool WritePDU(
      const Tools::SByteVector & pdu  /// PDU to write
    ) = 0;
  //@}

  /**@name Signalling Channel */
  //@{
    /**Wait for first PDU and find/create connection object.
       If returns false, then the transport is deleted by the calling thread.
      */
    bool HandleFirstSignallingChannelPDU();
  //@}

  /**@name Control Channel */
  //@{
    /**Begin the opening of a control channel.
       This sets up the channel so that the remote endpoint can connect back
       to this endpoint. This would be called on the signalling channel
       instance of a H323Transport.
      */
    virtual H323Transport * CreateControlChannel(
      H323Connection & connection
    );

    /**Begin the opening of a control channel.
       This sets up the channel so that the remote endpoint can connect back
       to this endpoint. This would be called on the signalling channel
       instance of a H323Transport.
       (SwTCL extension)
      */
    virtual H323Transport * CreateSwTCLControlChannel(
      H323Connection & connection
    );

    /**Finish the opening of a control channel.
       This waits for the connect backfrom the remote endpoint, completing the
       control channel open sequence.
      */
    virtual bool AcceptControlChannel(
      H323Connection & connection
    );

    /**Connect the control channel.
      */
    virtual bool ConnectControlChannel(
      H323Connection & connection
    );
  //@}

  /**@name RAS Channel */
  //@{
    /**Discover a Gatekeeper on the network.
       This locates a gatekeeper on the network and associates this transport
       object with packet exchange with that gatekeeper.
      */
    virtual bool DiscoverGatekeeper(
      H323Gatekeeper & gk,                  /// Gatekeeper to set on discovery.
      H323RasPDU & pdu,                     /// GatekeeperRequest PDU
      const H323TransportAddress & address  /// Address of gatekeeper (if present)
    );

    /**Discover a Gatekeeper on the network.
       This locates a gatekeeper on the network and associates this transport
       object with packet exchange with that gatekeeper.
      */
    virtual bool DiscoverGatekeeper(
      H323GkLink & gk,                      /// Gatekeeper to set on discovery.
      H323RasPDU & pdu,                     /// GatekeeperRequest PDU
      const H323TransportAddress & address  /// Address of gatekeeper (if present)
    );
  //@}


  /**@name Member variable access */
  //@{
    /**Attach a thread to the transport.
      */
    void AttachThread(
      Tools::SThread * thread
    );

    /**Wait for associated thread to terminate.
      */
    void CleanUpOnTermination();
  //@}

  protected:
    H323EndPoint    & endpoint; /// Endpoint that owns the listener.
    Tools::SThread  * thread;   /// Thread handling the transport

    bool m_bIsClosed; // Flag that shows if transport is closed.

    Tools::SMutex     m_Mutex;
};



//-----------------------------------------------------------------------------
// Transport classes for IP

/**This class represents a particular H323 transport using IP.
   It is used by the TCP and UDP transports.
 */
class H323TransportIP : public H323Transport
{
  public:
    /**Create a new transport channel.
     */
    H323TransportIP(
      H323EndPoint & endpoint,    /// H323 End Point object
      WORD remPort
    );

    /**Set local address
      */
    virtual void SetLocalAddress(Tools::SIPSocket::Address addr) { localAddress=addr; }

    /**Set the local port a priori for finding outgoing TLS connection to be reused
      */
    virtual void SetLocalPort(WORD a_localPort) { localPort=a_localPort; }

    /**Get the local port
      */
    virtual WORD GetLocalPort() const { return localPort; }

    /**Get the transport dependent name of the local endpoint.
      */
    virtual H323TransportAddress GetLocalAddress() const;

    /**Get the transport dependent name of the remote endpoint.
      */
    virtual H323TransportAddress GetRemoteAddress() const;

    /**Get the transport dependent name as string of the remote endpoint.
      */
    virtual Tools::SString GetRemoteAddressString() const;

    /**Get the transport dependent IP address of the remote endpoint.
      */
    Tools::SIPSocket::Address GetRemoteIpAddress() const { return remoteAddress; }

    /**Get the transport dependent port of the remote endpoint.
      */
    WORD GetRemotePort() const { return remotePort; }

    /**Check that the transport address PDU is compatible with transport.
      */
    virtual bool IsCompatibleTransport(
      const H225_TransportAddress & pdu
    ) const;

    /**Set up a transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_TransportAddress & pdu,
      bool bIsLocalTsap
    ) const;

    /**Set up a transport address PDU for logical channel.
      */
    virtual void SetUpTransportPDU(
      H245_TransportAddress & pdu,
      unsigned tsap
    ) const;

    /**Set up a transport address PDU for logical channel. (SwTCL extension)
      */
    virtual void SetUpTransportPDU(
      H245_TransportAddress & pdu,
      Tools::SIpAddress tsap
    ) const;

  protected:
    Tools::SIPSocket::Address localAddress;  // Address of the local interface
    WORD                      localPort;
    Tools::SIPSocket::Address remoteAddress; // Address of the remote host
    WORD                      remotePort;
};


//-----------------------------------------------------------------------------
// Transport classes for TCP/IP

/**This class manages H323 connections using TCP/IP transport.
 */
class H323ListenerTCP : public H323Listener
{
  public:
    enum {
      DefaultSignalPort = 1720
    };

    /**Create a new listener for the TCP/IP protocol.
     */
    H323ListenerTCP(
      H323EndPoint & endpoint,                        /// Endpoint instance for channel
      Tools::SIPSocket::Address binding = INADDR_ANY, /// Local interface to listen on
      WORD port = DefaultSignalPort                   /// TCP port to listen for connections
    );

    /** Destroy the listener thread.
      */
    ~H323ListenerTCP();
    
  // Overrides from H323Listener
    /** Open the listener.
      */
    virtual bool Open();

    /**Stop the listener thread and no longer accept incoming connections.
     */
    virtual bool Close();

    /**Set up a transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_ArrayOf_TransportAddress & pdu,  /// List of transport addresses listening on
      const H225_TransportAddress & first   /// Transport address to take precedence
    );

    /**Set up a only one transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_ArrayOf_TransportAddress & pdu,   /// List of transport addresses listening on
      const Tools::SIPSocket::Address & addr /// Transport address
    );

    /**Get port of listener.
     */
    WORD GetListenerPort() const { return listener.GetPort(); }

    /**Test if listener is listening on a specified interface.
     */
    virtual bool IsListeningOnAddress(
        const Tools::SIPSocket::Address & rAddress) const;
    
  protected:
    /**Handle incoming H.323 connections and dispatch them in new threads
       based on the H323Transport class. This is defined in the descendent
       class that knows what the low level transport is, eg H323ListenerIP
       for the TCP/IP protocol.

       Note this function does not return until the Close() function is called
       or there is some other error.
     */
    virtual void Main();


    Tools::STCPSocket         listener;
    Tools::SIPSocket::Address localAddress;
	bool					  bShutdownListener;
};


/**This class represents a particular H323 transport using TCP/IP.
 */
class H323TransportTCP : public H323TransportIP
{
  public:
    /**Create a new transport channel.
     */
    H323TransportTCP(
      H323EndPoint & endpoint    /// H323 End Point object
      , bool bIsIncoming
    );

    /**Create a new transport channel.
     */
    H323TransportTCP(
      H323EndPoint & endpoint,    /// H323 End Point object
      WORD port                   /// Specify remote port if it should not be H323ListenerTCP::DefaultSignalPort
    );

    /**Destroy transport channel.
     */
    virtual ~H323TransportTCP();

    /**Set default remote address to connect to.
       Note that this does not necessarily initiate a transport level
       connection, but only indicates where to connect to. The actual
       connection is made by the Connect() function.
      */
    virtual bool SetRemoteAddress(
      const H323TransportAddress & address
    );

    /**Connect to the remote party.
      */
    virtual bool Connect();

    /** Close connection
      */
    virtual bool Close();


    /** aborts connection associated to this transport in case the tcp connection closes or the transport is reused by a subsequent H323Connection.
    */
    virtual bool AbortH323Connection(bool bIsFromWire);

    /** this object is deleted if the transport is not yet associated to an H323Connection
    */
    virtual bool ClosedByWire();

    /** receive processing time from listener to process incoming messages
      */
    virtual bool ProcessRead();

    /**Read a protocol data unit from the transport in asyncronous matter       
      */
    virtual bool ReadPDU();

    /**Read a protocol data unit from the transport.
       This will read using the transports mechanism for PDU boundaries, for
       example UDP is a single Read() call, while for TCP there is a TPKT
       header that indicates the size of the PDU.
      */
    virtual bool ReadPDU(
      Tools::SByteVector & pdu   /// PDU read from transport
    );

    /**Write a protocol data unit from the transport.
       This will write using the transports mechanism for PDU boundaries, for
       example UDP is a single Write() call, while for TCP there is a TPKT
       header that indicates the size of the PDU.
      */
    virtual bool WritePDU(
      const Tools::SByteVector & pdu  /// PDU to write
    );

    /**Begin the opening of a control channel.
       This sets up the channel so that the remote endpoint can connect back
       to this endpoint.
      */
    virtual H323Transport * CreateControlChannel(
      H323Connection & connection
    );

    /**Begin the opening of a control channel.
       This sets up the channel so that the remote endpoint can connect back
       to this endpoint.
       (SwTCL extension)
      */
    virtual H323Transport * CreateSwTCLControlChannel(
      H323Connection & connection
    );

    /**Finish the opening of a control channel.
       This waits for the connect backfrom the remote endpoint, completing the
       control channel open sequence.
      */
    virtual bool AcceptControlChannel(
      H323Connection & connection
    );

    /**Set 'Early H245' cancel flag.
      */
    virtual void SetCancelEarlyH245(bool bCancelEarlyH245) { m_bCancelEarlyH245 = bCancelEarlyH245; }

    /**Get 'Early H245' cancel flag.
      */
    virtual bool IsCancelEarlyH245() const { return m_bCancelEarlyH245; }


    virtual void SetConnection(H323Connection* pConnection) { m_pConnection=pConnection; }


    enum TpktReadStates
    {
        e_ReadingVersion,
        e_ReadingHeader,
        e_ReadingBody,
        e_PduComplete,
        e_MaxTpktReadState
    };

    TpktReadStates      m_tpktState;
    BYTE                m_tpktHeader[3];
    unsigned            m_readHeaderBytes;
    
    Tools::SByteVector  m_rawData;
    unsigned            m_bytesLeftToRead;

    enum ConnectionStates
    {
        e_HandleFirstSignallingChannelPDU,
        e_HandleSignallingChannelPDU,
        e_MaxConnectionState
    };
        
    Tools::SAutoRefPtr<SRefCriticalSection> m_pRefMutex;

    ConnectionStates m_connectionState;
    H323Connection*  m_pConnection;

  protected:
    /**This callback is executed when the Open() function is called with
       open channels. It may be used by descendent channels to do any
       handshaking required by the protocol that channel embodies.

       The default behaviour is to simply return true.

       @return
       Returns true if the protocol handshaking is successful.
     */
    virtual bool OnOpen();

    Tools::STCPSocket * h245listener;

    bool m_bCancelEarlyH245;    // Flag if 'Early H.245' should be canceled
                                // (especially the early started H.245 thread).
};


//-----------------------------------------------------------------------------
// Transport classes for UDP/IP

/**This class represents a particular H323 transport using UDP/IP.
 */
class H323TransportUDP : public H323TransportIP
{
  public:
    enum {
        DefaultRasMulticastPort = 1718,
        DefaultRasPort = 1719
    };

    /**Create a new transport channel.
     */
    H323TransportUDP(
      H323EndPoint & endpoint,                        /// H323 End Point object
      Tools::SIPSocket::Address binding = INADDR_ANY, /// Local interface to listen on
      WORD LocalPort = DefaultRasPort,                /// Local port to listen on,
      bool Multicast = false
    );
    ~H323TransportUDP();

    /**Get address of remote endpoint, that sent the last upd-packet
       to the transport.
      */
    void GetLastReceiveAddress(
        Tools::SIPSocket::Address & address,    /// IP address to send packets.
        WORD & port);                           /// Port to send packets.

    /**Set default remote address to connect to.
       Note that this does not necessarily initiate a transport level
       connection, but only indicates where to connect to. The actual
       connection is made by the Connect() function.
      */
    virtual bool SetRemoteAddress(
      const H323TransportAddress & address
    );

    /**Connect to the remote party.
      */
    virtual bool Connect();

    /**Read a protocol data unit from the transport.
       This will read using the transports mechanism for PDU boundaries, for
       example UDP is a single Read() call, while for TCP there is a TPKT
       header that indicates the size of the PDU.
      */
    virtual bool ReadPDU(
      Tools::SByteVector & pdu   /// PDU read from transport
    );

    /**Write a protocol data unit from the transport.
       This will write using the transports mechanism for PDU boundaries, for
       example UDP is a single Write() call, while for TCP there is a TPKT
       header that indicates the size of the PDU.
      */
    virtual bool WritePDU(
      const Tools::SByteVector & pdu  /// PDU to write
    );

    /**Discover a Gatekeeper on the local network.
       This locates a gatekeeper on the network and associates this transport
       object with packet exchange with that gatekeeper. This broadcasts a UDP
       packet on the local network to find the gatekeeper's IP address.
      */
    virtual bool DiscoverGatekeeper(
      H323Gatekeeper & gk,                  /// Gatekeeper to set on discovery.
      H323RasPDU & pdu,                     /// GatekeeperRequest PDU
      const H323TransportAddress & address  /// Address of gatekeeper (if present)
    );

    /**Discover a Gatekeeper on the network.
       This locates a gatekeeper on the network and associates this transport
       object with packet exchange with that gatekeeper.
      */
    virtual bool DiscoverGatekeeper(
      H323GkLink & gk,                      /// Gatekeeper to set on discovery.
      H323RasPDU & pdu,                     /// GatekeeperRequest PDU
      const H323TransportAddress & address  /// Address of gatekeeper (if present)
    );
};


//-----------------------------------------------------------------------------
// Transport classes for TLS
class H323TransportTls;
/**This class manages H323 connections using TLS transport.
 */
class H323ListenerTls : public H323Listener
{
  public:
    enum {
      DefaultSignalTcpPort = 1720,
      DefaultSignalTlsPort = 1300
    };

    /**Create a new listener for the TLS protocol.
     */
    H323ListenerTls(
      H323EndPoint & endpoint,                        /// Endpoint instance for channel
      Tools::SIPSocket::Address binding = INADDR_ANY, /// Local interface to listen on
      WORD tcpPort = DefaultSignalTcpPort,                   /// Tls port to listen for connections
      WORD tlsPort = DefaultSignalTlsPort                   /// Tls port to listen for connections
    );

    /** Destroy the listener thread.
      */
    ~H323ListenerTls();
    
  // Overrides from H323Listener

    /** add a further socket to the listener to receive and process messages from wire
    */
    virtual bool AddConnectedConnection(Tools::SChannel* socket);

    /** Open the listener.
      */
    virtual bool Open();

    /**Stop the listener thread and no longer accept incoming connections.
     */
    virtual bool Close();

    /**Set up a transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_ArrayOf_TransportAddress & pdu,  /// List of transport addresses listening on
      const H225_TransportAddress & first   /// Transport address to take precedence
    );

    /**Set up a only one transport address PDU for RAS channel.
      */
    virtual void SetUpTransportPDU(
      H225_ArrayOf_TransportAddress & pdu,   /// List of transport addresses listening on
      const Tools::SIPSocket::Address & addr /// Transport address
    );

    /**Get port of listener.
     */
    WORD GetListenerPort() const { return listenerTcp.GetPort(); }

    /**Get port of listener.
     */
    WORD GetListenerPortTls() const { return listenerTls.GetPort(); }

    /**Test if listener is listening on a specified interface.
     */
    bool IsListeningOnAddress(
        const Tools::SIPSocket::Address & rAddress) const;
    
    Tools::STcpConnection* GetTcpConnectionByRemote(const Tools::SIPSocket::Address & remoteAddress, DWORD remotePort);
    Tools::STcpConnection* GetTcpConnectionByLocal(const Tools::SIPSocket::Address & localAddress, DWORD localPort);

    void SignalSelectInterrupter() { char dummy=42; selectInterrupter.Write(&dummy, sizeof(dummy)); }

    bool GetIsTlsServer() { return m_bIsTlsServer; }

  protected:
    /**Handle incoming H.323 connections and dispatch them. 

       Note this function does not return until the Close() function is called
       or there is some other error.
     */
    virtual void Main();

    virtual bool DeleteAllTcpConnections();

    void AcceptTcpConnection(Tools::STcpConnection* signallingChannel, Tools::STCPSocket& listenerSocket );

    Tools::STCPSocket         listenerTls;
    Tools::STCPSocket         listenerTcp;
    Tools::SIPSocket::Address localAddress;
    Tools::SUDPSocket         selectInterrupter; // dummy socket to interrupt select calls at any time necessary
	bool					  bShutdownListener;

    typedef std::set<Tools::STcpConnection*,lessthanSTcpConnection> ConnectionSet;

    Tools::SMutex         m_TcpConnectionListMutex;     // protect access to the below list and set 
    // the following list contains sockets of currently inactive H.323 TLS connections to SwyxPhones, which
    // should not be deleted/closed, but reused later on for the next H.323 call
    ConnectionSet         m_TcpConnectionList;          

    bool                  m_bIsTlsServer;
};


/**This class represents a particular H323 transport using TCP/IP.
 */
class H323TransportTls : public H323TransportTCP
{
    bool startTls;

public:
    /**Create a new transport channel.
     */
    H323TransportTls(
      H323EndPoint & endpoint,    /// H323 End Point object
      bool bIsIncoming
    );

    /**Create a new transport channel.
     */
    H323TransportTls(
      H323EndPoint & endpoint,    /// H323 End Point object
      WORD port                   /// Specify remote port if it should not be H323ListenerTCP::DefaultSignalPort
    );

    /**Destroy transport channel.
     */
    ~H323TransportTls();

    /**Connect to the remote party.
      */
    virtual bool Connect();

    /**Close the channel.
      */
    virtual bool Close();

    void StartTls() { startTls=true; }

  protected:
    /**This callback is executed when the Open() function is called with
       open channels. It may be used by descendent channels to do any
       handshaking required by the protocol that channel embodies.

       The default behaviour is to simply return true.

       @return
       Returns true if the protocol handshaking is successful.
     */
    virtual bool OnOpen();
};


#endif // __TRANSPORTS_H
