/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Channels
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/SwyxSrv v4.3.x.x.root/H323/source/channels.cpp       $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 19.11.04 17:02                                               $
   $Modtime:: 19.11.04 17:01                                               $
  $Revision:: 29                                                           $
  $Workfile:: channels.cpp                                                 $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "h323.h"
#include "codecs.h"
#include "channels.h"
#include "h323pdu.h"

using namespace std;
using namespace Tools;

//#define JITTER_DEBUG 10000


class H323LogicalChannelThread : public SThread
{
  public:
    H323LogicalChannelThread(H323EndPoint & endpoint, H323Channel & channel, bool rx);
    void Main();
  private:
    H323Channel & channel;
    bool receiver;
};


class RTP_JitterBuffer : public SThread
{
  public:
    RTP_JitterBuffer(
      H323EndPoint & endpoint,
      const H323Capability & capability,
      H323Codec & codec
    );
    ~RTP_JitterBuffer();

    virtual void Main();

    RTP_DataFrame * GetReadFrame();
    void QueueReadFrame();
    RTP_DataFrame * GetWriteFrame(DWORD timestamp);
    void FreeWriteFrame();

    int GetSize() const { return bufferSize; }

    void PreBuffering() { preBuffering = true; }

  protected:
    class Entry : public RTP_DataFrame
    {
      public:
        Entry(int size) { SetPayloadSize(size); }
        Entry * next;
        Entry * prev;
    };

    H323Codec & codec;
    int         bufferSize;
    DWORD       maxJitterTime;
    Entry     * oldestFrame;
    Entry     * newestFrame;
    Entry     * freeFrames;
    Entry     * currentReadFrame;
    Entry     * currentWriteFrame;
    SMutex      mutex;
    bool        running;
    bool        preBuffering;

#ifdef JITTER_DEBUG
    STimeInterval debugInTick[JITTER_DEBUG], debugOutTick[JITTER_DEBUG];
    DWORD debugInTime[JITTER_DEBUG], debugOutTime[JITTER_DEBUG];
    int debugInPos, debugOutPos;
#endif
};


#define new SNEW


/////////////////////////////////////////////////////////////////////////////

H323LogicalChannelThread::H323LogicalChannelThread(H323EndPoint & endpoint,
                                                   H323Channel & c,
                                                   bool rx)
  : SThread(endpoint.GetChannelThreadStackSize(), NoAutoDeleteThread),
    channel(c)
{
  receiver = rx;
  Resume();
}


void H323LogicalChannelThread::Main()
{
    SDEFFUNC("H323LogicalChannelThread::Main");

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0,
        "() Started thread (H323LogicalChannelThread)");

  if (receiver)
    channel.Receive();
  else
    channel.Transmit();

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0,
        "() Finished thread (H323LogicalChannelThread)");
}


/////////////////////////////////////////////////////////////////////////////

H323ChannelNumber::H323ChannelNumber(unsigned num, bool fromRem)
{
  SAssert(num < 0x10000, SInvalidParameter);
  number = num;
  fromRemote = fromRem;
}


SObject * H323ChannelNumber::Clone() const
{
  return new H323ChannelNumber(number, fromRemote);
}


int H323ChannelNumber::HashFunction() const
{
  if (fromRemote)
    return number;

  return number|0x10000;
}


void H323ChannelNumber::PrintOn(ostream & strm) const
{
  strm << (fromRemote ? "Rx" : "Tx") << '-' << number;
}


H323ChannelNumber & H323ChannelNumber::operator++(int)
{
    // increment and check for 16bit boundary to wrap around
  if ( (++number) > 65535)
    number = 100;
  return *this;
}

SObject::Comparison H323ChannelNumber::Compare(const SObject & obj) const
{
    const H323ChannelNumber & rNumber = dynamic_cast<const H323ChannelNumber &>(obj);

	// mh 6.8.04: I don't know why member number is defined as an int, but 
	// h323 channel numbers are 16-bit only. Therefore the following works. 
	int i1 = number | (fromRemote<<17);
	int i2 = rNumber | (rNumber.fromRemote<<17);

    if (i1<i2)
        return LessThan;
	else if (i1==i2)
		return EqualTo;
	else
		return GreaterThan; 
}

/////////////////////////////////////////////////////////////////////////////

H323Channel::H323Channel(H323Connection & conn, const H323Capability & cap)
  : endpoint(conn.GetEndPoint()),
    connection(conn),
    capability(cap),
    m_RemoteDataPort(0),
    m_RemoteControlPort(0)
{
    SDEFFUNC("H323Channel::H323Channel");

    codec = NULL;
    bandwidthUsed = 0;
    receiveThread = NULL;
    transmitThread = NULL;
    terminating = false;
    SetConnID( connection.GetConnID() );

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(), "()");
}


H323Channel::~H323Channel()
{
  SAssert(receiveThread == NULL && transmitThread == NULL, SLogicError);

  connection.UseBandwidth(bandwidthUsed, false);

  delete codec;
}


unsigned H323Channel::GetSessionID() const
{
  return 0;
}


void H323Channel::CleanUpOnTermination()
{
    SDEFFUNC("H323Channel::CleanUpOnTermination");

  if (terminating)
    return;

  STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
      "() Cleaning up " << number);

  terminating = true;

  // Signal to the connection that this channel is on the way out
  connection.OnClosedLogicalChannel(*this);

  // If we have a codec, then close it, this allows the transmitThread to be
  // broken out of any I/O block on reading the codec.
  if (codec != NULL)
    codec->Close();

  // If we have a receiver thread, wait for it to die.
  if (receiveThread != NULL) {
    receiveThread->WaitForTermination();
    delete receiveThread;
    receiveThread = NULL;
  }

  // If we have a transmitter thread, wait for it to die.
  if (transmitThread != NULL) {
    transmitThread->WaitForTermination();
    delete transmitThread;
    transmitThread = NULL;
  }

  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
      "() Cleaned up " << number);
}


bool H323Channel::IsRunning() const
{
  if (g_pH323Module->m_bUsingSwtcl)
  {
    // No LogicalChannel thread started if we use SwTCL.
    SAssertAlways(SLogicError);
  }
  if (receiveThread  != NULL && !receiveThread ->IsTerminated())
    return true;

  if (transmitThread != NULL && !transmitThread->IsTerminated())
    return true;

  return false;
}


bool H323Channel::OnReceivedPDU(const H245_OpenLogicalChannel & /*pdu*/,
                                unsigned & /*errorCode*/)
{
  return true;
}


bool H323Channel::OnReceivedAckPDU(const H245_OpenLogicalChannelAck & /*pdu*/)
{
  return true;
}


void H323Channel::OnSendOpenAck(const H245_OpenLogicalChannel & /*pdu*/,
                                H245_OpenLogicalChannelAck & /* pdu*/) const
{
}


void H323Channel::OnFlowControl(long bitRateRestriction)
{
    SDEFFUNC("H323Channel::OnFlowControl");

    if (codec != NULL)
    {
        codec->OnFlowControl(bitRateRestriction);
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "(" << bitRateRestriction << ")");
    }
}


void H323Channel::OnMiscellaneousCommand(const H245_MiscellaneousCommand_type & type)
{
    SDEFFUNC("H323Channel::OnMiscellaneousCmnd");

    if (codec != NULL)
    {
        codec->OnMiscellaneousCommand(type);
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "() chan=" << number << ", type=" << type.GetTagName());
    }
}


void H323Channel::OnMiscellaneousIndication(const H245_MiscellaneousIndication_type & type)
{
    SDEFFUNC("H323Channel::OnMiscellaneousInd");

    if (codec != NULL)
    {
        codec->OnMiscellaneousIndication(type);
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "() chan=" << number << ", type=" << type.GetTagName());
    }
}


void H323Channel::OnJitterIndication(DWORD jitter,
                                     int   skippedFrameCount,
                                     int   additionalBuffer)
{
    SDEFFUNC("H323Channel::OnJitterIndication");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
      "(jitter=" << jitter << ", skippedFrameCount=" << skippedFrameCount
      << ", additionalBuffer=" << additionalBuffer << ")");
}


bool H323Channel::SetInitialBandwidth()
{
    if (GetCodec() == NULL)
        return true;

    return SetBandwidthUsed(codec->GetBandwidth());
}


bool H323Channel::SetBandwidthUsed(unsigned bandwidth)
{
  connection.UseBandwidth(bandwidthUsed, false);
  bandwidthUsed = 0;

  if (!connection.UseBandwidth(bandwidth, true))
    return false;

  bandwidthUsed = bandwidth;
  return true;
}


H323Codec * H323Channel::GetCodec() const
{
    if (codec == NULL)
        ((H323Channel*)this)->codec = capability.CreateCodec(
            GetDirection() == IsReceiver ? H323Codec::Decoder : H323Codec::Encoder);

    return codec;
}


Tools::SIpAddress H323Channel::GetRemoteMediaTsap() const
{
    Tools::SIpAddress MediaTsap;
    MediaTsap.m_Address = m_RemoteAddress;
    MediaTsap.m_Port    = m_RemoteDataPort;
    return MediaTsap;
}


Tools::SIpAddress H323Channel::GetRemoteMediaControlTsap() const
{
    Tools::SIpAddress MediaControlTsap;
    MediaControlTsap.m_Address = m_RemoteAddress;
    MediaControlTsap.m_Port    = m_RemoteControlPort;
    return MediaControlTsap;
}


/////////////////////////////////////////////////////////////////////////////

H323UnidirectionalChannel::H323UnidirectionalChannel(H323Connection & conn,
                                                     const H323Capability & cap,
                                                     Directions direction)
  : H323Channel(conn, cap),
    receiver(direction == IsReceiver)
{
}


H323Channel::Directions H323UnidirectionalChannel::GetDirection() const
{
  return receiver ? IsReceiver : IsTransmitter;
}


void H323UnidirectionalChannel::Start()
{
  if (g_pH323Module->m_bUsingSwtcl)
  {
    // Don't start a LogicalChannel thread if we use SwTCL.
    SAssertAlways(SLogicError);
  }

  SThread * thread = new H323LogicalChannelThread(endpoint, *this, receiver);

  if (receiver)
    receiveThread  = thread;
  else
    transmitThread = thread;
}


/////////////////////////////////////////////////////////////////////////////

H323BidirectionalChannel::H323BidirectionalChannel(H323Connection & conn,
                                                   const H323Capability & cap)
  : H323Channel(conn, cap)
{
}


H323Channel::Directions H323BidirectionalChannel::GetDirection() const
{
  return IsBidirectional;
}


void H323BidirectionalChannel::Start()
{
  if (g_pH323Module->m_bUsingSwtcl)
  {
    // Don't start a LogicalChannel thread if we use SwTCL.
    SAssertAlways(SLogicError);
  }
  receiveThread  = new H323LogicalChannelThread(endpoint, *this, true);
  transmitThread = new H323LogicalChannelThread(endpoint, *this, false);
}


/////////////////////////////////////////////////////////////////////////////

H323_RTPChannel::H323_RTPChannel(H323Connection & conn,
                                 const H323Capability & cap,
                                 Directions direction,
                                 RTP_Session & r)
  : H323UnidirectionalChannel(conn, cap, direction),
    rtp(r)
{
    SDEFFUNC("H323_RTPChannel::H323_RTPChannel");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() " << (receiver ? "Receiver" : "Transmitter")
        << " created using session " << rtp.GetSessionID());
}


H323_RTPChannel::~H323_RTPChannel()
{
  // Finished with the RTP session, this will delete the session if it is no
  // longer referenced by any logical channels.
  connection.ReleaseSession(rtp.GetSessionID());
}


void H323_RTPChannel::CleanUpOnTermination()
{
    SDEFFUNC("H323_RTPChannel::CleanUpOnTermin");

  if (terminating)
    return;

  STRACE2_INFO3(g_pH323Module->m_RtpTraceId, this, GetConnID(),
      "() Cleaning up RTP " << number);

  // Break any I/O blocks and wait for the thread that uses this object to
  // terminate before we allow it to be deleted.
  rtp.Close(receiver);

  H323Channel::CleanUpOnTermination();
}


unsigned H323_RTPChannel::GetSessionID() const
{
  return rtp.GetSessionID();
}


bool H323_RTPChannel::OnSendingPDU(H245_OpenLogicalChannel & open) const
{
    SDEFFUNC("H323_RTPChannel::OnSendingPDU");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, connection.GetConnID(),
        "()");

    H323Codec::CodecType type = GetCodecType();

  open.m_forwardLogicalChannelNumber = (unsigned)number;

  if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
    open.m_reverseLogicalChannelParameters.IncludeOptionalField(
            H245_OpenLogicalChannel_reverseLogicalChannelParameters::e_multiplexParameters);
    // Set the communications information for unicast IPv4
    open.m_reverseLogicalChannelParameters.m_multiplexParameters.SetTag(
                H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters
                    ::e_h2250LogicalChannelParameters);

    return rtp.OnSendingPDU(open.m_reverseLogicalChannelParameters.m_multiplexParameters,
                            receiver,
                            type);
  }
  else {
    // Set the communications information for unicast IPv4
    open.m_forwardLogicalChannelParameters.m_multiplexParameters.SetTag(
                H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters
                    ::e_h2250LogicalChannelParameters);

    return rtp.OnSendingPDU(open.m_forwardLogicalChannelParameters.m_multiplexParameters,
                            receiver,
                            type);
  }
}


void H323_RTPChannel::OnSendOpenAck(const H245_OpenLogicalChannel & open,
                                    H245_OpenLogicalChannelAck & ack) const
{
    SDEFFUNC("H323_RTPChannel::OnSendOpenAck");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, connection.GetConnID(),
        "()");

    H323Codec::CodecType type = GetCodecType();

  // set forwardMultiplexAckParameters option
  ack.IncludeOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters);

  // select H225 choice
  ack.m_forwardMultiplexAckParameters.SetTag(
      H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters);

  // get H225 parms
  H245_H2250LogicalChannelAckParameters & param = ack.m_forwardMultiplexAckParameters;

  // set session ID
  param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_sessionID);
  const H245_H2250LogicalChannelParameters & openparam =
                          open.m_forwardLogicalChannelParameters.m_multiplexParameters;
  unsigned sessionID = openparam.m_sessionID;
  param.m_sessionID = sessionID;

  rtp.OnSendingAckPDU(param, type);

  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, connection.GetConnID(),
      "() Sending open logical channel ACK: sessionID=" << sessionID);
}


static bool ExtractTransport(const H245_TransportAddress & pdu,
                             SIPSocket::Address & ipnum, WORD & port,
                             unsigned & errorCode)
{
    SDEFFUNC("ExtractTransport");

    if (pdu.GetTag() != H245_TransportAddress::e_unicastAddress)
    {
        if (pdu.GetTag() == H245_TransportAddress::e_multicastAddress)
        {
            STRACE2_WARN(g_pH323Module->m_H245TraceId, 0, 0,
                "() Multicast address found but handling it as unicast address!");

            const H245_MulticastAddress & multicast = pdu;
            if (multicast.GetTag() != H245_MulticastAddress::e_iPAddress)
            {
                STRACE2_ERROR(g_pH323Module->m_H245TraceId, 0, 0,
                    "() Only multicast IP address supported!");
                errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
                return false;
            }
            const H245_MulticastAddress_iPAddress & addr = multicast;
            ipnum = SIPSocket::Address((DWORD)addr.m_network);
            port = (WORD)(unsigned)addr.m_tsapIdentifier;
            return true;
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_H245TraceId, 0, 0,
                "() Neither unicast nor multicast address tag!");
            errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
            return false;
        }
    }

  const H245_UnicastAddress & unicast = pdu;
  if (unicast.GetTag() != H245_UnicastAddress::e_iPAddress) {
    STRACE_ERROR(g_pH323Module->m_RtpTraceId, "UDP\tLogic error, UDP must be IP unicast");
    errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
    return false;
  }

  const H245_UnicastAddress_iPAddress & addr = unicast;

  ipnum = SIPSocket::Address((DWORD)addr.m_network);

  port = (WORD)(unsigned)addr.m_tsapIdentifier;
  return true;
}


bool H323_RTPChannel::OnReceivedPDU(const H245_OpenLogicalChannel & open,
                                    unsigned & errorCode)
{
    SDEFFUNC("H323_RTPChannel::OnReceivedPDU");

  if (receiver)
    number = H323ChannelNumber(open.m_forwardLogicalChannelNumber, true);

  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() channel: " << number);

    if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters))
    {
        if (open.m_reverseLogicalChannelParameters.m_multiplexParameters.GetTag() ==
             H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters::
                                     e_h2250LogicalChannelParameters)
        {
            if (g_pH323Module->m_bUsingSwtcl)
            {
                // If we use SwTCL we decode it here directly not in class RTP_UDP.
                // This is a workaround to handle multiple RTP/RTCP ports for audio sessions.
                const H245_H2250LogicalChannelParameters & param = open.m_reverseLogicalChannelParameters.m_multiplexParameters;
                if (param.m_sessionID != GetSessionID())
                {
                    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                        "() Open for invalid session: " << param.m_sessionID);
                    errorCode = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
                    return false;
                }

                if (!param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel))
                {
                    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                        "() No mediaControlChannel specified");
                    errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
                    return false;
                }

                if (!ExtractTransport(param.m_mediaControlChannel, m_RemoteAddress, m_RemoteControlPort, errorCode))
                    return false;

                if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel))
                    return ExtractTransport(param.m_mediaChannel, m_RemoteAddress, m_RemoteDataPort, errorCode);

                return true;
            }
            else    
            {
                return rtp.OnReceivedPDU(open.m_reverseLogicalChannelParameters.m_multiplexParameters, errorCode);
            }
        }
    }
    else
    {
        if (open.m_forwardLogicalChannelParameters.m_multiplexParameters.GetTag() ==
            H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters::
                                    e_h2250LogicalChannelParameters)
        {
            if (g_pH323Module->m_bUsingSwtcl)
            {
                // If we use SwTCL we decode it here directly not in class RTP_UDP.
                // This is a workaround to handle multiple RTP/RTCP ports for audio sessions.
                const H245_H2250LogicalChannelParameters & param = open.m_forwardLogicalChannelParameters.m_multiplexParameters;
                if (param.m_sessionID != GetSessionID())
                {
                    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                        "() Open for invalid session: " << param.m_sessionID);
                    errorCode = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
                    return false;
                }

                if (!param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel))
                {
                    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                        "() No mediaControlChannel specified");
                    errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
                    return false;
                }

                if (!ExtractTransport(param.m_mediaControlChannel, m_RemoteAddress, m_RemoteControlPort, errorCode))
                    return false;

                if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel))
                    return ExtractTransport(param.m_mediaChannel, m_RemoteAddress, m_RemoteDataPort, errorCode);

                return true;
            }
            else    
            {
                return rtp.OnReceivedPDU(open.m_forwardLogicalChannelParameters.m_multiplexParameters, errorCode);
            }
        }
    }

  STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() Only H.225.0 multiplex supported");
  errorCode = H245_OpenLogicalChannelReject_cause::e_unsuitableReverseParameters;
  return false;
}


bool H323_RTPChannel::OnReceivedAckPDU(const H245_OpenLogicalChannelAck & ack)
{
    SDEFFUNC("H323_RTPChannel::OnReceivedAckPDU");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() OLCAck received");

  if (!ack.HasOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters)) {
    STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() No forwardMultiplexAckParameters");
    return false;
  }

  if (ack.m_forwardMultiplexAckParameters.GetTag() !=
            H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters) {
    STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Only H.225.0 multiplex supported");
    return false;
  }

    if (g_pH323Module->m_bUsingSwtcl)
    {
        // If we use SwTCL we decode it here directly not in class RTP_UDP.
        // This is a workaround to handle multiple RTP/RTCP ports for audio sessions.
        const H245_H2250LogicalChannelAckParameters & param = ack.m_forwardMultiplexAckParameters;

        if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_sessionID))
        {
            STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                "() No session specified");
            //return false;  // sessionID is optional !
        }
        else
        {
            if (param.m_sessionID != GetSessionID())
            {
                STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                    "() Ack for invalid session: " << param.m_sessionID);
            }
        }

        if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel))
        {
            STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                "() No mediaControlChannel specified");
            return false;
        }

        unsigned errorCode;
        if (!ExtractTransport(param.m_mediaControlChannel, m_RemoteAddress, m_RemoteControlPort, errorCode))
            return false;

        if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel))
        {
            STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                "() No mediaChannel specified");
            return false;
        }

        if (!ExtractTransport(param.m_mediaChannel, m_RemoteAddress, m_RemoteDataPort, errorCode))
            return false;

        // If we use SwTCL we want to get parameters out of OLCAck
        if (!connection.OnReceivedLogicalChannelAckPDU(ack))
            return false;
    }
    else
    {
        if (!rtp.OnReceivedAckPDU(ack.m_forwardMultiplexAckParameters))
        {
            STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                "() Error handling multiplex PDU");
            return false;
        }
    }

  return true;
}


void H323_RTPChannel::Transmit()
{
    SDEFFUNC("H323_RTPChannel::Transmit");

  if (codec == NULL) {
    SAssertAlways(SLogicError);
    return;
  }

  // Open the codec
  if (!codec->Open(connection)) {
    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() Transmit thread ended (open fail)");
    return;
  }

  // Give the connection (or endpoint) a chance to do something with
  // the opening of the codec. Default calls codec->Open();
  if (!connection.OnStartLogicalChannel(*this)) {
    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() Transmit thread ended (OnStartLogicalChannel fail)");
    return;
  }

  // Get parameters from the codec on time and data sizes
  unsigned packetSize = capability.GetTxPacketSize();
  unsigned maxFrameSize = capability.GetMaxFrameSize();
  RTP_DataFrame frame(packetSize);
  unsigned codecTimestamps = codec->GetFrameRate();
  unsigned codecMilliseconds = codecTimestamps/codec->GetTimeUnits();

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() Transmit thread started:"
        " rate=" << codecTimestamps <<
        " time=" << codecMilliseconds << "ms"
        " size=" << packetSize);

  // This is real time so need to keep track of elapsed milliseconds
  STimeInterval nextTick = STimer::Tick();
  DWORD timestamp = 0;
  bool silent = true;
  unsigned length;
  unsigned frameOffset = 0;

  // Now keep getting encoded frames from the codec, we make sure we do not
  // call it any faster than codecMilliseconds, even though as a rule
  // the function takes exactly codecMilliseconds to complete anyway.

  while (codec->Read(frame.GetPayloadPtr()+frameOffset, length)) {
    // Check for silence from codec.
    if (length == 0) {
      silent = true;
      if (frameOffset > 0) {
        // Send the frame of coded data to RTP transport
        if (!rtp.WriteData(frame))
          break;
        frameOffset = 0;
      }
    }
    else {
      if (frameOffset == 0) {
#if STRACING
        if (silent)
        {
            STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                "() Sending start of talk burst: " << timestamp);
        }
#endif
        frame.SetMarker(silent);  // Set flag for start of sound
        frame.SetTimestamp(timestamp);

        // This allows for the fact that a codec may change the RTP payload
        // type on the fly. If it can be shown that this does not occur then
        // this can be moved to outside of the loop.
        frame.SetPayloadType(codec->GetRTPPayloadType());
      }

      frameOffset += length;

      if (frameOffset + maxFrameSize > packetSize) {
        frame.SetPayloadSize(frameOffset);

        // Send the frame of coded data to RTP transport
        if (!rtp.WriteData(frame))
          break;
        frameOffset = 0;
      }

      // Is no longer silent
      silent = false;
    }

    // Calculate the timestamp and real time to take in processing
    timestamp += codecTimestamps;
    nextTick += codecMilliseconds;

    // Wait for milliseconds to go past
    STimeInterval delay = nextTick - STimer::Tick();
    if (delay > 1)
      SThread::Current()->Sleep(delay);
  }

  if (!terminating)
    connection.CloseLogicalChannel(number, number.IsFromRemote());

  STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
      "() Transmit thread ended");
}


void H323_RTPChannel::Receive()
{
    SDEFFUNC("H323_RTPChannel::Receive");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() Receive thread started");

  if (codec == NULL) {
    SAssertAlways(SLogicError);
    return;
  }

  // Open the codec
  if (!codec->Open(connection)) {
    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() Receive thread ended (open fail)");
    return;
  }

  // Give the connection (or endpoint) a chance to do something with
  // the opening of the codec. Default calls codec->Open();
  if (!connection.OnStartLogicalChannel(*this)) {
    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() Receive thread ended (OnStartLogicalChannel fail)");
    return;
  }

    // if no jitter buffer required, then just write frames to the codec

    if (!codec->NeedsJitterBuffer()) {

      RTP_DataFrame frame;
      while (rtp.ReadData(frame)) {

        if (frame.GetPayloadType() != codec->GetRTPPayloadType())
        {
            STRACE2_WARN(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                "() Payload type mismatch: "
                 << frame.GetPayloadType() << " != " << codec->GetRTPPayloadType());
        }
        else {
          unsigned byteOffset = 0;
          unsigned size = frame.GetPayloadSize();
          unsigned written;
          while (byteOffset < size) {
            if (!codec->Write(frame.GetPayloadPtr()+byteOffset, size-byteOffset, frame, written))
              connection.CloseLogicalChannel(number, number.IsFromRemote());
            byteOffset += written;
          }
        }
      }
    }

    else {
    // if jitter buffer required, start the thread that is on the other end of the jitter buffer

      RTP_JitterBuffer * jitter = NULL;
      jitter = new RTP_JitterBuffer(endpoint, capability, *codec);

      // Get the first free frame in the jitter buffer
      RTP_DataFrame * frame = jitter->GetReadFrame();

      // Keep reading from the RTP transport frames
      while (rtp.ReadData(*frame)) {

        // Make sure is from the current payload type for the codec selected
        if (frame->GetPayloadType() != codec->GetRTPPayloadType())
        {
            STRACE2_WARN(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                "() Payload type mismatch: "
                << frame->GetPayloadType() << " != " << codec->GetRTPPayloadType());
        }
        else {
          if (frame->GetMarker()) {
            STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                "() Received start of talk burst: " << frame->GetTimestamp());
            jitter->PreBuffering();
          }

          // Queue the frame for playing by the thread at other end of jitter buffer
          jitter->QueueReadFrame();

          // Get next free frame out of the jitter buffer
          frame = jitter->GetReadFrame();
        }
      }

      delete jitter;
    }

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
        "() Receive thread ended");
}


H323Codec::CodecType H323_RTPChannel::GetCodecType() const
{
    H323Codec * pCodec = GetCodec();
    H323Codec::CodecType type = H323Codec::Undefined;
    if (!SBAD_READPTR(pCodec))
    {
        type = pCodec->GetCodecType();
    }
    return type;
}


unsigned long H323_RTPChannel::GetFramesize() const
{
    H323Codec * pCodec = GetCodec();
    if (!SBAD_READPTR(pCodec))
    {
        return pCodec->GetFramesize();
    }
    return 0;
}


Tools::SIpAddress H323_RTPChannel::GetRemoteMediaTsap() const
{
    if (g_pH323Module->m_bUsingSwtcl)
    {
        // If we use SwTCL we decode it in H323_RTPChannel not in class RTP_UDP.
        // This is a workaround to handle multiple RTP/RTCP ports for audio sessions.
        return H323Channel::GetRemoteMediaTsap();
    }
    return rtp.GetRemoteRtpTsap();
}


Tools::SIpAddress H323_RTPChannel::GetRemoteMediaControlTsap() const
{
    if (g_pH323Module->m_bUsingSwtcl)
    {
        // If we use SwTCL we decode it in H323_RTPChannel not in class RTP_UDP.
        // This is a workaround to handle multiple RTP/RTCP ports for audio sessions.
        return H323Channel::GetRemoteMediaControlTsap();
    }
    return rtp.GetRemoteRtcpTsap();
}


/////////////////////////////////////////////////////////////////////////////

RTP_JitterBuffer::RTP_JitterBuffer(H323EndPoint & endpoint,
                                   const H323Capability & capability,
                                   H323Codec & cod)
  : SThread(endpoint.GetChannelThreadStackSize(),
            NoAutoDeleteThread,
            HighPriority),
    codec(cod)
{
    SDEFFUNC("RTP_JitterBuffer::RTP_JitterBuff");

  // Jitter buffer is a queue of frames waiting for playback, a list of
  // free frames, and a couple of place holders for the frame that is
  // currently beeing read from the RTP transport or written to the codec.

  oldestFrame = newestFrame = currentReadFrame = currentWriteFrame = NULL;

  // Calculate the maximum amount of timestamp units for the jitter buffer
  maxJitterTime = endpoint.GetMaxAudioDelayJitter()*codec.GetTimeUnits();

  // Calculate number of frames to allocate
  bufferSize = (maxJitterTime/codec.GetFrameRate() + 1)*2;

  // Allocate the frames and put them all into the free list
  freeFrames = new Entry(capability.GetRxPacketSize());
  freeFrames->next = freeFrames->prev = NULL;

  for (int i = 0; i < bufferSize; i++) {
    Entry * frame = new Entry(capability.GetRxPacketSize());
    frame->prev = NULL;
    frame->next = freeFrames;
    freeFrames->prev = frame;
    freeFrames = frame;
  }
  STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() Jitter buffer created:"
        " size=" << bufferSize <<
        " timestamp=" << maxJitterTime);

#ifdef JITTER_DEBUG
  debugInPos = debugOutPos = 0;
#endif

  // Start reading data from jitter buffer and writing to codec
  running = true;
  Resume();
}


#ifdef JITTER_DEBUG
static void JitterDebugDump(const char * title, int count, STimeInterval * tick, DWORD * time)
{
  STimeInterval prevtick = tick[0];
  DWORD prevtime = time[0];
  int i;

  for (i = 0; i < count; i++) {
    STRACE_INFO(g_pH323Module->m_RtpTraceId, title
          << ' ' << tick[i] << ' ' << (tick[i]-prevtick)
          << ' ' << time[i] << ' ' << (time[i]-prevtime));
    prevtick = tick[i];
    prevtime = time[i];
  }
}
#endif


RTP_JitterBuffer::~RTP_JitterBuffer()
{
    SDEFFUNC("RTP_JitterBuffer::~RTP_JitterBuf");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() Removing jitter buffer");

  running = false;
  WaitForTermination();

  mutex.Wait();

  // Free up all the memory allocated
  while (oldestFrame != NULL) {
    Entry * frame = oldestFrame;
    oldestFrame = oldestFrame->next;
    delete frame;
  }

  while (freeFrames != NULL) {
    Entry * frame = freeFrames;
    freeFrames = freeFrames->next;
    delete frame;
  }

  delete currentReadFrame;
  delete currentWriteFrame;

  mutex.Signal();

#ifdef JITTER_DEBUG
  JitterDebugDump("In ", debugInPos, debugInTick, debugInTime);
  JitterDebugDump("Out", debugOutPos, debugOutTick, debugOutTime);
#endif
}


void RTP_JitterBuffer::Main()
{
    SDEFFUNC("RTP_JitterBuffer::Main");

    STRACE2_INFO3(g_pH323Module->m_RtpTraceId, this, 0,
        "() Started thread (RTP_JitterBuffer)");

  // Get the real time parameters from the codec.
  unsigned codecTimestamps = codec.GetFrameRate();
  unsigned codecMilliseconds = codecTimestamps/codec.GetTimeUnits();

  STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
      "() Jitter thread started:"
      " rate=" << codecTimestamps <<
      " time=" << codecMilliseconds << "ms");

  // This is real time so need to keep track of elapsed milliseconds
  STimeInterval nextTick = STimer::Tick();
  DWORD timestamp = 0;
  RTP_DataFrame * frame = NULL;
  unsigned byteOffset = 0;

  while (running) {
    // Get the next frame from buffer, if it is time to do so and have
    // finished the previous frame.
    if (frame == NULL) {
      frame = GetWriteFrame(timestamp);
      // If have a frame set timestamp to correct value (allows for lost data)
      if (frame != NULL)
        timestamp = frame->GetTimestamp();
    }

    // Not time yet, just wait
    if (frame == NULL) {
      // Output silence
      unsigned dummy;
      if (!codec.Write(NULL, 0, *frame, dummy))
        break;
    }
    else {
      // Output noise
      unsigned size = frame->GetPayloadSize();
      unsigned written;
      if (!codec.Write(frame->GetPayloadPtr()+byteOffset, size-byteOffset, *frame, written))
        break;

      // Move offset to next (possible) codec data block
      byteOffset += written;

      // Return the frame to the free list if got all of the data out of it
      if (byteOffset >= size) {
        frame = NULL;
        byteOffset = 0;
        FreeWriteFrame();
      }
    }

    // Calculate the timestamp value for next frame
    timestamp += codecTimestamps;

    // Calculate the real time tick we should wait for it to arrive
    nextTick += codecMilliseconds;
    STimeInterval delay = nextTick - STimer::Tick();
    if (delay > 0) {
      SAssert(delay < 1000, "Abnormally large jitter time.");
      Sleep(delay);
    }
  }

  STRACE2_INFO3(g_pH323Module->m_RtpTraceId, this, 0,
      "() Finished thread (RTP_JitterBuffer)");
}


RTP_DataFrame * RTP_JitterBuffer::GetReadFrame()
{
  // Get the next free frame available for use for reading from the RTP
  // transport. Place it into a parking spot.

  if (currentReadFrame != NULL)
    return currentReadFrame;

  mutex.Wait();

  if (freeFrames == NULL) {
    // We have a full jitter buffer, need a new frame so take the oldest one
    currentReadFrame = oldestFrame;
    oldestFrame = oldestFrame->next;
    if (oldestFrame != NULL)
      oldestFrame->prev = NULL;
  }
  else {
    // Take the next free frame and make it the current for reading
    currentReadFrame = freeFrames;
    freeFrames = freeFrames->next;
    if (freeFrames != NULL)
      freeFrames->prev = NULL;
  }

  currentReadFrame->next = NULL;

  mutex.Signal();

  return currentReadFrame;
}


void RTP_JitterBuffer::QueueReadFrame()
{
  // This should not happen, but just in case....
  if (currentReadFrame == NULL)
    return;

#ifdef JITTER_DEBUG
  if (debugInPos < PARRAYSIZE(debugInTime)) {
    debugInTick[debugInPos] = PTimer::Tick();
    debugInTime[debugInPos++] = currentReadFrame->GetTimestamp();
  }
#endif

  mutex.Wait();

  // Have been reading a frame, put it into the queue now, at correct position
  if (newestFrame == NULL)
    oldestFrame = newestFrame = currentReadFrame;
  else {
    DWORD time = currentReadFrame->GetTimestamp();

    if (time > newestFrame->GetTimestamp()) {
      currentReadFrame->prev = newestFrame;
      newestFrame->next = currentReadFrame;
      newestFrame = currentReadFrame;
    }
    else {
      Entry * frame = newestFrame->prev;
      while (frame != NULL && time > frame->GetTimestamp())
        frame = frame->prev;

      if (frame == NULL) {
        currentReadFrame->next = oldestFrame;
        oldestFrame->prev = currentReadFrame;
        oldestFrame = currentReadFrame;
      }
      else {
        currentReadFrame->prev = frame;
        currentReadFrame->next = frame->next;
        frame->next = currentReadFrame;
        frame->next->prev = currentReadFrame;
      }
    }
  }

  currentReadFrame = NULL;

  mutex.Signal();
}


RTP_DataFrame * RTP_JitterBuffer::GetWriteFrame(DWORD timestamp)
{
  // Get the next frame to write to the codec. Takes it from the oldest
  // position in the queue, if it is time to do so, and parks it in the
  // special member so can unlock the mutex while the writer thread has its
  // way with the buffer.

  if (currentWriteFrame != NULL)
    return currentWriteFrame;

  // No data yet to play!
  if (oldestFrame != NULL) {
    mutex.Wait();

    /* See if time for this packet, if our oldest frame is older than the
       required age, then use it. If it is not time yet, make sure that the
       writer thread isn't falling behind (not enough MIPS). If the time
       between the oldest and the newest entries in the jitter buffer is
       greater than the size specified for the buffer, then return the oldest
       entry regardless, making the writer thread catch up.
    */

    DWORD oldTimestamp = oldestFrame->GetTimestamp();
    DWORD deltaTime = newestFrame->GetTimestamp() - oldTimestamp;

    if (preBuffering && deltaTime > maxJitterTime/2)
      preBuffering = false;

    if (!preBuffering && (oldTimestamp <= timestamp || deltaTime > maxJitterTime)) {
      // Detatch oldest packet from the list, put into parking space
      currentWriteFrame = oldestFrame;
      oldestFrame = currentWriteFrame->next;
      if (oldestFrame != NULL)
        oldestFrame->prev = NULL;
      else
        newestFrame = NULL;
      currentWriteFrame->next = NULL;
    }

    mutex.Signal();
  }

#ifdef JITTER_DEBUG
  if (debugOutPos < PARRAYSIZE(debugOutTime)) {
    debugOutTick[debugOutPos] = STimer::Tick();
    debugOutTime[debugOutPos++] = currentWriteFrame != NULL ? currentWriteFrame->GetTimestamp() : 0;
  }
#endif

  return currentWriteFrame;
}


void RTP_JitterBuffer::FreeWriteFrame()
{
  // Free the frame just written to codec (or ignored), putting it back into
  // the free list and clearing the parking spot for it.

  // This should not happen, but just in case....
  if (currentWriteFrame == NULL)
    return;

  mutex.Wait();

  // Have been reading a frame, put it into the queue now
  currentWriteFrame->next = freeFrames;
  if (freeFrames != NULL)
    freeFrames->prev = currentWriteFrame;
  freeFrames = currentWriteFrame;

  currentWriteFrame = NULL;

  mutex.Signal();
}


/////////////////////////////////////////////////////////////////////////////

H323DataChannel::H323DataChannel(H323Connection & conn,
                                 const H323Capability & cap,
                                 Directions dir,
                                 unsigned id)
  : H323UnidirectionalChannel(conn, cap, dir)
{
    SDEFFUNC("H323DataChannel::H323DataChannel");

    sessionID = id;
    separateReverseChannel = false;

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "()");
}


H323DataChannel::~H323DataChannel()
{
    SDEFFUNC("H323DataChannel::~H323DataChannel");

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "()");
}


void H323DataChannel::CleanUpOnTermination()
{
    SDEFFUNC("H323DataChannel::CleanUpOnTermin");

    if (terminating)
        return;

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "() Cleaning up data channel " << number);

    H323UnidirectionalChannel::CleanUpOnTermination();
}


unsigned H323DataChannel::GetSessionID() const
{
    return sessionID;
}

bool H323DataChannel::OnSendingPDU(H245_OpenLogicalChannel & open) const
{
    SDEFFUNC("H323DataChannel::OnSendingPDU");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
        "() OnSendingPDU for channel: " << number);
    
    H323Codec::CodecType type = GetCodecType();

    open.m_forwardLogicalChannelNumber = (unsigned)number;

    if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters))
    {
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
            "() OnSendingPDU reverse");
    
        open.m_reverseLogicalChannelParameters.IncludeOptionalField(
                H245_OpenLogicalChannel_reverseLogicalChannelParameters::e_multiplexParameters);
        open.m_reverseLogicalChannelParameters.m_multiplexParameters.SetTag(
                H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters
                    ::e_h2250LogicalChannelParameters);
        H245_H2250LogicalChannelParameters & rparam = open.m_reverseLogicalChannelParameters.m_multiplexParameters;
        rparam.m_sessionID = GetSessionID();

        if (receiver)
        {
            Tools::SIpAddress SourceMediaTSAP = connection.GetSourceMediaTSAP(type);

            STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() Data TSAP: " << SourceMediaTSAP);

            rparam.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel);
            SetUpTransportPDU(rparam.m_mediaChannel, SourceMediaTSAP);
        }
        return capability.OnSendingPDU(open.m_reverseLogicalChannelParameters.m_dataType);
    }
    else
    {
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
            "() OnSendingPDU forward");

        open.m_forwardLogicalChannelParameters.m_multiplexParameters.SetTag(
            H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters
                ::e_h2250LogicalChannelParameters);
        H245_H2250LogicalChannelParameters & fparam = open.m_forwardLogicalChannelParameters.m_multiplexParameters;
        fparam.m_sessionID = GetSessionID();

        if (receiver && separateReverseChannel == true)
        {
            Tools::SIpAddress SourceMediaTSAP = connection.GetSourceMediaTSAP(type);

            STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() Data TSAP: " << SourceMediaTSAP);

            fparam.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel);
            SetUpTransportPDU(fparam.m_mediaChannel, SourceMediaTSAP);
        }
    }

    if (    separateReverseChannel == false
        && !open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters))
    {
        // Until now used for T.38 bidirectional TCP
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
            "() OnSendingPDU reverse (bidirectional)");

        open.IncludeOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters);
        open.m_reverseLogicalChannelParameters.IncludeOptionalField(
            H245_OpenLogicalChannel_reverseLogicalChannelParameters::e_multiplexParameters);
        open.m_reverseLogicalChannelParameters.m_multiplexParameters.SetTag(
            H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters
                ::e_h2250LogicalChannelParameters);
        H245_H2250LogicalChannelParameters & rparam = open.m_reverseLogicalChannelParameters.m_multiplexParameters;
        rparam.m_sessionID = GetSessionID();

        if (receiver)
        {
            Tools::SIpAddress SourceMediaTSAP = connection.GetSourceMediaTSAP(type);

            STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() Data TSAP: " << SourceMediaTSAP);

            rparam.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel);
            SetUpTransportPDU(rparam.m_mediaChannel, SourceMediaTSAP);
        }
        return capability.OnSendingPDU(open.m_reverseLogicalChannelParameters.m_dataType);
    }
    return true;
}


void H323DataChannel::OnSendOpenAck(const H245_OpenLogicalChannel & /*open*/,
                                    H245_OpenLogicalChannelAck & ack) const
{
    SDEFFUNC("H323DataChannel::OnSendOpenAck");

    H323Codec::CodecType type = GetCodecType();

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
        "() OnSendOpenAck for channel: " << number);

    H245_H2250LogicalChannelAckParameters * param;

    if (separateReverseChannel)
    {
        ack.IncludeOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters);
        ack.m_forwardMultiplexAckParameters.SetTag(
            H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters);
        param = (H245_H2250LogicalChannelAckParameters*)&ack.m_forwardMultiplexAckParameters.GetObject();
    }
    else
    {
        ack.IncludeOptionalField(H245_OpenLogicalChannelAck::e_reverseLogicalChannelParameters);
        ack.m_reverseLogicalChannelParameters.m_multiplexParameters.SetTag(
            H245_OpenLogicalChannelAck_reverseLogicalChannelParameters_multiplexParameters
                ::e_h2250LogicalChannelParameters);
        param = (H245_H2250LogicalChannelAckParameters*)
                    &ack.m_reverseLogicalChannelParameters.m_multiplexParameters.GetObject();
    }

    unsigned session = GetSessionID();
    if (session != 0)
    {
        param->IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_sessionID);
        param->m_sessionID = GetSessionID();
    }

    Tools::SIpAddress SourceMediaTSAP = connection.GetSourceMediaTSAP(type);

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
        "() Data TSAP: " << SourceMediaTSAP);

    param->IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);

    SetUpTransportPDU(param->m_mediaChannel, SourceMediaTSAP);
}


bool H323DataChannel::OnReceivedPDU(const H245_OpenLogicalChannel & open,
                                    unsigned & errorCode)
{
    SDEFFUNC("H323DataChannel::OnReceivedPDU");

    if (receiver)
        number = H323ChannelNumber(open.m_forwardLogicalChannelNumber, true);

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
        "() OnReceivedPDU for data channel: " << number);

    if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters))
    {
        if (open.m_reverseLogicalChannelParameters.m_multiplexParameters.GetTag() ==
             H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters::
                                     e_h2250LogicalChannelParameters)
        {
            const H245_H2250LogicalChannelParameters & param = open.m_reverseLogicalChannelParameters.m_multiplexParameters;
            if (param.m_sessionID != GetSessionID())
            {
                STRACE2_WARN(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                    "() OLC has invalid session id " << param.m_sessionID << " instead of " << GetSessionID());
                //errorCode = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
                //return false;
            }

            if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel))
            {
                if (!ExtractTransport(param.m_mediaChannel, m_RemoteAddress, m_RemoteDataPort, errorCode))
                    return false;
            }

            if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel))
                return ExtractTransport(param.m_mediaControlChannel, m_RemoteAddress, m_RemoteControlPort, errorCode);

            return true;
        }
    }
    else
    {
        if (open.m_forwardLogicalChannelParameters.m_multiplexParameters.GetTag() ==
            H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters::
                                    e_h2250LogicalChannelParameters)
        {
            const H245_H2250LogicalChannelParameters & param = open.m_forwardLogicalChannelParameters.m_multiplexParameters;
            if (param.m_sessionID != GetSessionID())
            {
                STRACE2_WARN(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                    "() OLC has invalid session id " << param.m_sessionID << " instead of " << GetSessionID());
                //errorCode = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
                //return false;
            }

            if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel))
            {
                if (!ExtractTransport(param.m_mediaChannel, m_RemoteAddress, m_RemoteDataPort, errorCode))
                    return false;
            }

            if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel))
                return ExtractTransport(param.m_mediaControlChannel, m_RemoteAddress, m_RemoteControlPort, errorCode);

            return true;
        }
    }

    STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
        "() Only H.225.0 multiplex supported");
    errorCode = H245_OpenLogicalChannelReject_cause::e_unsuitableReverseParameters;
    return false;
}


bool H323DataChannel::OnReceivedAckPDU(const H245_OpenLogicalChannelAck & ack)
{
    SDEFFUNC("H323DataChannel::OnReceivedAckPDU");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "()");

    const H245_TransportAddress * address;

    if (separateReverseChannel)
    {
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
            "() separateReverseChannels");
        if (!ack.HasOptionalField(H245_OpenLogicalChannelAck::e_forwardMultiplexAckParameters))
        {
            STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() No forwardMultiplexAckParameters");
            return false;
        }

        if (ack.m_forwardMultiplexAckParameters.GetTag() !=
                H245_OpenLogicalChannelAck_forwardMultiplexAckParameters::e_h2250LogicalChannelAckParameters)
        {
            STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() Only H.225.0 multiplex supported");
            return false;
        }

        const H245_H2250LogicalChannelAckParameters & param = ack.m_forwardMultiplexAckParameters;

        if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel))
        {
            STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() No media channel address provided");
            return false;
        }

        address = &param.m_mediaChannel;

        unsigned errorCode;
        if (!ExtractTransport(param.m_mediaChannel, m_RemoteAddress, m_RemoteDataPort, errorCode))
            return false;

        if (param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel))
        {
            if (!ExtractTransport(param.m_mediaControlChannel, m_RemoteAddress, m_RemoteControlPort, errorCode))
                return false;
        }

        // If we use SwTCL we want to get parameters out of OLCAck
        if (!connection.OnReceivedLogicalChannelAckPDU(ack))
            return false;

        if (ack.HasOptionalField(H245_OpenLogicalChannelAck::e_reverseLogicalChannelParameters))
        {
            STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, GetConnID(),
                "() reverseLogicalChannelParameters set");
//          reverseChannel = H323ChannelNumber(ack.m_reverseLogicalChannelParameters.m_reverseLogicalChannelNumber, true);
        }
    }
    else
    {
        if (!ack.HasOptionalField(H245_OpenLogicalChannelAck::e_reverseLogicalChannelParameters))
        {
            STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() No reverseLogicalChannelParameters");
            return false;
        }

        if (ack.m_reverseLogicalChannelParameters.m_multiplexParameters.GetTag() !=
                H245_OpenLogicalChannelAck_reverseLogicalChannelParameters_multiplexParameters
                    ::e_h2250LogicalChannelParameters)
        {
            STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() Only H.225.0 multiplex supported");
            return false;
        }

        const H245_H2250LogicalChannelParameters & param = ack.m_reverseLogicalChannelParameters.m_multiplexParameters;

        if (!param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel))
        {
            STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection.GetConnID(),
                "() No media channel address provided");
            return false;
        }

        address = &param.m_mediaChannel;

        unsigned errorCode;
        if (!ExtractTransport(param.m_mediaChannel, m_RemoteAddress, m_RemoteDataPort, errorCode))
            return false;

        if (param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel))
        {
            if (!ExtractTransport(param.m_mediaControlChannel, m_RemoteAddress, m_RemoteControlPort, errorCode))
                return false;
        }

        // If we use SwTCL we want to get parameters out of OLCAck
        if (!connection.OnReceivedLogicalChannelAckPDU(ack))
            return false;
    }
    return true;
}


void H323DataChannel::SetUpTransportPDU(H245_TransportAddress & pdu, Tools::SIpAddress tsap) const
{
    pdu.SetTag(H245_TransportAddress::e_unicastAddress);

    H245_UnicastAddress & unicast = pdu;
    unicast.SetTag(H245_UnicastAddress::e_iPAddress);

    H245_UnicastAddress_iPAddress & addr = unicast;
    addr.m_network.SetValue((DWORD)tsap.m_Address);
    addr.m_tsapIdentifier = tsap.m_Port;
}

/////////////////////////////////////////////////////////////////////////////
