/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Codecs
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/codecs.cpp                 $
    $Author:: Martin.hueser                                                $
      $Date:: 20.12.05 18:34                                               $
   $Modtime:: 19.12.05 9:40                                                $
  $Revision:: 3                                                            $
  $Workfile:: codecs.cpp                                                   $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include <ssvcproc.h>
#include "h323.h"
#include "codecs.h"
#include "channels.h"
#include "h323pdu.h"
#include "videoio.h"

using namespace std;
using namespace Tools;

const unsigned DefaultEnergyThreshold = 100; // 1% of max signal
const unsigned DefaultSignalDeadbandSamples = 480;  // 60 milliseconds
const unsigned DefaultSilenceDeadbandSamples = 4000; // 500 milliseconds


#define new SNEW

#ifdef H323_NO_AUDIOCODECS
// This is to prevent linkage of AudioCodecs.lib.
// This methods should never be called! 
extern "C" {
  unsigned char linear2ulaw(int pcm_val) { return 0; }
  int ulaw2linear(unsigned char u_val)   { return 0; }
  unsigned char linear2alaw(int pcm_val) { return 0; }
  int alaw2linear(unsigned char u_val)   { return 0; }
};
#else
extern "C" {
  unsigned char linear2ulaw(int pcm_val);
  int ulaw2linear(unsigned char u_val);
  unsigned char linear2alaw(int pcm_val);
  int alaw2linear(unsigned char u_val);
};
#endif // H323_NO_AUDIOCODECS



/////////////////////////////////////////////////////////////////////////////

const char * const H323Codec::CodecNames[MaxCodec] = {
    "Undefined",
    "G711a",
    "G711u",
    "G723",
    "G729",
    "G722",
    "GSM_FullRate",
    "T38",
};


H323Codec::H323Codec(Direction dir)
{
  direction = dir;
  rtpPayloadType = RTP_DataFrame::DynamicBase;
}


bool H323Codec::Open(H323Connection & /*connection*/)
{
  return true;
}


void H323Codec::OnFlowControl(long bitRateRestriction)
{
    SDEFFUNC("H323Codec::OnFlowControl");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "(" << bitRateRestriction << ")");
}


void H323Codec::OnMiscellaneousCommand(const H245_MiscellaneousCommand_type & type)
{
    SDEFFUNC("H323Codec::OnMiscellaneousCommand");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "(" << type.GetTagName() << ")");
}


void H323Codec::OnMiscellaneousIndication(const H245_MiscellaneousIndication_type & type)
{
    SDEFFUNC("H323Codec::OnMiscellaneousIndica");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "(" << type.GetTagName() << ")");
}


/////////////////////////////////////////////////////////////////////////////

H323VideoCodec::H323VideoCodec(Direction dir)
  : H323Codec(dir)
{
  rawDevice = NULL;
  deleteDevice = false;
}


H323VideoCodec::~H323VideoCodec()
{
  Close();

  if (deleteDevice)
    delete rawDevice;
  rawDevice = NULL;
}


bool H323VideoCodec::Open(H323Connection & connection)
{
  return connection.OpenVideoDevice(direction == Encoder, *this);
}


unsigned H323VideoCodec::GetTimeUnits() const
{
  return 90; // Assuming standard 90kHz sample rate
}


bool H323VideoCodec::NeedsJitterBuffer() const
{
  return false;
}


void H323VideoCodec::OnMiscellaneousCommand(const H245_MiscellaneousCommand_type & type)
{
  switch (type.GetTag()) {
    case H245_MiscellaneousCommand_type::e_videoFreezePicture :
      OnFreezePicture();
      break;

    case H245_MiscellaneousCommand_type::e_videoFastUpdatePicture :
      OnFastUpdatePicture();
      break;

    case H245_MiscellaneousCommand_type::e_videoFastUpdateGOB :
    {
      const H245_MiscellaneousCommand_type_videoFastUpdateGOB & fuGOB = type;
      OnFastUpdateGOB(fuGOB.m_firstGOB, fuGOB.m_numberOfGOBs);
      break;
    }

    case H245_MiscellaneousCommand_type::e_videoFastUpdateMB :
    {
      const H245_MiscellaneousCommand_type_videoFastUpdateMB & fuMB = type;
      OnFastUpdateMB(fuMB.HasOptionalField(H245_MiscellaneousCommand_type_videoFastUpdateMB::e_firstGOB) ? (int)fuMB.m_firstGOB : -1,
                     fuMB.HasOptionalField(H245_MiscellaneousCommand_type_videoFastUpdateMB::e_firstMB)  ? (int)fuMB.m_firstMB  : -1,
                     fuMB.m_numberOfMBs);
      break;
    }
  }

  H323Codec::OnMiscellaneousCommand(type);
}


void H323VideoCodec::OnFreezePicture()
{
}


void H323VideoCodec::OnFastUpdatePicture()
{
}


void H323VideoCodec::OnFastUpdateGOB(unsigned /*firstGOB*/, unsigned /*numberOfGOBs*/)
{
}


void H323VideoCodec::OnFastUpdateMB(int /*firstGOB*/, int /*firstMB*/, unsigned /*numberOfMBs*/)
{
}


void H323VideoCodec::OnMiscellaneousIndication(const H245_MiscellaneousIndication_type & type)
{
  switch (type.GetTag()) {
    case H245_MiscellaneousIndication_type::e_videoIndicateReadyToActivate :
      OnVideoIndicateReadyToActivate();
      break;

    case H245_MiscellaneousIndication_type::e_videoTemporalSpatialTradeOff :
      OnVideoTemporalSpatialTradeOff();
      break;

    case H245_MiscellaneousIndication_type::e_videoNotDecodedMBs :
    {
      const H245_MiscellaneousIndication_type_videoNotDecodedMBs & vndMB = type;
      OnVideoNotDecodedMBs(vndMB.m_firstMB, vndMB.m_numberOfMBs, vndMB.m_temporalReference);
      break;
    }
  }

  H323Codec::OnMiscellaneousIndication(type);
}


void H323VideoCodec::OnVideoIndicateReadyToActivate()
{
}


void H323VideoCodec::OnVideoTemporalSpatialTradeOff()
{
}


void H323VideoCodec::OnVideoNotDecodedMBs(unsigned /*firstMB*/,
                                          unsigned /*numberOfMBs*/,
                                          unsigned /*temporalReference*/)
{
}


bool H323VideoCodec::AttachDevice(H323VideoDevice * channel, bool autoDelete)
{
  if (deleteDevice)
    delete rawDevice;

  rawDevice    = channel;
  deleteDevice = autoDelete;

  return channel != NULL;
}


void H323VideoCodec::Close()
{
  if (rawDevice != NULL) {
    rawDevice->Close();
    rawDevice = NULL;
  }
}


/////////////////////////////////////////////////////////////////////////////

H323AudioCodec::H323AudioCodec(Direction dir, unsigned samples)
  : H323Codec(dir)
{
  samplesPerFrame = samples;
  rawDataChannel = NULL;
  deleteChannel = false;

  // The silence deadband is the number of frames of low energy that have to
  // occur before the system stops sending data over the RTP link.
  energyThreshold = DefaultEnergyThreshold;
  signalDeadbandFrames = (DefaultSignalDeadbandSamples+samples-1)/samples;
  silenceDeadbandFrames = (DefaultSilenceDeadbandSamples+samples-1)/samples;

  // Start off in silent mode
  framesReceived = 0;
  inTalkBurst = false;
}


H323AudioCodec::~H323AudioCodec()
{
  Close();

  if (deleteChannel)
    delete rawDataChannel;
  rawDataChannel = NULL;
}


bool H323AudioCodec::Open(H323Connection & connection)
{
  return connection.OpenAudioChannel(direction == Encoder, samplesPerFrame*2, *this);
}


unsigned H323AudioCodec::GetFrameRate() const
{
  return samplesPerFrame;
}


unsigned H323AudioCodec::GetTimeUnits() const
{
  return 8; // Assuming standard 8kHz sample rate
}


bool H323AudioCodec::NeedsJitterBuffer() const
{
  return true;
}


bool H323AudioCodec::AttachChannel(SChannel * channel, bool autoDelete)
{
  if (deleteChannel)
    delete rawDataChannel;

  rawDataChannel = channel;
  deleteChannel = autoDelete;

  return channel != NULL;
}


void H323AudioCodec::Close()
{
  if (rawDataChannel != NULL)
    rawDataChannel->Close();
}


void H323AudioCodec::EnableSilenceDetection(bool enable)
{
  if (enable)
    energyThreshold = DefaultEnergyThreshold;
  else
    energyThreshold = 0;
}


bool H323AudioCodec::DetectSilence()
{
  // Can never have silence if threshold zero
  if (energyThreshold == 0)
    return false;

  /* Future enhancement: need to dynamically adjust the threshold value as will
     probably vary greatly from system to system and background noise in the
     environment of the microphone.
   */

  // Now if energy above threshold we are "talking"
  bool stillTalking = CalculateEnergy() > energyThreshold;

  // If no change ie still talking or still silent, resent frame counter
  if (stillTalking == inTalkBurst)
    framesReceived = 0;
  else {
    framesReceived++;
    // If have had enought frames talking/silent, swap modes.
    if (framesReceived > (inTalkBurst ? silenceDeadbandFrames : signalDeadbandFrames))
      inTalkBurst = stillTalking;
  }

  return !inTalkBurst;
}


unsigned H323AudioCodec::CalculateEnergy()
{
  return 0;
}


/////////////////////////////////////////////////////////////////////////////

H323FramedAudioCodec::H323FramedAudioCodec(Direction dir,
                                           unsigned samples,
                                           unsigned bytes)
  : H323AudioCodec(dir, samples),
    sampleBuffer(samples)
{
  bytesPerFrame = bytes;
}


bool H323FramedAudioCodec::Read(BYTE * buffer, unsigned & length)
{
    SDEFFUNC("H323FramedAudioCodec::Read");

  if (rawDataChannel == NULL) {
    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() No audio channel for read");
    return false;
  }

  if (direction != Encoder) {
    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Attempt to decode from encoder");
    return false;
  }

  if (samplesPerFrame > sampleBuffer.size())
      sampleBuffer.resize(samplesPerFrame);

  if (!rawDataChannel->Read(&(*(sampleBuffer.begin())), samplesPerFrame*2)) {
    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Audio read failed: " << rawDataChannel->GetErrorText());
    return false;
  }

  length = DetectSilence() ? 0 : bytesPerFrame;

  // Still do encode even though data is not sent as the particular codec
  // algorithm may not like gaps in the PCM data stream.
  return EncodeFrame(buffer);
}


bool H323FramedAudioCodec::Write(const BYTE * buffer,
                                 unsigned length,
                                 const RTP_DataFrame & /*rtpFrame*/,
                                 unsigned & written)
{
    SDEFFUNC("H323FramedAudioCodec::Write");

  if (rawDataChannel == NULL) {
    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() No audio channel for write");
    return false;
  }

  if (direction != Decoder) {
    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Attempt to encode from decoder");
    return false;
  }

  // If length is zero then it indicates silence, do nothing.
  if (length == 0) {
    written = 0;
    return true;
  }

  if (length < bytesPerFrame) {
    /*Need to add something in here if not enough bytes (ie data in original
      packet was not evenly divisible by bytesPerFrame) to save the partial
      frame till the next call to Write(). Theoretically this is not
      necessary as the H323Capability should dictate the quanta to be used
      but may end up being needed for some codec types where it is supposed
      to be streamed but due to hardware is framed.

      At present this is an error condition.
     */
    return false;
  }

  // Decode the data
  if (!DecodeFrame(buffer))
    return false;

  written = bytesPerFrame;

  // Write as 16bit PCM to sound channel
  if (!rawDataChannel->Write(&(*(sampleBuffer.begin())), samplesPerFrame*2))
     return true;

  STRACE(STrace::Debug, g_pH323Module->m_CodecTraceId, "Write failed: " << rawDataChannel->GetErrorText());

  return true;  // This probably should be false as write error is fatal, maybe
}


unsigned H323FramedAudioCodec::CalculateEnergy()
{
  // Calculate the energy of this frame
  unsigned energy = 0;

  const short * pcm = &(*(sampleBuffer.begin()));
  const short * end = pcm + samplesPerFrame;
  while (pcm != end) {
    if (*pcm < 0)
      energy -= *pcm++;
    else
      energy += *pcm++;
  }

  return energy/samplesPerFrame;
}


unsigned H323FramedAudioCodec::GetBandwidth() const
{
  // Want 100's of bits per second from the number of bytes in a frame
  // the number of time units per frame and number of units per millisecond.
  return 80*bytesPerFrame*GetTimeUnits()/GetFrameRate();
}


/////////////////////////////////////////////////////////////////////////////

H323StreamedAudioCodec::H323StreamedAudioCodec(Direction dir,
                                               unsigned samples,
                                               unsigned bits)
  : H323FramedAudioCodec(dir, samples, (samples*bits+7)/8)
{
  bitsPerSample = bits;
}


bool H323StreamedAudioCodec::EncodeFrame(BYTE * buffer)
{
  int i;

  switch (bitsPerSample) {
    case 8 :
      for (i = 0; i < (int)samplesPerFrame; i++)
        *buffer++ = (BYTE)Encode(sampleBuffer[i]);
      break;

    case 4 :
      for (i = 0; i < (int)samplesPerFrame; i++) {
        if ((i&1) == 0)
          *buffer = (BYTE)Encode(sampleBuffer[i]);
        else
          *buffer |= (BYTE)(Encode(sampleBuffer[i]) << 4);
      }
      break;

    default :
      SAssertAlways("Unsupported bit size");
      return false;
  }

  return true;
}


bool H323StreamedAudioCodec::DecodeFrame(const BYTE * buffer)
{
  int i;
  short * out = &(*(sampleBuffer.begin()));

  if (samplesPerFrame > sampleBuffer.size())
      sampleBuffer.resize(samplesPerFrame);

  switch (bitsPerSample) {
    case 8 :
      for (i = 0; i < (int)samplesPerFrame; i++)
        *out++ = Decode(*buffer++);
      break;

    case 4 :
      for (i = 0; i < (int)samplesPerFrame; i++) {
        *out++ = Decode(*buffer & 15);
        *out++ = Decode(*buffer >> 4);
        buffer++;
      }
      break;

    default :
      SAssertAlways("Unsupported bit size");
      return false;
  }

  return true;
}


/////////////////////////////////////////////////////////////////////////////

H323_ALawCodec::H323_ALawCodec(Direction dir,
                               bool at56kbps,
                               unsigned frameSize)
  : H323StreamedAudioCodec(dir, frameSize, 8)
{
    SDEFFUNC("H323_ALawCodec::H323_ALawCodec");

  rtpPayloadType = RTP_DataFrame::PCMA;
  sevenBit = at56kbps;

  STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
      "() G711 ALaw " << (dir == Encoder ? "en" : "de")
      << "coder created for at "
      << (sevenBit ? "56k" : "64k") << ", " << frameSize << "samples");
}



int H323_ALawCodec::Encode(short sample) const
{
  return linear2alaw(sample);
}


short H323_ALawCodec::Decode(int sample) const
{
  return (short)alaw2linear((unsigned char)sample);
}


/////////////////////////////////////////////////////////////////////////////

H323_muLawCodec::H323_muLawCodec(Direction dir,
                                 bool at56kbps,
                                 unsigned frameSize)
  : H323StreamedAudioCodec(dir, frameSize, 8)
{
    SDEFFUNC("H323_muLawCodec::H323_muLawCodec");

  rtpPayloadType = RTP_DataFrame::PCMU;
  sevenBit = at56kbps;

  STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
      "() G711 uLaw " << (dir == Encoder ? "en" : "de")
      << "coder created for at "
      << (sevenBit ? "56k" : "64k") << ", frame of " << frameSize << " samples");
}


int H323_muLawCodec::Encode(short sample) const
{
  return linear2ulaw(sample);
}


short H323_muLawCodec::Decode(int sample) const
{
  return (short)ulaw2linear((unsigned char)sample);
}


//-----------------------------------------------------------------------------
// G.722 Codec (dummy!!!)
//-----------------------------------------------------------------------------

H323_G722Codec::H323_G722Codec(Direction direction)
  : H323FramedAudioCodec(direction, 160, 20)
{
    SDEFFUNC("H323_G722Codec::H323_G722Codec");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.722 dummy " << (direction == Encoder ? "en" : "de")
        << "coder created for " << bytesPerFrame << " bytes per frame");
}

//-----------------------------------------------------------------------------

bool H323_G722Codec::EncodeFrame(BYTE * buffer)
{
    SDEFFUNC("H323_G722Codec::EncodeFrame");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.722 dummy: No 'EncodeFrame' implemented!");
    return false;
}

//-----------------------------------------------------------------------------

bool H323_G722Codec::DecodeFrame(const BYTE * buffer)
{
    SDEFFUNC("H323_G722Codec::DecodeFrame");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.722 dummy: No 'DecodeFrame' implemented!");
    return false;
}

//-----------------------------------------------------------------------------
// G.723.1 Codec (dummy!!!)
//-----------------------------------------------------------------------------

H323_G7231Codec::H323_G7231Codec(Direction direction)
  : H323FramedAudioCodec(direction, 240, 24)
{
    SDEFFUNC("H323_G7231Codec::H323_G7231Codec");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.723.1 dummy " << (direction == Encoder ? "en" : "de")
        << "coder created for " << bytesPerFrame << " bytes per frame");
}

//-----------------------------------------------------------------------------

bool H323_G7231Codec::EncodeFrame(BYTE * buffer)
{
    SDEFFUNC("H323_G7231Codec::EncodeFrame");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.723.1 dummy: No 'EncodeFrame' implemented!");
    return false;
}

//-----------------------------------------------------------------------------

bool H323_G7231Codec::DecodeFrame(const BYTE * buffer)
{
    SDEFFUNC("H323_G7231Codec::DecodeFrame");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.723.1 dummy: No 'DecodeFrame' implemented!");
    return false;
}


//-----------------------------------------------------------------------------
// G.729 Codec (dummy!!!)
//-----------------------------------------------------------------------------

H323_G729Codec::H323_G729Codec(Direction direction)
  : H323FramedAudioCodec(direction, 160, 20)
{
    SDEFFUNC("H323_G729Codec::H323_G729Codec");

    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.729 dummy " << (direction == Encoder ? "en" : "de")
        << "coder created for " << bytesPerFrame << " bytes per frame");
}

//-----------------------------------------------------------------------------

bool H323_G729Codec::EncodeFrame(BYTE * buffer)
{
    SDEFFUNC("H323_G729Codec::EncodeFrame");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.729 dummy: No 'EncodeFrame' implemented!");
    return false;
}

//-----------------------------------------------------------------------------

bool H323_G729Codec::DecodeFrame(const BYTE * buffer)
{
    SDEFFUNC("H323_G729Codec::DecodeFrame");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() G.729 dummy: No 'DecodeFrame' implemented!");
    return false;
}

//-----------------------------------------------------------------------------
