/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Gatekeeper server
 
    Comments: Gatekeeper server protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/gkserver.cpp               $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 6.01.06 11:15                                                $
   $Modtime:: 6.01.06 10:50                                                $
  $Revision:: 19                                                           $
  $Workfile:: gkserver.cpp                                                 $
$NoKeywords: $
----------------------------------------------------------------------------*/

#include <stlib.h>

#include "h323pdu.h"
#include "gkserver.h"

using namespace std;
using namespace Tools;


//-------------------------------------------------------------------

const unsigned DefaultBandwidth = 1280;


//-------------------------------------------------------------------
//-------------------------------------------------------------------

static void AppendTransportAddress(H225_ArrayOf_TransportAddress & pdu,
                                   const H323TransportAddress & rAddr)
{
  int lastPos = pdu.GetSize();

  // Put new transport address into array
  pdu.SetSize(lastPos+1);
  H225_TransportAddress & taddr = pdu[lastPos];

  rAddr.SetPDU(taddr);
}


//-------------------------------------------------------------------
//-------------------------------------------------------------------

// This thread reads all the messages received at the
// ras port (socket).

class H225RasServerThread : public SThread
{
public:
    H225RasServerThread(H323EndPoint & rEndpoint,
                        H323GkServer & rGatekeeperServer);

protected:
    void Main();

    H323GkServer & m_rGatekeeper;
};


//-------------------------------------------------------------------
//-------------------------------------------------------------------

// This thread reads all the messages received at the
// multicast ras port

class H225RasMulticastServerThread : public SThread
{
public:
    H225RasMulticastServerThread(H323EndPoint & rEndpoint,
                                 H323GkServer & rGatekeeperServer);

protected:
    void Main();

    H323GkServer & m_rGatekeeper;
};


//-------------------------------------------------------------------
//-------------------------------------------------------------------

// This thread checks for timeouts if we do not get a response
// in time on requests. Timebased message sending is handled, too.
// E. g. TimeToLive mechanisms.

class H323GkTimeoutThread : public SThread
{
public:
    H323GkTimeoutThread(H323EndPoint & rEndpoint,
                        H323GkServer & rGatekeeperServer);

    void TriggerTimeoutThread() { m_GkTimeoutThreadSync.Signal(); }
    void Stop()                 { m_bStop = true; m_GkTimeoutThreadSync.Signal(); }

protected:
    void Main();

    bool                  m_bStop;
    H323GkServer        & m_rGatekeeper;

    Tools::SSyncPoint     m_GkTimeoutThreadSync; // 
};


//-------------------------------------------------------------------
//-------------------------------------------------------------------


#define new SNEW


//-------------------------------------------------------------------
//-------------------------------------------------------------------

H225RasServerThread::H225RasServerThread(H323EndPoint & rEndpoint,
                                         H323GkServer & rGatekeeperServer) :
        SThread(rEndpoint.GetRasThreadStackSize(), NoAutoDeleteThread),
        m_rGatekeeper(rGatekeeperServer)
{
    Resume();
}

//-------------------------------------------------------------------

void H225RasServerThread::Main()
{
    SDEFFUNC("H225RasServerThread::Main");

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Started H.323 RAS server thread");

    try
    {
        // if HandleRasChannel() returns 'false' (and NO shutdown in progress!)
        // -> we'll restart the _whole_ thread
        if ( m_rGatekeeper.HandleRasChannel() == false && 
             !m_rGatekeeper.GatekeeperShutdownInProgress() )
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Stopped H.323 RAS server thread due internal failure!");
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Trying to restart H.323 RAS server thread in 5 seconds...");
            Sleep(5000);
            m_rGatekeeper.SignalShutdown( H323GkServer::RASThread );
            return;
        }
    }
    catch( SAccessViolation av )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleRasChannel threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << av.What() << " (" << (av.AccessFlag() ? "R/W":"R")
            << ") access to 0x" << (void*)av.AccessedAddress() << " thrown from "
            << av.Where() << " ###");
    }
    catch( SStackOverflow so )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleRasChannel threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << so.What() << " ###");
    }
    catch( SStructuredException se )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleRasChannel threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << se.What() << " at 0x" << se.Where() << " ###");
    }
    catch(...)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() ##### Exception in HandleRasChannel() #####");
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Stopped H.323 RAS server thread");

    if (!m_rGatekeeper.GatekeeperShutdownInProgress())
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Trying to restart H.323 RAS server thread in 5 seconds...");
        Sleep(5000);
        m_rGatekeeper.SignalShutdown( H323GkServer::RASThread );
    }
}

//-------------------------------------------------------------------
//-------------------------------------------------------------------

H225RasMulticastServerThread::H225RasMulticastServerThread(
    H323EndPoint & rEndpoint,
    H323GkServer & rGatekeeperServer) :
        SThread(rEndpoint.GetRasThreadStackSize(), NoAutoDeleteThread),
        m_rGatekeeper(rGatekeeperServer)
{
    Resume();
}

//-------------------------------------------------------------------

void H225RasMulticastServerThread::Main()
{
    SDEFFUNC("H225RasMulticastSrvThread::Main");

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Started H.323 RAS multicast server thread");

    try
    {
        // if HandleRasMulticastChannel() returns 'false' (and NO shutdown in progress!)
        // -> we'll restart the _whole_ thread
        if ( m_rGatekeeper.HandleRasMulticastChannel() == false && 
             !m_rGatekeeper.GatekeeperShutdownInProgress() )
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Stopped H.323 RAS multicast server thread due internal failure!");
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Trying to restart H.323 RAS multicast server thread in 5 seconds...");
            Sleep(5000);
            m_rGatekeeper.SignalShutdown( H323GkServer::RASMulticastThread );
            return;
        }
    }
    catch( SAccessViolation av )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleRasMulticastChannel() threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << av.What() << " (" << (av.AccessFlag() ? "R/W":"R")
            << ") access to 0x" << (void*)av.AccessedAddress() << " thrown from "
            << av.Where() << " ###");
    }
    catch( SStackOverflow so )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleRasMulticastChannel() threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << so.What() << " ###");
    }
    catch( SStructuredException se )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleRasMulticastChannel() threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << se.What() << " at 0x" << se.Where() << " ###");
    }
    catch(...)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() ##### Exception in HandleRasMulticastChannel() #####");
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Stopped H.323 RAS multicast server thread");

    if (!m_rGatekeeper.GatekeeperShutdownInProgress())
    {
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() Trying to restart H.323 RAS multicast server thread in 5 seconds...");
        Sleep(5000);
        m_rGatekeeper.SignalShutdown( H323GkServer::RASMulticastThread );
    }
}

//-------------------------------------------------------------------
//-------------------------------------------------------------------

H323GkTimeoutThread::H323GkTimeoutThread(
    H323EndPoint & rEndpoint,
    H323GkServer & rGatekeeperServer) :
        SThread(rEndpoint.GetRasThreadStackSize(), NoAutoDeleteThread),
        m_rGatekeeper(rGatekeeperServer),
        m_bStop(false)
{
    Resume();
};

//-------------------------------------------------------------------

void H323GkTimeoutThread::Main()
{
    SDEFFUNC("H323GkTimeoutThread::Main");

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Started H.323 RAS server timeout thread");

    try
    {
        while ( !m_bStop )
        {
            STRACE2_INFO3( g_pH323Module->m_GkTraceId, this, 0,
                "() Process registered endpoints timer list...");

            STimeInterval SleepTime = m_rGatekeeper.GetRegisteredEndpoints().ProcessRegTimerList();
            if ( SleepTime > STimeInterval(0, 60) ) // wait max 1 minute!
                SleepTime = 60000;
            
            STRACE2_INFO3( g_pH323Module->m_GkTraceId, this, 0,
                "() Next timer list processing in " << SleepTime.GetSeconds() << "sec" );

            m_GkTimeoutThreadSync.Wait(SleepTime);
        }
    }
    catch( SAccessViolation av )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleTimeout() threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << av.What() << " (" << (av.AccessFlag() ? "R/W":"R")
            << ") access to 0x" << (void*)av.AccessedAddress() << " thrown from "
            << av.Where() << " ###");
    }
    catch( SStackOverflow so )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleTimeout() threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << so.What() << " ###");
    }
    catch( SStructuredException se )
    {
        STRACE2_FATAL(g_pH323Module->m_GkTraceId, this, 0,
            "() HandleTimeout() threw an exception:\n" << STrace::MultiLinePrefix
            << " ### " << se.What() << " at 0x" << se.Where() << " ###");
    }
    catch(...)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() ##### Exception in HandleTimeout() #####");
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() Stopped H.323 RAS server timeout thread");

    if (!m_rGatekeeper.GatekeeperShutdownInProgress())
        m_rGatekeeper.SignalShutdown( H323GkServer::TimeoutThread );
}

void H323GkServer::SignalRegistrationTimerListChange()
{
    if (!SBAD_READPTR(m_pTimeoutThread))
        m_pTimeoutThread->TriggerTimeoutThread();
}

//-------------------------------------------------------------------
//-------------------------------------------------------------------

// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H323GkServer::H323GkServer( H323EndPoint            & rEndpoint,
                            const Tools::SIpAddress & rTransportAddress,
                            const Tools::SIpAddress & rMulticastTransportAddress,
                            bool GkRoutedCallModel ) :
        m_rEndpoint(rEndpoint),
        m_TransportAddress(rTransportAddress),
        m_MulticastTransportAddress(rMulticastTransportAddress),
        m_LastTransportError(0),
        m_LastSequenceNumber(0),
        m_RegisteredEPList(*this, false),
        m_GkRoutedCallModel(GkRoutedCallModel),
        m_GatekeeperIdentifier(SProcess::Current().GetManufacturer() + " Gatekeeper")
#ifdef _DEBUG // add name to H323 gatekeeper mutex
      , m_Mutex(Tools::SString(Tools::SString::Printf, "H323GkServer::Mutex[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
#pragma warning(default:4355)
{
    SDEFFUNC("H323GkServer::H323GkServer");

    m_LastSequenceNumber = 0;
    m_RasTimeout = rEndpoint.GetRasRequestTimeout();
    m_RasRetries = rEndpoint.GetRasRequestRetries();
    
    m_ShutdownGatekeeper = false;

    m_pTransport          = new H323TransportUDP( m_rEndpoint,
                                                  m_TransportAddress.m_Address,
                                                  m_TransportAddress.m_Port );

    m_pMulticastTransport = new H323TransportUDP( m_rEndpoint,
                                                  m_MulticastTransportAddress.m_Address,
                                                  m_MulticastTransportAddress.m_Port,
                                                  true );
    if (m_pTransport)
    {
        m_pTransport->AttachThread(
            new H225RasServerThread(m_rEndpoint, *this));
    }

    if (m_pMulticastTransport)
    {
        m_pMulticastTransport->AttachThread(
            new H225RasMulticastServerThread(m_rEndpoint, *this));
    }
    
    m_pTimeoutThread = new H323GkTimeoutThread(m_rEndpoint, *this);

    m_pThreadPool = new SThreadPool();

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-------------------------------------------------------------------

H323GkServer::~H323GkServer()
{
    SDEFFUNC("H323GkServer::~H323GkServer");

    if (m_ShutdownGatekeeper == false)
        Shutdown();

    if (m_pTransport != 0)
    {
        m_pTransport->CleanUpOnTermination();
        delete m_pTransport;
    }

    if (m_pMulticastTransport != 0)
    {
        m_pMulticastTransport->CleanUpOnTermination();
        delete m_pMulticastTransport;
    }
    
    if (m_pTimeoutThread != 0)
    {
        m_pTimeoutThread->Stop();
        m_pTimeoutThread->WaitForTermination();
        delete m_pTimeoutThread;
    }

    // All threads of the GK are terminated, so no endpoints
    // can be deleted by reception of a ucf.
    // Set autodelete to true, for deleting all
    // the endpoints in the H323Registration destructor,
    // so we do not have to loop through the endpoints list
    // for ourself.
    m_RegisteredEPList.SetAutoDelete();
    // destructor of m_RegisteredEPList is called automatically!

    delete m_pThreadPool;

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0, "()");
}

//-------------------------------------------------------------------

void H323GkServer::PrintOn(ostream & rStream) const
{
    rStream << "Gatekeeper" << "\n" << STrace::MultiLinePrefix;
    rStream << "GatekeeperId = " << m_GatekeeperIdentifier << "\n" << STrace::MultiLinePrefix;

    rStream << m_RegisteredEPList;
}

//-------------------------------------------------------------------

void H323GkServer::ForceEraseClient(const Tools::SString    & szH323Alias,
                                    const Tools::SIpAddress & rIpAddress)
{
    m_Mutex.Wait();

    H323RemoteEndpointVector EPsToErase = m_RegisteredEPList.FindEndpointsByAlias(szH323Alias);
    H323RemoteEndpointVector::iterator it;

    for (it = EPsToErase.begin(); it != EPsToErase.end(); ++it)
    {
        H323RemoteEndpoint * pEndpoint = (*it);
        if (!SBAD_READPTR(pEndpoint))
        {
            // Unregister and...
            // (we do not wait for UCF - it will be discarded)
            UnregistrationRequest(*pEndpoint);
            // delete this endpoint
            m_RegisteredEPList.Erase(pEndpoint->m_EndpointIdentifier);
            // remove reference -> EP may be deleted NOW
            pEndpoint->DecRef(__FILE__, __LINE__);
        }
    }
    m_Mutex.Signal();
}

//-------------------------------------------------------------------

void H323GkServer::ForceEraseClient(const Tools::SString & szH323Alias)
{
    m_Mutex.Wait();

    Tools::SString szLowerAlias = szH323Alias.ToLower();

    H323RemoteEndpointHashmap & rEndpoints = m_RegisteredEPList.GetEndpoints();
    H323RemoteEndpointHashmap::iterator it = rEndpoints.begin();

    while (it != rEndpoints.end())
    {
        if (   !SBAD_READPTR((*it).second)
            && (   ((*it).second->m_H323GkAliases.m_RegisteredH323Alias.ToLower() == szLowerAlias)
                || ((*it).second->FindMixedAlias(szH323Alias) == true) ) )
        {
            // store the original iterator position. This one will be deleted inside
            // the call to the method Erase of the m_RegisteredEPList!
            H323RemoteEndpointHashmap::iterator TempIt = it;
            it++;

            // Unregister and...
            // (we do not wait for UCF - it will be discarded)
            H323RemoteEndpoint * pEndpoint = (*TempIt).second;
            UnregistrationRequest(*pEndpoint);
            // delete this endpoint
            m_RegisteredEPList.Erase(pEndpoint->m_EndpointIdentifier);
        }
        else
        {
            it++;
        }
    }

    m_Mutex.Signal();
}

//-------------------------------------------------------------------

void H323GkServer::UnregisterAllClients()
{
    m_Mutex.Wait();

    H323RemoteEndpoint * pEndpoint = 0;
    SMutex& rRegisteredEPListMutex = m_RegisteredEPList.GetRegTimerListMutex();
    H323RemoteEndpointHashmap & rEndpoints = m_RegisteredEPList.GetEndpoints();

    rRegisteredEPListMutex.Wait();
    H323RemoteEndpointHashmap::iterator EpIter = rEndpoints.begin();
    while ( EpIter != rEndpoints.end() )
    {
        pEndpoint = EpIter->second;
        if (!SBAD_READPTR(pEndpoint))
        {
            pEndpoint->IncRef(__FILE__, __LINE__);
            // Unregister and...
            // (we do not wait for UCF - it will be discarded)
            rRegisteredEPListMutex.Signal();
            UnregistrationRequest(*pEndpoint);
            rRegisteredEPListMutex.Wait();
            // delete this endpoint
            m_RegisteredEPList.Erase(pEndpoint->m_EndpointIdentifier);
            // reset iterator
            EpIter = rEndpoints.begin();
        }
    }
    rRegisteredEPListMutex.Signal();
    
    m_Mutex.Signal();
}

//-------------------------------------------------------------------

void H323GkServer::Shutdown()
{
    // set shutdown here, to reject all incoming rrq's after
    // sending the urq's to the endpoints
    m_ShutdownGatekeeper = true;

    UnregisterAllClients();
}

//-------------------------------------------------------------------

void H323GkServer::SignalShutdown( GkServerThreadType threadType )
{
    m_pThreadPool->QueueWorkItem( SCREATE_NOTIFIER(OnThreadShutdown),
                                  (int)threadType,
                                  SThreadPool::ExecLongFunction);
}

//-------------------------------------------------------------------

void H323GkServer::OnThreadShutdown(Tools::SThreadPool&, int threadtype)
{
    SDEFFUNC("H323GkServer::OnThreadShutdown");

    switch( threadtype )
    {
    case RASMulticastThread :
        if (m_pMulticastTransport != 0)
        {
            m_pMulticastTransport->CleanUpOnTermination();

            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RAS multicast thread stopped");

            delete m_pMulticastTransport;
            m_pMulticastTransport = 0;
        }
        if ( !m_ShutdownGatekeeper )
        {
            m_pMulticastTransport = new H323TransportUDP( m_rEndpoint,
                                                      m_MulticastTransportAddress.m_Address,
                                                      m_MulticastTransportAddress.m_Port,
                                                      true );
            if (m_pMulticastTransport)
            {
                m_pMulticastTransport->AttachThread(new H225RasMulticastServerThread(m_rEndpoint, *this));

                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() RAS multicast thread restarted");
            }
        }
        break;

    case RASThread :
        if (m_pTransport != 0)
        {
            m_pTransport->CleanUpOnTermination();

            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RAS thread stopped");

            delete m_pTransport;
            m_pTransport = 0;
        }

        if ( !m_ShutdownGatekeeper )
        {
            m_pTransport = new H323TransportUDP( m_rEndpoint,
                                             m_TransportAddress.m_Address,
                                             m_TransportAddress.m_Port );
            if (m_pTransport)
            {
                m_pTransport->AttachThread(new H225RasServerThread(m_rEndpoint, *this));

                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() RAS thread restarted");
            }
        }
        break;

    case TimeoutThread :
        if (m_pTimeoutThread != 0)
        {
            m_pTimeoutThread->Stop();
            m_pTimeoutThread->WaitForTermination();

            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RAS timeout thread stopped");

            delete m_pTimeoutThread;
            m_pTimeoutThread = 0;
        }
        if ( !m_ShutdownGatekeeper )
        {
            m_pTimeoutThread = new H323GkTimeoutThread(m_rEndpoint, *this);

            if (m_pTimeoutThread)
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() RAS timeout thread restarted");
            }
        }            
        break;

    default :
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Unknown thread type " << threadtype);
    }
}

//-------------------------------------------------------------------

bool H323GkServer::HandleRasRegistrationTimeout(H323RemoteEndpoint * pEndpoint)
{
    SDEFFUNC("H323GkServer::HdlRasRegTimeout");

    if ( SBAD_READPTR(pEndpoint) )
        return false;

    bool bDeleteEndpoint = false;

    if ( pEndpoint->m_RegistrationStatus != H323RemoteEndpoint::RegistrationTimerExpired )
    {
        pEndpoint->m_RegistrationStatus  = H323RemoteEndpoint::RegistrationTimerExpired;
        pEndpoint->m_uRequestRetryNumber = 0;

        // first off all we release EP from GK to free licenses etc.
        OnReleaseRemoteEndpoint(
            pEndpoint->m_IpAddress,
            pEndpoint->m_H323GkAliases,
            pEndpoint->m_pUserData,
            pEndpoint->m_TerminalType);

        if ( pEndpoint->m_TerminalType == H323EndPoint::e_GatewayOnly )
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Registration timeout - sent URQ to gateway with alias '"
                << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "' and line id "
                << pEndpoint->m_iLineId );
        }
        else if (!pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.empty())
        {
            SStringVector::const_iterator Iter;
            SStringStream Aliases;
            for ( Iter = pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.begin();
                  Iter != pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.end();
                  ++Iter )
            {
                Aliases << "'" << (*Iter).c_str() << "', ";
            }
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Registration timeout - sent URQ to endpoint with H.323 alias(es) "
                << Aliases );
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Registration timeout - sent URQ to endpoint with H.323 alias '"
                << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "'");
        }

        if (!UnregistrationRequest(*pEndpoint))
        {
            // transport object invalid OR last socket operation failed???
            if (SBAD_READPTR(m_pTransport) || m_LastTransportError != 0)
            {
                // kick failing endpoint from the hashmap!
                bDeleteEndpoint = true;
            }
        }

        if ( !bDeleteEndpoint )
        {
            // set timeout value to RAS default timeout for URQ retry and reinsert the endpoint!
            pEndpoint->m_TimeToLive = STimer::Tick() + m_RasTimeout;
            GetRegisteredEndpoints().AddEPToRegTimerList( pEndpoint );
        }
    }
    else if (    pEndpoint->m_RegistrationStatus == H323RemoteEndpoint::RegistrationTimerExpired
              && ++pEndpoint->m_uRequestRetryNumber <= m_RasRetries )
    {
        if ( SendRasPdu(pEndpoint->m_RasAddress,
                        pEndpoint->m_RasPdu) == true )
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Resent " << pEndpoint->m_RasPdu.GetTagName()
                << ", retry " << pEndpoint->m_uRequestRetryNumber );

            // set timeout value to RAS default timeout for URQ retry and reinsert the endpoint!
            pEndpoint->m_TimeToLive = STimer::Tick() + m_RasTimeout;
            GetRegisteredEndpoints().AddEPToRegTimerList( pEndpoint );
        }
        else // resending message failed!
        {
            // transport object invalid OR last socket operation failed???
            if (SBAD_READPTR(m_pTransport) || m_LastTransportError != 0)
            {
                STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                    "() Failed to resent " << pEndpoint->m_RasPdu.GetTagName() << " to endpoint with H.323 alias '"
                    << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "' -> will be deleted!");
                // kick failing endpoint from the hashmap!
                bDeleteEndpoint = true;
            }
        }
    }
    else // number of retries exceeded...
    {
        // ...kick failing endpoint from the hashmap!
        bDeleteEndpoint = true;
    }

    // if we never received an ucf at
    // our urq, and the ttl did a timeout
    // OR the transport socket failed
    // OR all the retries did not result in
    // an ucf, well delete the endpoint
    if ( bDeleteEndpoint == true )
    {
        if (!pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.empty())
        {
            SStringVector::const_iterator Iter;
            SStringStream Aliases;
            for ( Iter = pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.begin();
                  Iter != pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.end();
                  ++Iter )
            {
                Aliases << "'" << (*Iter).c_str() << "', ";
            }
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Unregistration failed - force release of endpoint with H.323 alias(es) "
                << Aliases );
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Unregistration failed - force release of endpoint with H.323 alias '"
                << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "'");
        }
        // kick failing endpoint from the hashmap...
        m_RegisteredEPList.Erase(pEndpoint->m_EndpointIdentifier);
    }

    return bDeleteEndpoint;    
}

//-------------------------------------------------------------------

bool H323GkServer::HandleRasChannel()
{
    SDEFFUNC("H323GkServer::HandleRasChannel");

    unsigned uCountDiscardedActions = 0;

    m_pTransport->SetReadTimeout(SMaxTimeInterval);

    H323TransportUDP *pTransportUDP = dynamic_cast<H323TransportUDP *>(m_pTransport);
    if (pTransportUDP == 0)
        return false;

    while(pTransportUDP->IsOpen())
    {
        RasPDU * pRasPDU = new RasPDU;
        if (pRasPDU == 0)
            break;

        switch(pRasPDU->m_H323RasPDU.Read(*pTransportUDP))
        {
            case H323RasPDU::Success:
            {
                // Get the address/port where this RAS PDU did come from.
                // With the help of this information we will find the
                // appropriate endpoint to communicate with (send a
                // response to).
                pTransportUDP->GetLastReceiveAddress( pRasPDU->m_ReceiveIpAddress.m_Address,
                                                      pRasPDU->m_ReceiveIpAddress.m_Port );
                Tools::SString szAddress("ip:");
                szAddress += pRasPDU->m_ReceiveIpAddress.AsString();
                pRasPDU->m_ReceiveTransportAddress = H323TransportAddress(szAddress);
                STRACE2_INFO3(g_pH323Module->m_RasTraceId, this, 0,
                    "() Transport address: " << pRasPDU->m_ReceiveTransportAddress );

                // push PDU object to work item
                pRasPDU->IncRef(__FILE__, __LINE__);
                m_pThreadPool->QueueWorkItem( SCREATE_NOTIFIER(OnRASRequest),
                                              (int)pRasPDU,
                                              SThreadPool::ExecLongFunction);
            }
            break;

            case H323RasPDU::TransportFailure:
            {
                if (pTransportUDP->GetErrorNumber() != 0)
                {
                    STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                        "() Read on RAS channel failed: error=0x"
                        << hex << pTransportUDP->GetErrorNumber() << dec << ", " << pTransportUDP->GetErrorText());
                }
                return false;
            }
            break;

            case H323RasPDU::DiscardLastAction:
            {
                // This can occure and is not handled as an error, because
                // when gatekeeper is down and we send e.g. a RRQ we get
                // an icmp-message from the machine running gatekeeper.
                // This will result in a failed read-operation on the socket,
                // but here we will ignore it and everything works fine!
                if ( ++uCountDiscardedActions == 1 )
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Ignoring last error on RAS transport channel!");
                }
                else if ( uCountDiscardedActions > 10 )
                    uCountDiscardedActions = 0;
            }
            break;
                
            case H323RasPDU::DecodeFailure:
            {
                STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
                    "() Decode error - discarding RAS message!");
            }
            break;

            default:
            {
                STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
                    "() Unknown error reading RAS message - ignoring it!");
            }
            break;
        }
    }
    return true;
}

//-------------------------------------------------------------------

bool H323GkServer::HandleRasMulticastChannel()
{
    SDEFFUNC("H323GkServer::HandleRasMCastCh");

    unsigned uCountDiscardedActions = 0;

    H323TransportUDP *pMCastTransportUDP = dynamic_cast<H323TransportUDP *>(m_pMulticastTransport);
    if (pMCastTransportUDP == 0)
        return false;

    pMCastTransportUDP->SetReadTimeout(SMaxTimeInterval);

    while(pMCastTransportUDP->IsOpen())
    {
        RasPDU * pRasPDU = new RasPDU;
        if (pRasPDU == 0)
            break;

        switch(pRasPDU->m_H323RasPDU.Read(*pMCastTransportUDP))
        {
            case H323RasPDU::Success:
            {
                // Get the address/port where this RAS PDU did come from.
                // With the help of this information we will find the
                // appropriate endpoint to communicate with (send a
                // response to).
                pMCastTransportUDP->GetLastReceiveAddress( pRasPDU->m_ReceiveIpAddress.m_Address,
                                                           pRasPDU->m_ReceiveIpAddress.m_Port );

                Tools::SString szAddress("ip:");
                szAddress += pRasPDU->m_ReceiveIpAddress.AsString();
                pRasPDU->m_ReceiveTransportAddress = H323TransportAddress(szAddress);
                STRACE2_INFO3(g_pH323Module->m_RasTraceId, this, 0,
                    "() Transport address: " << pRasPDU->m_ReceiveTransportAddress );

                // push PDU object to work item
                pRasPDU->IncRef(__FILE__, __LINE__);
                m_pThreadPool->QueueWorkItem( SCREATE_NOTIFIER(OnRASRequest),
                                              (int)pRasPDU,
                                              SThreadPool::ExecLongFunction);
            }
            break;

            case H323RasPDU::TransportFailure:
            {
                if (pMCastTransportUDP->GetErrorNumber() != 0)
                {
                    STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                        "() Read on RAS channel failed: error=0x"
                        << hex << pMCastTransportUDP->GetErrorNumber() << dec << ", " << pMCastTransportUDP->GetErrorText());
                }
                return false;
            }
            break;

            case H323RasPDU::DiscardLastAction:
            {
                // This can occure and is not handled as an error, because
                // when gatekeeper is down and we send e.g. a RRQ we get
                // an icmp-message from the machine running gatekeeper.
                // This will result in a failed read-operation on the socket,
                // but here we will ignore it and everything works fine!
                if ( ++uCountDiscardedActions == 1 )
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                        "() Ignoring last error on RAS transport channel!");
                }
                else if ( uCountDiscardedActions > 10 )
                    uCountDiscardedActions = 0;
            }
            break;
                
            case H323RasPDU::DecodeFailure:
            {
                STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
                    "() Decode error - discarding RAS message!");
            }
            break;

            default:
            {
                STRACE2_WARN(g_pH323Module->m_RasTraceId, this, 0,
                    "() Unknown error reading RAS message - ignoring it!");
            }
            break;
        }
    }
    return true;
}

//-------------------------------------------------------------------
void H323GkServer::OnRASRequest(Tools::SThreadPool&, int pdu)
{
    H323GkServer::RasPDU *pRasPDU = (H323GkServer::RasPDU *)pdu;
    HandleRasPDU(pRasPDU);
    pRasPDU->DecRef(__FILE__, __LINE__);
}

//-------------------------------------------------------------------

bool H323GkServer::HandleRasPDU(H323GkServer::RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::HandleRasPDU");

    if (SBAD_READPTR(pRasPDU))
        return false;

    switch (pRasPDU->m_H323RasPDU.GetTag())
    {
        case H225_RasMessage::e_gatekeeperRequest:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS GRQ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveGatekeeperRequest(pRasPDU);

        case H225_RasMessage::e_registrationRequest:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS RRQ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveRegistrationRequest(pRasPDU);
            
        case H225_RasMessage::e_unregistrationRequest:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS URQ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveUnregistrationRequest(pRasPDU);

        case H225_RasMessage::e_unregistrationConfirm:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS UCF from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveUnregistrationConfirm(pRasPDU);

        case H225_RasMessage::e_unregistrationReject:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS URJ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveUnregistrationReject(pRasPDU);
            
        case H225_RasMessage::e_admissionRequest:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS ARQ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveAdmissionRequest(pRasPDU);

        case H225_RasMessage::e_bandwidthRequest:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS BRQ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveBandwidthRequest(pRasPDU);

        case H225_RasMessage::e_bandwidthConfirm:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS BCF from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveBandwidthConfirm(pRasPDU);

        case H225_RasMessage::e_bandwidthReject:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS BRJ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveBandwidthReject(pRasPDU);

        case H225_RasMessage::e_infoRequestResponse:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS IRR from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveInfoRequestResponse(pRasPDU);

        case H225_RasMessage::e_disengageRequest:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS DRQ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveDisengageRequest(pRasPDU);

        case H225_RasMessage::e_disengageConfirm:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS DCF from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveDisengageConfirm(pRasPDU);

        case H225_RasMessage::e_disengageReject:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS DRJ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveDisengageReject(pRasPDU);
            
        case H225_RasMessage::e_locationRequest:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS LRQ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveLocationRequest(pRasPDU);

        case H225_RasMessage::e_locationConfirm:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS LCF from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveLocationConfirm(pRasPDU);

        case H225_RasMessage::e_locationReject:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS LRJ from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveLocationReject(pRasPDU);

        case H225_RasMessage::e_nonStandardMessage:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS NonStandardMessage from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveNonStandardMessage(pRasPDU);

        case H225_RasMessage::e_unknownMessageResponse:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS UnknownMessageResponse from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveUnknown(pRasPDU);

        case H225_RasMessage::e_requestInProgress:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS RIP from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveRequestInProgress(pRasPDU);

        case H225_RasMessage::e_resourcesAvailableIndicate:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS RAI from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveResourcesAvailableIndicate(pRasPDU);

        default:
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() RECV RAS unknown message from " << pRasPDU->m_ReceiveIpAddress); 
            return OnReceiveUnknown(pRasPDU);
    }

    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::GatekeeperConfirm()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::GatekeeperReject()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::RegistrationConfirm()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::RegistrationReject()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::UnregistrationRequest(H323RemoteEndpoint & rEndpoint)
{
    SDEFFUNC("H323GkServer::UnregistrationReq");
    H323RasPDU Request;
    unsigned uReqSeqNumber = GetNextSequenceNumber();
    H225_UnregistrationRequest & urq =
        Request.BuildUnregistrationRequest(uReqSeqNumber);
    rEndpoint.m_uSequenceNumber = uReqSeqNumber;

    AppendTransportAddress(urq.m_callSignalAddress,
                           rEndpoint.m_CallSignalAddress);

    if (!rEndpoint.m_EndpointIdentifier.empty())
    {
        urq.IncludeOptionalField(
            H225_UnregistrationRequest::e_endpointIdentifier);
        urq.m_endpointIdentifier = rEndpoint.m_EndpointIdentifier;

        // if we unregister the endpoint, because of the ttl mechanism
        // we tell the endpoint the appropriate reason
        if (rEndpoint.m_TimeToLive != 0 &&
            rEndpoint.m_RegistrationStatus == H323RemoteEndpoint::RegistrationTimerExpired)
        {
            urq.IncludeOptionalField(
                H225_UnregistrationRequest::e_reason);
            urq.m_reason.SetTag(H225_UnregRequestReason::e_ttlExpired);
        }
        else
        {
            urq.IncludeOptionalField(
                H225_UnregistrationRequest::e_reason);
            urq.m_reason.SetTag(H225_UnregRequestReason::e_reregistrationRequired);
        }
    }

    OnSendUnregistrationRequest(urq);

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS URQ to " << rEndpoint.m_IpAddress); 

    return MakeRequest(rEndpoint, Request);
}

//-------------------------------------------------------------------

bool H323GkServer::UnregistrationConfirm()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::UnregistrationReject()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::AdmissionConfirm()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::AdmissionReject()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::BandwidthRequest()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::BandwidthConfirm()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::BandwidthReject()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::InfoRequest()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::InfoRequestAck()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::InfoRequestNack()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::DisengageRequest()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::DisengageConfirm()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::DisengageReject()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::LocationRequest()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::LocationConfirm()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::LocationReject()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::NonStandardMessage()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::UnknownMessageResponse()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::RequestInProgress()
{
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::RequestAvailableConfirm()
{
    return false;
}
    
//-------------------------------------------------------------------


bool H323GkServer::OnReceiveGatekeeperRequest( RasPDU *pRasPDU )
{
    SDEFFUNC("H323GkServer::OnReceiveGRQ");

    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_GatekeeperRequest & grq = (const H225_GatekeeperRequest &)(pRasPDU->m_H323RasPDU);

    H323TransportAddress RecvRasAddress;
    SIPSocket::Address   RecvRasIP;
    WORD                 RecvRasPort;

    H323RasPDU           Response;

    H225_GatekeeperConfirm & gcf =
        Response.BuildGatekeeperConfirm(grq.m_requestSeqNum);

    if (!m_GatekeeperIdentifier.empty())
    {
        gcf.IncludeOptionalField(H225_GatekeeperConfirm::e_gatekeeperIdentifier);
        gcf.m_gatekeeperIdentifier = m_GatekeeperIdentifier;
    }

    m_pTransport->SetUpTransportPDU(gcf.m_rasAddress, true);

    // Get RAS address out of GRQ
    RecvRasAddress = grq.m_rasAddress;
    if (RecvRasAddress.GetIpAndPort(RecvRasIP, RecvRasPort) == false)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Could not get RAS IP address and port out of GRQ -> ignoring GRQ!");
        return false;
    }

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS GCF to " << RecvRasIP << ":" << RecvRasPort); 

    return SendRasPdu(RecvRasAddress, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveRegistrationRequest(RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::OnReceiveRRQ");

    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_RegistrationRequest & rrq = (const H225_RegistrationRequest &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = 0;
    bool IsNewEndpoint = false;
    bool SendConfirm   = false;
    bool bIsTA1100     = false;
    bool bIsBadTA1100  = false;
    bool bRequiresDelayedDisconnect  = false;   // this is a really BAD client (e.g. ip600) which needs much helpin disconnection a call (grrr)
    bool bAliasAlreadyRegistered     = false;

    SWString      szVersionInfo;
    SStringVector E164Aliases;
    SStringVector H323Aliases;
    SStringVector::const_iterator  Iter;
    SStringVector::const_iterator  Iter2;

    SwyxH225NonStandardData nsData(g_pH323Module->m_RasTraceId);

    STimeInterval TimeToLive;
    H323RasPDU           Response;
    H323TransportAddress RecvRasAddress;
    SIPSocket::Address   RecvRasIP;
    WORD                 RecvRasPort;
    H323GkSrv::Err err = H323GkSrv::Failure;
    // default -> we consider codecs G.711 a/u and T.38 are available for H.323 devices!
    int iSupportedCodecs =   H323RemoteEndpoint::CodecG711a
                           | H323RemoteEndpoint::CodecG711u
                           | H323RemoteEndpoint::CodecG729
                           | H323RemoteEndpoint::CodecT38;

    // get call signalling address from RRQ
    SIPSocket::Address    RecvCallSignalAddr;
    WORD                  RecvCallSignalAddrPort = 0;
    H323TransportAddress  RecvCallSignalAddress;
    Tools::SIpAddress     callSignalAddress;

    // Get RAS address out of RRQ
    if (rrq.m_rasAddress.GetSize() == 0)
    {
        RecvRasAddress = pRasPDU->m_ReceiveTransportAddress;

        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Size of rasAddress field is zero! Using last received IP address: " << pRasPDU->m_ReceiveTransportAddress);
    }
    else
    {
        RecvRasAddress = H323GetTransportAddress(rrq.m_rasAddress,
                                                 H225_TransportAddress::e_ipAddress);
    }

    if (RecvRasAddress.GetIpAndPort(RecvRasIP, RecvRasPort) == false)
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Could not get RAS IP address and port out of RRQ -> ignoring RRQ!");
        return false;
    }

    // get call signalling address from FULL RRQ 
    // NOTE:  Cisco ATA 186 / AddPac Voicefinder 'keep alive' RRQ messages do NOT contain valid call signalling address :-( !!!
    RecvCallSignalAddress = H323GetTransportAddress(rrq.m_callSignalAddress,
                                                    H225_TransportAddress::e_ipAddress);
    if ( rrq.m_callSignalAddress.GetSize() == 0 || RecvCallSignalAddress.GetIpAndPort(RecvCallSignalAddr, RecvCallSignalAddrPort) == false )
    {
        if ( rrq.m_keepAlive == false )
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Could not get CallSignalling channel IP address and port out of RRQ -> ignoring RRQ!");
            return false;
        }
        else // use last received address!
        {
            callSignalAddress = pRasPDU->m_ReceiveIpAddress;
        }
    }
    else
    {
        callSignalAddress.m_Address = RecvCallSignalAddr;
        callSignalAddress.m_Port    = RecvCallSignalAddrPort;
    }

    // Map terminalType from RRQ
    H323EndPoint::TerminalTypes TerminalType;
    if (rrq.m_terminalType.HasOptionalField(H225_EndpointType::e_gateway))
    {
        TerminalType = H323EndPoint::e_GatewayOnly;
/*
        // Workaround for terminal adapter using terminal type gateway, e.g. COM21
        H225_H221NonStandard h221ZeroVendor;
        h221ZeroVendor.m_t35CountryCode   = 0;
        h221ZeroVendor.m_t35Extension     = 0;
        h221ZeroVendor.m_manufacturerCode = 0;

        H225_H221NonStandard h221Com21Vendor;
        h221Com21Vendor.m_t35CountryCode   = 181; // 9;
        h221Com21Vendor.m_t35Extension     = 0;
        h221Com21Vendor.m_manufacturerCode = 21325; //62;

        // Compare H.221 NonStandardIdentifier
        if (   (   rrq.m_endpointVendor.m_vendor.m_t35CountryCode.GetValue()
                == h221ZeroVendor.m_t35CountryCode.GetValue() )
            && (   rrq.m_endpointVendor.m_vendor.m_t35Extension.GetValue()
                == h221ZeroVendor.m_t35Extension.GetValue() )
            && (   rrq.m_endpointVendor.m_vendor.m_manufacturerCode.GetValue()
                == h221ZeroVendor.m_manufacturerCode.GetValue() ) )
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() 1 Endpoint is using terminal type 'gateway' and a vendor id 0-0-0. Using terminal type 'terminal'.");

            TerminalType = H323EndPoint::e_TerminalOnly;
        }

        // The following could be used after bugfix in asnparser!
        if (rrq.m_endpointVendor.m_vendor == h221ZeroVendor)
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() 2 Endpoint is using terminal type gateway and a vendor id 0-0-0.");
        }
*/
    }
    else if (rrq.m_terminalType.HasOptionalField(H225_EndpointType::e_gatekeeper))
    {
        TerminalType = H323EndPoint::e_GatekeeperOnly;
    }
    else
    {
        TerminalType = H323EndPoint::e_TerminalOnly;

        // for terminalType == Terminal(-> simple endpoint) we have to check
        // the product- and vendor Id to determine which media codec are supported by device!
        if (rrq.m_endpointVendor.HasOptionalField(H225_VendorIdentifier::e_productId))
        {
            SString szProductIdString(rrq.m_endpointVendor.m_productId.AsString().ToLower());
            szVersionInfo = szProductIdString.ToWChar();
            // get version Id if available for terminal device
            SString szVersionIdString;
            if (rrq.m_endpointVendor.HasOptionalField(H225_VendorIdentifier::e_versionId))
                szVersionIdString = rrq.m_endpointVendor.m_versionId.AsString();

            // check for Kirk ip600/ip1500
            if ( szProductIdString.find("ip600") != SString::npos )
            {
                STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint is a single cell DECT ip600, version= " << szVersionIdString);
                // limit supported codec to G.711a/u
                iSupportedCodecs =   H323RemoteEndpoint::CodecG711a
                                   | H323RemoteEndpoint::CodecG711u;
                // delay the disconnect in some cases!
                bRequiresDelayedDisconnect  = true;

                if ( !szVersionIdString.empty() )
                    szVersionInfo = L"ip600; " + szVersionIdString.ToWChar();
            }
            else if ( szProductIdString.find("ipdect1500") != SString::npos )
            {
                STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint is a multi cell DECT ip1500, version= " << szVersionIdString);
                // limit supported codec to G.711a/u
                iSupportedCodecs =   H323RemoteEndpoint::CodecG711a
                                   | H323RemoteEndpoint::CodecG711u;
                // delay the disconnect in some cases!
                bRequiresDelayedDisconnect  = true;

                if ( !szVersionIdString.empty() )
                    szVersionInfo = L"ip1500; " + szVersionIdString.ToWChar();
            }
            else if ( szProductIdString.find("hinet ta 1100") != SString::npos )
            {
                STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint is a HiNet TA 1100.");
                // limit supported codec to G.711a/u OR G.723 (if configured via registry)
                iSupportedCodecs =   H323RemoteEndpoint::CodecG711a
                                   | H323RemoteEndpoint::CodecG711u;
#ifdef IPPBXSERVER
                // check if hidden G.723 registry setting is active.
                SString szCodecInfo;
                SMediaControl::GetCompressionCodec(szCodecInfo);
                if (   szCodecInfo.ToLower().find("comp")==0
                    || szCodecInfo.ToLower().find("lb")==0)
                {
                    //  we wan't to support G.723 too!
                    iSupportedCodecs |= H323RemoteEndpoint::CodecG723;
                }
#endif // IPPBXSERVER

                bIsTA1100 = true;
            }
        }
    }

    if (bIsTA1100 && m_rEndpoint.UseTA1100RegWorkaround() == true)
    {
        // Get endpointVendor - product and version id out of RRQ
        // and compare them with Siemens HiNet TA 1100 versions
        // which have a Time To Live (TTL) value that is too high (8888 hours).
        // The TA 1100 sends RRQ (keep alive flag set to false!) 1 to 2 minutes to late.
        // Therefore we send a RCF with a shorter one: Using TTL of configured
        // seconds + 120 seconds internally, but sending configured seconds in RCF!
        if (rrq.m_endpointVendor.HasOptionalField(H225_VendorIdentifier::e_versionId))
        {
            SAsnOctetString VersionId1;
            SAsnOctetString VersionId2;
            VersionId1.SetValue((BYTE *)"1.5.3.24", 8);
            VersionId2.SetValue((BYTE *)"1.7.0.21", 8);

            if (rrq.m_endpointVendor.m_versionId == VersionId1)
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint is a HiNet TA 1100 version 1.5.3.24 -> use shorter TTL value ("
                    << m_rEndpoint.GetTA1100RegisterTTLSeconds() << " seconds).");
                bIsBadTA1100 = true;
            }
            else if (rrq.m_endpointVendor.m_versionId == VersionId2)
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint is a HiNet TA 1100 version 1.7.0.21 -> use shorter TTL value ("
                    << m_rEndpoint.GetTA1100RegisterTTLSeconds() << " seconds).");
                bIsBadTA1100 = true;
            }
        }
    }

    // extract information from rrq
    if (rrq.HasOptionalField(H225_RegistrationRequest::e_terminalAlias))
    {
        H323GetAliasAddressStringVector(rrq.m_terminalAlias,
                                        H225_AliasAddress::e_dialedDigits,
                                        E164Aliases);
        H323GetAliasAddressStringVector(rrq.m_terminalAlias,
                                        H225_AliasAddress::e_h323_ID,
                                        H323Aliases);
    }

    if (rrq.HasOptionalField(H225_RegistrationRequest::e_timeToLive) )
    {
        if (bIsBadTA1100 == true)
        {
            // The Time To Live (TTL) value of Siemens TA 1100 is too high.
            // But TA 1100 sends RRQ (keep alive flag set to false!) 1 to 2 minutes to late.
            // => Using TTL of configured seconds + 120 seconds internally, but sending
            //    configured seconds in RCF!
            TimeToLive = STimeInterval(0, m_rEndpoint.GetTA1100RegisterTTLSeconds() + 120);
        }
        else
        {
            TimeToLive = STimeInterval(0, rrq.m_timeToLive);
        }
    }
    else
    {
        TimeToLive = STimeInterval(0, 120); // default TTL= 120 seconds
    }

    // work on NonStandardData present in RRQ to get SWYX specific information
    if ( rrq.HasOptionalField( H225_RegistrationRequest::e_nonStandardData ) )
    {
        SPerStream ExtStrm;
        // get NonStandardData out of RRJ
        GetRRQNonStandardData(rrq, ExtStrm);
        // Handle NonStandardData
        nsData.HandleSwH225NonStandardData(ExtStrm);
    }

    // check if it is a lightweight registration or a re-registration
    // (so we will find this endpoint in our list)

    // 1. Search for endpointIdentifier if present but only for keep-alive RRQ.
    //    Otherwise we may find a wrong device (e.g. Welltech sends endpointIdentifier
    //    although it is not registered).
    if (   rrq.HasOptionalField(H225_RegistrationRequest::e_endpointIdentifier)
        && rrq.m_keepAlive == true)
    {
        Tools::SString endpointId = rrq.m_endpointIdentifier;
        if (!endpointId.empty())
        {
            pEndpoint = m_RegisteredEPList.FindEndpointByEndpointId(endpointId);
            if (pEndpoint != 0)
                goto EndpointFound;
        }
    }

    // 2. Search for terminalAlias if present

    // search for plain H.323 alias
    if (!H323Aliases.empty())
    {
        for(size_t ii = 0; ii < H323Aliases.size(); ii++)
        {
            // in case if SwyxGate we have to search by alias and line id:
            if ( nsData.m_NSDataChoice == SwyxH225NonStandardData::CallInfo )
                pEndpoint = m_RegisteredEPList.FindEndpointByAliasAndLineId( H323Aliases[ii], nsData.m_uLineId, callSignalAddress, true);
            else
                pEndpoint = m_RegisteredEPList.FindEndpointByAlias( H323Aliases[ii], callSignalAddress, true, true);
            if (pEndpoint != 0)
                goto EndpointFound;
        }
    }
    // search for plain E.164 alias
    if (!E164Aliases.empty())
    {
        pEndpoint = m_RegisteredEPList.FindEndpointByAliases(E164Aliases, callSignalAddress, true, true);
        if (pEndpoint != 0)
            goto EndpointFound;

        // search for our 'mixed mode' alias 
        for (Iter =  H323Aliases.begin();
            Iter != H323Aliases.end();
            ++Iter)
        {
            for (Iter2 =  E164Aliases.begin();
                Iter2 != E164Aliases.end();
                ++Iter2)
            {
                pEndpoint = m_RegisteredEPList.FindEndpointByAlias((*Iter) + (*Iter2), callSignalAddress, true, true);
                if (pEndpoint != 0)
                    goto EndpointFound;
            }
        }
    }

    // no endpoint found? -> create a new one...
    if (pEndpoint == 0)
    {
        // create a new endpoint, because it is a new registration
        IsNewEndpoint = true;
        bool bIsSwyxGate = false;

        // Check endpoint vendor (is it SwyxGate or not)
        if (TerminalType == H323EndPoint::e_GatewayOnly)
        {
            // Get our own endpoints H.221 NonStandardIdentifier
            H225_H221NonStandard h221Endpoint;
            m_rEndpoint.SetH221NonStandardInfo(h221Endpoint);

            // Compare H.221 NonStandardIdentifier
            if (   (   rrq.m_endpointVendor.m_vendor.m_t35CountryCode.GetValue()
                    == h221Endpoint.m_t35CountryCode.GetValue() )
                && (   rrq.m_endpointVendor.m_vendor.m_t35Extension.GetValue()
                    == h221Endpoint.m_t35Extension.GetValue() )
                && (   rrq.m_endpointVendor.m_vendor.m_manufacturerCode.GetValue()
                    == h221Endpoint.m_manufacturerCode.GetValue() ) )
            {
                bIsSwyxGate = true;
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint is a IpPbxGate.");
                // add Speex codec for IpPbxGate!
                iSupportedCodecs |= H323RemoteEndpoint::CodecSpeex;
            }
        }
        else if (TerminalType == H323EndPoint::e_GatekeeperOnly)
        {
            // Get our own endpoints H.221 NonStandardIdentifier
            H225_H221NonStandard h221Endpoint;
            m_rEndpoint.SetH221NonStandardInfo(h221Endpoint);

            // Compare H.221 NonStandardIdentifier
            if (   (   rrq.m_endpointVendor.m_vendor.m_t35CountryCode.GetValue()
                    == h221Endpoint.m_t35CountryCode.GetValue() )
                && (   rrq.m_endpointVendor.m_vendor.m_t35Extension.GetValue()
                    == h221Endpoint.m_t35Extension.GetValue() )
                && (   rrq.m_endpointVendor.m_vendor.m_manufacturerCode.GetValue()
                    == h221Endpoint.m_manufacturerCode.GetValue() ) )
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint is a LinkManager.");
                // add Speex codec for Links!
                iSupportedCodecs |= H323RemoteEndpoint::CodecSpeex;
            }
            else
            {
                STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                    "() Endpoint of type gatekeeper is NOT a IpPbxServer -> reject registration!");
                SendConfirm = false;
                goto SendAnswer;
            }
        }

        pEndpoint = new H323RemoteEndpoint(
            *this,
            pRasPDU->m_ReceiveIpAddress,
            RecvCallSignalAddress,
            RecvRasAddress,
            E164Aliases,
            H323Aliases,
            TimeToLive,
            TerminalType,
            szVersionInfo,
            iSupportedCodecs,
            bRequiresDelayedDisconnect,
            nsData.m_uLineId,
            nsData.m_uFaxChannels );
    }

EndpointFound:

    if (SBAD_READPTR(pEndpoint))
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() No endpoint found/created!");
        return false;
    }

    // try to insert endpoint in the registration and implicitly
    // register the endpoint at the pbx, but only if the gatekeeper
    // is NOT shutting down!

    if (!m_ShutdownGatekeeper)
    {
        if (IsNewEndpoint)
        {
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Trying to insert new endpoint " << callSignalAddress
                << " in registration table...");

            err = m_RegisteredEPList.Insert(callSignalAddress, pEndpoint);
            if (err == H323GkSrv::Success)
                SendConfirm = true;
            else
                SendConfirm = false;

            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() Insertion of new endpoint " << callSignalAddress
                << (SendConfirm ? " was successful." : " failed.") );
        }
        else if (    rrq.m_keepAlive
                  && rrq.HasOptionalField(H225_RegistrationRequest::e_timeToLive) )
        {
            // it is a lightweight (re-)registration, so update the timeToLive value
            SendConfirm = m_RegisteredEPList.UpdateEPRegTimer( pEndpoint, TimeToLive );

            if ( pEndpoint->m_iLineId > 0 )
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() It is a lightweight RRQ of endpoint " << pRasPDU->m_ReceiveIpAddress
                    << " with alias '" << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "' and line Id " << pEndpoint->m_iLineId);
            }
            else
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() It is a lightweight RRQ of endpoint " << pRasPDU->m_ReceiveIpAddress
                    << " with alias '" << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "'");
            }
        }
        else if ( !rrq.m_keepAlive && callSignalAddress == pEndpoint->m_IpAddress ) // full RRQ from already known CallSigAddress ?
        {
            // We already have an endpoint registered with alias and same IP -> accept full RRQ as 'keep alive'

            // update call signalling and...
            pEndpoint->m_CallSignalAddress = RecvCallSignalAddress;
            // RAS address data and...
            pEndpoint->m_RasAddress        = RecvRasAddress;
            // update the timeToLive value
            if ( m_RegisteredEPList.UpdateEPRegTimer( pEndpoint, TimeToLive ) )
            {
                SendConfirm = true;

                if ( !pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.empty() )
                {
                    SStringVector::const_iterator Iter;
                    SStringStream Aliases;
                    for ( Iter = pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.begin();
                          Iter != pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.end();
                          ++Iter )
                    {
                        Aliases << "'" << (*Iter).c_str() << "', ";
                    }
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "() Follow up RRQ from endpoint with mixed alias(es) "
                        << Aliases << "' and IP "
                        << pEndpoint->m_IpAddress.m_Address );
                }
                else
                {
                    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "() Follow up RRQ from endpoint with alias '"
                        << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "' and IP "
                        << pEndpoint->m_IpAddress.m_Address );
                }
            }
            else // the endpoint is NOT in timer list -> insert it now!
            {
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Trying to re-insert endpoint " << callSignalAddress
                    << " in registration table...");

                err = m_RegisteredEPList.Insert(callSignalAddress, pEndpoint);
                if (err == H323GkSrv::Success)
                    SendConfirm = true;
                else
                    SendConfirm = false;

                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() Re-insertion of endpoint with alias '"
                    << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "' and IP "
                    << pEndpoint->m_IpAddress.m_Address
                    << (SendConfirm ? " was successful." : " failed.") );
            }
        }
        else
        {
            // We already have an endpoint registered with the alias -> reject RRQ!
            bAliasAlreadyRegistered = true;
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0, "() Endpoint with alias '"
                << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias << "' is already registered -> reject attempt from "
                << pRasPDU->m_ReceiveIpAddress );
        }
    }

SendAnswer:

    if (SendConfirm && !m_ShutdownGatekeeper)
    {
        // ok, so confirm registration to endpoint
        H225_RegistrationConfirm & rcf =
            Response.BuildRegistrationConfirm(rrq.m_requestSeqNum);

        // fill call signal address
        const H323ListenerList & rListeners = m_rEndpoint.GetListeners();
        H225_TransportAddress LocalAddress;
        if (SBAD_READPTR(m_pTransport))
        {
            STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                "() Invalid pointer to transport object!");
            // Until now only log message, no fix
        }
        m_pTransport->SetUpTransportPDU(LocalAddress, true);
        for (H323ListenerList::size_type i = 0; i < rListeners.size(); i++)
        {
            if (SBAD_READPTR(rListeners[i]))
            {
                STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
                    "() Invalid pointer to listener object " << i << "!"); 
                // Until now only log message, no fix
            }
            rListeners[i]->SetUpTransportPDU(rcf.m_callSignalAddress,
                                             LocalAddress);
        }

        // attach NonStandardData in case of SwyxGate/SwyxFax >= 2.0
        // and for e.g. LinkMgr to transmit SwyxWare version

        // create Swyx extension structure...
        SwyxH225NonStandardData nsData(g_pH323Module->m_GkTraceId);
        if ( pEndpoint->m_iLineId > 0 )
        {
            nsData.m_NSDataChoice = SwyxH225NonStandardData::CallInfo;
            nsData.m_uFaxType     = SwyxH225NonStandardData::UnknownFT;
            nsData.m_uLineId      = pEndpoint->m_iLineId;
        }
        else if ( pEndpoint->m_iNumOfChannels > 0 )
        {
            nsData.m_NSDataChoice = SwyxH225NonStandardData::FaxChannels;
            nsData.m_uFaxChannels = pEndpoint->m_iNumOfChannels;
        }
        else
        {
            // Use a default for e.g. LinkMgr (necessary for ASN choice!)
            nsData.m_NSDataChoice = SwyxH225NonStandardData::CallInfo;
            nsData.m_uFaxType     = SwyxH225NonStandardData::UnknownFT;
        }

        // Set endpoint vendor identifier including product and version ID
        m_rEndpoint.SetVendorIdentifierInfo(nsData.m_H225VendorIdentifier);

        // build and attach NonStandardData to RCF signal pdu
        SPerStream ExtStrm;
        nsData.BuildSwH225NonStandardData(ExtStrm);
        AttachRCFNonStandardData(ExtStrm, rcf);

        if (!rrq.m_keepAlive)
        {
            // fill the terminalAlias
            rcf.IncludeOptionalField(H225_RegistrationConfirm::e_terminalAlias);
            if ( pEndpoint->m_iNumOfChannels == 0 )
            {
                rcf.m_terminalAlias.SetSize(1 // m_RegisteredH323Alias
                                            + pEndpoint->m_H323GkAliases.m_RegisteredE164Aliases.size());
            }
            else
            {
                // in case RRQ was submitted by SwyxFax >= 2.0
                // we will NOT encode all registered extension because this might lead into troube if array count >= 128
                // refer to CQ#8587!
                rcf.m_terminalAlias.SetSize(1);  // m_RegisteredH323Alias _only_!!!
            }
            int Pos = 0;
            rcf.m_terminalAlias[Pos].SetTag(H225_AliasAddress::e_h323_ID);
            // in case of our special 'mixed mode' registration, we use the first H.323 alias from RRQ
            if ( !pEndpoint->m_H323GkAliases.m_RegisteredMixedAliases.empty() )
                (SAsnBmpString &)rcf.m_terminalAlias[Pos] = pEndpoint->m_H323GkAliases.m_H323Aliases[0];
            else 
                (SAsnBmpString &)rcf.m_terminalAlias[Pos] = pEndpoint->m_H323GkAliases.m_RegisteredH323Alias;
            Pos++;
            // in case RRQ was submitted by SwyxFax >= 2.0
            // we will NOT encode all registered extension because this might lead into troube if array count >= 128
            // refer to CQ#8587!
            if ( pEndpoint->m_iNumOfChannels == 0 )
            {
                for (Tools::SStringVector::size_type i = 0; i < pEndpoint->m_H323GkAliases.m_RegisteredE164Aliases.size(); i++)
                {
                    rcf.m_terminalAlias[Pos].SetTag(H225_AliasAddress::e_dialedDigits);
                    SAsnIA5String & rIA5 = (SAsnIA5String &)rcf.m_terminalAlias[Pos];
                    rIA5 = pEndpoint->m_H323GkAliases.m_RegisteredE164Aliases[i];
                    Pos++;
                }
            }
        }

        // fill gatekeeper identifier if present
        if (!m_GatekeeperIdentifier.empty())
        {
            rcf.IncludeOptionalField(
                H225_RegistrationConfirm::e_gatekeeperIdentifier);
            rcf.m_gatekeeperIdentifier = m_GatekeeperIdentifier;
        }
        
        // fill endpoint identifier
        rcf.m_endpointIdentifier = pEndpoint->m_EndpointIdentifier;

        // time to live
        if (pEndpoint->m_TimeToLive != 0)
        {
            rcf.IncludeOptionalField(H225_RegistrationConfirm::e_timeToLive);
            rcf.m_timeToLive = (unsigned int)pEndpoint->m_uTimeToLiveInSeconds;
        }

        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS RCF to " << RecvRasIP << ":" << RecvRasPort); 
    }
    else
    {
        // failure, so reject endpoint

        // set reason for reject
        unsigned uRRJreason = H225_RegistrationRejectReason::e_undefinedReason;
        if (err == H323GkSrv::LicenseLimitReached)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Rejecting registration with reason 'resourceUnavailable' because license limit was reached!");
            uRRJreason = H225_RegistrationRejectReason::e_resourceUnavailable;
        }
        else if (m_ShutdownGatekeeper)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Rejecting registration with reason 'fullRegistrationRequired' because GK shutdown in progress!");
            uRRJreason = H225_RegistrationRejectReason::e_fullRegistrationRequired;
        }
        else if (bAliasAlreadyRegistered || err == H323GkSrv::UserAlreadyRegistered)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Rejecting registration with reason 'e_duplicateAlias' because device(s) with alias already registered!");
            uRRJreason = H225_RegistrationRejectReason::e_duplicateAlias;
        }
        else if ( IsNewEndpoint )
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Rejecting registration with reason 'invalidAlias' because alias not found!");
            uRRJreason = H225_RegistrationRejectReason::e_invalidAlias;
        }
        else
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Rejecting registration with reason 'undefinedReason'!");
        }
        H225_RegistrationReject & rrj = Response.BuildRegistrationReject(rrq.m_requestSeqNum, uRRJreason);

        // set gatekeeper identifier if present
        if (!m_GatekeeperIdentifier.empty())
        {
            rrj.IncludeOptionalField(
                H225_RegistrationReject::e_gatekeeperIdentifier);
            rrj.m_gatekeeperIdentifier = m_GatekeeperIdentifier;
        }
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS RRJ to " << RecvRasIP << ":" << RecvRasPort); 

        if (IsNewEndpoint)
            delete pEndpoint;
    }

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "() " << m_RegisteredEPList);

    return SendRasPdu(RecvRasAddress, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveUnregistrationRequest(RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::OnReceiveURQ");

    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_UnregistrationRequest & urq = (const H225_UnregistrationRequest &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = 0;
    SStringVector E164Aliases;
    SStringVector H323Aliases;

    // Try to find endpoint
    // 1. Search for endpoint using endpointIdentifier if present
    if (urq.HasOptionalField(H225_UnregistrationRequest::e_endpointIdentifier))
    {
        Tools::SString endpointId = urq.m_endpointIdentifier;
        if (!endpointId.empty())
        {
            pEndpoint = m_RegisteredEPList.FindEndpointByEndpointId(urq.m_endpointIdentifier);
            if (pEndpoint != 0)
                goto EndpointFound;
        }
    }

    // 2. Search for endpoint using endpointAlias if present
    if (urq.HasOptionalField(H225_UnregistrationRequest::e_endpointAlias))
    {
        H323GetAliasAddressStringVector(urq.m_endpointAlias,
                                        H225_AliasAddress::e_dialedDigits,
                                        E164Aliases);
        H323GetAliasAddressStringVector(urq.m_endpointAlias,
                                        H225_AliasAddress::e_h323_ID,
                                        H323Aliases);
        if (!H323Aliases.empty())
        {
            pEndpoint = m_RegisteredEPList.FindEndpointByAliases(H323Aliases, pRasPDU->m_ReceiveIpAddress, false);
            if (pEndpoint)
                goto EndpointFound;
        }
        if (!E164Aliases.empty())
        {
            pEndpoint = m_RegisteredEPList.FindEndpointByAliases(E164Aliases, pRasPDU->m_ReceiveIpAddress, false);
            if (pEndpoint)
                goto EndpointFound;
        }
    }

    // 3. Search for endpoint using callSignalAddress
    // Until now not implemented!
    // Compare urq.m_callSignalAddress and find endpoint by m_CallSignalAddress

    // 4. Search for endpoint using IP address and port (TSAP)
    pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);

EndpointFound:

    H323TransportAddress Addr(pEndpoint ? 
        pEndpoint->m_RasAddress : 
        pRasPDU->m_ReceiveTransportAddress);

    // if search for endpoint was successful, deregister it, otherwise
    // reject unregistration
    H323RasPDU Response;
    if (pEndpoint != 0 && m_RegisteredEPList.Erase(pEndpoint->m_EndpointIdentifier))
    {
        // endpoint may be delete right now!
        pEndpoint = 0;
        // deregister endpoint
        H225_UnregistrationConfirm & ucf =
            Response.BuildUnregistrationConfirm(urq.m_requestSeqNum);
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS UCF to " << pRasPDU->m_ReceiveIpAddress); 
    }
    else
    {
        // reject unregistration request
        H225_UnregistrationReject & urj =
            Response.BuildUnregistrationReject(
                urq.m_requestSeqNum,
                H225_UnregRejectReason::e_notCurrentlyRegistered);
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS URJ to " << pRasPDU->m_ReceiveIpAddress); 
    }

    return SendRasPdu(Addr, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveUnregistrationConfirm(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_UnregistrationConfirm & ucf = (const H225_UnregistrationConfirm &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_unregistrationRequest,
                             ucf.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            // delete this endpoint
            m_RegisteredEPList.Erase(pEndpoint->m_EndpointIdentifier);
        }
    }

    return true;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveUnregistrationReject(RasPDU *pRasPDU)
{
    //! not implemented
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveAdmissionRequest(RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::OnReceiveARQ");

    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_AdmissionRequest & arq = (const H225_AdmissionRequest &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = 0;

    // Try to find endpoint
    // 1. Search for endpoint using endpointIdentifier
    Tools::SString endpointId = arq.m_endpointIdentifier;
    if (!endpointId.empty())
    {
        pEndpoint = m_RegisteredEPList.FindEndpointByEndpointId(endpointId);
        if (pEndpoint != 0)
            goto EndpointFound;
    }

    // 2. Search for endpoint using IP address and port (TSAP)
    pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);

EndpointFound:

    H323TransportAddress Addr(pEndpoint ? 
        pEndpoint->m_RasAddress :
        pRasPDU->m_ReceiveTransportAddress);

    // If we found the endpoint, confirm the request, otherwise
    // reject it. Unregistered endpoints are not allowed to
    // send an admission request to this gatekeeper!
    H323RasPDU Response;
    if (pEndpoint != 0)
    {
        if (OnAdmissionRequest(pEndpoint->m_H323GkAliases,
                               pEndpoint->m_pUserData,
                               pEndpoint->m_TerminalType) == false)
        {
            // Reject admission e.g. not further call allowed
            H225_AdmissionReject & arj = Response.BuildAdmissionReject(arq.m_requestSeqNum,
                                                                       H225_AdmissionRejectReason::e_requestDenied);
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                "() Caller " << pRasPDU->m_ReceiveIpAddress << " with alias '"
                << pEndpoint->m_H323GkAliases.m_RegisteredH323Alias
                << "' is not allowed to make a call (e.g. no further caller licence)!");
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() SEND RAS ARJ to " << pRasPDU->m_ReceiveIpAddress); 

            return SendRasPdu(Addr, Response);
        }

        // If we use callIdentifier to resolve H.323 client
        // check for call identifier to save in endpoint object
        if (   m_rEndpoint.UseCallIdentifier()
            && arq.HasOptionalField(H225_AdmissionRequest::e_callIdentifier) )
        {
            pEndpoint->m_H323CallIdentifier.push_back(arq.m_callIdentifier.m_guid);
        }
        else // call identifier is only available on H.323v2 and later -> fall back to conference ID for H.323v1
        {
            pEndpoint->m_H323CallIdentifier.push_back(arq.m_conferenceID);
        }

        if (IsGkRoutedCallModel())
        {
            H225_AdmissionConfirm & acf =
                Response.BuildAdmissionConfirm(
                    arq.m_requestSeqNum,
                    arq.m_bandWidth < DefaultBandwidth ?
                    DefaultBandwidth : (unsigned)arq.m_bandWidth);

            // Gatekeeper routed call model
            acf.m_callModel.SetTag(H225_CallModel::e_gatekeeperRouted);
            SetUpOwnCallSignalTransportPDU(acf.m_destCallSignalAddress);
            STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                "() SEND RAS ACF to " << pRasPDU->m_ReceiveIpAddress); 
        }
        else
        {
            // direct signalling
            bool Resolved = false;
            SIpAddress IpAddress;

            // get e164 alias of destination
            SStringVector E164Aliases;
            H323GetAliasAddressStringVector(arq.m_destinationInfo,
                                        H225_AliasAddress::e_dialedDigits,
                                        E164Aliases);
            if ((!E164Aliases.empty() && 
                ((Resolved = OnResolveAddress(E164Aliases[0], IpAddress)) == true) ) ||
                (arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress)))
            {
                // set up destination address
                H225_AdmissionConfirm & acf =
                    Response.BuildAdmissionConfirm(
                        arq.m_requestSeqNum,
                        arq.m_bandWidth < DefaultBandwidth ?
                        DefaultBandwidth : (unsigned)arq.m_bandWidth);
                
                acf.m_callModel.SetTag(H225_CallModel::e_direct);
                acf.m_destCallSignalAddress.SetTag(
                    H225_TransportAddress::e_ipAddress);
                H225_TransportAddress_ipAddress & rIp =
                    acf.m_destCallSignalAddress;
                if (Resolved)
                {
                    rIp.m_ip.SetValue((DWORD)IpAddress.m_Address);
                    rIp.m_port = 1720;
                }
                else
                {
                    rIp = arq.m_destCallSignalAddress;
                }
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() SEND RAS ACF to " << pRasPDU->m_ReceiveIpAddress); 
            }
            else
            {
                // reject admission
                H225_AdmissionReject & arj = Response.BuildAdmissionReject(
                    arq.m_requestSeqNum,
                    H225_AdmissionRejectReason::e_calledPartyNotRegistered);
                STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
                    "() SEND RAS ARJ to " << pRasPDU->m_ReceiveIpAddress); 
            }
        }
    }
    else
    {
        // reject admission
        H225_AdmissionReject & arj = Response.BuildAdmissionReject(
            arq.m_requestSeqNum,
            H225_AdmissionRejectReason::e_callerNotRegistered);
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Caller " << pRasPDU->m_ReceiveIpAddress << " not registered, sending ARJ!"); 
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS ARJ to " << pRasPDU->m_ReceiveIpAddress); 
    }

    return SendRasPdu(Addr, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveBandwidthRequest(RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::OnReceiveBRQ");

    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_BandwidthRequest & brq = (const H225_BandwidthRequest &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = 0;

    // Try to find endpoint
    // 1. Search for endpoint using endpointIdentifier
    Tools::SString endpointId = brq.m_endpointIdentifier;
    if (!endpointId.empty())
    {
        pEndpoint = m_RegisteredEPList.FindEndpointByEndpointId(endpointId);
        if (pEndpoint != 0)
            goto EndpointFound;
    }

    // 2. Search for endpoint using IP address and port (TSAP)
    pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);

EndpointFound:

    H323TransportAddress Addr(pEndpoint ? 
        pEndpoint->m_RasAddress :
        pRasPDU->m_ReceiveTransportAddress);

    H323RasPDU Response;
    if (pEndpoint != 0)
    {
        Response.BuildBandwidthConfirm(brq.m_requestSeqNum,
                                       brq.m_bandWidth);
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS BCF to " << pRasPDU->m_ReceiveIpAddress);
    }
    else
    {
        Response.BuildBandwidthReject(brq.m_requestSeqNum,
                                      H225_BandRejectReason::e_invalidPermission);
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS BRJ to " << Addr);
    }
    return SendRasPdu(Addr, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveBandwidthConfirm(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_BandwidthConfirm & bcf = (const H225_BandwidthConfirm &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_bandwidthRequest,
                             bcf.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            return true;
        }
    }
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveBandwidthReject(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_BandwidthReject & brj = (const H225_BandwidthReject &)(pRasPDU->m_H323RasPDU);
    
    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_bandwidthRequest,
                             brj.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            return true;
        }
    }
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveInfoRequestResponse(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_InfoRequestResponse & irr = (const H225_InfoRequestResponse &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_infoRequest,
                             irr.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            return true;
        }
    }
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveDisengageRequest(RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::OnReceiveDRQ");

    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_DisengageRequest & drq = (const H225_DisengageRequest &)(pRasPDU->m_H323RasPDU);

    H323GloballyUniqueID id;
    H323RemoteEndpoint * pEndpoint = 0;

    if (   m_rEndpoint.UseCallIdentifier()
        && drq.HasOptionalField(H225_DisengageRequest::e_callIdentifier) )
    {
        id = drq.m_callIdentifier.m_guid;
    }
    else // call identifier is only available on H.323v2 and later -> fall back to conference ID for H.323v1
    {
        id = drq.m_conferenceID;
    }

    // Try to find endpoint
    // 1. Search for endpoint using endpointIdentifier
    Tools::SString endpointId = drq.m_endpointIdentifier;
    if (!endpointId.empty())
    {
        pEndpoint = m_RegisteredEPList.FindEndpointByEndpointId(endpointId);
        if (pEndpoint != 0)
            goto EndpointFound;
    }

    // 2. Search for endpoint using IP address and port (TSAP)
    pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);

EndpointFound:

    H323TransportAddress Addr(pEndpoint ? 
        pEndpoint->m_RasAddress :
        pRasPDU->m_ReceiveTransportAddress);

    // If we found the endpoint, confirm the request, otherwise
    // reject it.
    H323RasPDU Response;
    if (pEndpoint != 0)
    {
        // Remove call identifier from list
        pEndpoint->m_H323CallIdentifier.remove(id);
        OnDisengageRequest(pEndpoint->m_H323GkAliases,
                           pEndpoint->m_pUserData,
                           pEndpoint->m_TerminalType);
        Response.BuildDisengageConfirm(drq.m_requestSeqNum);
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS DCF to " << Addr);
    }
    else
    {
        // Reject disengage request
        Response.BuildDisengageReject(drq.m_requestSeqNum,
                                      H225_DisengageRejectReason::e_notRegistered);
        STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
            "() SEND RAS DRJ to " << Addr);
    }
    return SendRasPdu(Addr, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveDisengageConfirm(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_DisengageConfirm & dcf = (const H225_DisengageConfirm &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_disengageRequest,
                             dcf.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            return true;
        }
    }
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveDisengageReject(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_DisengageReject & drj = (const H225_DisengageReject &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_disengageRequest,
                             drj.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            return true;
        }
    }    
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveLocationRequest(RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::OnReceiveLRQ");

    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_LocationRequest & lrq = (const H225_LocationRequest &)(pRasPDU->m_H323RasPDU);

    // not allowed till now !
    H323RasPDU Response;
    H225_LocationReject & lrj =
        Response.BuildLocationReject(
            lrq.m_requestSeqNum,
            H225_LocationRejectReason::e_invalidPermission);
    
    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);

    H323TransportAddress Addr(pEndpoint ? 
        pEndpoint->m_RasAddress :
        pRasPDU->m_ReceiveTransportAddress);

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS LRJ to " << pRasPDU->m_ReceiveIpAddress);

    return SendRasPdu(Addr, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveLocationConfirm(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_LocationConfirm & lcf = (const H225_LocationConfirm &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_locationRequest,
                             lcf.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            return true;
        }
    }
    return false;

    // get call signal and ras address...
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveLocationReject(RasPDU *pRasPDU)
{
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_LocationReject & lrj = (const H225_LocationReject &)(pRasPDU->m_H323RasPDU);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);
    if (pEndpoint != 0)
    {
        if (CheckForResponse(*pEndpoint,
                             H225_RasMessage::e_locationRequest,
                             lrj.m_requestSeqNum))
        {
            pEndpoint->m_uRequestRetryNumber = 0;
            return true;
        }
    }
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveNonStandardMessage(RasPDU *pRasPDU)
{
/*
    if (SBAD_READPTR(pRasPDU))
        return false;

    const H225_NonStandardMessage & nsm = (const H225_NonStandardMessage &)(pRasPDU->m_H323RasPDU);

    // this message contains our gateway proprietary stuff!
    SPerStream Stream((const BYTE *)nsm.m_nonStandardData.m_data,
                      nsm.m_nonStandardData.m_data.GetSize());
    
    SwyxExt_Message SwyxMsg;
    if (SwyxMsg.Decode(Stream))
    {
        // analyse the message
        // todo
    }
*/
    return false;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveUnknown(RasPDU *pRasPDU)
{
    SDEFFUNC("H323GkServer::OnReceiveUnknown");

    if (SBAD_READPTR(pRasPDU))
        return false;

    H323RasPDU Response;
    Response.BuildUnknownMessageResponse(0);

    H323RemoteEndpoint * pEndpoint = m_RegisteredEPList.FindEndpointByTsap(pRasPDU->m_ReceiveIpAddress);

    H323TransportAddress Addr(pEndpoint ? 
        pEndpoint->m_RasAddress :
        pRasPDU->m_ReceiveTransportAddress);

    STRACE2_INFO(g_pH323Module->m_GkTraceId, this, 0,
        "() SEND RAS UnknownMessageResponse to " << pRasPDU->m_ReceiveIpAddress);

    return SendRasPdu(Addr, Response);
}

//-------------------------------------------------------------------

bool H323GkServer::OnReceiveRequestInProgress(RasPDU *pRasPDU)
{
    //if (SBAD_READPTR(pRasPDU))
    //    return false;
    //const H225_RequestInProgress & = (const H225_RequestInProgress &)(pRasPDU->m_H323RasPDU);
    return false;
}
//-------------------------------------------------------------------

bool H323GkServer::OnReceiveResourcesAvailableIndicate(RasPDU *pRasPDU)
{
    //if (SBAD_READPTR(pRasPDU))
    //    return false;
    //const H225_ResourcesAvailableIndicate = (const H225_ResourcesAvailableIndicate &)(pRasPDU->m_H323RasPDU);
    return false;
}

//-------------------------------------------------------------------

void H323GkServer::OnSendGatekeeperConfirm(H225_GatekeeperConfirm &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendGatekeeperReject(H225_GatekeeperReject &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendRegistrationConfirm(H225_RegistrationConfirm &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendRegistrationReject(H225_RegistrationReject &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendUnregistrationRequest(H225_UnregistrationRequest &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendUnregistrationConfirm(H225_UnregistrationConfirm &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendUnregistrationReject(H225_UnregistrationReject &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendAdmissionConfirm(H225_AdmissionConfirm &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendAdmissionReject(H225_AdmissionReject &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendBandwidthRequest(H225_BandwidthRequest &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendBandwidthConfirm(H225_BandwidthConfirm &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendBandwidthReject(H225_BandwidthReject &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendInfoRequest(H225_InfoRequest &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendInfoRequestAck(H225_InfoRequestAck &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendInfoRequestNak(H225_InfoRequestNak &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendDisengageRequest(H225_DisengageRequest &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendDisengageConfirm(H225_DisengageConfirm &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendDisengageReject(H225_DisengageReject &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendLocationRequest(H225_LocationRequest &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendLocationConfirm(H225_LocationConfirm &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendLocationReject(H225_LocationReject &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendNonStandardMessage(H225_NonStandardMessage &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendUnknownMessageResponse(H225_UnknownMessageResponse &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendRequestInProgress(H225_RequestInProgress &)
{
}

//-------------------------------------------------------------------

void H323GkServer::OnSendResourcesAvailableConfirm(H225_ResourcesAvailableConfirm &)
{
}

//-------------------------------------------------------------------

H323GkSrv::Err H323GkServer::OnRegisterRemoteEndpoint(
        const Tools::SIpAddress           &  rAddress,
        H323GkAliases                     &  rAliases,
        Tools::SObject                    ** ppUserData,
        const H323EndPoint::TerminalTypes &  rTerminalType,
        int                                  iSupportedCodecs,
        Tools::SWString                   &  rszVersion,
        int                                  iLineId,
        int                                  iNumOfChannels )
{
    SDEFFUNC("H323GkServer::OnRegisterRemoteEp");

    STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
        "(" << rAddress << ",,, "
        << rTerminalType << ", "
        << iSupportedCodecs << ", "
        << rszVersion << ", "
        << iLineId << ", "
        << iNumOfChannels
        << ") Base class method called: OnRegisterRemoteEndpoint!");
    return H323GkSrv::Success;
}

//-------------------------------------------------------------------

bool H323GkServer::OnReleaseRemoteEndpoint(const Tools::SIpAddress           & rAddress,
                                           H323GkAliases                     & rAliases,
                                           Tools::SObject                  * & prUserData,
                                           const H323EndPoint::TerminalTypes & rTerminalType)
{
    SDEFFUNC("H323GkServer::OnReleaseRemoteEp");

    STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
        "(" << rAddress << ",,, "
        << rTerminalType
        << ") Base class method called: OnReleaseRemoteEndpoint!");
    return true;
}

//-------------------------------------------------------------------

bool H323GkServer::OnResolveAddress(
        const Tools::SIpAddress & rAddress,
        Tools::SObject ** ppUserData)
{
    SDEFFUNC("H323GkServer::OnResolveAddress");

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "(" << rAddress << ",)");

    return true;
}

//-------------------------------------------------------------------

bool H323GkServer::OnResolveAddress(
    const Tools::SString & rAlias,
    Tools::SIpAddress & rIpAddress)
{
    SDEFFUNC("H323GkServer::OnResolveAddress");

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "(" << rAlias << ", " << rIpAddress << ")");

    return true;
}

//-------------------------------------------------------------------

bool H323GkServer::OnAdmissionRequest(H323GkAliases                       & rAliases,
                                      Tools::SObject                    * & pUserData,
                                      const H323EndPoint::TerminalTypes   & rTerminalType)
{
    SDEFFUNC("H323GkServer::OnAdmissionRequest");

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "()");

    return true;
}

//-------------------------------------------------------------------

bool H323GkServer::OnDisengageRequest(H323GkAliases                       & rAliases,
                                      Tools::SObject                    * & pUserData,
                                      const H323EndPoint::TerminalTypes   & rTerminalType)
{
    SDEFFUNC("H323GkServer::OnDisengageRequest");

    STRACE2_INFO3(g_pH323Module->m_GkTraceId, this, 0,
        "()");

    return true;
}

//-------------------------------------------------------------------

SIPSocket::Address H323GkServer::GetLocalAddress() const
{
    H225_TransportAddress pdu;
    m_pTransport->SetUpTransportPDU(pdu, true);
    H323TransportAddress TAddr(pdu);
    SIPSocket::Address Addr;
    WORD Port;
    TAddr.GetIpAndPort(Addr, Port);
    return Addr;
}

//-------------------------------------------------------------------

bool H323GkServer::SetUpOwnCallSignalTransportPDU(
    H225_TransportAddress & pdu) const
{
    SDEFFUNC("H323GkSrv::SetUpOwnCallSigTrspPDU");

    if (SBAD_READPTR(m_pTransport))
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Invalid pointer to transport object!");
        // Until now only log message, no fix.
        // Access violation will be catched and cause restart
        // of thread and transport object.
    }
    m_pTransport->SetUpTransportPDU(pdu, true);
    if (pdu.GetTag() == H225_TransportAddress::e_ipAddress)
    {
        H225_TransportAddress_ipAddress & ip = pdu;
        const H323ListenerList & rListeners = m_rEndpoint.GetListeners();
        for (H323ListenerList::size_type i = 0; i < rListeners.size(); i++)
        {
            H323ListenerTCP * pListener =
                dynamic_cast<H323ListenerTCP *>(rListeners[i]);

            if (!SBAD_READPTR(pListener) && pListener->IsListeningOnAddress(GetLocalAddress()))
            {
                ip.m_port = pListener->GetListenerPort();
                return true;
            }
        }
        ip.m_port = H323ListenerTCP::DefaultSignalPort;
        return true;
    }

    return false;
}

//-----------------------------------------------------------------------------

bool H323GkServer::AttachRCFNonStandardData(const Tools::SByteVector & data,
                                            H225_RegistrationConfirm & rcf)
{
    SDEFFUNC("H323GkServer::AttachRCFNSData");

    rcf.IncludeOptionalField(H225_RegistrationConfirm::e_nonStandardData );

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        rcf.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_object);
        ((SAsnObjectId &)rcf.m_nonStandardData.m_nonStandardIdentifier) = m_szObjectIdentifier;
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        rcf.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_h221NonStandard);
        H225_H221NonStandard & h221 = rcf.m_nonStandardData.m_nonStandardIdentifier;
        m_rEndpoint.SetH221NonStandardInfo(h221);
    }

    // Adding nonStandard data
    rcf.m_nonStandardData.m_data.SetValue(data);

    return true;
}
//-----------------------------------------------------------------------------

bool H323GkServer::GetRRQNonStandardData(const H225_RegistrationRequest & rrq,
                                         Tools::SByteVector             & data)
{
    SDEFFUNC("H323GkServer::GetRRQNSData");

    // Get NonStandardData part of RCF
    if (!rrq.HasOptionalField(H225_RegistrationRequest::e_nonStandardData))
        return false;

    const H225_NonStandardIdentifier & nonStdId
        = rrq.m_nonStandardData.m_nonStandardIdentifier;

    if (!m_szObjectIdentifier.empty())
    {
        // Connection is using object identifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_object)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using object identifier "
                         << m_szObjectIdentifier);
            return false;
        }
        const SAsnObjectId & oid = nonStdId;
        if (oid != m_szObjectIdentifier.c_str())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received object identifier " << oid
                         << " != " << m_szObjectIdentifier);
            return false;
        }
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_h221NonStandard)
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because not using h221NonStandard identifier!");
            return false;
        }
        const H225_H221NonStandard & h221Msg = nonStdId;

        // Get endpoints H.221 NonStandardIdentifier
        H225_H221NonStandard h221Endpoint;
        m_rEndpoint.SetH221NonStandardInfo(h221Endpoint);

        // Compare H.221 NonStandardIdentifier
        if (h221Msg.m_t35CountryCode.GetValue() != h221Endpoint.m_t35CountryCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35CountryCode="
                         << h221Msg.m_t35CountryCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35CountryCode.GetValue());
            return false;
        }
        if (h221Msg.m_t35Extension.GetValue() != h221Endpoint.m_t35Extension.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received t35Extension="
                         << h221Msg.m_t35Extension.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35Extension.GetValue());
            return false;
        }
        if (h221Msg.m_manufacturerCode.GetValue() != h221Endpoint.m_manufacturerCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
                         "() Ignoring nonStandardData because received manufacturerCode="
                         << h221Msg.m_manufacturerCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_manufacturerCode.GetValue());
            return false;
        }
    }

    // Get nonStandard data
    data.clear();
    data = rrq.m_nonStandardData.m_data.GetValue();

    return true;
}

//-------------------------------------------------------------------

unsigned H323GkServer::GetNextSequenceNumber()
{
    SWaitAndSignal lock(m_LastSequenceNumberCritSec);

    m_LastSequenceNumber++;
    if (m_LastSequenceNumber >= 65536)
        m_LastSequenceNumber = 1;
    return m_LastSequenceNumber;
}

//-------------------------------------------------------------------

bool H323GkServer::MakeRequest(H323RemoteEndpoint & rEndpoint,
                               const H323RasPDU & rRequest)
{
//    rEndpoint.m_NextTimeout = STime() + m_RasTimeout;
    rEndpoint.m_RasPdu = rRequest;
    bool bReqResult = SendRasPdu(rEndpoint.m_RasAddress, rRequest);
    if (bReqResult)
        ++rEndpoint.m_uRequestRetryNumber;
    return bReqResult;
}

//-------------------------------------------------------------------

bool H323GkServer::CheckForResponse(H323RemoteEndpoint & rEndpoint,
                                    unsigned Tag,
                                    unsigned SequenceNumber)
{
    SDEFFUNC("H323GkServer::CheckForResponse");

    // check Tag
    if (rEndpoint.m_RasPdu.GetTag() != Tag)
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Wrong tag in received ras message: "
            << Tag << rEndpoint.m_RasPdu);
        
        return false;
    }
    
    // check sequence number
    if (rEndpoint.m_uSequenceNumber != SequenceNumber)
    {
        STRACE2_WARN(g_pH323Module->m_GkTraceId, this, 0,
            "() Wrong sequence number in received ras message: "
            << SequenceNumber << rEndpoint.m_RasPdu);
        
        return false;
    }
        
    return true;
}

//-------------------------------------------------------------------

bool H323GkServer::SendRasPdu(const H323TransportAddress & rDest,
                              const H323RasPDU & rPdu)
{
    SDEFFUNC("H323GkServer::SendRasPdu");
 
    H323RasPDU::H323RasPDUError Result;

    if (SBAD_READPTR(m_pTransport))
    {
        STRACE2_ERROR(g_pH323Module->m_GkTraceId, this, 0,
            "() Invalid pointer to transport object!");
        return false;
    }
    m_pTransport->SetRemoteAddress(rDest);
    m_pTransport->Connect(); // activates the new remote address
    
    STRACE2_INFO2(g_pH323Module->m_RasTraceId, this, 0,
        "() <-Sent RAS PDU to " << rDest);

    Result = rPdu.Write(*m_pTransport);
    if ( Result == H323RasPDU::TransportFailure && m_pTransport == 0 ) // m_pTransport may be deleted now!
    {
        // get last error from socket operation. Otherwise an encode failure happend
        m_LastTransportError = m_pTransport->GetErrorNumber();
    }
    return ( Result == H323RasPDU::Success );
}

//-------------------------------------------------------------------

Tools::SString H323GkServer::GetH323Alias(const H323GloballyUniqueID & rCallIdentifier,
                                          const Tools::SStringVector & rE164Aliases)
{
    return m_RegisteredEPList.FindH323AliasByCallIdentifier(rCallIdentifier, rE164Aliases);
}
