/*---------------------------------------------------------------------------

   Swyx Solutions AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) Swyx Solutions AG
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/LinkMgr.root/H323/source/h323.cpp                    $
    $Author:: Heiner.kath                                                  $
      $Date:: 21.09.05 13:54                                               $
   $Modtime:: 21.09.05 13:36                                               $
  $Revision:: 2                                                            $
  $Workfile:: h323.cpp                                                     $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "h323.h"
#include "h323neg.h"
#include "channels.h"
#include "h323t38.h"

using namespace std;
using namespace Tools;

#define new SNEW


/////////////////////////////////////////////////////////////////////////////

//#if STRACING
const char * const H323Connection::ConnectionStatesNames[NumConnectionStates] = {
  "NoConnectionActive",
  "AwaitingGatekeeperAdmission",
  "AwaitingTransportConnect",
  "AwaitingSignalConnect",
  "HasExecutedSignalConnect",
  "EstablishedConnection",
  "TransmitterSidePaused",
  "ShuttingDownConnection"
};

const char * const H323Connection::FastStartStateNames[NumFastStartStates] = {
  "FastStartDisabled", "FastStartInitiate", "FastStartResponse", "FastStartAcknowledged"
};

const char * const H323Connection::AnswerCallResponseNames[NumAnswerCallResponses] = {
  "AnswerCallNow", "AnswerCallDenied", "AnswerCallPending"
};

const char * const H323Connection::TerminationReasonNames[NumTerminationReasons] = {
      "LocalTermination",
      "NoAcceptTermination",
      "AnswerDeniedTermination",
      "RemoteTermination",
      "RefusalTermination",
      "NoAnswerTermination",
      "CallAbortTermination",
      "AbnormalTermination",
      "ConnectFailTermination",
      "GatekeeperForcedTermination",
      "EndedByCapabilityExchange"
};
//#endif

// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H323Connection::H323Connection(H323EndPoint & ep, unsigned ref, const Tools::SString & token, const unsigned long ulConnId)
  : endpoint(ep),
  m_bControlDataHandled(false)
#ifdef _DEBUG // add name to H323 connection in use flag mutex
  , inUseFlag( 1, 1, (const char*) Tools::SString(Tools::SString::Printf, "H323Connection::inUseFlag[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
  , answerWaitFlag(Tools::SString(Tools::SString::Printf, "H323Connection::answerWaitFlag[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#else
  , inUseFlag( 1, 1 )
#endif
#pragma warning(default:4355)
{
  SDEFFUNC("H323Connection::H323Connection");

  // valid connection ID available?
  if ( ulConnId == 0 )
  {
    // Nope! -> assign connection ID for tracing purpose
    AssignConnID();
  }
  else
    SetConnID(ulConnId);

  // set connection ID to trace capility handling
  localCapabilities.SetConnID( GetConnID() );
  // get local capilities from endpoint
  localCapabilities = endpoint.GetCapabilities();

  callAnswered = false;
  callReference = ref;
  callToken = token;

  signallingChannel = NULL;
  controlChannel = NULL;
  h245Tunneling = true;
  h245TunnelPDU = NULL;
  h245ParallelPDU = NULL;

  connectionState = NoConnectionActive;
  terminationReason = NumTerminationReasons;
  answerResponse = NumAnswerCallResponses;

  bandwidthAvailable = endpoint.GetInitialBandwidth();

  fastStartState = FastStartDisabled;

  remoteMaxAudioDelayJitter = 0;

  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(), "()");
}


H323Connection::~H323Connection()
{
  SDEFFUNC("H323Connection::~H323Connection");

  delete signallingChannel;
  delete controlChannel;

  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(), "()");
}


bool H323Connection::Init()
{
    SDEFFUNC("H323Connection::Init");

    masterSlaveDeterminationProcedure = new H245NegMasterSlaveDetermination(endpoint, *this);
    if (SBAD_READPTR(masterSlaveDeterminationProcedure))
        return false;
    masterSlaveDeterminationProcedure->IncRef(__FILE__,__LINE__);

    capabilityExchangeProcedure = new H245NegTerminalCapabilitySet(endpoint, *this);
    if (SBAD_READPTR(capabilityExchangeProcedure))
    {
        // delete previosly allocated stuff
        masterSlaveDeterminationProcedure->DecRef(__FILE__,__LINE__);
        masterSlaveDeterminationProcedure = 0;
        return false;
    }
    capabilityExchangeProcedure->IncRef(__FILE__,__LINE__);

    logicalChannels = new H245NegLogicalChannels(endpoint, *this);
    if (SBAD_READPTR(logicalChannels))
    {
        // delete previosly allocated stuff
        masterSlaveDeterminationProcedure->DecRef(__FILE__,__LINE__);
        capabilityExchangeProcedure->DecRef(__FILE__,__LINE__);
        masterSlaveDeterminationProcedure = 0;
        capabilityExchangeProcedure = 0;
        return false;
    }
    logicalChannels->IncRef(__FILE__,__LINE__);   

    requestModeProcedure = new H245NegRequestMode(endpoint, *this);
    if (SBAD_READPTR(requestModeProcedure))
    {
        // delete previosly allocated stuff
        masterSlaveDeterminationProcedure->DecRef(__FILE__,__LINE__);
        capabilityExchangeProcedure->DecRef(__FILE__,__LINE__);
        logicalChannels->DecRef(__FILE__,__LINE__);
        masterSlaveDeterminationProcedure = 0;
        capabilityExchangeProcedure = 0;
        logicalChannels = 0;
        return false;
    }
    requestModeProcedure->IncRef(__FILE__,__LINE__);           

    roundTripDelayProcedure = new H245NegRoundTripDelay(endpoint, *this);
    if (SBAD_READPTR(roundTripDelayProcedure))
    {
        // delete previosly allocated stuff
        masterSlaveDeterminationProcedure->DecRef(__FILE__,__LINE__);
        capabilityExchangeProcedure->DecRef(__FILE__,__LINE__);
        logicalChannels->DecRef(__FILE__,__LINE__);
        requestModeProcedure->DecRef(__FILE__,__LINE__);
        masterSlaveDeterminationProcedure = 0;
        capabilityExchangeProcedure = 0;
        logicalChannels = 0;
        requestModeProcedure = 0;
        return false;
    }
    roundTripDelayProcedure->IncRef(__FILE__,__LINE__);

    h450dispatcher = new H450xDispatcher(*this);
    h4502handler   = new H4502Handler(*this, *h450dispatcher);
    h4504handler   = new H4504Handler(*this, *h450dispatcher);
    
    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(), "()");

     // okay -> let's rock and roll!
    return true;
}


void H323Connection::Delete()
{
    SDEFFUNC("H323Connection::Delete");

    // Call of Delete method instead of delete itself to prevent crashes in timeouts!
    if (!SBAD_READPTR(masterSlaveDeterminationProcedure))
    {
        masterSlaveDeterminationProcedure->Delete();
        masterSlaveDeterminationProcedure->DecRef(__FILE__,__LINE__);
    }
    masterSlaveDeterminationProcedure = 0;

    if (!SBAD_READPTR(capabilityExchangeProcedure))
    {
        capabilityExchangeProcedure->Delete();
        capabilityExchangeProcedure->DecRef(__FILE__,__LINE__);
    }
    capabilityExchangeProcedure = 0;

    if (!SBAD_READPTR(requestModeProcedure))
    {
        requestModeProcedure->Delete();
        requestModeProcedure->DecRef(__FILE__,__LINE__);
    }
    requestModeProcedure = 0;

    if (!SBAD_READPTR(roundTripDelayProcedure))
    {
        roundTripDelayProcedure->Delete();
        roundTripDelayProcedure->DecRef(__FILE__,__LINE__);
    }
    roundTripDelayProcedure = 0;

    delete h4502handler;
    h4502handler = 0;
    delete h4504handler;
    h4504handler = 0;
    delete h450dispatcher;
    h450dispatcher = 0;

	if (STrace::CanTrace(g_pH323Module->m_H323TraceId, STrace::Info3))
    {
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "fast start channels: " << fastStartChannels.size()
            << ", logical channels: " << logicalChannels->GetSize());
		for (H323LogicalChannelList::size_type i = 0; i < fastStartChannels.size(); i++)
        {
            H323Channel *pLogCh = fastStartChannels[i];
            if (!SBAD_READPTR(pLogCh))
            {
                STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                    "fast start channel " << i << ": " << pLogCh
                    << " No: " << pLogCh->GetNumber());
            }
        }
        for (int j = 0; j < logicalChannels->GetSize(); j++)
        {
            H323Channel *pLogCh = logicalChannels->GetChannelAt(j);
            if (!SBAD_READPTR(pLogCh))
            {
                STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                    "logical channel " << j << ": " << pLogCh
                    << " No: " << pLogCh->GetNumber());
            }
        }
    }

    // check for remaining fast start channel entries
    H323LogicalChannelList::iterator pos = fastStartChannels.begin();
    while (!fastStartChannels.empty())
    {
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "channel: " << (*pos));

        // Look for a matching pointer in the logical channels hash map
        if ( !logicalChannels->Erase( (*pos) ) )
        {
            // Channel not found in logicalChannels, delete it here
            STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                "deleting channel: " << (*pos));
            delete (*pos);
        }
        // Channel was deleted, in logicalChannels->Erase() or above.
        // Therefore erase it here in fast start list.
        pos = fastStartChannels.erase(pos);
    }
    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "fast start channels: " << fastStartChannels.size()
        << ", logical channels: " << logicalChannels->GetSize());

    // delete logical channel hash map AFTER locking for pointers matching with fast start channels hash map
    logicalChannels->DecRef(__FILE__,__LINE__);
    logicalChannels = 0;
    //delete logicalChannels;

    //! check for remaining entries in the RTP session table(is this table REALLY needed? RB 22.2.2000)
    if ( rtpSessions.size() > 0 )
    {
        RTPSessionDict::iterator sessionDictIter;
        for( sessionDictIter =  rtpSessions.begin();
             sessionDictIter != rtpSessions.end();
            ++sessionDictIter )
        {
            delete (*sessionDictIter).second;
        }
    }

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                  "()");
}


void H323Connection::SetTerminationReason(TerminationReason reason)
{
  // Only set reason if not already set to something
  if (terminationReason == NumTerminationReasons)
    terminationReason = reason;
}


bool H323Connection::ClearCall(H323Connection::TerminationReason reason)
{
  return endpoint.ClearCall(callToken, reason);
}


void H323Connection::CleanUpOnTermination()
{
    SDEFFUNC("H323Connection::CleanUpOnTermin");

  // Make sure no one is still using this object before we go to town on it
  Lock();

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
                 "() Connection \"" << callToken << "\" closing: "
                 << "reason=" << terminationReason
                 << ", states: " << connectionState << " -> " << ShuttingDownConnection);

  connectionState = ShuttingDownConnection;

  // Check for gatekeeper and do disengage if have one
  if (terminationReason != GatekeeperForcedTermination) {
    H323Gatekeeper * gatekeeper = endpoint.GetGatekeeper();
    if (gatekeeper != NULL)
    {
      gatekeeper->DisengageRequest(*this, H225_DisengageReason::e_normalDrop);
      gatekeeper->DecRef(__FILE__,__LINE__);
    }      
  }

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                  "() fast start channels: " << fastStartChannels.size()
                  << ", logical channels: " << logicalChannels->GetSize());

  // Clean up any fast start "pending" channels we may have running.
  H323LogicalChannelList::size_type i;

  // channel objects stored in the container hash map will be destroy during the destructor!
  for (i = 0; i < fastStartChannels.size(); i++)
    fastStartChannels[i]->CleanUpOnTermination();

  // Dispose of all the logical channels
  for (int j = 0; j < logicalChannels->GetSize(); j++) {
    H323Channel * channel = logicalChannels->GetChannelAt(j);
    if (!SBAD_READPTR(channel))
      channel->CleanUpOnTermination();
  }

//! // Commented out: We should send EndSessionCommand not only on local termination ?
  //if (   terminationReason == LocalTermination
  //    || terminationReason == EndedByCapabilityExchange) {
    // Send an H.245 end session to the remote endpoint.
  //}


  /* If we have a signalling channel, then we should send a release complete.
     This may actually be redundent if we a terminating via the remote end
     sending release complete or if there is a transport failure, but it does
     not hurt to try and send it anyway.
   */
  if (signallingChannel != NULL) {
    if (signallingChannel->IsOpen() && terminationReason!=CallAbortTermination) {
      STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                   "() Sending release complete PDU: callRef=" << callReference);

      // if cause not yet set --> set to normal call clearing
      if( GetCause().cause==Q931::Cause::e_UnknownCause ) 
          SetCause(Q931::Cause(Q931::Cause::e_NormalCallClearing));


      H323SignalPDU releaseCompletePDU;
      releaseCompletePDU.BuildReleaseComplete(*this, GetReleaseCompleteReason(), GetCause());

      /*
      // if we use TLS connection, set the Siemens proprietary nonStandardControl information
      // establishTLSConnection
      if (dynamic_cast<H323TransportTLS *>(signallingChannel) != 0)
      {
          replyPDU.IncludeOptionalField(H225_H323_UU_PDU::e_nonStandardControl);
          int last = replyPDU.
      }
      */

      // write end session command into the ReleaseComplete
      h245TunnelPDU=&releaseCompletePDU;

      H323ControlPDU pdu;
      pdu.BuildEndSessionCommand(H245_EndSessionCommand::e_disconnect);
      WriteControlPDU(pdu);

      OnSendReleaseComplete(releaseCompletePDU, callReference, GetReleaseCompleteReason(), GetCause());
      WriteSignalPDU(releaseCompletePDU);

      h245TunnelPDU=0; //forget tunnel pdu as it is sent out
    }
    signallingChannel->CleanUpOnTermination();
  }

  // Wait for control channel to be cleaned up (thread ended).
  if (controlChannel != NULL)
    controlChannel->CleanUpOnTermination();

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
                 "() Connection " << callToken << " terminated.");

  // There is intentionally no Unlock() at this point because this connection is
  // going to be deleted in the next time and we want to prevent that e.g. a call
  // to H323EndPoint::FindConnectionWithLock() could access this connection!
  // Unlock() is implicitly in delete of the H323Connection!
}


void H323Connection::AttachSignalChannel(H323Transport * channel)
{
  if (signallingChannel != NULL && signallingChannel->IsOpen()) {
    SAssertAlways(SLogicError);
    return;
  }

  delete signallingChannel;
  signallingChannel = channel;
}


bool H323Connection::WriteSignalPDU(H323SignalPDU & pdu)
{
  if ( signallingChannel && signallingChannel->IsOpen())
  {
        pdu.m_h323_uu_pdu.m_h245Tunneling = h245Tunneling;
        return pdu.Write(*signallingChannel);
  }

  return false;
}


void H323Connection::HandleSignallingChannel()
{
  SDEFFUNC("H323Connection::HandleSigChannel");

  SAssert(signallingChannel != NULL, SLogicError);

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Reading PDUs: callRef=" << callReference);

  H323SignalPDU pdu;
  while (signallingChannel->IsOpen()) {
    if (pdu.Read(*signallingChannel)) {
      if (!HandleSignalPDU(pdu)) {
        ClearCall(AbnormalTermination);
        break;
      }
    }
#ifdef ROUND_TRIP_DELAY
    else if (connectionState != EstablishedConnection ||
             signallingChannel->GetErrorCode() != SChannel::Timeout) {
      if (controlChannel == NULL || !controlChannel->IsOpen())
        ClearCall(signallingChannel->GetErrorCode() != SChannel::Timeout
                                  ? AbnormalTermination : NoAnswerTermination);
      signallingChannel->Close();
      break;
    }

    if (controlChannel == NULL)
      StartRoundTripDelay();
#else
    else {
    //if (controlChannel == NULL || !controlChannel->IsOpen())
        ClearCall(signallingChannel->GetErrorCode() != SChannel::Timeout
                                  ? AbnormalTermination : NoAnswerTermination);
      signallingChannel->Close();
    }
#endif // ROUND_TRIP_DELAY
  }

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Signal channel closed.");
}


bool H323Connection::HandleSignalPDU(H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::HandleSignalPDU");

    const Q931& q931 = pdu.GetQ931();

    m_bControlDataHandled=false; // set flag to find out if HandleControlData is called in some message handling (e.g. OnRecvAlerting)

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Handling PDU: " << q931.GetMessageTypeName()
                 << " callRef=" << q931.GetCallReference());

    if( q931.GetCallReference() != callReference ) 
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                     "() callRef mismatch msg.callRef=" 
                     << q931.GetCallReference() <<"!="<<callReference
                     <<" --> ignore message, might be a trailing message through a reused TLS connection!" );
        return true;
    }

    // Do not check tunneling if e.g. SetupAck because it has no UUIE,
    // but the member variables are present and initialised!
    // -> Leads to problems if tunneling and overlapped sending is used!
    switch (q931.GetMessageType())
    {
        // Some messages have no Q.931 UUIE
        case Q931::SetupAckMsg:
        case Q931::NotifyMsg:    
        case Q931::InformationMsg:
        case Q931::StatusMsg:
        case Q931::StatusEnquiryMsg:
            break;

        default:
            // If remote does not do tunneling, so we don't either. Note that if it
            // gets turned off once, it stays off for good.
            h245Tunneling = h245Tunneling && pdu.m_h323_uu_pdu.m_h245Tunneling;
            break;
    }

  switch (q931.GetMessageType()) {
    case Q931::SetupMsg :
      if (!OnReceivedSignalSetup(pdu))
        return false;
      break;

    case Q931::SetupAckMsg :
      if (!OnReceivedSetupAck(pdu))
        return false;
      break;

    case Q931::InformationMsg :
      if (!OnReceivedInformation(pdu))
        return false;
      break;

    case Q931::CallProceedingMsg :
      if (!OnReceivedCallProceeding(pdu))
        return false;
      break;

    case Q931::AlertingMsg :
      if (!OnReceivedAlerting(pdu))
        return false;
      break;

    case Q931::ConnectMsg :
      if (!OnReceivedSignalConnect(pdu))
        return false;
      break;

    case Q931::ProgressMsg :
      if (!OnReceivedProgress(pdu))
        return false;
      break;

    case Q931::NotifyMsg :
      if (!OnReceivedNotify(pdu))
        return false;
      break;

    case Q931::FacilityMsg :
      if (!OnReceivedFacility(pdu))
        return false;
      break;

    case Q931::StatusMsg :
      if (!OnReceivedStatus(pdu))
        return false;
      break;

    case Q931::StatusEnquiryMsg :
      if (!OnReceivedStatusEnquiry(pdu))
        return false;
      break;

    case Q931::ReleaseCompleteMsg :
      OnReceivedReleaseComplete(pdu);
      if (   connectionState == EstablishedConnection
          || connectionState == TransmitterSidePaused)
        ClearCall(RemoteTermination);
      else if (answerResponse == AnswerCallPending)
        ClearCall(CallAbortTermination);
      else
        ClearCall(RefusalTermination);
      return false;

    default :
      if (!OnUnknownSignalPDU(pdu))
        return false;
  }

  // if tunneled PDU was already processed do not process it again, it causes useless errors in trace
  if( m_bControlDataHandled==false )
  {
      // Process tunnelled H245 PDU, if present.
      HandleTunnelPDU(pdu, NULL);
  }

  return true;
}


void H323Connection::HandleTunnelPDU(const H323SignalPDU & pdu, H323SignalPDU * reply)
{
    SDEFFUNC("H323Connection::HandleTunnelPDU");

  if (!h245Tunneling)
    return;

  h245TunnelPDU = reply;

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Handling " << pdu.m_h323_uu_pdu.m_h245Control.GetSize() << " tunnel PDUs.");

  for (int i = 0; i < pdu.m_h323_uu_pdu.m_h245Control.GetSize(); i++) {
    SPerStream strm = pdu.m_h323_uu_pdu.m_h245Control[i].GetValue();
    HandleControlData(strm);
  }

  h245TunnelPDU = NULL;

  // Make sure does not get repeated, break const and clear tunnelled H.245 PDU's
  ((H323SignalPDU &)pdu).m_h323_uu_pdu.m_h245Control.SetSize(0);
}


bool H323Connection::BuildFastStartList(const H323Channel & channel,
                                        H225_ArrayOf_SAsnOctetString & array,
                                        H323Channel::Directions reverseDirection)
{
    SDEFFUNC("H323Connection::BuildFastStrtList");

  H245_OpenLogicalChannel open;
  const H323Capability & capability = channel.GetCapability();

  if (channel.GetDirection() != reverseDirection) {
    if (!capability.OnSendingPDU(open.m_forwardLogicalChannelParameters.m_dataType))
      return false;
  }
  else {
    if (!capability.OnSendingPDU(open.m_reverseLogicalChannelParameters.m_dataType))
      return false;

    open.m_forwardLogicalChannelParameters.m_multiplexParameters.SetTag(
                H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters::e_none);
    open.m_forwardLogicalChannelParameters.m_dataType.SetTag(H245_DataType::e_nullData);
    open.IncludeOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters);
  }

  if (!channel.OnSendingPDU(open))
    return false;

  STRACE2_INFO2(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Build fastStart:\n" << STrace::MultiLinePrefix << "  "
               << setfill(' ') << setprecision(2) << open);
  SPerStream strm;
  open.Encode(strm);
  strm.CompleteEncoding();

  int last = array.GetSize();
  array.SetSize(last+1);
  array[last] = strm;
  return true;
}


bool H323Connection::OnReceivedSignalSetup(const H323SignalPDU & setupPDU)
{
    SDEFFUNC("H323Connection::OnRecvSigSetup");

	H323LogicalChannelList::size_type i;

  const H225_Setup_UUIE & setup = setupPDU.m_h323_uu_pdu.m_h323_message_body;

  // Indicate the direction of call.
  callAnswered = true;

  // Save the identifiers sent by caller
  if (setup.HasOptionalField(H225_Setup_UUIE::e_callIdentifier))
    callIdentifier = setup.m_callIdentifier.m_guid;
  conferenceIdentifier = setup.m_conferenceID;

    if (endpoint.IsOverlappedSendingEnabled())
    {
        // Overlapped sending is enabled
        if (   setup.HasOptionalField(H225_Setup_UUIE::e_canOverlapSend)
            && setup.m_canOverlapSend == true
            && !setupPDU.GetQ931().HasIE(Q931::SendingCompleteIE) )
        {
            // Send back a H.225 Setup Acknowledge PDU
            H323SignalPDU setupAckPDU;
            setupAckPDU.BuildSetupAck(callReference);
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Sending setup acknowledge PDU");
            if (!WriteSignalPDU(setupAckPDU))
                return false;
//          m_bIsOverlappedSending = true;
        }
        else
        {
            // Send back a H.225 Call Proceeding PDU
            H323SignalPDU callProceedingPDU;
            callProceedingPDU.BuildCallProceeding(*this, callReference, callIdentifier);
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Sending call proceeding PDU");
            if (!WriteSignalPDU(callProceedingPDU))
                return false;
        }
    }
    else
    {
        // Send back a H323 Call Proceeding PDU in case OnIncomingCall() takes a while
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                     "() Sending call proceeding PDU");
        H323SignalPDU callProceedingPDU;
        callProceedingPDU.BuildCallProceeding(*this, callReference, callIdentifier);
        if (!WriteSignalPDU(callProceedingPDU))
            return false;
    }

  // Determine the remote parties name as best we can
  if (setup.HasOptionalField(H225_Setup_UUIE::e_sourceAddress) &&
      setup.m_sourceAddress.GetSize() > 0) {
    remotePartyName = H323GetAliasAddressString(setup.m_sourceAddress[0]);
    if (setup.m_sourceAddress.GetSize() > 1) {
      remotePartyName += " (";
	  for (int j = 1; j < setup.m_sourceAddress.GetSize(); j++) {
        if (j > 1)
          remotePartyName += ", ";
        remotePartyName += H323GetAliasAddressString(setup.m_sourceAddress[j]);
      }
      remotePartyName += ')';
    }
  }
  else {
    remotePartyName = setupPDU.GetQ931().GetDisplayName();
    if (remotePartyName.empty())
      remotePartyName = signallingChannel->GetRemoteAddressString();
  }

  // See if remote endpoint wants to start fast
  if (setup.HasOptionalField(H225_Setup_UUIE::e_fastStart))
  {
/*
    H323CapabilityList::iterator CapListIter;
    for ( CapListIter = remoteCapabilityTable.begin(); CapListIter != remoteCapabilityTable.end(); ++ CapListIter )
    {
        delete (*CapListIter);
    }
*/
    remoteCapabilities.RemoveAll();
    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Fast start detected");

    try
    {

        // Extract capabilities from the fast start OpenLogicalChannel structures
        for (int i = 0; i < setup.m_fastStart.GetSize(); i++) {
          H245_OpenLogicalChannel open;
          SPerStream strm = setup.m_fastStart[i].GetValue();
          if (open.Decode(strm)) {
            STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Fast start open:\n" << STrace::MultiLinePrefix
                         <<"  " << setprecision(2) << open);
            unsigned error;
            H323Channel * channel = CreateLogicalChannel(open, error);
            if (channel != NULL) {
              if (channel->GetDirection() == H323Channel::IsTransmitter)
                channel->SetNumber(logicalChannels->GetNextChannelNumber());
              fastStartChannels.push_back(channel);
            }
          }
        }
    }

    catch(...)
    {
        STRACE2_FATAL(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() ### Exception in ASN decode of Setup fast start entries thrown! ###");
        return false;
    }

    // If we are incapable of ANY of the fast start channels, don't do fast start
    if (!fastStartChannels.empty())
      fastStartState = FastStartResponse;
  }

  // Check for gatekeeper and do admission check if have one
  {
      H323Gatekeeper * gatekeeper = endpoint.GetGatekeeper();
      if (gatekeeper != NULL) {
        if (!gatekeeper->AdmissionRequest(*this)) {
          ClearCall(GatekeeperForcedTermination);
          gatekeeper->DecRef(__FILE__,__LINE__);
          return false;
        }
        gatekeeper->DecRef(__FILE__,__LINE__);
      }

  }

  // if the application indicates not to contine, then send a Q931 Release Complete PDU
  H323SignalPDU alertingPDU;
  alertingPDU.BuildAlerting(*this, callReference, callIdentifier, GetProgressIndicator());
  if (!OnIncomingCall(setupPDU, alertingPDU)) {
    ClearCall(NoAcceptTermination);
    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                  "() Application not accepting calls");
    return false;
  }

  // send Q931 Alerting PDU
  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Incoming call accepted - sending Alerting PDU");
  HandleTunnelPDU(setupPDU, &alertingPDU);
  if (!WriteSignalPDU(alertingPDU))
    return false;

  // Build the reply with the channels we are actually using
  H323SignalPDU connectPDU;
  H225_Connect_UUIE & connect = connectPDU.BuildConnect(*this,
                                                        callReference,
                                                        callIdentifier,
                                                        conferenceIdentifier);

  // call the application callback to determine if to answer the call or not
  answerResponse = OnAnswerCall(remotePartyName, setupPDU, connectPDU);
  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Answer call response: " << answerResponse);

  // If pending response wait
  if (answerResponse == AnswerCallPending) {
    // Wait for answer from application, but check for call being cleared in
    // the mean time.
    while (!answerWaitFlag.Wait(500)) {
      STimeInterval oldTimeout = signallingChannel->GetReadTimeout();
      signallingChannel->SetReadTimeout(0);

      H323SignalPDU pdu;
      if (pdu.Read(*signallingChannel)) {
        if (!HandleSignalPDU(pdu))
          return false;
      }
      else {
        if (signallingChannel->GetErrorCode() != SChannel::Timeout)
          return false;
      }

      if (connectionState == ShuttingDownConnection)
        return false;

      signallingChannel->SetReadTimeout(oldTimeout);
    }
  }

  // If response is denied, abort the call
  if (answerResponse == AnswerCallDenied) {
    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
        "() Application has declined to answer incoming call");
    ClearCall(AnswerDeniedTermination);
    return false;
  }

  // Now ask the application to select which channels to start
  if (fastStartState == FastStartResponse)
    OnSelectLogicalChannels();

  // Remove any channels that were not started by OnSelectLogicalChannels(),
  // those that were started are put into the logical channel dictionary
  for (H323LogicalChannelList::iterator pos =  fastStartChannels.begin();
                                        pos != fastStartChannels.end(); ++pos)
  {
        if (!SBAD_READPTR((*pos)))
        {
            if (g_pH323Module->m_bUsingSwtcl)
            {
                // No LogicalChannel thread started if we use SwTCL.
                // Therefore we do not check for running channels and just
                // add them to the list.
                logicalChannels->Add(*(*pos));
            }
            else
            {
                if ((*pos)->IsRunning())
                    logicalChannels->Add(*(*pos));
                else
                {
                    delete (*pos);
                    fastStartChannels.erase(pos);
                }
            }
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                "() skipped invalid logical channel!");
        }
  }

  // The channels we just transferred to the logical channels dictionary
  // should not be deleted via this structure now.
  //fastStartChannels.DisallowDeleteObjects();

  // Set flag that we are up to CONNECT stage
  if (connectionState == ShuttingDownConnection)
    return false;
  connectionState = HasExecutedSignalConnect;

  // No channels selected (or never provided) so do traditional H245 start
  if (fastStartChannels.empty()) {
    fastStartState = FastStartDisabled;
    if (h245Tunneling) { // Piggy back H245 on this reply
      h245TunnelPDU = &connectPDU;
      capabilityExchangeProcedure->Start();
      HandleTunnelPDU(setupPDU, &connectPDU);
      return WriteSignalPDU(connectPDU); // Send H323 Connect PDU
    }
    controlChannel = signallingChannel->CreateControlChannel(*this);
    return controlChannel != NULL;
  }

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Accepting fastStart for " << fastStartChannels.size() << " channels");

  connect.IncludeOptionalField(H225_Connect_UUIE::e_fastStart);

  for (i = 0; i < fastStartChannels.size(); i++)
    BuildFastStartList(*fastStartChannels[i], connect.m_fastStart, H323Channel::IsTransmitter);

  // Have moved open channels to logicalChannels structure, remove all others.
  fastStartChannels.clear();

  // If have some acceptable channels, do the fast start
  if (connect.m_fastStart.GetSize() > 0) {
    connectionState = EstablishedConnection;
    connectionStartTime = STime();
    endpoint.OnConnectionEstablished(*this, callToken);
    return WriteSignalPDU(connectPDU); // Send H323 Connect PDU
  }

  // Otherwise something really odd just happened, abort call
  STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                "() Fast start error!");
  return false;
}


void H323Connection::SetRemotePartyName(const H323SignalPDU & pdu)
{
  SString newRemotePartyName = pdu.GetQ931().GetDisplayName();
  if (!newRemotePartyName.empty())
    remotePartyName = newRemotePartyName;
  if (remotePartyName.empty())
    remotePartyName = signallingChannel->GetRemoteAddressString();
}


bool H323Connection::OnReceivedSetupAck(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnRecvSetupAck");

    SetRemotePartyName(pdu);

    if (pdu.GetQ931().GetProgressIndicator(m_ProgressIndicator))
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                     "() Found progress indicator: "
                     << pdu.GetQ931().GetProgressIndicatorInfo(m_ProgressIndicator));
    }

    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                  "() H323Connection::OnReceivedSetupAck not implemented!");

/* for tests
    // Send H.225 Information message
    STRACE_INFO(g_pH323Module->m_H225TraceId,
                "Sending Information message");
    H323SignalPDU informationPDU;
    informationPDU.BuildInformation(callReference, false, "123", callIdentifier);
    if (!WriteSignalPDU(informationPDU))
    {
        STRACE_ERROR(g_pH323Module->m_H225TraceId,
                     " Sending Information message failed!");
        return false;
    }        
*/
    return true;
}


bool H323Connection::OnReceivedInformation(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnRecvInformat");

    SetRemotePartyName(pdu);

    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                  "() H323Connection::OnReceivedInformation not implemented!");

    return true;
}


bool H323Connection::OnReceivedCallProceeding(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnRecvCallProc");

  SetRemotePartyName(pdu);

    if (pdu.GetQ931().GetProgressIndicator(m_ProgressIndicator))
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                     "() Found progress indicator: "
                     << pdu.GetQ931().GetProgressIndicatorInfo(m_ProgressIndicator));
    }

  const H225_CallProceeding_UUIE & call = pdu.m_h323_uu_pdu.m_h323_message_body;

  // Check for fastStart data and start fast
  if (call.HasOptionalField(H225_CallProceeding_UUIE::e_fastStart))
    HandleFastStartAcknowledge(call.m_fastStart);

  // Check that it has the H.245 channel connection info
  if (call.HasOptionalField(H225_CallProceeding_UUIE::e_h245Address))
    return StartControlChannel(call.m_h245Address);

  return true;
}


bool H323Connection::OnReceivedAlerting(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnRecvAlerting");

  SetRemotePartyName(pdu);

    if (pdu.GetQ931().GetProgressIndicator(m_ProgressIndicator))
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                     "() Found progress indicator: "
                     << pdu.GetQ931().GetProgressIndicatorInfo(m_ProgressIndicator));
    }
    Q931::SignalInfo signalInfo = pdu.GetQ931().GetSignalInfo();
    if (signalInfo != Q931::SignalErrorInIE)
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() Found signal info element: "
            << pdu.GetQ931().GetSignalInfoName(signalInfo));
    }

  const H225_Alerting_UUIE & alert = pdu.m_h323_uu_pdu.m_h323_message_body;

  // Check for fastStart data and start fast
  if (alert.HasOptionalField(H225_Alerting_UUIE::e_fastStart))
    HandleFastStartAcknowledge(alert.m_fastStart);

  // Check that it has the H.245 channel connection info
  if (alert.HasOptionalField(H225_Alerting_UUIE::e_h245Address))
    if (!StartControlChannel(alert.m_h245Address))
      return false;

  return OnAlerting(pdu, remotePartyName);
}


bool H323Connection::OnReceivedSignalConnect(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnRecvSigConnect");

  if (connectionState == ShuttingDownConnection)
    return false;
  connectionState = HasExecutedSignalConnect;

  SetRemotePartyName(pdu);

    if (pdu.GetQ931().GetProgressIndicator(m_ProgressIndicator))
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                     "() Found progress indicator: "
                     << pdu.GetQ931().GetProgressIndicatorInfo(m_ProgressIndicator));
    }

  if (!OnOutgoingCall(pdu)) {
    ClearCall(NoAcceptTermination);
    return false;
  }

  const H225_Connect_UUIE & connect = pdu.m_h323_uu_pdu.m_h323_message_body;

#ifdef ROUND_TRIP_DELAY
  // have answer, so set timeout to interval for pinging remote
  signallingChannel->SetReadTimeout(endpoint.GetRoundTripDelayRate());
#else
  // have answer, so set timeout to not happen
  signallingChannel->SetReadTimeout(SMaxTimeInterval);
#endif // ROUND_TRIP_DELAY

  // Check for fastStart data and start fast
  if (connect.HasOptionalField(H225_Connect_UUIE::e_fastStart))
    HandleFastStartAcknowledge(connect.m_fastStart);

  // Check that it has the H.245 channel connection info
  if (connect.HasOptionalField(H225_Connect_UUIE::e_h245Address) && 
      pdu.m_h323_uu_pdu.m_h245Tunneling.GetValue()==false) // do not open control channel if fast start already was successful
    if (!StartControlChannel(connect.m_h245Address))
      if (fastStartState != FastStartAcknowledged)
        return false;

  // Fast started some channels, so we are connected straight away
  if (fastStartState == FastStartAcknowledged) {
    fastStartState = FastStartDisabled;
    connectionState = EstablishedConnection;
    connectionStartTime = STime();
    endpoint.OnConnectionEstablished(*this, callToken);
    return true;
  }

  // Abandon fast start, if were attempting to do it though we need
  // to do capabilty exchange later
  fastStartState = FastStartDisabled;
  bool needsCapabilityExchange = fastStartChannels.size() > 0;
  fastStartChannels.clear();

  if (!h245Tunneling && controlChannel == NULL) {
    STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() No H245 address in remote PDU");
    return false;
  }

  // Tried to do fast start and was refused. But are using h245 tunnelling
  // so need to initiate a capability exchange which we hadn't before as we
  // were fast starting.
  if (needsCapabilityExchange) {
    capabilityExchangeProcedure->Start();
    return true;
  }

  // Check for if all the 245 conditions are met so can call
  // OnControlChannelOpen() to start up logical channels and complete the
  // connection establishment.
  if (masterSlaveDeterminationProcedure->IsDetermined() &&
      capabilityExchangeProcedure->HasSentCapabilities() &&
      capabilityExchangeProcedure->HasReceivedCapabilities()) {
    if (!OnControlChannelOpen())
      return false;
    connectionState = EstablishedConnection;
  }
  
  return true;
}


bool H323Connection::OnReceivedProgress(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnRecvProgress");
    
    if (pdu.GetQ931().GetProgressIndicator(m_ProgressIndicator))
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() Found progress indicator: "
            << pdu.GetQ931().GetProgressIndicatorInfo(m_ProgressIndicator));
    }

    Q931::Cause cause;
    if (pdu.GetQ931().GetCause(cause) == true)
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() Found CauseIE, cause: " << cause.cause);
        SetCause(cause);
    }
    
    const H225_Progress_UUIE & progress = pdu.m_h323_uu_pdu.m_h323_message_body;

    // Check for fastStart data and start fast
    if (progress.HasOptionalField(H225_Progress_UUIE::e_fastStart))
        HandleFastStartAcknowledge(progress.m_fastStart);

    // Check that it has the H.245 channel connection info
    if (progress.HasOptionalField(H225_Progress_UUIE::e_h245Address))
        return StartControlChannel(progress.m_h245Address);

    return true;
}


bool H323Connection::OnReceivedFacility(const H323SignalPDU & pdu)
{
  if (pdu.m_h323_uu_pdu.m_h323_message_body.GetTag() ==
                            H225_H323_UU_PDU_h323_message_body::e_empty)
    return true;

  const H225_Facility_UUIE & fac = pdu.m_h323_uu_pdu.m_h323_message_body;

  // Check for fastStart data and start fast
  if (fac.HasOptionalField(H225_Facility_UUIE::e_fastStart))
    HandleFastStartAcknowledge(fac.m_fastStart);

  // Check that it has the H.245 channel connection info
  if (fac.HasOptionalField(H225_Facility_UUIE::e_h245Address))
    return StartControlChannel(fac.m_h245Address);

  return true;
}


bool H323Connection::OnReceivedNotify(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnReceivedNotify");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Received Notify PDU");
    return true;
}


bool H323Connection::OnReceivedStatus(const H323SignalPDU & pdu)
{
    SDEFFUNC("H323Connection::OnReceivedStatus");

    Q931::Cause cause;
    if (pdu.GetQ931().GetCause(cause) == true)
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() Received Status PDU, cause: " << cause.cause);
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() Received Status PDU");
    }
    return true;
}


bool H323Connection::OnReceivedStatusEnquiry(const H323SignalPDU & pdu)
{
  H323SignalPDU reply;
  reply.BuildStatus(callReference, !pdu.GetQ931().IsFromDestination());
  return reply.Write(*signallingChannel);
}


bool H323Connection::OnReceivedReleaseComplete(const H323SignalPDU & pdu)
{
    // This is a SwTCL extension
    return OnReleaseComplete(pdu);
}


bool H323Connection::OnIncomingCall(const H323SignalPDU &, H323SignalPDU &)
{
  return true;
}


H323Connection::AnswerCallResponse
     H323Connection::OnAnswerCall(const SString & STRACE_caller,
                                  const H323SignalPDU &,
                                  H323SignalPDU &)
{
    SDEFFUNC("H323Connection::OnAnswerCall");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() OnAnswerCall from \"" << STRACE_caller << '"');
    return AnswerCallNow;
}


void H323Connection::AnsweringCall(AnswerCallResponse response)
{
  if (response == AnswerCallPending)
    return;

  answerResponse = response;
  answerWaitFlag.Signal();
}


bool H323Connection::SendSignalSetup(const SString & alias)
{
    SDEFFUNC("H323Connection::SendSignalSetup");

  connectionState = AwaitingGatekeeperAdmission;

  // Indicate the direction of call.
  callAnswered = false;
  remotePartyName = alias;

    // Start building the setup PDU to get various ID's
    H323SignalPDU setupPDU;
    SString szAlias;
    if (!endpoint.GetH323IdTerminalAlias().empty())
    {
        szAlias = endpoint.GetH323IdTerminalAlias();
    }
    else
    {
        szAlias = endpoint.GetLocalUserName();
    }

    SString szCalledPartyNumber;
    if (endpoint.IsE164(remotePartyName))
    {
        szCalledPartyNumber = remotePartyName;
    }

    H225_Setup_UUIE & setup = setupPDU.BuildSetup(*this,
                                                  callReference,
                                                  szAlias,
                                                  endpoint.GetCallingPartyNumber(),
                                                  szCalledPartyNumber,
                                                  remotePartyName,
                                                  szAlias,
                                                  endpoint.IsOverlappedSendingEnabled());

  // Save the identifiers generated by BuildSetup
  callIdentifier = setup.m_callIdentifier.m_guid;
  conferenceIdentifier = setup.m_conferenceID;

  // Check for gatekeeper and do admission check if have one
  {
      H323Gatekeeper * gatekeeper = endpoint.GetGatekeeper();
      if (gatekeeper != NULL) {
        H323TransportAddress gatekeeperRoute;
        if (!gatekeeper->AdmissionRequest(*this, gatekeeperRoute)) {
          STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                        "() Gatekeeper refused admission.");
          gatekeeper->DecRef(__FILE__,__LINE__);
          return false;
        }
        if (!gatekeeperRoute.empty() && !signallingChannel->SetRemoteAddress(gatekeeperRoute)) {
          STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                        "() Invalid gatekeeper route address: \"" << gatekeeperRoute << '"');
          connectionState = AwaitingTransportConnect;
          gatekeeper->DecRef(__FILE__,__LINE__);
          return false;
        }
        gatekeeper->DecRef(__FILE__,__LINE__);
      }
  }

  // Do the transport connect
  connectionState = AwaitingTransportConnect;
  if (!signallingChannel->Connect()) {
    ClearCall(ConnectFailTermination);
    connectionState = NoConnectionActive;
    return false;
  }

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Sending Setup PDU");
  connectionState = AwaitingSignalConnect;

  // Put in all the signalling addresses for link
  setup.IncludeOptionalField(H225_Setup_UUIE::e_sourceCallSignalAddress);
  SIPSocket::Address localAddr=GetLocalInterfaceAddress();
  if( localAddr.IsValid() )
    signallingChannel->SetLocalAddress(localAddr); // use the correct address of local interface according to LinkManager address determination
  signallingChannel->SetUpTransportPDU(setup.m_sourceCallSignalAddress, true);
  setup.IncludeOptionalField(H225_Setup_UUIE::e_destCallSignalAddress);
  signallingChannel->SetUpTransportPDU(setup.m_destCallSignalAddress, false);

  // Ask the application what channels to open
  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Check for Fast start by local endpoint");
  fastStartState = FastStartInitiate;
  fastStartChannels.clear();
  OnSelectLogicalChannels();

  // If application called OpenLogicalChannel, put in the fastStart field
  if (!fastStartChannels.empty()) {
    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Fast start begun by local endpoint");
    for (H323LogicalChannelList::size_type i = 0; i < fastStartChannels.size(); i++)
    {
        BuildFastStartList(*fastStartChannels[i], setup.m_fastStart, H323Channel::IsReceiver);
    }
    if (setup.m_fastStart.GetSize() > 0)
      setup.IncludeOptionalField(H225_Setup_UUIE::e_fastStart);
  }

  if (h245Tunneling) {

    //!
    //// Cannot do capability exchange if if fast starting though.
    //if (fastStartChannels.empty()) {
    //  // Try and start the master/slave and capability exchange through the tunnel
    //  if (!StartControlNegotiations())
    //    return false;
    //}

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
        "() h245Tunneling, start TCS+MSD");

    h245ParallelPDU = &setupPDU;

    // put TCS + MSD into setup (parallelH245Control), altough we are using tunneling!
    if (!StartControlNegotiations())
        return false;

    h245ParallelPDU = NULL;
  }

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
      "() call OnSendSignalSetup");

  if (!OnSendSignalSetup(setupPDU))
    return false;

  fastStartState = FastStartDisabled;

  // Send the initial PDU
  if (!WriteSignalPDU(setupPDU))
    return false;

  // Set timeout for remote party to answer the call
  signallingChannel->SetReadTimeout(endpoint.GetSignallingChannelCallTimeout());

  return true;
}


bool H323Connection::OnSendSignalSetup(H323SignalPDU & /*setupPDU*/)
{
  return true;
}


bool H323Connection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
                                const SString & /*username*/)
{
    SDEFFUNC("H323Connection::OnAlerting");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Received alerting PDU.");
    return true;
}


bool H323Connection::OnOutgoingCall(const H323SignalPDU & /*connectPDU*/)
{
    SDEFFUNC("H323Connection::OnOutgoingCall");

    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Received connect PDU.");
    return true;
}


bool H323Connection::OnSendReleaseComplete(const H323SignalPDU               & /*ReleaseCompletePDU*/,
                                           int                                 /*callReference*/,
                                           const H225_ReleaseCompleteReason  & /*reason*/,
                                           const Q931::Cause                 & /*cause*/)
{
    // This is a SwTCL extension
    return true;
}


bool H323Connection::OnReleaseComplete(const H323SignalPDU & releasePDU)
{
    SDEFFUNC("H323Connection::OnRelComplete");

    Q931::Cause cause;
    if (releasePDU.GetQ931().GetCause(cause) == true)
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() Received ReleaseComplete with Q.931 cause: '"
            << cause.GetCauseString() << "' (" << cause.cause << ")");
    }
    else
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() Received ReleaseComplete.");
    }

    const H225_ReleaseComplete_UUIE & release = releasePDU.m_h323_uu_pdu.m_h323_message_body;
    if (release.HasOptionalField(H225_ReleaseComplete_UUIE::e_reason))
    {
         STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
             "() Reason: " << release.m_reason.GetTag());

         SetReleaseCompleteReason(release.m_reason);
    }

    return true;
}


bool H323Connection::HandleFastStartAcknowledge(const H225_ArrayOf_SAsnOctetString & array)
{
    SDEFFUNC("H323Connection::HdlFastStartAck");

  if (fastStartChannels.empty()) {
    STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Fast start response with no channels to open");
    return true;
  }

  switch(fastStartState)
  {
      case FastStartAcknowledged:
          STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
                       "() fast start already done --> no further processing");
          return true;
  }

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Fast start accepted by remote endpoint");

  int i;

    try
    {
      // Go through provided list of structures, if can decode it and match it up
      // with a channel we requested AND it has all the information needed in the
      // m_multiplexParameters, then we can start the channel.
      for (i = 0; i < array.GetSize(); i++) {
        H245_OpenLogicalChannel open;
        SPerStream strm = array[i].GetValue();
        if (open.Decode(strm)) {
          STRACE2_INFO2(g_pH323Module->m_H225TraceId, this, GetConnID(),
                       "() Fast start open:\n" << STrace::MultiLinePrefix << "  "
                       << setprecision(2) << open);
          bool reverse = open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters);
          H323Capability * capability;
          if (reverse)
            capability = localCapabilities.FindCapability(open.m_reverseLogicalChannelParameters.m_dataType);
          else
            capability = localCapabilities.FindCapability(open.m_forwardLogicalChannelParameters.m_dataType);
          if (capability != NULL) {
            for (H323LogicalChannelList::size_type ch = 0; ch < fastStartChannels.size(); ch++) {
              H323Channel & chan = *fastStartChannels[ch];
              if (   chan.GetCapability() == *capability
                  && (  (chan.GetDirection() == H323Channel::IsReceiver) == reverse
                      || chan.GetCapability().IsBidirectional() ) )
              {
                unsigned error;
                if (chan.OnReceivedPDU(open, error))
                {
                    if (!g_pH323Module->m_bUsingSwtcl)
                    {
                        // Only start a LogicalChannel thread if we not use SwTCL.
                        chan.Start();
                    }
                    else
                    {
                        // Move this channel to logical channels list
                        logicalChannels->Add(chan);
                    }
                }
                else
                {
                  STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                                "() Fast start channel open error: " << error);
                }
                break;
              }
            }
          }
        }
      }
    }

    catch(...)
    {
        STRACE2_FATAL(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() ### Exception in ASN decode of fast start entries thrown! ###");
        return false;
    }

    if (!g_pH323Module->m_bUsingSwtcl)
    {
        // Remove any channels that were not started by above, those that were
        // started are put into the logical channel dictionary
        for (H323LogicalChannelList::iterator pos =  fastStartChannels.begin();
                                              pos != fastStartChannels.end(); ++pos)
        {
            if (!SBAD_READPTR((*pos)))
            {
                if ((*pos)->IsRunning())
                    logicalChannels->Add(*(*pos));
                else
                    fastStartChannels.erase(pos);
            }
            else
            {
                STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                    "() skipped invalid logical channel!");
            }
        }
    }

  // The channels we just transferred to the logical channels dictionary
  // should not be deleted via this structure now.
  //fastStartChannels.DisallowDeleteObjects();

  STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
               "() Fast starting " << fastStartChannels.size() << " channels");
  if (fastStartChannels.empty())
    return false;

  // Have moved open channels to logicalChannels structure, remove them now.
  DeleteUnusedFastStartChannels();

    if (logicalChannels->GetSize() == 0)
    {
        STRACE2_INFO(g_pH323Module->m_H225TraceId, this, GetConnID(),
            "() No matching fast start channels.");
        fastStartState = FastStartDisabled;
    }
    else
    {
        fastStartState = FastStartAcknowledged;
    }

  return true;
}


bool H323Connection::StartControlChannel(const H225_TransportAddress & h245Address)
{
    SDEFFUNC("H323Connection::StartCtrlChannel");

  // Check that it is an IP address, all we support at the moment
  if (h245Address.GetTag() != H225_TransportAddress::e_ipAddress) {
    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                  "() Connect of H245 failed: Unsupported transport");
    return false;
  }

  // Already have the H245 channel up.
  if (controlChannel != NULL)
    return true;

  H323TransportTCP * channel = new H323TransportTCP(endpoint,true);
  if (channel->SetRemoteAddress(h245Address) && channel->ConnectControlChannel(*this)) {
    controlChannel = channel;
    return true;
  }

  STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, GetConnID(),
                "() Connect of H245 failed: " << channel->GetErrorText());
  delete channel;
  return false;
}


bool H323Connection::AttachH225NonStandardData(H323SignalPDU     & pdu,
                                               const SByteVector & data)
{
    pdu.m_h323_uu_pdu.IncludeOptionalField(H225_H323_UU_PDU::e_nonStandardData);

    if (!objectIdentifier.empty())
    {
        // Connection is using object identifier
        pdu.m_h323_uu_pdu.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_object);
        ((SAsnObjectId &)pdu.m_h323_uu_pdu.m_nonStandardData.m_nonStandardIdentifier) = objectIdentifier;
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        pdu.m_h323_uu_pdu.m_nonStandardData.m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_h221NonStandard);
        H225_H221NonStandard & h221 = pdu.m_h323_uu_pdu.m_nonStandardData.m_nonStandardIdentifier;
        endpoint.SetH221NonStandardInfo(h221);
    }

    // Adding nonStandard data
    pdu.m_h323_uu_pdu.m_nonStandardData.m_data.SetValue(data);

    return true;
}


bool H323Connection::HandleH225NonStandardData(const H323SignalPDU & pdu,
                                               SByteVector         & data)
{
    SDEFFUNC("H323Connection::HdlH225NonStdDta");

    // Get NonStandardData part of H323-UU-PDU
    if (!pdu.m_h323_uu_pdu.HasOptionalField(H225_H323_UU_PDU::e_nonStandardData))
        return false;

    const H225_NonStandardIdentifier & nonStdId
        = pdu.m_h323_uu_pdu.m_nonStandardData.m_nonStandardIdentifier;

    if (!objectIdentifier.empty())
    {
        // Connection is using object identifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_object)
        {
            STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Ignoring nonStandardData because not using object identifier "
                         << objectIdentifier);
            return false;
        }
        const SAsnObjectId & oid = nonStdId;
        if (oid != objectIdentifier.c_str())
        {
            STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Ignoring nonStandardData because received object identifier " << oid
                         << " != " << objectIdentifier);
            return false;
        }
    }
    else
    {
        // Connection is using H.221 NonStandardIdentifier
        if (nonStdId.GetTag() != H225_NonStandardIdentifier::e_h221NonStandard)
        {
            STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Ignoring nonStandardData because not using h221NonStandard identifier!");
            return false;
        }
        const H225_H221NonStandard & h221Msg = nonStdId;

        // Get endpoints H.221 NonStandardIdentifier
        H225_H221NonStandard h221Endpoint;
        endpoint.SetH221NonStandardInfo(h221Endpoint);

        // Compare H.221 NonStandardIdentifier
        if (h221Msg.m_t35CountryCode.GetValue() != h221Endpoint.m_t35CountryCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Ignoring nonStandardData because received t35CountryCode="
                         << h221Msg.m_t35CountryCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35CountryCode.GetValue());
            return false;
        }
        if (h221Msg.m_t35Extension.GetValue() != h221Endpoint.m_t35Extension.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Ignoring nonStandardData because received t35Extension="
                         << h221Msg.m_t35Extension.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_t35Extension.GetValue());
            return false;
        }
        if (h221Msg.m_manufacturerCode.GetValue() != h221Endpoint.m_manufacturerCode.GetValue())
        {
            STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                         "() Ignoring nonStandardData because received manufacturerCode="
                         << h221Msg.m_manufacturerCode.GetValue()
                         << " does not match endpoints: " << h221Endpoint.m_manufacturerCode.GetValue());
            return false;
        }
    }

    // Get nonStandard data
    data.clear();
    data = pdu.m_h323_uu_pdu.m_nonStandardData.m_data.GetValue();

    return true;
}


bool H323Connection::AttachH225NonStandardControl(H323SignalPDU                           & pdu,
                                                  const std::vector<NonStandardParameter> & rNonStdParamVector)
{
    pdu.m_h323_uu_pdu.IncludeOptionalField(H225_H323_UU_PDU::e_nonStandardControl);

    pdu.m_h323_uu_pdu.m_nonStandardControl.SetSize(rNonStdParamVector.size());

    for (unsigned i = 0; i < rNonStdParamVector.size(); i++)
    {
        if (rNonStdParamVector[i].m_bObjectIdUsed)
        {
            pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_object);
            ((SAsnObjectId &)pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_nonStandardIdentifier) = rNonStdParamVector[i].m_ObjectId;
            pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_data.SetValue(rNonStdParamVector[i].m_Data);
        }
        else
        {
            pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_nonStandardIdentifier.SetTag(H225_NonStandardIdentifier::e_h221NonStandard);
            ((H225_H221NonStandard &)pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_nonStandardIdentifier) = rNonStdParamVector[i].m_H221NonStandard;
            pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_data.SetValue(rNonStdParamVector[i].m_Data);
        }
    }
    return true;
}


bool H323Connection::HandleH225NonStandardControl(const H323SignalPDU               & pdu,
                                                  std::vector<NonStandardParameter> & rNonStdParamVector)
{
    SDEFFUNC("H323Connection::HdlH225NonStdCtrl");

    // Get NonStandardData part of H323-UU-PDU
    if (!pdu.m_h323_uu_pdu.HasOptionalField(H225_H323_UU_PDU::e_nonStandardControl))
        return false;

    for (int i = 0; i < pdu.m_h323_uu_pdu.m_nonStandardControl.GetSize(); i++)
    {
        const H225_NonStandardIdentifier & nonStdId
            = pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_nonStandardIdentifier;

        if (nonStdId.GetTag() == H225_NonStandardIdentifier::e_object)
        {
            const SAsnObjectId & objectId = nonStdId;
            rNonStdParamVector.push_back(
                NonStandardParameter(objectId, pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_data.GetValue()));
        }
        else if (nonStdId.GetTag() == H225_NonStandardIdentifier::e_h221NonStandard)
        {
            const H225_H221NonStandard & h221NonStd = nonStdId;
            rNonStdParamVector.push_back(
                NonStandardParameter(h221NonStd, pdu.m_h323_uu_pdu.m_nonStandardControl[i].m_data.GetValue()));
        }
        else
        {
            STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                "() Unknown H225_NonStandardIdentifier tag: " << nonStdId.GetTag());
            return false;
        }
    }
    return true;
}


bool H323Connection::OnUnknownSignalPDU(const H323SignalPDU & STRACE_pdu)
{
    SDEFFUNC("H323Connection::OnUnknownSigPDU");

    STRACE2_WARN(g_pH323Module->m_H225TraceId, this, GetConnID(),
                 "() Unknown signalling PDU: " << STRACE_pdu);
    return true;
}


bool H323Connection::WriteControlPDU(const H323ControlPDU & pdu)
{
    SDEFFUNC("H323Connection::WriteControlPDU");
    SPerStream strm;
    pdu.Encode(strm);
    strm.CompleteEncoding();

    if (h245TunnelPDU != NULL)
    {
        STRACE2_INFO2(g_pH323Module->m_H245TraceId, this, GetConnID(),
                     "() Building PDU for H.245 tunneling (h245Control):\n" << STrace::MultiLinePrefix
                     << "  " << setprecision(2) << setfill(' ') << pdu
                     << "\n" << STrace::MultiLinePrefix << setprecision(2) << hex
                     << setfill('0') << strm << dec << setfill(' '));
    }
    if (h245ParallelPDU != NULL)
    {
        STRACE2_INFO2(g_pH323Module->m_H245TraceId, this, GetConnID(),
                     "() Building PDU for H.245 tunneling (parallelH245Control):\n" << STrace::MultiLinePrefix
                     << "  " << setprecision(2) << setfill(' ') << pdu
                     << "\n" << STrace::MultiLinePrefix << setprecision(2) << hex
                     << setfill('0') << strm << dec << setfill(' '));
    }
    if (h245TunnelPDU == NULL && h245ParallelPDU == NULL)
    {
        STRACE2_INFO2(g_pH323Module->m_H245TraceId, this, GetConnID(),
                     "() Sending PDU:\n" << STrace::MultiLinePrefix
                     << "  " << setprecision(2) << setfill(' ') << pdu
                     << "\n" << STrace::MultiLinePrefix << setprecision(2) << hex
                     << setfill('0') << strm << dec << setfill(' '));
    }

    if (!h245Tunneling)
    {
        if (controlChannel == NULL)
        {
            STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                " Sending H.245 message failed (channel not initialised).");
            return false;
        }
        if (!controlChannel->IsOpen())
        {
            STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                " Sending H.245 message failed (channel not open).");
            return false;
        }
        if (!controlChannel->WritePDU(strm))
        {
            STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                " Sending H.245 message failed (write failed).");
            return false;
        }
        return true;
    }

  // If have a pending signalling PDU, use it rather than separate write
  H323SignalPDU localTunnelPDU;
  H323SignalPDU * tunnelPDU = NULL;
  if (h245TunnelPDU != NULL)
  {
      tunnelPDU = h245TunnelPDU;
  }
  else if (h245ParallelPDU == NULL)
  {
      localTunnelPDU.BuildFacility(*this, false);
      tunnelPDU = &localTunnelPDU;
  }

  if (tunnelPDU != NULL)
  {
      tunnelPDU->m_h323_uu_pdu.IncludeOptionalField(H225_H323_UU_PDU::e_h245Control);
      int last = tunnelPDU->m_h323_uu_pdu.m_h245Control.GetSize();
      tunnelPDU->m_h323_uu_pdu.m_h245Control.SetSize(last+1);
      tunnelPDU->m_h323_uu_pdu.m_h245Control[last] = strm;
  }
  if (h245ParallelPDU != NULL)
  {
      H225_Setup_UUIE & setup = h245ParallelPDU->m_h323_uu_pdu.m_h323_message_body;
      setup.IncludeOptionalField(H225_Setup_UUIE::e_parallelH245Control);
      int last = setup.m_parallelH245Control.GetSize();
      setup.m_parallelH245Control.SetSize(last+1);
      setup.m_parallelH245Control[last] = strm;
  }

  if (h245TunnelPDU != NULL || h245ParallelPDU != NULL)
    return true;

  return WriteSignalPDU(localTunnelPDU);
}


bool H323Connection::StartControlNegotiations()
{
    SDEFFUNC("H323Connection::StartControlNeg");
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Started control channel");

  // Begin the capability exchange procedure
  if (!capabilityExchangeProcedure->HasSentCapabilities() &&
      !capabilityExchangeProcedure->Start()) {
    STRACE2_ERROR(g_pH323Module->m_H245TraceId,  this, 0,
                  "() Start of Capability Exchange failed");
    return false;
  }

  // Begin the Master/Slave determination procedure
  if (!masterSlaveDeterminationProcedure->IsDetermined() &&
      !masterSlaveDeterminationProcedure->Start()) {
    STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                  "() Start of Master/Slave determination failed");
    return false;
  }

  return true;
}


void H323Connection::HandleControlChannel()
{
    SDEFFUNC("H323Connection::HandleCtrlChannel");

    if (!StartControlNegotiations())
    {
        if (controlChannel != NULL)
        {
            STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                "() Starting H.245 negotiation failed -> closing control channel!");
            controlChannel->Close();
            delete controlChannel;
            controlChannel = NULL;
        }
        else
        {
            STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                "() Starting H.245 negotiation failed (NO control channel)!");
        }
        return;
    }

#ifdef ROUND_TRIP_DELAY
  controlChannel->SetReadTimeout(endpoint.GetRoundTripDelayRate());
  
  for (;;) {
    SPerStream strm;
    if (controlChannel->ReadPDU(strm)) {
      STRACE2_INFO2(g_pH323Module->m_H245TraceId, this, GetConnID(),
                   "() Received TPKT: \n" << STrace::MultiLinePrefix << hex << setfill('0')
                   << setprecision(2) << strm << dec << setfill(' '));
      if (!HandleControlData(strm))
        break;
    }
    else if (connectionState != EstablishedConnection ||
             controlChannel->GetErrorCode() != SChannel::Timeout) {
      STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                    "() Read error: " << controlChannel->GetErrorText());
      ClearCall(AbnormalTermination);
      break;
    }

    StartRoundTripDelay();
  }
#else
  bool ok = true;
  while (ok) {
    SPerStream strm;
    if (controlChannel->ReadPDU(strm)) {
      STRACE2_INFO2(g_pH323Module->m_H245TraceId, this, GetConnID(),
                   "() Received TPKT: \n" << STrace::MultiLinePrefix << hex << setfill('0')
                   << setprecision(2) << strm << dec << setfill(' '));
      ok = HandleControlData(strm);
    }
    else {
      STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                    "() Read error: " << controlChannel->GetErrorText());
      ClearCall(AbnormalTermination);
      ok = false;
    }
  }
#endif // ROUND_TRIP_DELAY

  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
               "() Control channel closed.");
}


bool H323Connection::HandleControlData(SPerStream & strm)
{
    SDEFFUNC("H323Connection::HandleCtrlData");

    m_bControlDataHandled=true;


  while (!strm.IsAtEnd()) {
    H323ControlPDU pdu;
    strm.ResetDecoder();
    try
    {
        if (!pdu.Decode(strm))
        {
            STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                "() Invalid PDU decode.");
            return true;
        }
    }
    catch(...)
    {
        STRACE2_FATAL(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() ### Exception in ASN decode of H.245 message thrown! ###");
        return false;
    }

    if (!HandleControlPDU(pdu))
      return false;

    if (connectionState == HasExecutedSignalConnect &&
        masterSlaveDeterminationProcedure->IsDetermined() &&
        capabilityExchangeProcedure->HasSentCapabilities() &&
        capabilityExchangeProcedure->HasReceivedCapabilities()) {
      if (fastStartChannels.empty() && !OnControlChannelOpen())
        return false;
      connectionState = EstablishedConnection;
    }

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
         "() connectionState: " << connectionState 
         << " MSD: " << masterSlaveDeterminationProcedure->IsDetermined()
         << " TCS sent: " << capabilityExchangeProcedure->HasSentCapabilities()
         << " TCS recv: " << capabilityExchangeProcedure->HasReceivedCapabilities()
         << " fastStartChannels: " << fastStartChannels.size());

    if (connectionState == TransmitterSidePaused &&
        masterSlaveDeterminationProcedure->IsDetermined() &&
        capabilityExchangeProcedure->HasSentCapabilities() &&
        capabilityExchangeProcedure->HasReceivedCapabilities())
    {
        // H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing:
        // Receiving a fully populated TerminalCapabilitySet in
        // "transmitter side paused" state means leaving this state.
        // We change to 'EstablishedConnection' (Is this always o.k.?, UF)
        ConnectionStates oldConnectionState = connectionState;
        connectionState = EstablishedConnection;
        STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Changing connection state from '" << oldConnectionState
            << "' to '" << connectionState << "'.");

        // If we had done fastStart reset the state to make it possible
        // to call OpenLogicalChannel() method!
        fastStartState = FastStartDisabled;

        // Start our logical channels again.
        RestartSelectLogicalChannels();
    }

    strm.ByteAlign();
  }

  return true;
}


bool H323Connection::HandleControlPDU(const H323ControlPDU & pdu)
{
    SDEFFUNC("H323Connection::HandleControlPDU");

    STRACE2_INFO2(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Received PDU (connectionState=" << connectionState << "):\n"
                 << STrace::MultiLinePrefix << "  " << setprecision(2) << setfill(' ') << pdu);

  switch (pdu.GetTag()) {
    case H245_MultimediaSystemControlMessage::e_request :
      return OnH245Request(pdu);

    case H245_MultimediaSystemControlMessage::e_response :
      return OnH245Response(pdu);

    case H245_MultimediaSystemControlMessage::e_command :
      return OnH245Command(pdu);

    case H245_MultimediaSystemControlMessage::e_indication :
      return OnH245Indication(pdu);
  }

  return OnUnknownControlPDU(pdu);
}


bool H323Connection::OnUnknownControlPDU(const H323ControlPDU & pdu)
{
    SDEFFUNC("H323Connection::OnUnknownCtrlPDU");

    STRACE2_WARN(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Unknown Control PDU: " << pdu);

  H323ControlPDU reply;
  reply.BuildFunctionNotUnderstood(pdu);
  return WriteControlPDU(reply);
}


bool H323Connection::OnH245Request(const H323ControlPDU & pdu)
{
  SDEFFUNC("H323Connection::OnH245Request");

  const H245_RequestMessage & request = pdu;

  switch (request.GetTag()) {
    case H245_RequestMessage::e_masterSlaveDetermination :
      return masterSlaveDeterminationProcedure->HandleIncoming(request);

    case H245_RequestMessage::e_terminalCapabilitySet :
      return capabilityExchangeProcedure->HandleIncoming(request);

    case H245_RequestMessage::e_openLogicalChannel :
      return logicalChannels->HandleOpen(request);

    case H245_RequestMessage::e_closeLogicalChannel :
      return logicalChannels->HandleClose(request);

    case H245_RequestMessage::e_requestChannelClose :
      return logicalChannels->HandleRequestClose(request);

    case H245_RequestMessage::e_requestMode :
      return requestModeProcedure->HandleRequest(request);

    case H245_RequestMessage::e_roundTripDelayRequest :
      return roundTripDelayProcedure->HandleRequest(request);
  }

  return OnUnknownControlPDU(pdu);
}


bool H323Connection::OnH245Response(const H323ControlPDU & pdu)
{
  const H245_ResponseMessage & response = pdu;

  switch (response.GetTag()) {
    case H245_ResponseMessage::e_masterSlaveDeterminationAck :
      return masterSlaveDeterminationProcedure->HandleAck(response);

    case H245_ResponseMessage::e_masterSlaveDeterminationReject :
      return masterSlaveDeterminationProcedure->HandleReject(response);

    case H245_ResponseMessage::e_terminalCapabilitySetAck :
      return capabilityExchangeProcedure->HandleAck(response);

    case H245_ResponseMessage::e_terminalCapabilitySetReject :
      return capabilityExchangeProcedure->HandleReject(response);

    case H245_ResponseMessage::e_openLogicalChannelAck :
      return logicalChannels->HandleOpenAck(response);

    case H245_ResponseMessage::e_openLogicalChannelReject :
      return logicalChannels->HandleReject(response);

    case H245_ResponseMessage::e_closeLogicalChannelAck :
      return logicalChannels->HandleCloseAck(response);

    case H245_ResponseMessage::e_requestChannelCloseAck :
      return logicalChannels->HandleRequestCloseAck(response);

    case H245_ResponseMessage::e_requestChannelCloseReject :
      return logicalChannels->HandleRequestCloseReject(response);

    case H245_ResponseMessage::e_requestModeAck :
      return requestModeProcedure->HandleAck(response);

    case H245_ResponseMessage::e_requestModeReject :
      return requestModeProcedure->HandleReject(response);

    case H245_ResponseMessage::e_roundTripDelayResponse :
      return roundTripDelayProcedure->HandleResponse(response);
  }

  return OnUnknownControlPDU(pdu);
}


bool H323Connection::OnH245Command(const H323ControlPDU & pdu)
{
  const H245_CommandMessage & command = pdu;

  switch (command.GetTag()) {
    case H245_CommandMessage::e_sendTerminalCapabilitySet :
      return OnH245_SendTerminalCapabilitySet(command);

    case H245_CommandMessage::e_flowControlCommand :
      return OnH245_FlowControlCommand(command);

    case H245_CommandMessage::e_miscellaneousCommand :
      return OnH245_MiscellaneousCommand(command);

    case H245_CommandMessage::e_endSessionCommand :
      OnH245_EndSessionCommand(command);
      if (   connectionState == EstablishedConnection
          || connectionState == TransmitterSidePaused)
        ClearCall(RemoteTermination);
      else if (answerResponse == AnswerCallPending)
        ClearCall(CallAbortTermination);
      else
        ClearCall(RefusalTermination);
      return false;
  }

  return OnUnknownControlPDU(pdu);
}


bool H323Connection::OnH245Indication(const H323ControlPDU & pdu)
{
  const H245_IndicationMessage & indication = pdu;

  switch (indication.GetTag()) {
    case H245_IndicationMessage::e_masterSlaveDeterminationRelease :
      return masterSlaveDeterminationProcedure->HandleRelease(indication);

    case H245_IndicationMessage::e_terminalCapabilitySetRelease :
      return capabilityExchangeProcedure->HandleRelease(indication);

    case H245_IndicationMessage::e_openLogicalChannelConfirm :
      return logicalChannels->HandleOpenConfirm(indication);

    case H245_IndicationMessage::e_requestChannelCloseRelease :
      return logicalChannels->HandleRequestCloseRelease(indication);

    case H245_IndicationMessage::e_requestModeRelease :
      return requestModeProcedure->HandleRelease(indication);

    case H245_IndicationMessage::e_miscellaneousIndication :
      return OnH245_MiscellaneousIndication(indication);

    case H245_IndicationMessage::e_jitterIndication :
      return OnH245_JitterIndication(indication);

    case H245_IndicationMessage::e_userInput :
      OnUserInputIndication(indication);
      break;
  }

  return true; // Do NOT call OnUnknownControlPDU for indications
}


bool H323Connection::OnH245_SendTerminalCapabilitySet(
                 const H245_SendTerminalCapabilitySet & pdu)
{
    SDEFFUNC("H323Connection::OnH245_SendTCS");

  if (pdu.GetTag() == H245_SendTerminalCapabilitySet::e_genericRequest)
    return capabilityExchangeProcedure->Start();

  STRACE2_WARN(g_pH323Module->m_H245TraceId, this, GetConnID(),
               "() Unhandled SendTerminalCapabilitySet: " << pdu);
  return true;
}


bool H323Connection::OnH245_FlowControlCommand(
                 const H245_FlowControlCommand & pdu)
{
    SDEFFUNC("H323Connection::OnH245_FlCtrlCmd");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() FlowControlCommand: scope=" << pdu.m_scope.GetTagName());

  long restriction;
  if (pdu.m_restriction.GetTag() == H245_FlowControlCommand_restriction::e_maximumBitRate)
    restriction = (const SAsnInteger &)pdu.m_restriction;
  else
    restriction = -1; // H245_FlowControlCommand_restriction::e_noRestriction

  switch (pdu.m_scope.GetTag()) {
    case H245_FlowControlCommand_scope::e_wholeMultiplex :
      OnLogicalChannelFlowControl(NULL, restriction);
      break;

    case H245_FlowControlCommand_scope::e_logicalChannelNumber :
    {
      H323Channel * chan = logicalChannels->FindChannel((unsigned)(const H245_LogicalChannelNumber &)pdu.m_scope, false);
      if (chan != NULL)
        OnLogicalChannelFlowControl(chan, restriction);
    }
  }

  return true;
}


bool H323Connection::OnH245_MiscellaneousCommand(
                 const H245_MiscellaneousCommand & pdu)
{
    SDEFFUNC("H323Connection::OnH245_MiscCmd");

  H323Channel * chan = logicalChannels->FindChannel((unsigned)pdu.m_logicalChannelNumber, false);
  if (chan != NULL)
    chan->OnMiscellaneousCommand(pdu.m_type);
  else
  {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() MiscellaneousCommand: chan=" << pdu.m_logicalChannelNumber
                 << ", type=" << pdu.m_type.GetTagName());
  }
  return true;
}


bool H323Connection::OnH245_EndSessionCommand(const H245_EndSessionCommand & pdu)
{
    // This is a SwTCL extension
    return true;
}


bool H323Connection::OnH245_MiscellaneousIndication(
                 const H245_MiscellaneousIndication & pdu)
{
    SDEFFUNC("H323Connection::OnH245_MiscInd");

  H323Channel * chan = logicalChannels->FindChannel((unsigned)pdu.m_logicalChannelNumber, true);
  if (chan != NULL)
    chan->OnMiscellaneousIndication(pdu.m_type);
  else
  {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() MiscellaneousIndication: chan=" << pdu.m_logicalChannelNumber
                 << ", type=" << pdu.m_type.GetTagName());
  }
  return true;
}


bool H323Connection::OnH245_JitterIndication(
                 const H245_JitterIndication & pdu)
{
    SDEFFUNC("H323Connection::OnH245_JitterInd");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() JitterIndication: scope=" << pdu.m_scope.GetTagName());

  static const DWORD mantissas[8] = { 0, 1, 10, 100, 1000, 10000, 100000, 1000000 };
  static const DWORD exponents[8] = { 10, 25, 50, 75 };
  DWORD jitter = mantissas[pdu.m_estimatedReceivedJitterMantissa]*
                 exponents[pdu.m_estimatedReceivedJitterExponent]/10;

  int skippedFrameCount = -1;
  if (pdu.HasOptionalField(H245_JitterIndication::e_skippedFrameCount))
    skippedFrameCount = pdu.m_skippedFrameCount;

  int additionalBuffer = -1;
  if (pdu.HasOptionalField(H245_JitterIndication::e_additionalDecoderBuffer))
    additionalBuffer = pdu.m_additionalDecoderBuffer;

  switch (pdu.m_scope.GetTag()) {
    case H245_JitterIndication_scope::e_wholeMultiplex :
      OnLogicalChannelJitter(NULL, jitter, skippedFrameCount, additionalBuffer);
      break;

    case H245_JitterIndication_scope::e_logicalChannelNumber :
    {
      H323Channel * chan = logicalChannels->FindChannel((unsigned)(const H245_LogicalChannelNumber &)pdu.m_scope, false);
      if (chan != NULL)
        OnLogicalChannelJitter(chan, jitter, skippedFrameCount, additionalBuffer);
    }
  }

  return true;
}


H323Channel * H323Connection::FindChannel(unsigned rtpSessionId, bool fromRemote) const
{
    return logicalChannels->FindChannelBySession(rtpSessionId, fromRemote);
}


void H323Connection::TransferCall(const Tools::SString & remoteParty,
                                  const Tools::SString & callIdentity)
{
  // According to H.450.4, if prior to consultation the primary call has been put on hold, the 
  // transferring endpoint shall first retrieve the call before Call Transfer is invoked.
//  if (!callIdentity.empty() && IsLocalHold())
//    RetrieveCall();
  h4502handler->TransferCall(remoteParty, callIdentity);
}

void H323Connection::ConsultationTransfer(const Tools::SString & primaryCallToken)
{
  h4502handler->ConsultationTransfer(primaryCallToken);
}

void H323Connection::HandleConsultationTransfer(const Tools::SString & callIdentity,
                                                H323Connection& incoming)
{
  h4502handler->HandleConsultationTransfer(callIdentity, incoming);
}

bool H323Connection::IsTransferringCall() const
{
  switch (h4502handler->GetState()) {
    case H4502Handler::e_ctAwaitIdentifyResponse :
    case H4502Handler::e_ctAwaitInitiateResponse :
    case H4502Handler::e_ctAwaitSetupResponse :
      return TRUE;

    default :
      return FALSE;
  }
}

bool H323Connection::IsTransferredCall() const
{
   return (h4502handler->GetInvokeId() != 0 &&
           h4502handler->GetState() == H4502Handler::e_ctIdle) ||
           h4502handler->isConsultationTransferSuccess();
}

void H323Connection::HandleTransferCall(const Tools::SString & callTtoken,
                                        const Tools::SString & callIdentity)
{
  if (!callToken.empty() && !callIdentity.empty())
    h4502handler->AwaitSetupResponse(callToken, callIdentity);
}

int H323Connection::GetCallTransferInvokeId()
{
  return h4502handler->GetInvokeId();
}

void H323Connection::HandleCallTransferFailure(const int returnError)
{
  h4502handler->HandleCallTransferFailure(returnError);
}

void H323Connection::SetAssociatedCallToken(const Tools::SString& token)
{
  h4502handler->SetAssociatedCallToken(token);
}

void H323Connection::OnConsultationTransferSuccess(H323Connection& /*secondaryCall*/)
{
   h4502handler->SetConsultationTransferSuccess();
}


bool H323Connection::DeleteUnusedFastStartChannels()
{
    SDEFFUNC("H323Conn::DeleteUnusedFastStartCh");

    if (STrace::CanTrace(g_pH323Module->m_H323TraceId, STrace::Info3))
    {
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "() fast start channels: " << fastStartChannels.size()
            << ", logical channels: " << logicalChannels->GetSize());
        for (H323LogicalChannelList::size_type i = 0; i < fastStartChannels.size(); i++)
        {
            H323Channel *pLogCh = fastStartChannels[i];
            if (!SBAD_READPTR(pLogCh))
            {
                STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                    "() fast start channel " << i << ": " << pLogCh
                    << " No: " << pLogCh->GetNumber()
                    << ", Codec: " << pLogCh->GetCodecType()
                    << ", Framesize: " << pLogCh->GetFramesize()
                    << ", Encrypt: " << pLogCh->GetCapability().GetUseEncryption()
                    << ", Media[" << pLogCh->GetRemoteMediaTsap()
                    << "], MediaCtrl[" << pLogCh->GetRemoteMediaControlTsap() << "]");
            }
        }
        for (int j = 0; j < logicalChannels->GetSize(); j++)
        {
            H323Channel *pLogCh = logicalChannels->GetChannelAt(j);
            if (!SBAD_READPTR(pLogCh))
            {
                STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                    "() logical channel " << j << ": " << pLogCh
                    << " No: " << pLogCh->GetNumber()
                    << ", Codec: " << pLogCh->GetCodecType()
                    << ", Framesize: " << pLogCh->GetFramesize()
                    << ", Encrypt: " << pLogCh->GetCapability().GetUseEncryption()
                    << ", Media[" << pLogCh->GetRemoteMediaTsap()
                    << "], MediaCtrl[" << pLogCh->GetRemoteMediaControlTsap() << "]");
            }
        }
    }

    // Check for remaining fast start channel entries
    // that are not in logical channels list
    H323LogicalChannelList::iterator pos = fastStartChannels.begin();
    while (!fastStartChannels.empty())
    {
        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "() fast start channel "  << (*pos));

        // Look for a matching pointer in the logical channels hash map
        if (!logicalChannels->IsChannelInMap((*pos)))
        {
            // Channel not found in logicalChannels, delete it here
            STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                "() Deleting fast start channel " << (*pos))
            delete (*pos);
        }
        pos = fastStartChannels.erase(pos);
    }
    return true;
}


bool H323Connection::OnControlProtocolError(ControlProtocolErrors /*errorSource*/,
                                            const void * /*errorData*/)
{
  return true;
}


void H323Connection::OnSendCapabilitySet(H245_TerminalCapabilitySet & /*pdu*/)
{
}


bool H323Connection::OnReceivedCapabilitySet(const H323Capabilities & remoteCaps,
                                             const H245_MultiplexCapability * muxCap,
                                             H245_TerminalCapabilitySetReject & /*rejectPDU*/)
{
    SDEFFUNC("H323Connection::OnReceivedCapSet");

  if (muxCap != NULL) {
    if (muxCap->GetTag() != H245_MultiplexCapability::e_h2250Capability) {
      STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, GetConnID(),
                    "() CapabilitySet contains unsupported multiplex.");
      return false;
    }

    const H245_H2250Capability & h225_0 = *muxCap;
    remoteMaxAudioDelayJitter = h225_0.m_maximumAudioDelayJitter;
  }

/*
    // We have to check for audio capability matches because it is possible that
    // e.g. UserInput caps are o.k. but not the audio caps!
    bool bAudioChannelMatch = false;
    for(H323CapabilityList::iterator capListIter =  remoteCapabilityTable.begin();
                                     capListIter != remoteCapabilityTable.end();
                                   ++capListIter)
    {
        if (   (*capListIter)
            && ((*capListIter)->GetDefaultSessionID() == RTP_Session::DefaultAudioSessionID))
        {
            // One audio cap is enough for this check
            bAudioChannelMatch = true;
            break;
        }
    }

    if (bAudioChannelMatch == true)
    {
        STRACE_INFO(g_pH323Module->m_H245TraceId, "Received capability set, is accepted.");
        return true;
    }
    else
    {
        if (remoteTable.size() > 0)
        {
            STRACE_INFO(g_pH323Module->m_H245TraceId,
                        "Remote capability table is not empty (size=" << remoteTable.size()
                        << ") but no matching audio capability found!")
        }
        STRACE_INFO(g_pH323Module->m_H245TraceId, "Received capability set, is rejected.");
        return false;
    }
*/
    if (remoteCapabilities.Merge(remoteCaps) == false)
        return false;

    if (!capabilityExchangeProcedure->HasSentCapabilities())
        capabilityExchangeProcedure->Start();
 
    return true;
}


bool H323Connection::SendEmptyCapabilitySet()
{
    SDEFFUNC("H323Connection::SendEmptyCapSet");

    // Make peer side closing its logical channels:
    // Sending empty TerminalCapabilitySet
    // (H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing)
    if (SBAD_READPTR(capabilityExchangeProcedure))
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Invalid pointer to capabilityExchangeProcedure!");
        return false;
    }

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Sending H.245 empty TerminalCapabilitySet to make peer side closing its logical channels.");

    if (capabilityExchangeProcedure->SendEmptyTerminalCapabilitySet() == false)
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Could not send empty TerminalCapabilitySet!");
        return false;
    }

    return true;
}


bool H323Connection::SendFullCapabilitySet()
{
    SDEFFUNC("H323Connection::SendFullCapSet");

    // Make peer side opening its logical channel:
    // Sending new full TerminalCapabilitySet
    // (H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing)
    if (SBAD_READPTR(capabilityExchangeProcedure))
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Invalid pointer to capabilityExchangeProcedure!");
        return false;
    }

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Sending H.245 new full TerminalCapabilitySet to make peer side open its logical channel.");

    capabilityExchangeProcedure->Reset();

    if (capabilityExchangeProcedure->Start() == false)
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Could not start capability exchange procedure!");
        return false;
    }

    // Sending MasterSlaveDetermination
    // (H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing)
    if (SBAD_READPTR(masterSlaveDeterminationProcedure))
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Invalid pointer to masterSlaveDeterminationProcedure!");
        return false;
    }

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Starting MasterSlaveDetermination on activate.");

    if (masterSlaveDeterminationProcedure->Start() == false)
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Could not start master slave determination procedure!");
        return false;
    }

    return true;
}


bool H323Connection::OnReceivedEmptyCapabilitySet()
{
    SDEFFUNC("H323Connection::OnRcvEmptyCapSet");

    // Received empty TerminalCapabilitySet
    // (H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing)

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Received empty TerminalCapabilitySet.");

    H323Channel * pH323Channel;
    H245NegLogicalChannel * pH245Channel;
    STRACE(STrace::Debug, g_pH323Module->m_H245TraceId,
           "logicalChannels size = " << logicalChannels->GetSize());
    for (int i = 0; i < logicalChannels->GetSize(); i++)
    {
        pH323Channel = logicalChannels->GetChannelAt(i);
        if (!SBAD_READPTR(pH323Channel))
        {
            STRACE(STrace::Debug, g_pH323Module->m_H245TraceId,
                   "Channel " << i << ": "
                   << "direction: "    << pH323Channel->GetDirection()
                   << ", capability: " << pH323Channel->GetCapability()
                   << ", number: "     << pH323Channel->GetNumber()
                   << ", session id: " << pH323Channel->GetSessionID());
        }
        else
        {
            STRACE2_WARN(g_pH323Module->m_H245TraceId, this, GetConnID(),
                "() Channel " << i << " is invalid!");
        }

        if (   !SBAD_READPTR(pH323Channel)
            && pH323Channel->GetDirection() == H323Channel::IsTransmitter)
        {
            // Close logical channel opened by this endpoint
            // Stop transmitting on this channel is done if CLCAck is received!

            // Find associated H.245 channel
            pH245Channel = logicalChannels->FindNegLogicalChannel(pH323Channel->GetNumber(), false);
            if (!SBAD_READPTR(pH245Channel))
            {
                // Close this logical channel (sending CLC)
                if (!pH245Channel->Close())
                {
                    STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                                  "() Closing logical channel " << pH323Channel->GetNumber() << " failed!");
                    continue;
                }
            }
            else
            {
                STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
                              "() Could not find logical channel " << pH323Channel->GetNumber());
                continue;
            }
        }
    }

    ConnectionStates oldConnectionState = connectionState;
    connectionState = TransmitterSidePaused;
    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
                 "() Changing connection state from '" << oldConnectionState
                 << "' to '" << connectionState << "'.");
    return true;
}


void H323Connection::ResetCapabilities()
{
    // Reset the capability table of this connection.
    // It is necessary when the remote endpoint has send an empty capability
    // set to initiate pause and re-routing handling.

    remoteCapabilities.RemoveAll();
}


bool H323Connection::IsH245Master() const
{
  return masterSlaveDeterminationProcedure->IsMaster();
}


#ifdef ROUND_TRIP_DELAY
void H323Connection::StartRoundTripDelay()
{
    SDEFFUNC("H323Connection::StartRoundTripDly");

    if (endpoint.IsSendRoundTripDelay() == false)
    {
        // Sending RoundTripDelayRequests is disabled.
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
                     "() Sending RoundTripDelayRequests is disabled.");
        return;
    }

  if (connectionState != EstablishedConnection)
    return;

  if (roundTripDelayTimer.IsRunning())
    return;

  roundTripDelayTimer = endpoint.GetRoundTripDelayRate();

  if (roundTripDelayProcedure->IsRemoteOffline()) {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Remote failed to respond to PDU.");
    if (endpoint.ShouldClearCallOnRoundTripFail())
      ClearCall(AbnormalTermination);
  }
  else
    roundTripDelayProcedure->StartRequest();
}


STimeInterval H323Connection::GetRoundTripDelay() const
{
  return roundTripDelayProcedure->GetRoundTripDelay();
}
#endif // ROUND_TRIP_DELAY


bool H323Connection::OnControlChannelOpen()
{
    SDEFFUNC("H323Connection::OnCtrlChannelOpen");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Default OnControlChannelOpen()");

  connectionStartTime = STime();
  endpoint.OnConnectionEstablished(*this, callToken);
  OnSelectLogicalChannels();
  return true;
}


static void StartFastStartChannel(H323LogicalChannelList & fastStartChannels,
                                  unsigned sessionID, H323Channel::Directions direction)
{
  for (H323LogicalChannelList::size_type i = 0; i < fastStartChannels.size(); i++) {
    H323Channel & channel = *fastStartChannels[i];
    if (channel.GetSessionID() == sessionID && channel.GetDirection() == direction) {
      if (!g_pH323Module->m_bUsingSwtcl)
      {
        // Only start a LogicalChannel thread if we not use SwTCL.
        fastStartChannels[i]->Start();
      }
      break;
    }
  }
}


void H323Connection::OnSelectLogicalChannels(bool bStartDataSession)
{
    SDEFFUNC("H323Connection::OnSelectLogChan");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Default OnSelectLogicalChannels, " << fastStartState);

  // Select the first codec that uses the "standard" audio session.
  switch (fastStartState) {
    default : //FastStartDisabled :
      SelectDefaultLogicalChannel(RTP_Session::DefaultAudioSessionID);
      if (bStartDataSession == true)
          SelectDefaultLogicalChannel(RTP_Session::DefaultDataSessionID);
      break;

    case FastStartInitiate :
      SelectFastStartChannels(RTP_Session::DefaultAudioSessionID);
      if (bStartDataSession == true)
          SelectFastStartChannels(RTP_Session::DefaultDataSessionID);
      break;

    case FastStartResponse :
      StartFastStartChannel(fastStartChannels, RTP_Session::DefaultAudioSessionID, H323Channel::IsTransmitter);
      StartFastStartChannel(fastStartChannels, RTP_Session::DefaultAudioSessionID, H323Channel::IsReceiver);
      if (bStartDataSession == true)
      {
          StartFastStartChannel(fastStartChannels, RTP_Session::DefaultDataSessionID, H323Channel::IsTransmitter);
          StartFastStartChannel(fastStartChannels, RTP_Session::DefaultDataSessionID, H323Channel::IsReceiver);
      }
      break;
  }
}


void H323Connection::SelectDefaultLogicalChannel(unsigned sessionID)
{
    SDEFFUNC("H323Connection::SelectDefLogChan");

  for (int i = 0; i < localCapabilities.GetSize(); i++) {
    H323Capability & localCapability = localCapabilities[i];
    if (localCapability.GetDefaultSessionID() == sessionID) {
      H323Capability * remoteCapability = remoteCapabilities.FindCapability(localCapability);
      if (remoteCapability != NULL) {
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(), 
                     "() Selecting " << *remoteCapability);
        if (OpenLogicalChannel(*remoteCapability, sessionID, H323Channel::IsTransmitter))
          break;
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
                     "() OnSelectLogicalChannels, OpenLogicalChannel failed: "
                     << *remoteCapability);
        break;
      }
    }
  }
}


void H323Connection::SelectFastStartChannels(unsigned sessionID)
{
    SDEFFUNC("H323Connection::SlctFastStartChn");

  // Select all of the fast start channels to offer to the remote when initiating a call.
  for (int i = 0; i < localCapabilities.GetSize(); i++) {
    H323Capability & capability = localCapabilities[i];
    if (capability.GetDefaultSessionID() == sessionID) {
      if (!capability.IsBidirectional()) {
        if (!OpenLogicalChannel(capability, sessionID, H323Channel::IsReceiver)) {
          STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
              "() OnSelectLogicalChannels, OpenLogicalChannel rx failed: " << capability);
        }
      }
      if (!OpenLogicalChannel(capability, sessionID, H323Channel::IsTransmitter)) {
        STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
                     "() OnSelectLogicalChannels, OpenLogicalChannel tx failed: " << capability);
      }
    }
  }
}


void H323Connection::RestartSelectLogicalChannels()
{
    SDEFFUNC("H323Connection::RstrtSlctLogChan");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                 "() Restart and select of logical channels...");

    SelectDefaultLogicalChannel(RTP_Session::DefaultAudioSessionID);
}


bool H323Connection::OpenLogicalChannel(const H323Capability & capability,
                                        unsigned sessionID,
                                        H323Channel::Directions dir)
{
    SDEFFUNC("H323Conn::OpenLogicalChannel");
              
  H323ChannelNumber dummyNumber;

  switch (fastStartState) {
    default : // FastStartDisabled
      if (dir == H323Channel::IsReceiver)
        return false;

      // Traditional H245 handshake
/*
      if (   capability.GetMainType() == H323Capability::e_Data
          && capability.GetSubType() == H245_DataApplicationCapability_application::e_t38fax)
      {
          if (!SBAD_READPTR(requestModeProcedure))
          {
              STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
                  "() Sending RequestMode for T.38 ...");
              if (requestModeProcedure->StartRequest(capability) == false)
              {
                  STRACE2_WARN(g_pH323Module->m_H245TraceId, this, GetConnID(),
                      "() Could not send RequestMode for T.38!");
              }
          }
      }
*/
      return logicalChannels->Open(capability, sessionID, dummyNumber);

    case FastStartResponse :
      // Do not use OpenLogicalChannel for starting these.
      return false;

    case FastStartInitiate :
      break;
  }

  /*If starting a receiver channel and are initiating the fast start call,
    indicated by the remoteCapabilities being empty, we do a "trial"
    listen on the channel. That is, for example, the UDP sockets are created
    to receive data in the RTP session, but no thread is started to read the
    packets and pass them to the codec. This is because at this point in time,
    we do not know which of the codecs is to be used, and more than one thread
    cannot read from the RTP ports at the same time.
  */
  H323Channel * channel = capability.CreateChannel(*this, dir, sessionID, NULL);
  if (channel == NULL)
    return false;

  if (!channel->SetInitialBandwidth()) {
    delete channel;
    return false;
  }

  if (dir != H323Channel::IsReceiver)
    channel->SetNumber(logicalChannels->GetNextChannelNumber());

  fastStartChannels.push_back(channel);
  return true;
}


bool H323Connection::OnOpenLogicalChannel(const H245_OpenLogicalChannel & /*openPDU*/,
                                         H245_OpenLogicalChannelAck & /*ackPDU*/,
                                         unsigned & /*errorCode*/)
{
  //errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
  return true;
}


void H323Connection::OnOLCAfterCreation(H323Channel             * pChannel,
                                        const H323ChannelNumber & channelNumber,
                                        unsigned                  uErrorCode)
{
    return;
}


H323Channel * H323Connection::CreateLogicalChannel(const H245_OpenLogicalChannel & open,
                                                   unsigned & errorCode)
{
    SDEFFUNC("H323Connection::CreateLogicChan");

  const H245_H2250LogicalChannelParameters * param;
  const H245_DataType * dataType;
  H323Channel::Directions direction;

  if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
    if (open.m_reverseLogicalChannelParameters.m_multiplexParameters.GetTag() !=
              H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters
                                                      ::e_h2250LogicalChannelParameters) {
      errorCode = H245_OpenLogicalChannelReject_cause::e_unsuitableReverseParameters;
      STRACE2_WARN(g_pH323Module->m_H323TraceId, this, GetConnID(),
          "() CreateLogicalChannel - reverse channel, H225.0 only supported");
      return NULL;
    }

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "() CreateLogicalChannel - reverse channel");
    dataType = &open.m_reverseLogicalChannelParameters.m_dataType;
    param = &(const H245_H2250LogicalChannelParameters &)
                      open.m_reverseLogicalChannelParameters.m_multiplexParameters;
    direction = H323Channel::IsTransmitter;
  }
  else {
    if (open.m_forwardLogicalChannelParameters.m_multiplexParameters.GetTag() !=
              H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters
                                                      ::e_h2250LogicalChannelParameters) {
      STRACE2_WARN(g_pH323Module->m_H323TraceId, this, GetConnID(),
          "() CreateLogicalChannel - forward channel, H225.0 only supported");
      errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
      return NULL;
    }

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "() CreateLogicalChannel - forward channel");
    dataType = &open.m_forwardLogicalChannelParameters.m_dataType;
    param = &(const H245_H2250LogicalChannelParameters &)
                      open.m_forwardLogicalChannelParameters.m_multiplexParameters;
    direction = H323Channel::IsReceiver;
  }

  // See if datatype is supported
  H323Capability * capability = localCapabilities.FindCapability(*dataType);
  if (capability == NULL) {
    errorCode = H245_OpenLogicalChannelReject_cause::e_unknownDataType;
    return NULL; // If codec not supported, return error
  }

  if (!capability->OnReceivedPDU(*dataType)) {
    errorCode = H245_OpenLogicalChannelReject_cause::e_dataTypeNotSupported;
    return NULL; // If codec not supported, return error
  }

    // Check for bidirectional channel capability (e.g. T.38) because it has forward and reverse part
    // in fast start channels and direction is not set correct then
    if (   open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)
        && HadAnsweredCall()
        && capability->IsBidirectional())
    {
        // Set correct direction
        direction = H323Channel::IsReceiver;
    }

  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
      "() call capability->CreateChannel");

  H323Channel * channel = capability->CreateChannel(*this, direction, param->m_sessionID.GetValue(), param);
  if (channel == NULL) {
    errorCode = H245_OpenLogicalChannelReject_cause::e_dataTypeNotAvailable;
    return NULL;
  }

  if (!channel->SetInitialBandwidth())
    errorCode = H245_OpenLogicalChannelReject_cause::e_insufficientBandwidth;
  else if (channel->OnReceivedPDU(open, errorCode))
    return channel;

  delete channel;
  return NULL;
}


bool H323Connection::OnReceivedLogicalChannelAckPDU(const H245_OpenLogicalChannelAck & /*ack*/)
{
    // This is a SwTCL extension
    return true;
}


bool H323Connection::OnStartLogicalChannel(H323Channel & channel)
{
  return endpoint.OnStartLogicalChannel(*this, channel);
}


bool H323Connection::OpenAudioChannel(bool isEncoding, unsigned bufferSize, H323AudioCodec & codec)
{
  return endpoint.OpenAudioChannel(*this, isEncoding, bufferSize, codec);
}


bool H323Connection::OpenVideoDevice(bool isEncoding, H323VideoCodec & codec)
{
  return endpoint.OpenVideoDevice(*this, isEncoding, codec);
}


void H323Connection::CloseLogicalChannel(unsigned number, bool fromRemote)
{
  logicalChannels->Close(number, fromRemote);
}


void H323Connection::CloseAllLogicalChannels(bool fromRemote)
{
    SDEFFUNC("H323Conn::CloseAllLogicalChannels");

    // Close transmitting (Tx) logical channels if fromRemote is false.
    // Close receiving (Rx) logical channels if fromRemote is true.
    H323Channel * pH323Channel;
    H245NegLogicalChannel * pH245Channel;
    for (int i = 0; i < logicalChannels->GetSize(); i++)
    {
        pH323Channel = logicalChannels->GetChannelAt(i);
        if (pH323Channel != NULL)
        {
            STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                "(" << fromRemote << ") Channel " << i << ": "
                << "number: "     << pH323Channel->GetNumber()
                << ", session id: " << pH323Channel->GetSessionID()
                << ", IsFromRemote: " << pH323Channel->GetNumber().IsFromRemote());
        }
        else
        {
            STRACE2_WARN(g_pH323Module->m_H323TraceId, this, GetConnID(),
                "() Channel " << i << " is zero!");
        }

        if (   pH323Channel != NULL
            && pH323Channel->GetNumber().IsFromRemote() == fromRemote)
        {
            // Find associated H.245 channel
            pH245Channel = logicalChannels->FindNegLogicalChannel(pH323Channel->GetNumber(), fromRemote);
            if (pH245Channel != NULL)
            {
                // Close this logical channel (sending CLC)
                if (!pH245Channel->Close())
                {
                    STRACE2_WARN(g_pH323Module->m_H323TraceId, this, GetConnID(),
                        "() Closing logical channel " << pH323Channel->GetNumber() << " failed!");
                    continue;
                }
            }
            else
            {
                STRACE2_WARN(g_pH323Module->m_H323TraceId, this, GetConnID(),
                    "() Could not find logical channel " << pH323Channel->GetNumber());
                continue;
            }
        }
    }
}


bool H323Connection::OnClosingLogicalChannel(H323Channel & /*channel*/)
{
  return true;
}


void H323Connection::OnClosedLogicalChannel(const H323Channel & channel)
{
  endpoint.OnClosedLogicalChannel(*this, channel);
}


void H323Connection::OnLogicalChannelFlowControl(H323Channel * channel,
                                                 long bitRateRestriction)
{
  if (channel != NULL)
    channel->OnFlowControl(bitRateRestriction);
}


void H323Connection::OnLogicalChannelJitter(H323Channel * channel,
                                            DWORD jitter,
                                            int skippedFrameCount,
                                            int additionalBuffer)
{
  if (channel != NULL)
    channel->OnJitterIndication(jitter, skippedFrameCount, additionalBuffer);
}


unsigned H323Connection::GetBandwidthUsed() const
{
  unsigned used = 0;

  for (int i = 0; i < logicalChannels->GetSize(); i++) {
    H323Channel * channel = logicalChannels->GetChannelAt(i);
    if (!SBAD_READPTR(channel))
      used += channel->GetBandwidthUsed();
  }

  STRACE(STrace::Debug2, g_pH323Module->m_H323TraceId, "Bandwidth used: " << used);

  return used;
}


bool H323Connection::UseBandwidth(unsigned bandwidth, bool removing)
{
    SDEFFUNC("H323Connection::UseBandwidth");

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
                  "() Bandwidth request: " << (removing ? '+' : '-') << bandwidth);

  if (removing)
    bandwidthAvailable += bandwidth;
  else {
    if (bandwidth > bandwidthAvailable) {
      STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, GetConnID(),
                    "() Available bandwidth exceeded");
      return false;
    }

    bandwidthAvailable -= bandwidth;
  }

  return true;
}


bool H323Connection::SetBandwidthAvailable(unsigned newBandwidth, bool force)
{
  unsigned used = GetBandwidthUsed();
  if (used > newBandwidth) {
    if (!force)
      return false;

    // Go through logical channels and close down some.
    int chanIdx = logicalChannels->GetSize();
    while (used > newBandwidth && chanIdx-- > 0) {
      H323Channel * channel = logicalChannels->GetChannelAt(chanIdx);
      if (!SBAD_READPTR(channel)) {
        used -= channel->GetBandwidthUsed();
        const H323ChannelNumber & number = channel->GetNumber();
        CloseLogicalChannel(number, number.IsFromRemote());
      }
    }
  }

  bandwidthAvailable = newBandwidth - used;
  return true;
}


void H323Connection::SendUserInputIndication(const H245_UserInputIndication & indication)
{
  H323ControlPDU pdu;
  H245_UserInputIndication & ind = pdu.Build(H245_IndicationMessage::e_userInput);
  ind = indication;
  WriteControlPDU(pdu);
}


void H323Connection::SendUserInput(const SString & value)
{
  H323ControlPDU pdu;
  pdu.BuildUserInputIndication(value);
  WriteControlPDU(pdu);
}


void H323Connection::SendUserInputTone(const SString & tone,
                                       unsigned duration,
                                       unsigned logicalChannel,
                                       unsigned rtpTimestamp)
{
  H323ControlPDU pdu;
  pdu.BuildUserInputIndication(tone, duration, logicalChannel, rtpTimestamp);
  WriteControlPDU(pdu);
}


void H323Connection::OnUserInputIndication(const H245_UserInputIndication & ind)
{
  switch (ind.GetTag()) {
    case H245_UserInputIndication::e_alphanumeric :
      OnUserInputString((const SAsnGeneralString &)ind);
      break;
    case H245_UserInputIndication::e_signal :
      const H245_UserInputIndication_signal & sig = ind;
      OnUserInputTone(sig.m_signalType,
                      sig.m_duration,
                      sig.m_rtp.m_logicalChannelNumber,
                      sig.m_rtp.m_timestamp);
  }
}


void H323Connection::OnUserInputString(const SString & value)
{
  endpoint.OnUserInputString(*this, value);
}


void H323Connection::OnUserInputTone(const SString & tone,
                                     unsigned duration,
                                     unsigned logicalChannel,
                                     unsigned rtpTimestamp)
{
  endpoint.OnUserInputTone(*this, tone, duration, logicalChannel, rtpTimestamp);
}


RTP_Session * H323Connection::GetSession(unsigned sessionID) const
{
    SDEFFUNC("H323Connection::GetSession");

  SWaitAndSignal mutex(((H323Connection*)this)->rtpSessionMutex);
  if (!(rtpSessions.find(sessionID) != rtpSessions.end()))
    return NULL;

  STRACE_INFO(g_pH323Module->m_RtpTraceId,
              "() Found existing session " << sessionID);
  return (*rtpSessions.find(sessionID)).second;
}


SString H323Connection::GetSessionCodecNames(unsigned sessionID) const
{
  SStringStream name;

  for (int i = 0; i < logicalChannels->GetSize(); i++) {
    H323Channel * channel = logicalChannels->GetChannelAt(i);
    if (!SBAD_READPTR(channel) && channel->GetSessionID() == sessionID) {
      H323Codec * codec = channel->GetCodec();
      if (!SBAD_READPTR(codec)) {
        if (name.empty())
          name << *codec;
        else
          name << " / " << *codec;
      }
    }
  }

  return name;
}


RTP_Session * H323Connection::UseSession(unsigned sessionID,
                                         const H245_TransportAddress & taddr)
{
    SDEFFUNC("H323Connection::UseSession");

  SWaitAndSignal mutex(rtpSessionMutex);

  if (rtpSessions.find(sessionID) != rtpSessions.end()) {
    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                 "() Found existing session " << sessionID);
    rtpSessions[sessionID]->IncrementReference();
    return rtpSessions[sessionID];
  }

  // We only support unicast IP at this time.
  if (!taddr.IsValid() || taddr.GetTag() != H245_TransportAddress::e_unicastAddress)
    return NULL;

  const H245_UnicastAddress & uaddr = taddr;
  if (uaddr.GetTag() != H245_UnicastAddress::e_iPAddress)
    return NULL;

//  H245_UnicastAddress_iPAddress & ipaddr = uaddr;
//  PIPSocket::Address ipnum(ipaddr.m_network[0], ipaddr.m_network[1],
//                           ipaddr.m_network[2], ipaddr.m_network[3]);

  STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
               "() Creating session " << sessionID);
  RTP_Session * session = new RTP_UDP(*this, sessionID, endpoint.GetRtpPortBase());
  rtpSessions.insert(pair<const int, RTP_Session *>(sessionID, session));
  return session;
}


void H323Connection::ReleaseSession(unsigned sessionID)
{
    SDEFFUNC("H323Connection::ReleaseSession");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                 "() Releasing session " << sessionID);

  rtpSessionMutex.Wait();

//  if (rtpSessions.Contains(POrdinalKey(sessionID))) {
  if (rtpSessions.find(sessionID) != rtpSessions.end()) {
    if (rtpSessions[sessionID]->DecrementReference()) {
      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, GetConnID(),
                   "() Deleting session " << sessionID);
      rtpSessions.insert(pair<const int, RTP_Session *>(sessionID, nullptr));
    }
  }

  rtpSessionMutex.Signal();
}


void H323Connection::RequestModeChange()
{
}


bool H323Connection::OnRequestModeChange(const H245_RequestMode & pdu,
                                         H245_RequestModeAck & /*ack*/,
                                         H245_RequestModeReject & /*reject*/,
                                         int & selectedMode)
{
    SDEFFUNC("H323Conn::OnRequestModeChange");

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() local capabilities: " << localCapabilities);

    for (selectedMode = 0; selectedMode < pdu.m_requestedModes.GetSize(); selectedMode++)
    {
        bool ok = true;
        for (int i = 0; i < pdu.m_requestedModes[selectedMode].GetSize(); i++)
        {
            if (localCapabilities.FindCapability(pdu.m_requestedModes[selectedMode][i]) == NULL)
            {
                ok = false;
                break;
            }
        }
        if (ok)
            return true;
    }
    STRACE2_WARN(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Mode change rejected as does not have capabilities");
    return false;
}


void H323Connection::OnModeChanged(const H245_ModeDescription & newMode)
{
    SDEFFUNC("H323Connection::OnModeChanged");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Completed request for mode change from remote.");

//  CloseAllLogicalChannels(false);

/* Copied from OpenH323:

  CloseAllLogicalChannels(FALSE);

  // Start up the new ones
  for (PINDEX i = 0; i < newMode.GetSize(); i++) {
    H323Capability * capability = localCapabilities.FindCapability(newMode[i]);
    PAssertNULL(capability); // Should not occur as OnRequestModeChange checks them
    if (!OpenLogicalChannel(*capability,
                            capability->GetDefaultSessionID(),
                            H323Channel::IsTransmitter)) {
      PTRACE(1, "H245\tCould not open channel after mode change: " << *capability);
    }
  }
*/
}


void H323Connection::OnAcceptModeChange(const H245_RequestModeAck & /*pdu*/)
{
    SDEFFUNC("H323Conn::OnAcceptModeChange");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received acceptance of last mode change request from remote.");
}


void H323Connection::OnRefusedModeChange(const H245_RequestModeReject * /*pdu*/)
{
    SDEFFUNC("H323Conn::OnRefusedModeChange");

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received reject of last mode change request from remote.");
}

bool H323Connection::OnReceivedHoldRequest()
{
    SDEFFUNC("H323Connection::OnRecvHoldReq");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "() Received hold request in base class!");

	return false;
}

bool H323Connection::OnReceivedActivateRequest()
{
    SDEFFUNC("H323Connection::OnRecvActReq");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "() Received activate request in base class!");

	return false;
}

bool H323Connection::OnReceivedCallTransferRequest(H323Connection *pTransferredToCall)
{
    SDEFFUNC("H323Connection::OnRecvCTRequest");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnID(),
        "() Received CT initiate in base class!");

	return false;
}

const H323Transport & H323Connection::GetControlChannel() const
{
  return *(controlChannel != NULL ? controlChannel : signallingChannel);
}

/////////////////////////////////////////////////////////////////////////////
