/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Capabilities
 
    Comments: H.323 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/h323caps.cpp               $
    $Author:: Ulrich.findeisen                                             $
      $Date:: 24.06.05 13:44                                               $
   $Modtime:: 24.06.05 11:27                                               $
  $Revision:: 3                                                            $
  $Workfile:: h323caps.cpp                                                 $
$NoKeywords: $
----------------------------------------------------------------------------*/

#include <stlib.h>
#include "h323.h"
#include "h323caps.h"
#include "h225.h"
#include "h245.h"

using namespace std;
using namespace Tools;

#define new SNEW


/////////////////////////////////////////////////////////////////////////////

H323Capability::H323Capability( unsigned long ulConnId, bool useEncrypt )
{
  SetConnID(ulConnId);
  assignedCapabilityNumber = 1;
  useEncryption = useEncrypt;
}


SObject::Comparison H323Capability::Compare(const SObject & obj) const
{
  SAssert(obj.IsDescendant(typeid(H323Capability)), SInvalidCast);
  const H323Capability & other = dynamic_cast<const H323Capability &>(obj);

  int mt = GetMainType();
  int omt = other.GetMainType();
  if (mt < omt)
    return LessThan;
  if (mt > omt)
    return GreaterThan;

  int st = GetSubType();
  int ost = other.GetSubType();
  if (st < ost)
    return LessThan;
  if (st > ost)
    return GreaterThan;

  if (useEncryption == 0 && other.useEncryption == 1)
      return LessThan;
  if (useEncryption == 1 && other.useEncryption == 0)
      return GreaterThan;

  return EqualTo;
}


void H323Capability::PrintOn(ostream & strm) const
{
  strm << "<" << assignedCapabilityNumber << "> "
       << (useEncryption ? "encrypt=true" : "encrypt=false");
}


unsigned H323Capability::GetDefaultSessionID() const
{
  return 0;
}


unsigned H323Capability::GetRxPacketSize() const
{
  return 0;
}


unsigned H323Capability::GetTxPacketSize() const
{
  return 0;
}

void H323Capability::SetTxPacketSize(unsigned uTxPacketSize)
{
}

unsigned H323Capability::GetMaxFrameSize() const
{
  return 0;
}

void H323Capability::SetPacketSizes(unsigned uPacketSizes)
{
}

bool H323Capability::OnReceivedPDU(const H245_Capability & cap)
{
  switch (cap.GetTag()) {
    case H245_Capability::e_receiveVideoCapability:
    case H245_Capability::e_receiveAudioCapability:
    case H245_Capability::e_receiveDataApplicationCapability:
    case H245_Capability::e_h233EncryptionReceiveCapability:
    case H245_Capability::e_receiveUserInputCapability:
      capabilityDirection = e_Receive;
      break;

    case H245_Capability::e_transmitVideoCapability:
    case H245_Capability::e_transmitAudioCapability:
    case H245_Capability::e_transmitDataApplicationCapability:
    case H245_Capability::e_h233EncryptionTransmitCapability:
    case H245_Capability::e_transmitUserInputCapability:
      capabilityDirection = e_Transmit;
      break;

    case H245_Capability::e_receiveAndTransmitVideoCapability:
    case H245_Capability::e_receiveAndTransmitAudioCapability:
    case H245_Capability::e_receiveAndTransmitDataApplicationCapability:
    case H245_Capability::e_receiveAndTransmitUserInputCapability:
      capabilityDirection = e_ReceiveAndTransmit;
      break;

    case H245_Capability::e_conferenceCapability:
    case H245_Capability::e_h235SecurityCapability:
    case H245_Capability::e_maxPendingReplacementFor:
      capabilityDirection = e_NoDirection;
  }

  return true;
}


//#if STRACING
const char * const H323Capability::MainTypesNames[] = {
  "Audio", "Video", "Data", "UserInput"
};

ostream & operator<<(ostream & o , H323Capability::CapabilityDirection d)
{
  const char * const names[] = {
    "Unknown", "Receive", "Transmit", "ReceiveAndTransmit", "NoDirection"
  };
  return o << names[d];
}
//#endif


/////////////////////////////////////////////////////////////////////////////

H323Channel * H323RealTimeCapability::CreateChannel(H323Connection & connection,
                                                    H323Channel::Directions dir,
                                                    unsigned sessionID,
                                 const H245_H2250LogicalChannelParameters * param) const
{
  SDEFFUNC("H323RltTimeCap::CreateChannel");

  RTP_Session * session;

  STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0, "()");

  if (param != NULL)
    session = connection.UseSession(param->m_sessionID, param->m_mediaControlChannel);
  else {
    // Make a fake transport address from the connection so gets initialised with
    // the transport type (IP, IPX, multicast etc).
    H245_TransportAddress addr;
    connection.GetControlChannel().SetUpTransportPDU(addr, 0);
    session = connection.UseSession(sessionID, addr);
  }

  if (session == NULL)
    return NULL;

  return new H323_RTPChannel(connection, *this, dir, *session);
}


/////////////////////////////////////////////////////////////////////////////

H323NonStandardCapabilityInfo::H323NonStandardCapabilityInfo(H323EndPoint & endpoint,
                                                             const BYTE * dataPtr,
                                                             int dataSize)
  : fixedData(dataSize, *dataPtr)
{
  H225_H221NonStandard h221;
  endpoint.SetH221NonStandardInfo(h221);
  t35CountryCode = (BYTE)(unsigned)h221.m_t35CountryCode;
  t35Extension = (BYTE)(unsigned)h221.m_t35Extension;
  manufacturerCode = (WORD)(unsigned)h221.m_manufacturerCode;
}


H323NonStandardCapabilityInfo::H323NonStandardCapabilityInfo(const SString & _oid,
                                                             const BYTE * dataPtr,
                                                             int dataSize)
  : oid(_oid),
    fixedData(dataSize, *dataPtr)
{
}


H323NonStandardCapabilityInfo::H323NonStandardCapabilityInfo(BYTE country,
                                                             BYTE extension,
                                                             WORD maufacturer,
                                                             const BYTE * dataPtr,
                                                             int dataSize)
  : t35CountryCode(country),
    t35Extension(extension),
    manufacturerCode(maufacturer),
    fixedData(dataSize, *dataPtr)
{
}


H323NonStandardCapabilityInfo::~H323NonStandardCapabilityInfo()
{
}


bool H323NonStandardCapabilityInfo::OnSendingPDU(SByteVector & data) const
{
  data = fixedData;
  return data.size() > 0;
}


bool H323NonStandardCapabilityInfo::OnReceivedPDU(const SByteVector & data)
{
  return data == fixedData;
}


bool H323NonStandardCapabilityInfo::OnSendingNonStandardPDU(SAsnChoice & pdu,
                                                            unsigned nonStandardTag) const
{
  SByteVector data;
  if (!OnSendingPDU(data))
    return false;

  pdu.SetTag(nonStandardTag);
  H245_NonStandardParameter & param = (H245_NonStandardParameter &)pdu.GetObject();

  if (!oid.empty()) {
    param.m_nonStandardIdentifier.SetTag(H245_NonStandardIdentifier::e_object);
    ((SAsnObjectId &)param) = oid;
  }
  else {
    param.m_nonStandardIdentifier.SetTag(H245_NonStandardIdentifier::e_h221NonStandard);
    H245_NonStandardIdentifier_h221NonStandard & h221 = param.m_nonStandardIdentifier;
    h221.m_t35CountryCode = (unsigned)t35CountryCode;
    h221.m_t35Extension = (unsigned)t35Extension;
    h221.m_manufacturerCode = (unsigned)manufacturerCode;
  }

  param.m_data = data;
  return data.size() > 0;
}


bool H323NonStandardCapabilityInfo::OnReceivedNonStandardPDU(const SAsnChoice & pdu,
                                                             unsigned nonStandardTag)
{
  if (pdu.GetTag() != nonStandardTag)
    return false;

  const H245_NonStandardParameter & param = (const H245_NonStandardParameter &)pdu.GetObject();

  if (!oid.empty()) {
    if (param.m_nonStandardIdentifier.GetTag() != H245_NonStandardIdentifier::e_object)
      return false;
    if (((SAsnObjectId &)param) != oid.c_str())
      return false;
  }
  else {
    if (param.m_nonStandardIdentifier.GetTag() != H245_NonStandardIdentifier::e_h221NonStandard)
      return false;
    const H245_NonStandardIdentifier_h221NonStandard & h221 = param.m_nonStandardIdentifier;
    if (h221.m_t35CountryCode != (unsigned)t35CountryCode)
      return false;
    if (h221.m_t35Extension != (unsigned)t35Extension)
      return false;
    if (h221.m_manufacturerCode != (unsigned)manufacturerCode)
      return false;
  }

  return OnReceivedPDU(param.m_data.GetValue());
}


/////////////////////////////////////////////////////////////////////////////

H323AudioCapability::H323AudioCapability(unsigned rx, unsigned tx, bool useEncrypt)
{
  rxPacketSize = rx;
  txPacketSize = tx;
  maxFrameSize = tx;
  useEncryption = useEncrypt;
}


H323Capability::MainTypes H323AudioCapability::GetMainType() const
{
  return e_Audio;
}


unsigned H323AudioCapability::GetDefaultSessionID() const
{
  return RTP_Session::DefaultAudioSessionID;
}


unsigned H323AudioCapability::GetRxPacketSize() const
{
  return rxPacketSize;
}


unsigned H323AudioCapability::GetTxPacketSize() const
{
  return txPacketSize;
}

void H323AudioCapability::SetTxPacketSize(unsigned uTxPacketSize)
{
    txPacketSize = uTxPacketSize;
}

unsigned H323AudioCapability::GetMaxFrameSize() const
{
  return maxFrameSize;
}

void H323AudioCapability::SetPacketSizes(unsigned uPacketSizes)
{
    rxPacketSize = uPacketSizes;
    txPacketSize = uPacketSizes;
    maxFrameSize = uPacketSizes;
}


bool H323AudioCapability::OnSendingPDU(H245_Capability & cap) const
{
  cap.SetTag(H245_Capability::e_receiveAudioCapability);
  return OnSendingPDU((H245_AudioCapability &)cap, rxPacketSize);
}


bool H323AudioCapability::OnSendingPDU(H245_DataType & dataType) const
{
  if (useEncryption)
  {
      dataType.SetTag(H245_DataType::e_h235Media);
      H245_H235Media & h235Media = (H245_H235Media &)dataType;

      h235Media.m_encryptionAuthenticationAndIntegrity.IncludeOptionalField(H245_EncryptionAuthenticationAndIntegrity::e_genericH235SecurityCapability);
      H245_GenericCapability & genCap = (H245_GenericCapability &)h235Media.m_encryptionAuthenticationAndIntegrity.m_genericH235SecurityCapability;
      genCap.m_capabilityIdentifier.SetTag(H245_CapabilityIdentifier::e_standard);
      SAsnObjectId & objId = (SAsnObjectId &)genCap.m_capabilityIdentifier;
      objId.SetValue("0.0.8.235.0.3.76"); // MIKEY, H.235.7, Table 1

      h235Media.m_mediaType.SetTag(H245_H235Media_mediaType::e_audioData);
      return OnSendingPDU((H245_AudioCapability &)h235Media.m_mediaType, txPacketSize);
  }
  else
  {
    dataType.SetTag(H245_DataType::e_audioData);
    return OnSendingPDU((H245_AudioCapability &)dataType, txPacketSize);
  }
}


bool H323AudioCapability::OnSendingPDU(H245_ModeElement & mode) const
{
  mode.m_type.SetTag(H245_ModeElementType::e_audioMode);
  return OnSendingPDU((H245_AudioMode &)mode);
}


bool H323AudioCapability::OnSendingPDU(H245_AudioMode & pdu) const
{
  static const H245_AudioMode::Choices AudioTable[] = {
    H245_AudioMode::e_nonStandard,
    H245_AudioMode::e_g711Alaw64k,
    H245_AudioMode::e_g711Alaw56k,
    H245_AudioMode::e_g711Ulaw64k,
    H245_AudioMode::e_g711Ulaw56k,
    H245_AudioMode::e_g722_64k,
    H245_AudioMode::e_g722_56k,
    H245_AudioMode::e_g722_48k,
    H245_AudioMode::e_g7231,
    H245_AudioMode::e_g728,
    H245_AudioMode::e_g729,
    H245_AudioMode::e_g729AnnexA,
    H245_AudioMode::e_is11172AudioMode,
    H245_AudioMode::e_is13818AudioMode,
    H245_AudioMode::e_g729wAnnexB,
    H245_AudioMode::e_g729AnnexAwAnnexB,
    H245_AudioMode::e_g7231AnnexCMode,
    H245_AudioMode::e_gsmFullRate,
    H245_AudioMode::e_gsmHalfRate,
    H245_AudioMode::e_gsmEnhancedFullRate,
    H245_AudioMode::e_genericAudioMode,
    H245_AudioMode::e_g729Extensions
  };

  unsigned subType = GetSubType();

  if (subType >= (int)(sizeof(AudioTable)/sizeof(AudioTable[0])))
    return false;

  pdu.SetTag(AudioTable[subType]);
  return true;
}


bool H323AudioCapability::OnReceivedPDU(const H245_Capability & cap)
{
  H323Capability::OnReceivedPDU(cap);

  if (cap.GetTag() != H245_Capability::e_receiveAudioCapability &&
      cap.GetTag() != H245_Capability::e_receiveAndTransmitAudioCapability)
    return false;

  unsigned packetSize;
  if (!OnReceivedPDU((const H245_AudioCapability &)cap, packetSize))
    return false;

  // Clamp our transmit size to maximum allowed
  if (txPacketSize > packetSize)
    txPacketSize = packetSize;

  return true;
}


bool H323AudioCapability::OnReceivedPDU(const H245_DataType & dataType)
{
  if (dataType.GetTag() == H245_DataType::e_audioData)
  {
      return OnReceivedPDU((const H245_AudioCapability &)dataType, rxPacketSize);
  }
  else if (dataType.GetTag() == H245_DataType::e_h235Media)
  {
      const H245_H235Media & h235Media = (const H245_H235Media &)dataType;
      const H245_AudioCapability & audio = (const H245_AudioCapability &)h235Media.m_mediaType;
      return OnReceivedPDU((const H245_AudioCapability &)audio, rxPacketSize);
  }

  return false;
}


/////////////////////////////////////////////////////////////////////////////

H323NonStandardAudioCapability::H323NonStandardAudioCapability(unsigned max,
                                                               unsigned desired,
                                                               H323EndPoint & endpoint,
                                                               const BYTE * fixedData,
                                                               int dataSize)
  : H323AudioCapability(max, desired),
    H323NonStandardCapabilityInfo(endpoint, fixedData, dataSize)
{
}


H323NonStandardAudioCapability::H323NonStandardAudioCapability(unsigned max,
                                                               unsigned desired,
                                                               const SString & oid,
                                                               const BYTE * fixedData,
                                                               int dataSize)
  : H323AudioCapability(max, desired),
    H323NonStandardCapabilityInfo(oid, fixedData, dataSize)
{
}


H323NonStandardAudioCapability::H323NonStandardAudioCapability(unsigned max,
                                                               unsigned desired,
                                                               BYTE country,
                                                               BYTE extension,
                                                               WORD maufacturer,
                                                               const BYTE * fixedData,
                                                               int dataSize)
  : H323AudioCapability(max, desired),
    H323NonStandardCapabilityInfo(country, extension, maufacturer, fixedData, dataSize)
{
}


unsigned H323NonStandardAudioCapability::GetSubType() const
{
  return H245_AudioCapability::e_nonStandard;
}


bool H323NonStandardAudioCapability::OnSendingPDU(H245_AudioCapability & pdu,
                                                  unsigned) const
{
  return OnSendingNonStandardPDU(pdu, H245_AudioCapability::e_nonStandard);
}


bool H323NonStandardAudioCapability::OnReceivedPDU(const H245_AudioCapability & pdu,
                                                   unsigned &)
{
  return OnReceivedNonStandardPDU(pdu, H245_AudioCapability::e_nonStandard);
}


/////////////////////////////////////////////////////////////////////////////

H323Capability::MainTypes H323VideoCapability::GetMainType() const
{
  return e_Video;
}


bool H323VideoCapability::OnSendingPDU(H245_Capability & cap) const
{
  cap.SetTag(H245_Capability::e_receiveVideoCapability);
  return OnSendingPDU((H245_VideoCapability &)cap);
}


bool H323VideoCapability::OnSendingPDU(H245_DataType & dataType) const
{
  dataType.SetTag(H245_DataType::e_videoData);
  return OnSendingPDU((H245_VideoCapability &)dataType);
}


bool H323VideoCapability::OnSendingPDU(H245_ModeElement & mode) const
{
  mode.m_type.SetTag(H245_ModeElementType::e_videoMode);
  return OnSendingPDU((H245_VideoMode &)mode.m_type);
}


bool H323VideoCapability::OnReceivedPDU(const H245_Capability & cap)
{
  H323Capability::OnReceivedPDU(cap);

  if (cap.GetTag() != H245_Capability::e_receiveVideoCapability &&
      cap.GetTag() != H245_Capability::e_receiveAndTransmitVideoCapability)
    return false;

  return OnReceivedPDU((const H245_VideoCapability &)cap);
}


bool H323VideoCapability::OnReceivedPDU(const H245_DataType & dataType)
{
  if (dataType.GetTag() != H245_DataType::e_videoData)
    return false;

  return OnReceivedPDU((const H245_VideoCapability &)dataType);
}


unsigned H323VideoCapability::GetDefaultSessionID() const
{
  return RTP_Session::DefaultVideoSessionID;
}


/////////////////////////////////////////////////////////////////////////////

H323NonStandardVideoCapability::H323NonStandardVideoCapability(H323EndPoint & endpoint,
                                                               const BYTE * fixedData,
                                                               int dataSize)
  : H323NonStandardCapabilityInfo(endpoint, fixedData, dataSize)
{
}


H323NonStandardVideoCapability::H323NonStandardVideoCapability(const SString & oid,
                                                               const BYTE * fixedData,
                                                               int dataSize)
  : H323NonStandardCapabilityInfo(oid, fixedData, dataSize)
{
}


H323NonStandardVideoCapability::H323NonStandardVideoCapability(BYTE country,
                                                               BYTE extension,
                                                               WORD maufacturer,
                                                               const BYTE * fixedData,
                                                               int dataSize)
  : H323NonStandardCapabilityInfo(country, extension, maufacturer, fixedData, dataSize)
{
}


unsigned H323NonStandardVideoCapability::GetSubType() const
{
  return H245_VideoCapability::e_nonStandard;
}


bool H323NonStandardVideoCapability::OnSendingPDU(H245_VideoCapability & pdu) const
{
  return OnSendingNonStandardPDU(pdu, H245_VideoCapability::e_nonStandard);
}


bool H323NonStandardVideoCapability::OnReceivedPDU(const H245_VideoCapability & pdu)
{
  return OnReceivedNonStandardPDU(pdu, H245_VideoCapability::e_nonStandard);
}


/////////////////////////////////////////////////////////////////////////////

H323DataCapability::H323DataCapability(unsigned rate)
  : maxBitRate(rate)
{
}


H323Capability::MainTypes H323DataCapability::GetMainType() const
{
  return e_Data;
}


unsigned H323DataCapability::GetDefaultSessionID() const
{
    return RTP_Session::DefaultDataSessionID;
}


H323Codec * H323DataCapability::CreateCodec(H323Codec::Direction) const
{
  return NULL;
}


bool H323DataCapability::OnSendingPDU(H245_Capability & cap) const
{
    SDEFFUNC("H323DataCapability::OnSendingPDU1");

    unsigned long ulConnId = const_cast<H323DataCapability *>(this)->GetConnID();
    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId, "()");

    cap.SetTag(H245_Capability::e_receiveAndTransmitDataApplicationCapability);
    H245_DataApplicationCapability & app = cap;
    app.m_maxBitRate = maxBitRate;
    return OnSendingPDU(app);
}


bool H323DataCapability::OnSendingPDU(H245_DataType & dataType) const
{
    SDEFFUNC("H323DataCapability::OnSendingPDU2");

    unsigned long ulConnId = const_cast<H323DataCapability *>(this)->GetConnID();
    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId, "()");

    dataType.SetTag(H245_DataType::e_data);
    H245_DataApplicationCapability & app = dataType;
    app.m_maxBitRate = maxBitRate;
    return OnSendingPDU(app);
}


bool H323DataCapability::OnSendingPDU(H245_ModeElement & mode) const
{
    SDEFFUNC("H323DataCapability::OnSendingPDU3");

    unsigned long ulConnId = const_cast<H323DataCapability *>(this)->GetConnID();
    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId, "()");

    mode.m_type.SetTag(H245_ModeElementType::e_dataMode);
    H245_DataMode & type = mode.m_type;
    type.m_bitRate = maxBitRate;
    return OnSendingPDU(type);
}


bool H323DataCapability::OnReceivedPDU(const H245_Capability & cap)
{
    SDEFFUNC("H323DataCapability::OnRecvPDU cap");

    unsigned long ulConnId = const_cast<H323DataCapability *>(this)->GetConnID();
    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId, "()");

    H323Capability::OnReceivedPDU(cap);

    if (cap.GetTag() != H245_Capability::e_receiveDataApplicationCapability &&
        cap.GetTag() != H245_Capability::e_receiveAndTransmitDataApplicationCapability)
        return false;

    const H245_DataApplicationCapability & app = cap;
    maxBitRate = app.m_maxBitRate;
    return OnReceivedPDU(app);
}


bool H323DataCapability::OnReceivedPDU(const H245_DataType & dataType)
{
    SDEFFUNC("H323DataCapability::OnRecvPDU dt");

    unsigned long ulConnId = const_cast<H323DataCapability *>(this)->GetConnID();
    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId, "()");

    if (dataType.GetTag() != H245_DataType::e_data)
        return false;

    const H245_DataApplicationCapability & app = dataType;
    maxBitRate = app.m_maxBitRate;
    return OnReceivedPDU(app);
}


/////////////////////////////////////////////////////////////////////////////

H323NonStandardDataCapability::H323NonStandardDataCapability(H323EndPoint & endpoint,
                                                             const BYTE * fixedData,
                                                             int dataSize)
  : H323NonStandardCapabilityInfo(endpoint, fixedData, dataSize)
{
}


H323NonStandardDataCapability::H323NonStandardDataCapability(const SString & oid,
                                                             const BYTE * fixedData,
                                                             int dataSize)
  : H323NonStandardCapabilityInfo(oid, fixedData, dataSize)
{
}


H323NonStandardDataCapability::H323NonStandardDataCapability(BYTE country,
                                                             BYTE extension,
                                                             WORD maufacturer,
                                                             const BYTE * fixedData,
                                                             int dataSize)
  : H323NonStandardCapabilityInfo(country, extension, maufacturer, fixedData, dataSize)
{
}


unsigned H323NonStandardDataCapability::GetSubType() const
{
  return H245_DataApplicationCapability_application::e_nonStandard;
}


bool H323NonStandardDataCapability::OnSendingPDU(H245_DataApplicationCapability & pdu) const
{
  return OnSendingNonStandardPDU(pdu.m_application, H245_DataApplicationCapability_application::e_nonStandard);
}


bool H323NonStandardDataCapability::OnReceivedPDU(const H245_DataApplicationCapability & pdu)
{
  return OnReceivedNonStandardPDU(pdu.m_application, H245_DataApplicationCapability_application::e_nonStandard);
}


/////////////////////////////////////////////////////////////////////////////

H323_G711Capability::H323_G711Capability(Mode m, Speed s, unsigned uRxPacketSize, unsigned uTxPacketSize, bool useEncrypt)
  : H323AudioCapability(uRxPacketSize*8, uTxPacketSize*8, useEncrypt)
{
  mode = m;
  speed = s;
}


SObject * H323_G711Capability::Clone() const
{
  return new H323_G711Capability(mode, speed, rxPacketSize/8, txPacketSize/8, useEncryption);
}


void H323_G711Capability::PrintOn(ostream & strm) const
{
  strm << "G.711 "
       << (mode == muLaw ? "u-Law " : "A-Law ")
       << (speed == At64k ? "64k " : "56k ");
  H323AudioCapability::PrintOn(strm);
}


static const unsigned G711SubType[2][2] = {
  { H245_AudioCapability::e_g711Alaw64k, H245_AudioCapability::e_g711Alaw56k },
  { H245_AudioCapability::e_g711Ulaw64k, H245_AudioCapability::e_g711Ulaw56k }
};


unsigned H323_G711Capability::GetSubType() const
{
  return G711SubType[mode][speed];
}


H323Codec * H323_G711Capability::CreateCodec(H323Codec::Direction direction) const
{
  unsigned packetSize = direction == H323Codec::Encoder ? txPacketSize : rxPacketSize;

  if (mode == muLaw)
    return new H323_muLawCodec(direction, speed == H323_G711Capability::At56k, packetSize);
  else
    return new H323_ALawCodec(direction, speed == H323_G711Capability::At56k, packetSize);
}


bool H323_G711Capability::OnSendingPDU(H245_AudioCapability & cap,
                                       unsigned packetSize) const
{
  cap.SetTag(G711SubType[mode][speed]);

  SAsnInteger & value = cap;
  value = packetSize/8;
  return true;
}


bool H323_G711Capability::OnReceivedPDU(const H245_AudioCapability & cap,
                                        unsigned & packetSize)
{
  switch (cap.GetTag()) {
    case H245_AudioCapability::e_g711Alaw64k :
      mode = H323_G711Capability::ALaw;
      speed = H323_G711Capability::At64k;
      break;

    case H245_AudioCapability::e_g711Alaw56k :
      mode = H323_G711Capability::ALaw;
      speed = H323_G711Capability::At56k;
      break;

    case H245_AudioCapability::e_g711Ulaw64k :
      mode = H323_G711Capability::muLaw;
      speed = H323_G711Capability::At64k;
      break;

    case H245_AudioCapability::e_g711Ulaw56k :
      mode = H323_G711Capability::muLaw;
      speed = H323_G711Capability::At56k;
      break;

    default :
      return false;
  }

  const SAsnInteger & value = cap;

  // Get the maximum number of milliseconds
  packetSize = value*8;
  return true;
}


/////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// G.723.1 Capability
//-----------------------------------------------------------------------------

H323_G7231Capability::H323_G7231Capability(bool useEncrypt)
  : H323AudioCapability(1, 1, useEncrypt)
{
    maxFrameSize = 24;
}

//-----------------------------------------------------------------------------

SObject * H323_G7231Capability::Clone() const
{
    return new H323_G7231Capability(useEncryption);
}

//-----------------------------------------------------------------------------

void H323_G7231Capability::PrintOn(ostream & strm) const
{
    strm << "G.723.1 ";
    H323AudioCapability::PrintOn(strm);
}

//-----------------------------------------------------------------------------

unsigned H323_G7231Capability::GetSubType() const
{
    return H245_AudioCapability::e_g7231;
}

//-----------------------------------------------------------------------------

H323Codec * H323_G7231Capability::CreateCodec(H323Codec::Direction direction) const
{
    return new H323_G7231Codec(direction);
}

//-----------------------------------------------------------------------------

bool H323_G7231Capability::OnSendingPDU(H245_AudioCapability & pdu,
                                        unsigned packetSize) const
{
    pdu.SetTag(H245_AudioCapability::e_g7231);

    H245_AudioCapability_g7231 & g7231 = pdu;

    // Max. number of audio frames per packet we could handle (until now only one!)
    g7231.m_maxAl_sduAudioFrames = 1;

    // No silence suppression
    g7231.m_silenceSuppression = false;

    return true;
}

//-----------------------------------------------------------------------------

bool H323_G7231Capability::OnReceivedPDU(const H245_AudioCapability & pdu,
                                         unsigned & packetSize)
{
    if (pdu.GetTag() != H245_AudioCapability::e_g7231)
        return false;

    const H245_AudioCapability_g7231 & g7231 = pdu;
    packetSize = g7231.m_maxAl_sduAudioFrames;

    return true;
}


//-----------------------------------------------------------------------------
// G.729 Capability
//-----------------------------------------------------------------------------

H323_G729Capability::H323_G729Capability(bool useEncrypt)
  : H323AudioCapability(160, 160, useEncrypt)
{
    maxFrameSize = 160;
}

//-----------------------------------------------------------------------------

SObject * H323_G729Capability::Clone() const
{
    return new H323_G729Capability(useEncryption);
}

//-----------------------------------------------------------------------------

void H323_G729Capability::PrintOn(ostream & strm) const
{
    strm << "G.729 ";
    H323AudioCapability::PrintOn(strm);
}

//-----------------------------------------------------------------------------

unsigned H323_G729Capability::GetSubType() const
{
    return H245_AudioCapability::e_g729;
}

//-----------------------------------------------------------------------------

H323Codec * H323_G729Capability::CreateCodec(H323Codec::Direction direction) const
{
    return new H323_G729Codec(direction);
}

//-----------------------------------------------------------------------------

bool H323_G729Capability::OnSendingPDU(H245_AudioCapability & pdu,
                                       unsigned packetSize) const
{
    pdu.SetTag(H245_AudioCapability::e_g729);

    SAsnInteger & value = pdu;
    // value = maximum number of frames per packet, each G.729 frame represents 10 ms
    value = packetSize/8/10;

    return true;
}

//-----------------------------------------------------------------------------

bool H323_G729Capability::OnReceivedPDU(const H245_AudioCapability & pdu,
                                        unsigned & packetSize)
{
    if (pdu.GetTag() != H245_AudioCapability::e_g729)
        return false;

    const SAsnInteger & value = pdu;

    // value = maximum number of frames per packet, each G.729 frame represents 10 ms
    packetSize = value*8*10;

    return true;
}

//-----------------------------------------------------------------------------
// G.729A Capability
//-----------------------------------------------------------------------------

H323_G729ACapability::H323_G729ACapability(bool useEncrypt)
  : H323AudioCapability(160, 160, useEncrypt)
{
    maxFrameSize = 160;
}

//-----------------------------------------------------------------------------

SObject * H323_G729ACapability::Clone() const
{
    return new H323_G729ACapability(useEncryption);
}

//-----------------------------------------------------------------------------

void H323_G729ACapability::PrintOn(ostream & strm) const
{
    strm << "G.729A ";
    H323AudioCapability::PrintOn(strm);
}

//-----------------------------------------------------------------------------

unsigned H323_G729ACapability::GetSubType() const
{
    return H245_AudioCapability::e_g729AnnexA;
}

//-----------------------------------------------------------------------------

H323Codec * H323_G729ACapability::CreateCodec(H323Codec::Direction direction) const
{
    return new H323_G729Codec(direction);
}

//-----------------------------------------------------------------------------

bool H323_G729ACapability::OnSendingPDU(H245_AudioCapability & pdu,
                                        unsigned packetSize) const
{
    pdu.SetTag(H245_AudioCapability::e_g729AnnexA);

    SAsnInteger & value = pdu;
    // value = maximum number of frames per packet, each G.729 frame represents 10 ms
    value = packetSize/8/10;

    return true;
}

//-----------------------------------------------------------------------------

bool H323_G729ACapability::OnReceivedPDU(const H245_AudioCapability & pdu,
                                         unsigned & packetSize)
{
    if (pdu.GetTag() != H245_AudioCapability::e_g729AnnexA)
        return false;

    const SAsnInteger & value = pdu;

    // value = maximum number of frames per packet, each G.729 frame represents 10 ms
    packetSize = value*8*10;

    return true;
}

//-----------------------------------------------------------------------------
// G.722 Capability
//-----------------------------------------------------------------------------

H323_G722Capability::H323_G722Capability(bool useEncrypt)
  : H323AudioCapability(160, 160, useEncrypt)
{
    maxFrameSize = 160;
}

//-----------------------------------------------------------------------------

SObject * H323_G722Capability::Clone() const
{
    return new H323_G722Capability(useEncryption);
}

//-----------------------------------------------------------------------------

void H323_G722Capability::PrintOn(ostream & strm) const
{
    strm << "G.722 ";
    H323AudioCapability::PrintOn(strm);
}

//-----------------------------------------------------------------------------

unsigned H323_G722Capability::GetSubType() const
{
    return H245_AudioCapability::e_g722_64k;
}

//-----------------------------------------------------------------------------

H323Codec * H323_G722Capability::CreateCodec(H323Codec::Direction direction) const
{
    return new H323_G722Codec(direction);
}

//-----------------------------------------------------------------------------

bool H323_G722Capability::OnSendingPDU(H245_AudioCapability & pdu,
                                        unsigned packetSize) const
{
    pdu.SetTag(H245_AudioCapability::e_g722_64k);

    SAsnInteger & value = pdu;
    // value = maximum number of frames per packet
    value = packetSize/8;

    return true;
}

//-----------------------------------------------------------------------------

bool H323_G722Capability::OnReceivedPDU(const H245_AudioCapability & pdu,
                                         unsigned & packetSize)
{
    if (pdu.GetTag() != H245_AudioCapability::e_g722_64k)
        return false;

    const SAsnInteger & value = pdu;

    // value = maximum number of frames per packet
    packetSize = value*8;

    return true;
}

/////////////////////////////////////////////////////////////////////////////

#define T120_MAX_BIT_RATE 825000


H323_T120Capability::H323_T120Capability()
  : H323DataCapability(T120_MAX_BIT_RATE)
{
}


SObject * H323_T120Capability::Clone() const
{
  return new H323_T120Capability();
}


unsigned H323_T120Capability::GetSubType() const
{
  return H245_DataApplicationCapability_application::e_t120;
}


H323Channel * H323_T120Capability::CreateChannel(H323Connection &,
                                                 H323Channel::Directions,
                                                 unsigned,
                                const H245_H2250LogicalChannelParameters *) const
{
  return NULL;
}


bool H323_T120Capability::OnSendingPDU(H245_DataApplicationCapability & pdu) const
{
  pdu.m_application.SetTag(H245_DataApplicationCapability_application::e_t120);
  return OnSendingPDU((H245_DataProtocolCapability &)pdu.m_application);
}


bool H323_T120Capability::OnSendingPDU(H245_DataMode & pdu) const
{
  pdu.m_application.SetTag(H245_DataMode_application::e_t120);
  return OnSendingPDU((H245_DataProtocolCapability &)pdu.m_application);
}


bool H323_T120Capability::OnSendingPDU(H245_DataProtocolCapability & pdu) const
{
  pdu.SetTag(H245_DataProtocolCapability::e_separateLANStack);
  return true;
}


bool H323_T120Capability::OnReceivedPDU(const H245_DataApplicationCapability & cap)
{
  if (cap.m_application.GetTag() != H245_DataApplicationCapability_application::e_t120)
    return false;

  const H245_DataProtocolCapability & dataCap = cap.m_application;

  return dataCap.GetTag() == H245_DataProtocolCapability::e_separateLANStack;
}


/////////////////////////////////////////////////////////////////////////////

H323_UserInputCapability::H323_UserInputCapability(unsigned _subType)
{
  subType = _subType;
}


SObject * H323_UserInputCapability::Clone() const
{
  return new H323_UserInputCapability(subType);
}


void H323_UserInputCapability::PrintOn(ostream & strm) const
{
  switch (subType) {
    case H245_UserInputCapability::e_basicString :
      strm << "UserInput basicString ";
      break;
    case H245_UserInputCapability::e_iA5String :
      strm << "UserInput iA5String ";
      break;
    case H245_UserInputCapability::e_generalString :
      strm << "UserInput generalString ";
      break;
    case H245_UserInputCapability::e_dtmf :
      strm << "UserInput dtmf ";
      break;
    case H245_UserInputCapability::e_hookflash :
      strm << "UserInput hookflash ";
      break;
    default :
      strm << "UserInput SubType(" << subType << ") ";
  }
  H323Capability::PrintOn(strm);
}


H323Capability::MainTypes H323_UserInputCapability::GetMainType() const
{
  return e_UserInput;
}


unsigned  H323_UserInputCapability::GetSubType()  const
{
  return subType;
}


H323Channel * H323_UserInputCapability::CreateChannel(H323Connection & connection,
                                                      H323Channel::Directions,
                                                      unsigned,
                                                      const H245_H2250LogicalChannelParameters *) const
{
    SDEFFUNC("H323_UserInputCap::CreateChannel");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, connection.GetConnID(),
        "() Cannot create UserInputCapability channel");
    return NULL;
}


H323Codec * H323_UserInputCapability::CreateCodec(H323Codec::Direction) const
{
    SDEFFUNC("H323_UserInputCap::CreateCodec");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Cannot create UserInputCapability codec");
    return NULL;
}


bool H323_UserInputCapability::OnSendingPDU(H245_Capability & pdu) const
{
  pdu.SetTag(H245_Capability::e_receiveUserInputCapability);
  H245_UserInputCapability & ui = pdu;
  ui.SetTag(subType);
  return true;
}


bool H323_UserInputCapability::OnSendingPDU(H245_DataType &) const
{
    SDEFFUNC("H323_UserInputCap::OnSendingPDU_D");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Cannot have UserInputCapability in DataType");
    return false;
}


bool H323_UserInputCapability::OnSendingPDU(H245_ModeElement &) const
{
    SDEFFUNC("H323_UserInputCap::OnSendingPDU_M");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Cannot have UserInputCapability in ModeElement");
    return false;
}


bool H323_UserInputCapability::OnReceivedPDU(const H245_Capability & pdu)
{
  if (pdu.GetTag() != H245_Capability::e_receiveUserInputCapability)
    return false;

  const H245_UserInputCapability & ui = pdu;
  subType = ui.GetTag();
  return true;
}


bool H323_UserInputCapability::OnReceivedPDU(const H245_DataType &)
{
    SDEFFUNC("H323_UserInputCap::OnReceivedPDU");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Cannot have UserInputCapability in DataType");
    return false;
}


/////////////////////////////////////////////////////////////////////////////

#ifdef S_NO_TRACE // Use Print() only in Debug version
#define CAP_PRINT(iLine)
#else
#define CAP_PRINT(iLine) H323Capabilities::Print(iLine)
#endif // !S_NO_TRACE


bool H323SimultaneousCapabilities::SetSize(int newSize)
{
  int oldSize = H323CapabilitiesListArray::size();

  while (oldSize < newSize) {
    H323CapabilitiesList * list = new H323CapabilitiesList;
    // The lowest level list should not delete codecs on destruction
//  list->DisallowDeleteObjects();
//  SetAt(oldSize++, list);
    H323CapabilitiesListArray::push_back(list);
    oldSize++;
  }

  return true;
}


void H323SimultaneousCapabilities::Clear()
{
  if ( H323CapabilitiesListArray::size() > 0 )
  {
    H323CapabilitiesListArray::iterator capListArrayIter;
    for( capListArrayIter  = H323CapabilitiesListArray::begin();
         capListArrayIter != H323CapabilitiesListArray::end();
         ++capListArrayIter )
    {
        delete (*capListArrayIter);
    }
  }
}

//-----------------------------------------------------------------------------

bool H323CapabilitiesSet::SetSize(int newSize)
{
  int oldSize = H323CapabilitiesSetArray::size();

  while (oldSize < newSize)
  {
      H323CapabilitiesSetArray::push_back(new H323SimultaneousCapabilities);
      oldSize++;
  }

  return true;
}

void H323CapabilitiesSet::Clear()
{
  if ( H323CapabilitiesSetArray::size() > 0 )
  {
    H323CapabilitiesSetArray::iterator capSetArrayIter;
    for( capSetArrayIter  = H323CapabilitiesSetArray::begin();
         capSetArrayIter != H323CapabilitiesSetArray::end();
         ++capSetArrayIter )
    {
        if ( (*capSetArrayIter) )
        {
            (*capSetArrayIter)->Clear();
            delete (*capSetArrayIter);
        }
    }
  }
}

//-----------------------------------------------------------------------------

H323Capabilities::H323Capabilities( unsigned long ulConnId )
{
    SDEFFUNC("H323Capabilities::H323Capabilitie");
    
    SetConnID(ulConnId);

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(), "()"); 

    CAP_PRINT(__LINE__);
}


H323Capabilities::H323Capabilities(const H323Connection & connection,
                                   const H245_TerminalCapabilitySet & pdu)
{
    SDEFFUNC("H323Capabilities::H323Capabilitie");

    unsigned long ulConnId = const_cast<H323Connection *>(&connection)->GetConnID();
    SetConnID(ulConnId);

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(), "(conn, pdu)"); 

    CAP_PRINT(__LINE__);

  // Decode out of the PDU, the list of known codecs.
  if (pdu.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityTable)) {
    for (int i = 0; i < pdu.m_capabilityTable.GetSize(); i++) {
      if (pdu.m_capabilityTable[i].HasOptionalField(H245_CapabilityTableEntry::e_capability)) {
        H323Capability * capability = connection.GetLocalCapabilities().FindCapability(pdu.m_capabilityTable[i].m_capability);
        if (capability == NULL )
        {
            capability = connection.GetLocalCapabilities().FindCapability(pdu.m_capabilityTable[i].m_capability, true);
        }
        if (capability != NULL) {
          H323Capability * copy = (H323Capability *)capability->Clone();
          copy->SetCapabilityNumber(pdu.m_capabilityTable[i].m_capabilityTableEntryNumber);
          if (copy->OnReceivedPDU(pdu.m_capabilityTable[i].m_capability))
            table.push_back(copy);
          else
            delete copy;
        }
      }
    }
  }

  int outerSize = pdu.m_capabilityDescriptors.GetSize();
  set.SetSize(outerSize);
  for (int outer = 0; outer < outerSize; outer++) {
    H245_CapabilityDescriptor & desc = pdu.m_capabilityDescriptors[outer];
    if (desc.HasOptionalField(H245_CapabilityDescriptor::e_simultaneousCapabilities)) {
      int middleSize = desc.m_simultaneousCapabilities.GetSize();
      set[outer]->SetSize(middleSize);
      for (int middle = 0; middle < middleSize; middle++) {
        H245_AlternativeCapabilitySet & alt = desc.m_simultaneousCapabilities[middle];
        for (int inner = 0; inner < alt.GetSize(); inner++) {
			for (H323CapabilitiesList::size_type cap = 0; cap < table.size(); cap++) {
            if (table[cap]->GetCapabilityNumber() == alt[inner]) {
              (*set[outer])[middle]->push_back(&(*table[cap]));
              break;
            }
          }
        }
      }
    }
  }
    CAP_PRINT(__LINE__);
}


H323Capabilities::H323Capabilities(const H323Capabilities & original)
{
    SDEFFUNC("H323Capabilities::H323Capabilitie");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "(orig)");

    operator=(original);
}


H323Capabilities & H323Capabilities::operator=(const H323Capabilities & original)
{
    SDEFFUNC("H323Capabilities::operator=");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "(orig)");
  RemoveAll();

  for (int i = 0; i < original.GetSize(); i++)
    Copy(original[i]);

  const H323CapabilitiesSet & origSet = original.set;
  int outerSize = origSet.size();
  set.SetSize(outerSize);
  for (int outer = 0; outer < outerSize; outer++) {
    int middleSize = origSet[outer]->size();
    set[outer]->SetSize(middleSize);
    for (int middle = 0; middle < middleSize; middle++) {
      int innerSize = (*origSet[outer])[middle]->size();
      for (int inner = 0; inner < innerSize; inner++)
        (*set[outer])[middle]->push_back(FindCapability((*(*origSet[outer])[middle])[inner]->GetCapabilityNumber()));
    }
  }
  return *this;
}


H323Capabilities::~H323Capabilities()
{
    SDEFFUNC("H323Capabilities::~H323Capabiliti");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "()");

    RemoveAll();
    CAP_PRINT(__LINE__);
}


void H323Capabilities::PrintOn(ostream & strm) const
{
  long long indent = strm.precision()-1;
  for (H323CapabilitiesList::size_type i = 0; i < table.size(); i++)
    strm << setw(indent) << " " << *table[i] << " ";
  strm << " mikey-size: " << mikey.size();
}


void H323Capabilities::Print(int iLine) const
{
    if (STrace::CanTrace(g_pH323Module->m_CodecTraceId, STrace::Debug))
    {
        STRACE(STrace::Debug, g_pH323Module->m_CodecTraceId, "(" << this << ") " << iLine
               << " table size=" << table.size() << ", set size=" << set.size());

		for(H323CapabilitiesList::size_type ii = 0; ii < table.size(); ii++)
        {
            if (table[ii])
            {
                STRACE(STrace::Debug, g_pH323Module->m_CodecTraceId, "(" << this << ") " << iLine
                       << " table: entry " << ii << " (" << table[ii] << ")=" << *table[ii]);
            }
        }
        for (H323CapabilitiesList::size_type outer = 0; outer < set.size(); outer++)
        {
            STRACE(STrace::Debug, g_pH323Module->m_CodecTraceId, "(" << this << ") " << iLine
                   << " set: outer entry " << outer << " :" << set[outer]);
            for (H323CapabilitiesList::size_type middle = 0; middle < set[outer]->size(); middle++)
            {
                STRACE(STrace::Debug, g_pH323Module->m_CodecTraceId, "(" << this << ") " << iLine
                       << " set: middle entry " << middle << " :" << (*set[outer])[middle]);
                for (H323CapabilitiesList::size_type inner = 0; inner < (*set[outer])[middle]->size(); inner++)
                {
                    STRACE(STrace::Debug, g_pH323Module->m_CodecTraceId, "(" << this << ") " << iLine
                           << " set: inner entry " << inner << " :" << ((*(*set[outer])[middle])[inner]));
                }
            }
        }
    }
}


void H323Capabilities::SetCapability(int descriptorNum,
                                     int simultaneousNum,
                                     H323Capability * capability)
{
    SDEFFUNC("H323Capabilities::SetCapability");

    if (SBAD_READPTR(capability))
    {
        STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, GetConnID(),
            "() capability invalid -> discard set operation!");
        return;
    }

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "(descriptorNum=" << descriptorNum
        << ", simultaneousNum=" << simultaneousNum
        << ", cap=" << capability << ", *cap=" << *capability << ")");

    // Make sure capability has been added to table.
    Add(capability);

    // Now add the same capability to the set.
    // Make sure the outer array is big enough
    if (descriptorNum+1 > (int)set.size())
    {
        while (descriptorNum+1 - set.size())
        {
            H323SimultaneousCapabilities *pSimCap = new H323SimultaneousCapabilities;
            if (!SBAD_READPTR(pSimCap))
            {
                set.push_back(pSimCap);
            }
        }
    }

    // Make sure the middle array is big enough
    if (simultaneousNum+1 > (int)set[descriptorNum]->size())
    {
        while (simultaneousNum+1 - set[descriptorNum]->size())
        {
            H323CapabilitiesList *pCapList = new H323CapabilitiesList;
            if (!SBAD_READPTR(pCapList))
            {
                set[descriptorNum]->push_back(pCapList);
            }
        }
    }

    // Now we can put the new entry in (with paranoid check).
    if (!SBAD_READPTR((*set[descriptorNum])[simultaneousNum]))
        (*set[descriptorNum])[simultaneousNum]->push_back(capability);

    CAP_PRINT(__LINE__);
}


static unsigned MergeCapabilityNumber(const H323CapabilitiesList & table,
                                      unsigned newCapabilityNumber)
{
  // Assign a unique number to the codec, check if the user wants a specific
  // value and start with that.
  if (newCapabilityNumber == 0)
    newCapabilityNumber = 1;

  H323CapabilitiesList::size_type i = 0;
  while (i < table.size()) {
    if (table[i]->GetCapabilityNumber() != newCapabilityNumber)
      i++;
    else {
      // If it already in use, increment it
      newCapabilityNumber++;
      i = 0;
    }
  }

  return newCapabilityNumber;
}


void H323Capabilities::Add(H323Capability * capability)
{
    SDEFFUNC("H323Capabilities::Add");

    if (SBAD_READPTR(capability))
    {
        STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, GetConnID(),
            "() capability invalid -> discard add operation!");
        return;
    }

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "(cap=" << capability << ", *cap=" << *capability << ")");

  // See if already added, confuses things if you add the same instance twice
  if (find(table.begin(), table.end(), capability) != table.end())
    return;

  capability->SetCapabilityNumber(MergeCapabilityNumber(table, 1));
  table.push_back(capability);

  STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, GetConnID(),
      "() Added capability: " << *capability);
    CAP_PRINT(__LINE__);
}


H323Capability * H323Capabilities::Copy(const H323Capability & capability)
{
    SDEFFUNC("H323Capabilities::Copy");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "(cap=" << capability << ", " << &capability << ")");
    CAP_PRINT(__LINE__);

  H323Capability * newCapability = (H323Capability *)capability.Clone();
  if (SBAD_READPTR(newCapability))
    return NULL;
  newCapability->SetCapabilityNumber(MergeCapabilityNumber(table, capability.GetCapabilityNumber()));
  table.push_back(newCapability);

  STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, GetConnID(),
      "() Added capability: " << *newCapability);
    CAP_PRINT(__LINE__);
  return newCapability;
}


void H323Capabilities::ClearTableWithoutDelete()
{
    SDEFFUNC("H323Cap::ClearTableWithoutDelete");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "()");
    CAP_PRINT(__LINE__);
    table.clear();
}


void H323Capabilities::ClearSetWithoutDelete()
{
    SDEFFUNC("H323Cap::ClearSetWithoutDelete");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "()");
    CAP_PRINT(__LINE__);

    set.Clear();
    set.clear();

    CAP_PRINT(__LINE__);
}


/*
void H323Capabilities::Remove(H323Capability * capability)
{
    SDEFFUNC("H323Capabilities::Remove");

    if (SBAD_READPTR(capability))
        return;

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "(" << capability << ")");

    CAP_PRINT(__LINE__);

    unsigned capabilityNumber = capability->GetCapabilityNumber();

    H323CapabilitiesSetArray::iterator  outerIter;
    H323CapabilitiesListArray::iterator middleIter;
    H323CapabilitiesList::iterator      innerIter;

    for(outerIter = set.begin(); outerIter != set.end(); ++outerIter)
    {
        if (!SBAD_READPTR(*outerIter))
        {
            for(middleIter = (*outerIter)->begin(); middleIter != (*outerIter)->end(); ++middleIter)
            {
                if (!SBAD_READPTR(*middleIter))
                {
                    for(innerIter = (*middleIter)->begin(); innerIter != (*middleIter)->end(); ++innerIter)
                    {
                        if (   !SBAD_READPTR(*innerIter)
                            && (*innerIter)->GetCapabilityNumber() == capabilityNumber)
                        {
                            STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
                                "() Removing cap entry from cap set (" << *innerIter << ")");
                            (*middleIter)->erase(innerIter);
                            // delete will be done in table (see bleow)
                        }
                    }
                    if ((*middleIter)->empty())
                    {
                        (*outerIter)->erase(middleIter);
                    }
                }
            }
            if ((*outerIter)->empty())
            {
                set.erase(outerIter);
            }
        }
    }

    H323CapabilitiesList::iterator tableIter;
    for(tableIter = table.begin(); tableIter != table.end(); ++tableIter)
    {
        if (   !SBAD_READPTR(*tableIter)
            && (*tableIter)->GetCapabilityNumber() == capabilityNumber)
        {
            STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
                "() Removing and deleting cap entry from table (" << *tableIter << ")");
            delete *tableIter;
            table.erase(tableIter);
        }
    }

    CAP_PRINT(__LINE__);
}


void H323Capabilities::Remove(const PString & codecName)
{
  H323Capability * cap = FindCapability(codecName);
  while (cap != NULL) {
    Remove(cap);
    cap = FindCapability(codecName);
  }
}


void H323Capabilities::Remove(const PStringArray & codecNames)
{
  for (int i = 0; i < codecNames.GetSize(); i++)
    Remove(codecNames[i]);
}
*/


void H323Capabilities::RemoveAllWithoutDelete()
{
    SDEFFUNC("H323Capabil::RemoveAllWithoutDel");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "()");

    CAP_PRINT(__LINE__);

    ClearTableWithoutDelete();
    ClearSetWithoutDelete();

    CAP_PRINT(__LINE__);
}


void H323Capabilities::RemoveAll()
{
    SDEFFUNC("H323Capabilities::RemoveAll");

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
        "()");

    CAP_PRINT(__LINE__);

    // Check for remaining entries in the capabilities table
    if (table.size() > 0)
    {
        H323CapabilitiesList::iterator capListIter;
        for(capListIter = table.begin(); capListIter != table.end(); ++capListIter)
        {
            if (!SBAD_READPTR(*capListIter))
            {
                STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, GetConnID(),
                    "() deleting table entry (" << (*capListIter) << ")=" << *(*capListIter));
                delete (*capListIter);
            }
        }
        table.clear();
    }
    if (set.size() > 0)
    {
        set.Clear();
        set.clear();
    }

    CAP_PRINT(__LINE__);
}


H323Capability * H323Capabilities::FindCapability(unsigned capabilityNumber) const
{
    SDEFFUNC("H323Capabilities::FindCapability1");

  unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
  STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
      "(cap num=" << capabilityNumber << ")");

  for (H323CapabilitiesList::size_type i = 0; i < table.size(); i++) {
    if (table[i]->GetCapabilityNumber() == capabilityNumber) {
      STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
          "() Found capability: " << *table[i]);
      return table[i];
    }
  }

  return NULL;
}

/*
H323Capability * H323Capabilities::FindCapability(const SString & formatName) const
{
    SDEFFUNC("H323Capabilities::FindCapability");

  unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
  STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
      "(" << formatName << ")");

  for (int i = 0; i < table.size(); i++) {
    PCaselessString str = table[i].GetFormatName();
    if (str.Find(formatName) != S_MAX_INDEX) {
      PTRACE(2, "H323\tFound capability: " << table[i]);
      return &table[i];
    }
  }

  return NULL;
}
*/

H323Capability * H323Capabilities::FindCapability(const H323Capability & capability) const
{
    SDEFFUNC("H323Capabilities::FindCapability2");

  unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
    STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
        "(" << capability << ")");

  for (H323CapabilitiesList::size_type i = 0; i < table.size(); i++) {
    if (*table[i] == capability) {
      STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
          "() Found capability: " << *table[i]);
      return table[i];
    }
  }

  return NULL;
}


H323Capability * H323Capabilities::FindCapability(const H245_Capability & cap, bool useEncryption) const
{
    SDEFFUNC("H323Capabilities::FindCapability3");

  unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
  STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
      "(" << cap.GetTagName() << ")");

  switch (cap.GetTag()) {
    case H245_Capability::e_receiveAudioCapability :
    case H245_Capability::e_transmitAudioCapability :
    case H245_Capability::e_receiveAndTransmitAudioCapability :
    {
      const H245_AudioCapability & audio = cap;
      return FindCapability(H323Capability::e_Audio, audio.GetTag(), useEncryption);
    }

    case H245_Capability::e_receiveVideoCapability :
    case H245_Capability::e_transmitVideoCapability :
    case H245_Capability::e_receiveAndTransmitVideoCapability :
    {
      const H245_VideoCapability & video = cap;
      return FindCapability(H323Capability::e_Video, video.GetTag(), useEncryption);
    }

    case H245_Capability::e_receiveDataApplicationCapability :
    case H245_Capability::e_transmitDataApplicationCapability :
    case H245_Capability::e_receiveAndTransmitDataApplicationCapability :
    {
      const H245_DataApplicationCapability & data = cap;
      return FindCapability(H323Capability::e_Data, data.m_application.GetTag(), useEncryption);
    }

    case H245_Capability::e_receiveUserInputCapability :
    case H245_Capability::e_transmitUserInputCapability :
    case H245_Capability::e_receiveAndTransmitUserInputCapability :
    {
      const H245_UserInputCapability & ui = cap;
      return FindCapability(H323Capability::e_UserInput, ui.GetTag(), useEncryption);
    }

    default :
      break;
  }

  return NULL;
}


H323Capability * H323Capabilities::FindCapability(const H245_DataType & dataType) const
{
    SDEFFUNC("H323Capabilities::FindCapability4");

  unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
  STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
      "(" << dataType.GetTagName() << ")");

  switch (dataType.GetTag()) {
    case H245_DataType::e_audioData :
      {
        const H245_AudioCapability & audio = dataType;
        return FindCapability(H323Capability::e_Audio, audio.GetTag());
      }

    case H245_DataType::e_videoData :
      {
        const H245_VideoCapability & video = dataType;
        return FindCapability(H323Capability::e_Video, video.GetTag());
      }

    case H245_DataType::e_data :
      {
        const H245_DataApplicationCapability & data = dataType;
        return FindCapability(H323Capability::e_Data, data.m_application.GetTag());
      }

    case H245_DataType::e_h235Media :
        {
          const H245_H235Media & h235Media = (const H245_H235Media &)dataType;
          const H245_AudioCapability & audio = (const H245_AudioCapability &)h235Media.m_mediaType;
          return FindCapability(H323Capability::e_Audio, audio.GetTag(), true);
        }

    default :
      break;
  }

  return NULL;
}


H323Capability * H323Capabilities::FindCapability(const H245_ModeElement & modeElement) const
{
    SDEFFUNC("H323Capabilities::FindCapability5");

    unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
    STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
        "(" << modeElement.m_type.GetTagName() << ")");

    switch (modeElement.m_type.GetTag())
    {
        case H245_ModeElementType::e_audioMode :
        {
            const H245_AudioMode & audio = modeElement.m_type;
            return FindCapability(H323Capability::e_Audio, audio.GetTag());
        }

        case H245_ModeElementType::e_videoMode :
        {
            const H245_VideoMode & video = modeElement.m_type;
            return FindCapability(H323Capability::e_Video, video.GetTag());
        }

        case H245_ModeElementType::e_dataMode :
        {
            const H245_DataMode & data = modeElement.m_type;
            return FindCapability(H323Capability::e_Data, data.m_application.GetTag(), true);
        }

        default :
            break;
    }

    return NULL;
}


H323Capability * H323Capabilities::FindCapability(H323Capability::MainTypes mainType,
                                                  unsigned subType,
                                                  bool useEncryption) const
{
  SDEFFUNC("H323Capabilities::FindCapability6");

  unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
  STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
      "(mainType=" << mainType << ", subtype=" << subType << ", encrypt=" << useEncryption << ")");

  for (H323CapabilitiesList::size_type i = 0; i < table.size(); i++) {
    if (table[i]->GetMainType() == mainType &&
        (subType == UINT_MAX || table[i]->GetSubType() == subType) &&
        table[i]->GetUseEncryption() == useEncryption) {
      STRACE2_INFO2(g_pH323Module->m_H323TraceId, this, ulConnId,
          "() Found capability: " << *table[i]);
      return table[i];
    }
  }

  return NULL;
}


void H323Capabilities::BuildPDU(H245_TerminalCapabilitySet & pdu) const
{
    SDEFFUNC("H323Capabilities::BuildPDU");

    unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId,
        "(pdu)");
    CAP_PRINT(__LINE__);

  pdu.IncludeOptionalField(H245_TerminalCapabilitySet::e_capabilityTable);

  bool addEncryptionInfo = false;
  int j = 0;
  int tableSize = table.size();
  pdu.m_capabilityTable.SetSize(tableSize);

  std::vector<H323Capability *> tempCaps;

  for (int i = 0; i < tableSize; i++) 
  {
      if (table[i]->GetUseEncryption() == true)
      {
          addEncryptionInfo = true;
        // don't add h323caps with encryption, just set flag to add mikey later on
      }

      // prevent to add caps twice - decide only on mainType and subType, don't check for encryption!
      bool found = false;
      for (unsigned ii = 0; ii < tempCaps.size() && !found; ii++)
      {
        H323Capability * pCap = tempCaps[ii];
        if (pCap->GetMainType() == table[i]->GetMainType() &&
            pCap->GetSubType()  == table[i]->GetSubType())
        {
            found = true;
        }
      }

      if (!found)
      {
        pdu.m_capabilityTable[j].m_capabilityTableEntryNumber = table[i]->GetCapabilityNumber();
        pdu.m_capabilityTable[j].IncludeOptionalField(H245_CapabilityTableEntry::e_capability);
        table[i]->OnSendingPDU(pdu.m_capabilityTable[j].m_capability);
        tempCaps.push_back(table[i]);
        j++;
      }
  }
  pdu.m_capabilityTable.SetSize(j);

  if ( !mikey.empty() && addEncryptionInfo )
  {
    H245_TerminalCapabilitySet & termCapSet = pdu;

    // add MIKEY
    if (!termCapSet.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityTable))
    {
        termCapSet.IncludeOptionalField(H245_TerminalCapabilitySet::e_capabilityTable);
    }
    int last = termCapSet.m_capabilityTable.GetSize();
    termCapSet.m_capabilityTable.SetSize(last+1);
    termCapSet.m_capabilityTable[last].m_capabilityTableEntryNumber = last+1; //! ???
    termCapSet.m_capabilityTable[last].IncludeOptionalField(H245_CapabilityTableEntry::e_capability);

    H245_Capability & cap = termCapSet.m_capabilityTable[last].m_capability;
    cap.SetTag(H245_Capability::e_h235SecurityCapability);

    H245_H235SecurityCapability & secCap = cap;
    H245_EncryptionAuthenticationAndIntegrity & encr = secCap.m_encryptionAuthenticationAndIntegrity;

    encr.IncludeOptionalField(H245_EncryptionAuthenticationAndIntegrity::e_genericH235SecurityCapability);
    H245_GenericCapability & genCap = encr.m_genericH235SecurityCapability;

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId, "() add mikey");

    genCap.m_capabilityIdentifier.SetTag(H245_CapabilityIdentifier::e_standard);
    SAsnObjectId & objectId = genCap.m_capabilityIdentifier;
    objectId.SetValue("0.0.8.235.0.3.76"); // MIKEY //! ??? => ok, H.235.7, Table 1

    genCap.IncludeOptionalField(H245_GenericCapability::e_nonCollapsing);
    genCap.m_nonCollapsing.SetSize(1);
    H245_GenericParameter & genParam = genCap.m_nonCollapsing[0];

    genParam.m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
    SAsnInteger & asnInt = genParam.m_parameterIdentifier;
    asnInt.SetConstraints(SAsnObject::FixedConstraint, 0, 127);
    
    //genParam.m_parameterIdentifier = 0; //! ???
    asnInt.SetValue(0);

    H245_ParameterValue & paramValue = genParam.m_parameterValue;
    paramValue.SetTag(H245_ParameterValue::e_octetString);

    SAsnOctetString & mikeyString = paramValue;

    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId, "() mikey-length: " << mikey.size());
    mikeyString.SetValue(mikey);

    secCap.m_mediaCapability = 1; // CapabilityTableEntryNumber; //! ???
  }

  // Set the sets of compatible capabilities
  pdu.IncludeOptionalField(H245_TerminalCapabilitySet::e_capabilityDescriptors);

  int outerSize = set.size();
  pdu.m_capabilityDescriptors.SetSize(outerSize);
  for (int outer = 0; outer < outerSize; outer++) {
    H245_CapabilityDescriptor & desc = pdu.m_capabilityDescriptors[outer];
    desc.m_capabilityDescriptorNumber = (unsigned)(outer + 1);
    desc.IncludeOptionalField(H245_CapabilityDescriptor::e_simultaneousCapabilities);
    int middleSize = set[outer]->size();
    desc.m_simultaneousCapabilities.SetSize(middleSize);
    for (int middle = 0; middle < middleSize; middle++) {
      H245_AlternativeCapabilitySet & alt = desc.m_simultaneousCapabilities[middle];
      int innerSize = (*set[outer])[middle]->size();
      alt.SetSize(innerSize);
      for (int inner = 0; inner < innerSize; inner++)
        alt[inner] = (*(*set[outer])[middle])[inner]->GetCapabilityNumber();
    }
  }

  SAssert((tableSize > 0) == (outerSize > 0), SLogicError);
}


bool H323Capabilities::Merge(const H323Capabilities & newCaps)
{
    SDEFFUNC("H323Capabilities::Merge");

    unsigned long ulConnId = const_cast<H323Capabilities *>(this)->GetConnID();
    STRACE2_INFO3(g_pH323Module->m_CodecTraceId, this, ulConnId,
        "(cap=" << newCaps << ", " << &newCaps << ")");
    CAP_PRINT(__LINE__);

  // Add any new capabilities not already in set.
  int i;
  for (i = 0; i < newCaps.GetSize(); i++) {
    if (!FindCapability(newCaps[i]))
      Copy(newCaps[i]);
  }

  // This should merge instead of just replacing the set.
//  set = newCaps.set;

    // The above would copy the outer vector but not the inner objects!
    // We do it here:
    const H323CapabilitiesSet & newSet = newCaps.set;
    int outerSize = newSet.size();
    set.SetSize(outerSize);
    for (int outer = 0; outer < outerSize; outer++)
    {
        int middleSize = newSet[outer]->size();
        set[outer]->SetSize(middleSize);
        for (int middle = 0; middle < middleSize; middle++)
        {
            int innerSize = (*newSet[outer])[middle]->size();
            for (int inner = 0; inner < innerSize; inner++)
            {
                for (H323CapabilitiesList::size_type cap = 0; cap < table.size(); cap++)
                {
                    if (table[cap]->GetCapabilityNumber() == newCaps[inner].GetCapabilityNumber())
                    {
                        (*set[outer])[middle]->push_back(&(*table[cap]));
                        break;
                    }
                }
            }
        }
    }

  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, ulConnId,
      "() Received capability set, is " << (table.empty() ? "rejected" : "accepted"));
    CAP_PRINT(__LINE__);
  return !table.empty();
}

/*
void H323Capabilities::Reorder(const PStringArray & preferenceOrder)
{
  if (preferenceOrder.IsEmpty())
    return;

  table.DisallowDeleteObjects();

  PINDEX preference = 0;

  for (PINDEX base = 0; base < table.GetSize(); base++) {
    PString codecName = preferenceOrder[preference];
    for (PINDEX idx = base; idx < table.GetSize(); idx++) {
      PCaselessString str = table[idx].GetFormatName();
      if (str.Find(codecName) != S_MAX_INDEX) {
        if (base != idx) // Move capability to the current position list.
          table.InsertAt(base, table.RemoveAt(idx));
        preference++;
        break;
      }
    }
    if (preference >= preferenceOrder.GetSize())
      break;
  }

  table.AllowDeleteObjects();
}


bool H323Capabilities::IsAllowed(const H323Capability & capability1,
                                 const H323Capability & capability2)
{
  // This needs more work, at present only sees if capability1 is in set at all.

  PINDEX outerSize = set.GetSize();
  for (PINDEX outer = 0; outer < outerSize; outer++) {
    PINDEX middleSize = set[outer].GetSize();
    for (PINDEX middle = 0; middle < middleSize; middle++) {
      PINDEX innerSize = set[outer][middle].GetSize();
      for (PINDEX inner = 0; inner < innerSize; inner++)
        if (capability1.GetCapabilityNumber() == set[outer][middle][inner].GetCapabilityNumber())
          return TRUE;
    }
  }

  return FALSE;
}
*/

/////////////////////////////////////////////////////////////////////////////

