/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323 Endpoint
 
    Comments: H.323 protocol handler

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxGate.root/H323/source/h323ep.cpp                $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 10.08.05 16:46                                               $
   $Modtime:: 10.08.05 16:45                                               $
  $Revision:: 3                                                            $
  $Workfile:: h323ep.cpp                                                   $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "h323.h"
#include "h323pdu.h"

using namespace std;
using namespace Tools;

// UF: class H225CallThread moved to h323.h


#define new SNEW



/////////////////////////////////////////////////////////////////////////////

H225CallThread::H225CallThread(H323EndPoint & endpoint,
                               H323Connection & c,
                               H323Transport & t,
                               const SString & r)
  : SThread(endpoint.GetSignallingThreadStackSize(), NoAutoDeleteThread),
    connection(c),
    transport(t),
    remoteParty(r)
{
    // use connection's ID for tracing
    SetConnID( c.GetConnID() );

    transport.AttachThread(this);
    if (!g_pH323Module->m_bUsingSwtcl)
    {
        // Make Resume() only if not using SwTCL!
        Resume();
    }
}


void H225CallThread::Main()
{
    SDEFFUNC("H225CallThread::Main");

    STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, GetConnID(),
        "() Started thread (H225CallThread)");

  if (connection.SendSignalSetup(remoteParty))
    connection.HandleSignallingChannel();
  else
    connection.ClearCall();

    STRACE2_INFO3(g_pH323Module->m_H225TraceId, this, GetConnID(),
        "() Finished thread (H225CallThread)");
}


////////////////////////////////////////////////////////////////////
// SH323TcpConnection

bool SH323TcpConnection::ProcessRead()
{
    SDEFFUNC("SH323TcpConnection::ProcessRead");
    
    if( m_ClosedFromApp )
    {
        // read all the rest from socket and drop it
        BYTE dummyBuffer[1024];
        if( Read(dummyBuffer,1024)==false )
        {
            unsigned lastError = GetLastError();
            if( lastError==0 || lastError==WSAEWOULDBLOCK )
                return true;

            return false;
        }
        return true;
    }

    try
    {
        H323TransportTCP* pTransport=m_pTransport;

        if( pTransport==NULL ) 
        {
            m_pTransport = new H323TransportTCP(m_endpoint,true/*is incoming*/);

            // no autoDelete for Tls connections!
            if (m_pTransport->Open(this, false))
            {
                SWaitAndSignal lock(*(m_pTransport->m_pRefMutex));
                return m_pTransport->ProcessRead();
            }
            else
            {
                STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, GetConnectionID(), "() failed to open signallingChannel " << this);
                return false;
            }    
        }
        else
        {
            SWaitAndSignal lock(*(pTransport->m_pRefMutex));

            return pTransport->ProcessRead();
        }
    }
    catch (...)
    {
        STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, GetConnectionID(), "ProcessRead failed - exception");
    }
    
    return false;
}



////////////////////////////////////////////////////////////////////
// SH323TlsConnection

bool SH323TlsConnection::ProcessRead()
{
    SDEFFUNC("SH323TlsConnection::ProcessRead");
    bool bReadRet = true;

    try
    {
        H323TransportTls* pTransport=m_pTransport;

        if( pTransport==NULL ) 
        {
            // m_pTransport is NULL --> means there is no connection associated to this connection
            //                      --> read first message using an provisional transport object
            //                          this provisional one becomes actual if read message is a setup message
            if( m_pProvisionalTransport==NULL)
            {
                m_pProvisionalTransport = new H323TransportTls(m_endpoint,true/*is incoming*/);

                // no autoDelete for Tls connections!
                if (m_pProvisionalTransport->Open(this, false))
                {
                    bReadRet = m_pProvisionalTransport->ProcessRead();
                }
                else
                {
                    STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, GetConnectionID(), "() failed to open signallingChannel " << this);
                    bReadRet = false;
                }
            }
            else
            {
                //note: provisional transport does not need lock as there is just the listener thread using this object
                bReadRet = m_pProvisionalTransport->ProcessRead();
            }

            // delete transport object if ProcessRead() failed
            if (bReadRet == false)
            {
                STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnectionID(), "() deleting provisional transport object");
                delete m_pProvisionalTransport;
                m_pProvisionalTransport=NULL;
            }
        }
        else
        {
            // call ProcessRead() under mutex protection
            // but don't delete the object under mutex protection
            {
                SWaitAndSignal lock(*(pTransport->m_pRefMutex.RawPtr()));

                bReadRet = pTransport->ProcessRead();
            }

            // delete transport object if ProcessRead() failed
            // the transport object mustn't be deleted while pTransport->m_pRefMutex is in use
            if (bReadRet == false)
            {
                STRACE2_INFO(g_pH323Module->m_H323TraceId, this, GetConnectionID(), "() deleting transport object");
                delete m_pTransport;
                m_pTransport=NULL;
            }
        }

        // always keep tls connection but transport may be deleted on any error
        return true;
    }
    catch (...)
    {
        STRACE2_ERROR( g_pH323Module->m_H323TraceId, this, GetConnectionID(), "ProcessRead failed - exception");
    }
    
    return false;
}



/////////////////////////////////////////////////////////////////////////////

// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H323EndPoint::H323EndPoint(SH323TcpConnectionFactory& tcpConFab, SH323TlsConnectionFactory& tlsConFab)
  : localUserName(SProcess::Current().GetUserName()),
    soundChannelPlayDevice(SSoundChannel::GetDefaultDevice(SSoundChannel::Player)),
    soundChannelRecordDevice(SSoundChannel::GetDefaultDevice(SSoundChannel::Recorder)),
    signallingChannelCallTimeout(0, 0, 3),  // Minutes
    masterSlaveDeterminationTimeout(0, 30), // Seconds
    capabilityExchangeTimeout(0, 30),       // Seconds
    logicalChannelTimeout(0, 30),           // Seconds
    requestModeTimeout(0, 30),              // Seconds
#ifdef ROUND_TRIP_DELAY
    roundTripDelayTimeout(0, 10),           // Seconds
//    roundTripDelayRate(0, 0, 1),            // Minutes
    roundTripDelayRate(0, 15),              // Seconds
#else
    roundTripDelayTimeout(0, 30),           // Seconds
#endif // ROUND_TRIP_DELAY
    gatekeeperRequestTimeout(0, 5),         // Seconds
    rasRequestTimeout(0, 5),                // Seconds
    rasRegistrationTTL(300),                // Seconds
    rasReRegisterTimeout(60),               // Seconds
    callTransferT1(0,10),                   // Seconds
    callTransferT2(0,10),                   // Seconds
    callTransferT3(0,10),                   // Seconds
    callTransferT4(0,10)                    // Seconds
#ifdef _DEBUG // add name to H323 endpoint connections mutex
  , connectionsMutex(Tools::SString(Tools::SString::Printf, "H323Endpoint::connectionsMutex[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
  , m_tcpConFactory(tcpConFab)
  , m_tlsConFactory(tlsConFab)
#pragma warning(default:4355)
{
  SDEFFUNC("H323EndPoint::H323EndPoint");

  terminalType                    = e_TerminalOnly;
  maxAudioDelayJitter             = 50;      // milliseconds
  initialBandwidth                = 100000;  // Standard 10base LAN in 100's of bits/sec
#ifdef ROUND_TRIP_DELAY
  clearCallOnRoundTripFail        = false;
  m_bSendRoundTripDelay           = true;
#endif // ROUND_TRIP_DELAY

  m_bIsOverlappedSendingEnabled   = false;
  m_bUseCallIdentifier            = false;
  m_bInGKDiscovery                = false;
  m_uTA1100RegisterTTLSeconds     = 0;
  
  // The following values were assigned for Swyx Communications AG
  // by German Telecommunication Administration (RegTP) on 2000-11-29.
  // Usage as recommended by RegTP and T.35. 
  t35CountryCode                  = 4;       // Country code for Germany
  t35Extension                    = 129;     // Manufacturer Code (T.35)
  manufacturerCode                = 6;       // Manufacturer oriented Code (T.35)

  rtpPortBase                     = 5000;

  masterSlaveDeterminationRetries = 10;
  gatekeeperRequestRetries        = 2;
  rasRequestRetries               = 2;
  rasMaxRequestRetries            = 3;      // maximum number of full RRQ retries before considering a gatekeeper connection broken

  signallingThreadStackSize       = 30000;
  controlThreadStackSize          = 30000;
  logicalThreadStackSize          = 30000;
  listenerThreadStackSize         = 30000;
  rasThreadStackSize              = 30000;

  gatekeeper                      = NULL;
  alternativeGatekeeper           = NULL;

//  connectionsActive.DisallowDeleteObjects();
//  connectionsToBeCleaned.DisallowDeleteObjects();

  srand((unsigned)time(NULL)+clock());

  STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, "()");
}


H323EndPoint::~H323EndPoint()
{
  SDEFFUNC("H323EndPoint::~H323EndPoint");

  // Shut down the listeners as soon as possible to avoid race conditions
  listeners.clear();

  // Clear any pending calls on this endpoint
  ClearAllCalls();

  // Clean up any connections that the cleaner thread missed
  CleanUpConnections();

  // And shut down the gatekeeper (if there was one)
  RemoveGatekeeper();

  // And shut down the gatekeeper (if there was one)
  RemoveAlternativeGatekeeper();

  STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, 0, "()");
}


void H323EndPoint::SetEndpointTypeInfo(H225_EndpointType & info, const SString sLinkName) const
{
  info.IncludeOptionalField(H225_EndpointType::e_vendor);
  SetVendorIdentifierInfo(info.m_vendor);

///*hk: this is to add supported tunnelled protocols 
  info.IncludeOptionalField(H225_EndpointType::e_supportedTunnelledProtocols);
  SetSupportedProtocolsToTunnelledProtocol(info.m_supportedTunnelledProtocols);


  switch (terminalType) {
    case e_TerminalOnly :
    case e_TerminalAndMC :
      info.IncludeOptionalField(H225_EndpointType::e_terminal);
      break;
    case e_GatewayOnly :
    case e_GatewayAndMC :
    case e_GatewayAndMCWithDataMP :
    case e_GatewayAndMCWithAudioMP :
    case e_GatewayAndMCWithAVMP :
      info.IncludeOptionalField(H225_EndpointType::e_gateway);
      break;
    case e_GatekeeperOnly :
    case e_GatekeeperWithDataMP :
    case e_GatekeeperWithAudioMP :
    case e_GatekeeperWithAVMP :
      if(!sLinkName.empty()&&IsRemoteSideLinkToThirdParty(sLinkName))
      {
          // if SwyxLink goes to foreign gatekeeper, the foreign gatekeeper don't understand the
          // fake we do for SwyxLinks.
          info.IncludeOptionalField(H225_EndpointType::e_gateway);
      }
      else
      {
          // for SwyxLink we fake the terminal type 'gatekeeper' in order for the SwyxServer to 
          // recognize a registration from a link
          info.IncludeOptionalField(H225_EndpointType::e_gatekeeper);
      }
      break;
    case e_MCUOnly :
    case e_MCUWithDataMP :
    case e_MCUWithAudioMP :
    case e_MCUWithAVMP :
      info.IncludeOptionalField(H225_EndpointType::e_mcu);
      info.m_mc = true;
  }
}


void H323EndPoint::SetVendorIdentifierInfo(H225_VendorIdentifier & info) const
{
  SetH221NonStandardInfo(info.m_vendor);

  info.IncludeOptionalField(H225_VendorIdentifier::e_productId);
  info.m_productId = SProcess::Current().GetManufacturer() + " " + SProcess::Current().GetName();

  info.IncludeOptionalField(H225_VendorIdentifier::e_versionId);
  info.m_versionId = SProcess::Current().GetVersion(true);
}

///*hk: this is to add supported tunnelled protocols 
void H323EndPoint::SetSupportedProtocolsToTunnelledProtocol(H225_ArrayOf_TunnelledProtocol & info) const
{    
    info.SetSize(1);

    H225_TunnelledProtocol& tunProt = info[0];

    SAsnObjectId objectId;
    objectId.SetValue("1.3.12.9"); // tunnelled object ID

    tunProt.m_id.SetTag(H225_TunnelledProtocol_id::e_tunnelledProtocolObjectID);
    ((SAsnObjectId &)tunProt.m_id)=objectId;
}


void H323EndPoint::SetH221NonStandardInfo(H225_H221NonStandard & info) const
{
  info.m_t35CountryCode = t35CountryCode;
  info.m_t35Extension = t35Extension;
  info.m_manufacturerCode = manufacturerCode;
}


H323Capability * H323EndPoint::FindCapability(const H245_Capability & cap) const
{
  return capabilities.FindCapability(cap);
}    


H323Capability * H323EndPoint::FindCapability(const H245_DataType & dataType) const
{
  return capabilities.FindCapability(dataType);
}


H323Capability * H323EndPoint::FindCapability(H323Capability::MainTypes mainType,
                                              unsigned subType, 
                                              bool useEncryption)  const
{
  return capabilities.FindCapability(mainType, subType, useEncryption);
}


void H323EndPoint::AddCapability(H323Capability * capability)
{
  capabilities.Add(capability);
}


void H323EndPoint::SetCapability(int descriptorNum,
                                 int simultaneousNum,
                                 H323Capability * capability)
{
    capabilities.SetCapability(descriptorNum, simultaneousNum, capability);
}


bool H323EndPoint::SetGatekeeper(const Tools::SString    & address,
                                 const Tools::SString    & szH323Alias,
                                 const Tools::SIpAddress & transportAddress,
                                 const bool                bIsAlternativeGatekeeper,
                                 const int                 iLineId,
                                 const int                 iNumOfChannels)
{
    if (bIsAlternativeGatekeeper)
    {
        if (!RemoveAlternativeGatekeeper(H225_UnregRequestReason::e_reregistrationRequired))
            return false;
    }
    else
    {
        if (!RemoveGatekeeper(H225_UnregRequestReason::e_reregistrationRequired))
            return false;
    }

    H323Gatekeeper * gk = CreateGatekeeper(transportAddress);
    if ( gk==NULL )
        return false;

    m_bInGKDiscovery = true;
    if ( gk->Discover(address) )
    {
        m_bInGKDiscovery = false;
        gk->UnregistrationRequest(szH323Alias, H225_UnregRequestReason::e_reregistrationRequired);

        gk->IncRef(__FILE__,__LINE__);
        if (bIsAlternativeGatekeeper)
            gk->SetAlternativeGatekeeper();
        if ( gk->RegistrationRequest(szH323Alias, false, iLineId, iNumOfChannels, true) )
        {
            // remove temp reference for RegistrationRequests()
            gk->DecRef(__FILE__,__LINE__);

            SWaitAndSignal lock(m_EndpointGkCritSec);
            if (bIsAlternativeGatekeeper)
                alternativeGatekeeper = gk; // the needed reference was done in CreateGatekeeper()
            else
                gatekeeper = gk;
            return true;
        }
        else
            gk->DecRef(__FILE__,__LINE__);
    }
    m_bInGKDiscovery = false;

    gk->Shutdown();
    gk->DecRef(__FILE__,__LINE__); // GK obj should be deleted NOW
    return false;
}

H323Gatekeeper * H323EndPoint::GetGatekeeper()
{
    SWaitAndSignal lock(m_EndpointGkCritSec);

    if (alternativeGatekeeper!=NULL)
    {
        alternativeGatekeeper->IncRef(__FILE__,__LINE__);
        return alternativeGatekeeper;
    }
    else if (gatekeeper!=NULL)
    {
        gatekeeper->IncRef(__FILE__,__LINE__);
        return gatekeeper;
    }

    return NULL;
}


bool H323EndPoint::DiscoverGatekeeper(const Tools::SIpAddress & transportAddress)
{
  if (!RemoveGatekeeper(H225_UnregRequestReason::e_reregistrationRequired))
    return false;

  H323Gatekeeper * gk = CreateGatekeeper(transportAddress);

  if (gk->Discover("") )
  {
    gk->UnregistrationRequest(Tools::SString(), H225_UnregRequestReason::e_reregistrationRequired);
    if ( gk->RegistrationRequest(Tools::SString()) )
    {
      SWaitAndSignal lock(m_EndpointGkCritSec);

      gk->IncRef(__FILE__,__LINE__);
      gatekeeper = gk;
      return true;
    }
  }

  gk->Shutdown();
  gk->DecRef(__FILE__,__LINE__); // GK obj should be deleted NOW
  return false;
}


H323Gatekeeper * H323EndPoint::CreateGatekeeper(const Tools::SIpAddress & transportAddress)
{
    H323Gatekeeper * pGk = new H323Gatekeeper(*this, transportAddress);
    if (!SBAD_READPTR(pGk))
        pGk->IncRef(__FILE__,__LINE__);

    return pGk;
}


bool H323EndPoint::RemoveGatekeeper(int reason)
{
    bool bResult = true;

    if (gatekeeper == NULL)
        return true;

    ClearAllCalls();

    if (!gatekeeper->UnregisterAllClients(reason))
        bResult = false;

    // protect pointer to gatekeeper objects
    SWaitAndSignal lock(m_EndpointGkCritSec);

    gatekeeper->DecRef(__FILE__,__LINE__);
    gatekeeper = NULL;
  
    return bResult;
}


bool H323EndPoint::RemoveAlternativeGatekeeper(int reason)
{
    bool bResult = true;

    if (alternativeGatekeeper == NULL)
        return true;

    ClearAllCalls();

    if (!alternativeGatekeeper->UnregisterAllClients(reason))
        bResult = false;

    // protect pointer to gatekeeper objects
    SWaitAndSignal lock(m_EndpointGkCritSec);

    alternativeGatekeeper->DecRef(__FILE__,__LINE__);
    alternativeGatekeeper = NULL;

    return bResult;
}


bool H323EndPoint::OnGatekeeperRegistrationLost(int iClientId)
{
    SDEFFUNC("H323EndPoint::OnGKRegLost");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0,
        "() Gatekeeper registration for client id=" << iClientId << " lost -> try to reregister...");
    return true;
}


bool H323EndPoint::OnGatekeeperRegistrationDone(int iClientId)
{
    SDEFFUNC("H323EndPoint::OnGKRegDone");

    STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0,
        "() Gatekeeper reregistration for client id=" << iClientId << " done!");
    return true;
}


bool H323EndPoint::StartListener(H323Listener * listener)
{
  if (!listener->Open())
    return false;

  listeners.push_back(listener);
  listener->Resume();
  return true;
}



H323Connection * H323EndPoint::MakeCall(const SString & remoteParty,
                                        H323Transport * transport,
                                        SString & token,
                                        const unsigned long ulConnId)
{
  SDEFFUNC("H323EndPoint::MakeCall");

  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, ulConnId,
    "() Making call to: " << remoteParty);

  connectionsMutex.Wait();

  unsigned lastReference;
  do
  {
    lastReference = Q931::GenerateCallReference();
    token = BuildConnectionToken(*transport, lastReference, false);
  } while (connectionsActive.find(token) != connectionsActive.end());

  H323Connection * connection = CreateConnection(lastReference, token, ulConnId);
  if (connection == NULL)
  {
      STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, connection->GetConnID(),
          "() FAILED to create new connection for token: " << token);
      connectionsMutex.Signal();
      return NULL;
  }
  connectionsActive.insert(pair<const SString, H323Connection *>(token, connection));

  connectionsMutex.Signal();

  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection->GetConnID(),
      "() Created new connection: " << token);

  connection->AttachSignalChannel(transport);
  new H225CallThread(*this, *connection, *transport, remoteParty);
  return connection;
}


bool H323EndPoint::ClearCall(const SString & token,
                             H323Connection::TerminationReason terminationReason)
{
    SDEFFUNC("H323EndPoint::ClearCall");

  /*The hugely multi-threaded nature of the H323Connection objects means that
    to avoid many forms of race condition, a call is cleared by moving it from
    the "active" call dictionary to a list of calls to be cleared that will be
    processed by a background thread specifically for the purpose of cleaning
    up cleared connections. So that is all that this function actually does.
    The real work is done in the H323ConnectionsCleaner thread.
   */

  SWaitAndSignal wait(connectionsMutex);

  // Find the connection by token, callid or conferenceid
  H323Connection * connection = FindConnectionWithoutLocks(token);
  if (connection == NULL)
  {
//      STRACE(STrace::Error, g_pH323Module->m_H323TraceId,
//             "Cannot find connection '" << token << "'");
      return false;
  }
  STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, connection->GetConnID(),
        "() Deleting connection '" << token << "' from active list.");

  // Remove from active dictionary and put into clean up list
  connectionsActive.erase(connection->GetCallToken());
  connectionsToBeCleaned.push_back(connection);

  // Now set reason for the connection close
  connection->SetTerminationReason(terminationReason);

  WakeUpListener();

  return true;
}

bool H323EndPoint::ClearCall( H323Connection * connection )
{
  SWaitAndSignal wait(connectionsMutex);

  if ( !SBAD_READPTR(connection) )
  {
    // Check if conneciton is still in list of active connections by token
    // in this case we will use a different mechanism!
    H323Connection * c = FindConnectionWithoutLocks(connection->GetCallToken());
    if ( c != 0 )
        ClearCall(c->GetCallToken());
    else
    {
        connectionsToBeDeleted.push_back(connection);
    }

    WakeUpListener();

    return true;
  }

  return false;
}

void H323EndPoint::ClearAllCalls(H323Connection::TerminationReason terminationReason)
{
    SDEFFUNC("H323EndPoint::ClearAllCalls");

    /*The hugely multi-threaded nature of the H323Connection objects means that
    to avoid many forms of race condition, a call is cleared by moving it from
    the "active" call dictionary to a list of calls to be cleared that will be
    processed by a background thread specifically for the purpose of cleaning
    up cleared connections. So that is all that this function actually does.
    The real work is done in the H323ConnectionsCleaner thread.
   */

  unsigned uNumOfConnectionToClear = 0;
  unsigned uNumOfRetries           = 0; 

  // aquire connection mutex to block all other operations
  connectionsMutex.Wait();

  // Remove all connections from the active list first
  for (H323ConnectionDict::iterator posDict = connectionsActive.begin();
                                    posDict != connectionsActive.end(); ++posDict)
  {
      if ((*posDict).second != NULL)
      {
          ++uNumOfConnectionToClear;
          connectionsToBeCleaned.push_back((*posDict).second);
      }
  }
  connectionsActive.clear();

  // Now set reason for the connection close
  for (H323ConnectionList::iterator posList = connectionsToBeCleaned.begin();
                                    posList != connectionsToBeCleaned.end(); ++posList)
  {
      if ((*posList) != NULL)
      {
          (*posList)->SetTerminationReason(terminationReason);
      }
  }

  if ( uNumOfConnectionToClear > 0 )
  {
      STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0,
            "() Clean up " << uNumOfConnectionToClear << " connections");
  }
  else
  {
      STRACE2_INFO(g_pH323Module->m_H323TraceId, this, 0,
            "() NO connections to clear");
  }

  // free connection mutex to ensure correct deletion process
  connectionsMutex.Signal();
}


void H323EndPoint::OnConnectionDeletion(H323Connection * pConnection)
{
    SDEFFUNC("H323EndPoint::OnCnnctionDeletion");

    if (SBAD_READPTR(pConnection))
    {
        STRACE2_ERROR(g_pH323Module->m_H323TraceId, this, 0,
            "() Invalid connection pointer!");

        return;
    }

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, pConnection->GetConnID(),
        "() Cleaning connection: " << pConnection);

    OnConnectionCleared(*pConnection, pConnection->GetCallToken());
        
    pConnection->CleanUpOnTermination();

    // The following function deletes the H323Connection.
    // It is a virtual method that could be overloaded by a derived
    // class of H323EndPoint but the implementation in the base
    // class H323EndPoint should exist because e.g. CleanUpConnections
    // may be called when the derived instance does not exist
    // any longer. (Destructor of a derived class is called first!)
    DestroyConnection(pConnection);

    return;
}


void H323EndPoint::CleanUpConnections()
{
    SDEFFUNC("H323EndPoint::CleanUpConnections");

  H323ConnectionList connections;

  connectionsMutex.Wait();

  // Now get all the connections that have been closed
  while (connectionsToBeCleaned.size() > 0)
  {
      connections.push_back(*connectionsToBeCleaned.begin());
      connectionsToBeCleaned.erase(connectionsToBeCleaned.begin());
  }

  connectionsMutex.Signal();

  // Now, outside the connectionsMutex, we can clean up connections
  for (H323ConnectionList::iterator posList = connections.begin();
                                    posList != connections.end(); ++posList)
  {
      H323Connection *pConnection = (*posList);
      if (pConnection != NULL)
      {

          Tools::SAutoRefPtr<SRefCriticalSection> pRefMutex = pConnection->GetRefMutex();

          if( pRefMutex!=NULL )
          {
             // call OnConnectionDeletion with lock
             SWaitAndSignal lock(*pRefMutex);

             OnConnectionDeletion(pConnection);
          }
          else
          {
             // call OnConnectionDeletion without lock
             OnConnectionDeletion(pConnection);
          }
      }
  }

  // Going out of scope does the delete of all this connections

  // Now work on all the connections that have been queued for deletetion
  connectionsMutex.Wait();

  H323ConnectionList::iterator conDelIter;
  for ( conDelIter = connectionsToBeDeleted.begin();
        conDelIter != connectionsToBeDeleted.end();
        ++conDelIter )
  {
    H323Connection *pConnection = (*conDelIter);
    connectionsMutex.Signal();
    if ( !SBAD_READPTR(pConnection) )
    {
        pConnection->DecRef(__FILE__, __LINE__);
    }
    connectionsMutex.Wait();
  }
  connectionsToBeDeleted.clear();

  connectionsMutex.Signal();
}


H323Connection * H323EndPoint::SetupTransfer(const SString & oldToken,
                                             const SString & callIdentity,
                                             const SString & remoteParty,
                                             SString & newToken,
                                             void * userData)
{
    // 'no op' in base class!
    return 0;
}

void H323EndPoint::TransferCall(const SString & token, 
                                const SString & remoteParty,
                                const SString & callIdentity)
{
  H323Connection * connection = FindConnectionWithLock(token);
  if (connection != NULL) {
    connection->TransferCall(remoteParty, callIdentity);
    connection->Unlock();
  }
}

void H323EndPoint::ConsultationTransfer(const SString & primaryCallToken,   
                                        const SString & secondaryCallToken)
{
  H323Connection * secondaryCall = FindConnectionWithLock(secondaryCallToken);
  if (secondaryCall != NULL) {
    secondaryCall->ConsultationTransfer(primaryCallToken);
    secondaryCall->Unlock();
  }
}


void H323EndPoint::OnConnectionEstablished(H323Connection & /*connection*/,
                                           const SString & /*token*/)
{
}


void H323EndPoint::OnConnectionCleared(H323Connection & /*connection*/,
                                       const SString & /*token*/)
{
}


H323Connection * H323EndPoint::FindConnectionWithLock(const SString & token)
{
  SWaitAndSignal wait(connectionsMutex);

  H323Connection * connection = FindConnectionWithoutLocks(token);
  if (connection == NULL)
    return NULL;

  // Do not lock or allow access if connection is in the process of
  // being cleaned up.
  if (   find(connectionsToBeCleaned.begin(), connectionsToBeCleaned.end(), connection)
      != connectionsToBeCleaned.end())
    return NULL;

  connection->Lock();
  return connection;
}


H323Connection * H323EndPoint::FindConnectionWithoutLocks(const SString & token)
{
    if (token.empty())
      return NULL;

    H323Connection * conn_ptr = NULL;
    H323ConnectionDict::iterator ConIter = connectionsActive.find(token);
    if ( ConIter != connectionsActive.end())
    {
        conn_ptr = (*ConIter).second;
    }

    if (conn_ptr != NULL)
      return conn_ptr;
/*
    UF, 07.08.2001:
    Until now the following was never used because the token was not based on
    call identifier or conference identifier. Commented it out to prevent
    unnecessary searches and possible access violations in 'GUID.AsString()'.

    H323ConnectionDict::iterator pos;
    for (pos = connectionsActive.begin(); pos != connectionsActive.end(); ++pos)
    {
        if ((*pos).second != NULL)
        {
            H323Connection & conn = *(*pos).second;
            if (conn.GetCallIdentifier().AsString() == token)
            {
                return &conn;
            }
        }
    }

    for (pos = connectionsActive.begin(); pos != connectionsActive.end(); ++pos)
    {
        if ((*pos).second != NULL)
        {
            H323Connection & conn = *(*pos).second;
            if (conn.GetConferenceIdentifier().AsString() == token)
            {
                return &conn;
            }
        }
    }
*/
    return NULL;
}


void H323EndPoint::DestroyConnection(H323Connection * pConnection)
{
    SDEFFUNC("H323EndPoint::DestroyConnection");

    STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, pConnection->GetConnID(),
        "(" << pConnection << ")");

    if (pConnection)
    {
        pConnection->Delete();
        delete pConnection;
        pConnection = 0;
    }

} // End of DestroyConnection


bool H323EndPoint::GatekeeperConnectionIsBroken(int iClientID)
{
    SDEFFUNC("H323Endpoint::GKConBroken");

    STRACE2_WARN(g_pH323Module->m_H323TraceId, this, 0,
                 "() Gatekeeper connection broken on client ID " << iClientID );

    return true;
} // End of GatekeeperConnectionIsBroken


SString H323EndPoint::BuildConnectionToken(const H323Transport & transport,
                                           unsigned callReference,
                                           bool fromRemote)
{
  SString token;

  if (fromRemote)
  {
    // UF 06.09.2001:
    // Changed to simple creation of address string 'ip:IPAddress:Port' because
    // name service resolution is called when transport.GetRemoteAddress() is used
    // which may take too much time (primarily when using Linux)!
    token = transport.GetRemoteAddressString();
  }
  else
    token = "ip:localhost";

  token.sprintf("%s/%u", token.c_str(), callReference);

  return token;
}


H323Connection * H323EndPoint::GetConnection(const H323Transport & transport,
                                             unsigned callReference,
                                             bool fromRemote)
{
  SDEFFUNC("H323EndPoint::GetConnection");

  SString token = BuildConnectionToken(transport, callReference, fromRemote);

  SWaitAndSignal wait(connectionsMutex);

  if (connectionsActive.find(token) != connectionsActive.end())
    return connectionsActive[token];

  H323Connection * connection = CreateConnection(callReference, token, 0);
  if (connection == NULL)
    return NULL;

  connectionsActive.insert(pair<const SString, H323Connection *>(token, connection));

  STRACE2_INFO(g_pH323Module->m_H323TraceId, this, connection->GetConnID(),
      "() Created new connection: " << token);

  return connection;
}


bool H323EndPoint::OnStartLogicalChannel(H323Connection & /*connection*/,
                                         H323Channel & /*channel*/)
{
  return true;
}


void H323EndPoint::OnClosedLogicalChannel(H323Connection & /*connection*/,
                                          const H323Channel & /*channel*/)
{
}


bool H323EndPoint::OpenAudioChannel(H323Connection & /*connection*/,
                                    bool isEncoding,
                                    unsigned bufferSize,
                                    H323AudioCodec & codec)
{
    SDEFFUNC("H323EndPoint::OpenAudioChannel");

  SString deviceName;
  if (isEncoding)
    deviceName = GetSoundChannelRecordDevice();
  else
    deviceName = GetSoundChannelPlayDevice();

  SSoundChannel * soundChannel = new SSoundChannel;
  soundChannel->SetBuffers(bufferSize, 8);

  if (soundChannel->Open(deviceName,
                         isEncoding ? SSoundChannel::Recorder
                                    : SSoundChannel::Player,
                         1, 8000, 16)) {
    STRACE2_INFO(g_pH323Module->m_CodecTraceId, this, 0,
        "() Opened sound channel \"" << deviceName
         << "\" for " << (isEncoding ? "record" : "play") << "ing.");
    return codec.AttachChannel(soundChannel);
  }

  STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
      "() Could not open sound channel \"" << deviceName
      << "\" for " << (isEncoding ? "record" : "play")
      << "ing: " << soundChannel->GetErrorText());

  delete soundChannel;
  return false;
}


bool H323EndPoint::OpenVideoDevice(H323Connection & /*connection*/,
                                   bool isEncoding,
                                   H323VideoCodec & /*codec*/)
{
    SDEFFUNC("H323EndPoint::OpenVideoDevice");

    STRACE2_ERROR(g_pH323Module->m_CodecTraceId, this, 0,
        "() Could not open video channel for "
        << (isEncoding ? "captur" : "display")
        << "ing: not yet implemented");
    return false;
}


void H323EndPoint::OnUserInputString(H323Connection & /*connection*/,
                                     const SString & /*value*/)
{
}


void H323EndPoint::OnUserInputTone(H323Connection & connection,
                                   const SString & tone,
                                   unsigned /*duration*/,
                                   unsigned /*logicalChannel*/,
                                   unsigned /*rtpTimestamp*/)
{
  connection.OnUserInputString(tone);
}


void H323EndPoint::SetLocalUserName(const SString & name)
{
  localUserName = name;
}


bool H323EndPoint::SetSoundChannelPlayDevice(const SString & name)
{
    // Search for device name
    SStringVector::iterator pos;
    for (pos =  SSoundChannel::GetDeviceNames(SSoundChannel::Player).begin();
         pos != SSoundChannel::GetDeviceNames(SSoundChannel::Player).end();
         ++pos)
    {
        if ((*pos).c_str() == name)
        {
            // Device name found, take this name
            soundChannelPlayDevice = name;
            return true;
        }
    }
    // No device name found, use default
    return false;
}


bool H323EndPoint::SetSoundChannelRecordDevice(const SString & name)
{
    // Search for device name
    SStringVector::iterator pos;
    for (pos =  SSoundChannel::GetDeviceNames(SSoundChannel::Recorder).begin();
         pos != SSoundChannel::GetDeviceNames(SSoundChannel::Recorder).end();
         ++pos)
    {
        if ((*pos).c_str() == name)
        {
            // Device name found, take this name
            soundChannelRecordDevice = name;
            return true;
        }
    }
    // No device name found, use default
    return false;
}


bool H323EndPoint::IsTerminal() const
{
  switch (terminalType) {
    case e_TerminalOnly :
    case e_TerminalAndMC :
      return true;

    default :
      return false;
  }
}


bool H323EndPoint::IsGateway() const
{
  switch (terminalType) {
    case e_GatewayOnly :
    case e_GatewayAndMC :
    case e_GatewayAndMCWithDataMP :
    case e_GatewayAndMCWithAudioMP :
    case e_GatewayAndMCWithAVMP :
      return true;

    default :
      return false;
  }
}


bool H323EndPoint::IsGatekeeper() const
{
  switch (terminalType) {
    case e_GatekeeperOnly :
    case e_GatekeeperWithDataMP :
    case e_GatekeeperWithAudioMP :
    case e_GatekeeperWithAVMP :
      return true;

    default :
      return false;
  }
}


bool H323EndPoint::IsMCU() const
{
  switch (terminalType) {
    case e_MCUOnly :
    case e_MCUWithDataMP :
    case e_MCUWithAudioMP :
    case e_MCUWithAVMP :
      return true;

    default :
      return false;
  }
}


bool H323EndPoint::IsE164(const SString & number) const
{
    static const char E164StringSet[] = "1234567890*#";

    if (number.find_first_not_of(E164StringSet) == SString::npos)
        return true;
    return false;
}


/////////////////////////////////////////////////////////////////////////////
