/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: H.323 Negotiation
 
    Comments: H.323 PDU definitions

  --------------------------------------------------------------------------

   $Archive:: /Server/SwyxSrv v4.3.x.x.root/H323/source/h323neg.cpp        $
    $Author:: Ralf.bonenkamp                                               $
      $Date:: 12.09.04 14:30                                               $
   $Modtime:: 12.09.04 14:29                                               $
  $Revision:: 47                                                           $
  $Workfile:: h323neg.cpp                                                  $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "h323neg.h"

using namespace std;
using namespace Tools;

#define new SNEW


/////////////////////////////////////////////////////////////////////////////

// we're storing the this pointer only. Therefore the warning can be ignored
#pragma warning(disable:4355)
H245Negotiator::H245Negotiator(H323EndPoint & end, H323Connection & conn)
  : endpoint(end),
    connection(conn)
#ifdef _DEBUG // add name to H323 negotiator mutex
    , mutex(Tools::SString(Tools::SString::Printf, "H245Negotiator::mutex[%#.8x]%#.8x",Tools::SProcess::GetProcessID(),this).c_str())
#endif
#pragma warning(default:4355)
{
    SDEFFUNC("H245Negotiator::H245Negotiator");

    replyTimer.SetNotifier(SCREATE_NOTIFIER(HandleTimeout));

    connection.IncRef(__FILE__,__LINE__);
    SetConnID( connection.GetConnID() );

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ")");
}


H245Negotiator::~H245Negotiator()
{
    SDEFFUNC("H245Negotiator::~H245Negotiator");

    connection.DecRef(__FILE__,__LINE__);

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "()");
}


void H245Negotiator::Delete()
{
    SDEFFUNC("H245Negotiator::Delete");

    if (replyTimer.Stop() == true)
    {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
    }

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "()");
}


void H245Negotiator::HandleTimeout(STimer &, INT)
{
    SDEFFUNC("H245Negotiator::HandleTimeout");

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "()");

    DecRef(__FILE__,__LINE__); // reference counting for timer
}

/////////////////////////////////////////////////////////////////////////////

H245NegMasterSlaveDetermination::H245NegMasterSlaveDetermination(H323EndPoint & end,
                                                                 H323Connection & conn)
  : H245Negotiator(end, conn)
{
    SDEFFUNC("H245NegMSD::H245NegMSD");

  retryCount = 1;
  state = e_Idle;
  status = e_Indeterminate;

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ")");
}


bool H245NegMasterSlaveDetermination::Start()
{
  SWaitAndSignal wait(mutex);

  retryCount = 1;
  return Restart();
}


bool H245NegMasterSlaveDetermination::Restart()
{
    SDEFFUNC("H245NegMasterSlaveDeterm::Restart");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Starting MasterSlaveDetermination ...");

  // Begin the Master/Slave determination procedure
  SRandom::Seed();
  determinationNumber = SRandom::GetDword()%16777216;
  if (!replyTimer.IsRunning())
    IncRef(__FILE__,__LINE__); // reference counting for timer
  replyTimer = endpoint.GetMasterSlaveDeterminationTimeout();
  state = e_Outgoing;

  H323ControlPDU pdu;
  pdu.BuildMasterSlaveDetermination(endpoint.GetTerminalType(), determinationNumber);
  return connection.WriteControlPDU(pdu);
}


bool H245NegMasterSlaveDetermination::HandleIncoming(const H245_MasterSlaveDetermination & pdu)
{
    SDEFFUNC("H245NegMasterSlaveDet::HdlIncom");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received MasterSlaveDetermination: state=" << state);

  if (state == e_Incoming) {
    if (replyTimer.Stop() == true)
    {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
    }
    state = e_Idle;
    return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                             "Duplicate MasterSlaveDetermination");
  }

  if (!replyTimer.IsRunning())
    IncRef(__FILE__,__LINE__); // reference counting for timer
  replyTimer = endpoint.GetMasterSlaveDeterminationTimeout();

  // Determine the master and slave
  MasterSlaveStatus newStatus;
  if (pdu.m_terminalType < (unsigned)endpoint.GetTerminalType())
    newStatus = e_DeterminedMaster;
  else if (pdu.m_terminalType > (unsigned)endpoint.GetTerminalType())
    newStatus = e_DeterminedSlave;
  else if (pdu.m_statusDeterminationNumber < determinationNumber) {
    if ((determinationNumber - pdu.m_statusDeterminationNumber) > 8388608)
      newStatus = e_DeterminedMaster;
    else
      newStatus = e_DeterminedSlave;
  }
  else if (pdu.m_statusDeterminationNumber > determinationNumber) {
    if ((pdu.m_statusDeterminationNumber - determinationNumber) > 8388608)
      newStatus = e_DeterminedSlave;
    else
      newStatus = e_DeterminedMaster;
  }
  else
    newStatus = e_Indeterminate;

  H323ControlPDU reply;

  if (newStatus != e_Indeterminate) {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() MasterSlaveDetermination: local is "
        << (newStatus == e_DeterminedMaster ? "master" : "slave"));
    reply.BuildMasterSlaveDeterminationAck(newStatus == e_DeterminedMaster);
    state = e_Incoming;
    status = newStatus;
  }
  else if (state == e_Outgoing) {
    retryCount++;
    if (retryCount < endpoint.GetMasterSlaveDeterminationRetries())
      return Restart(); // Try again

    if (replyTimer.Stop() == true)
    {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
    }
    state = e_Idle;
    return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                             "Retries exceeded");
  }
  else {
    reply.BuildMasterSlaveDeterminationReject(H245_MasterSlaveDeterminationReject_cause::e_identicalNumbers);
  }

  return connection.WriteControlPDU(reply);
}


bool H245NegMasterSlaveDetermination::HandleAck(const H245_MasterSlaveDeterminationAck & pdu)
{
    SDEFFUNC("H245NegMasterSlaveDeterm::HdlAck");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received MasterSlaveDeterminationAck: state=" << state);

  if (state == e_Idle)
    return true;

  if (!replyTimer.IsRunning())
    IncRef(__FILE__,__LINE__); // reference counting for timer
  replyTimer = endpoint.GetMasterSlaveDeterminationTimeout();

  MasterSlaveStatus newStatus;
  if (pdu.m_decision.GetTag() == H245_MasterSlaveDeterminationAck_decision::e_master)
    newStatus = e_DeterminedMaster;
  else
    newStatus = e_DeterminedSlave;

  H323ControlPDU reply;
  if (state == e_Outgoing) {
    status = newStatus;
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() MasterSlaveDetermination: remote is "
        << (newStatus == e_DeterminedSlave ? "master" : "slave"));
    reply.BuildMasterSlaveDeterminationAck(newStatus == e_DeterminedMaster);
    if (!connection.WriteControlPDU(reply))
      return false;
  }

  if (replyTimer.Stop() == true)
  {
    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Timer stop ok.");
    DecRef(__FILE__,__LINE__); // reference counting for timer
  }
  state = e_Idle;

  if (status != newStatus)
    return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                             "Master/Slave mismatch");

  return true;
}


bool H245NegMasterSlaveDetermination::HandleReject(const H245_MasterSlaveDeterminationReject & pdu)
{
    SDEFFUNC("H245NegMasterSlaveDet::HdlReject");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received MasterSlaveDeterminationReject: state=" << state);

  switch (state) {
    case e_Idle :
      return true;

    case e_Outgoing :
      if (pdu.m_cause.GetTag() == H245_MasterSlaveDeterminationReject_cause::e_identicalNumbers) {
        retryCount++;
        if (retryCount < endpoint.GetMasterSlaveDeterminationRetries())
          return Restart();
      }

    default :
      break;
  }

  if (replyTimer.Stop() == true)
  {
      STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
          "() Timer stop ok.");
      DecRef(__FILE__,__LINE__); // reference counting for timer
  }
  state = e_Idle;

  return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                           "Retries exceeded");
}


bool H245NegMasterSlaveDetermination::HandleRelease(const H245_MasterSlaveDeterminationRelease & /*pdu*/)
{
    SDEFFUNC("H245NegMasterSlaveDet::HdlRelease");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received MasterSlaveDeterminationRelease: state=" << state);

  if (state == e_Idle)
    return true;

  if (replyTimer.Stop() == true)
  {
      STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
          "() Timer stop ok.");
      DecRef(__FILE__,__LINE__); // reference counting for timer
  }
  state = e_Idle;

  return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                           "Aborted");
}


void H245NegMasterSlaveDetermination::HandleTimeout(STimer &, INT)
{
    SDEFFUNC("H245NegMasterSlaveDet::HdlTimeout");

    { // hk:
      // make sure SWaitAndSignal(mutex) shall be destroyed BEFORE DecRef(__FILE__,__LINE__) is called.
      // because the mutex is part of this object an may be deleted by DecRef()
      // in this case the local object wait(mutex) is destroyed after deletion of this object.
        SWaitAndSignal wait(mutex);

        STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timeout on MasterSlaveDetermination: state=" << state);

      if (state == e_Outgoing) {
        H323ControlPDU reply;
        reply.Build(H245_IndicationMessage::e_masterSlaveDeterminationRelease);
        connection.WriteControlPDU(reply);
      }

      state = e_Idle;

      connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                        "Timeout");
    }
    DecRef(__FILE__,__LINE__); // reference counting for timer
}


//#if STRACING
const char * const H245NegMasterSlaveDetermination::StateNames[] = {
  "Idle", "Outgoing", "Incoming"
};


const char * const H245NegMasterSlaveDetermination::StatusNames[] = {
  "Indeterminate", "DeterminedMaster", "DeterminedSlave"
};
//#endif


/////////////////////////////////////////////////////////////////////////////

H245NegTerminalCapabilitySet::H245NegTerminalCapabilitySet(H323EndPoint & end,
                                                           H323Connection & conn)
  : H245Negotiator(end, conn)
{
    SDEFFUNC("H245NegTCS::H245NegTCS");

  inSequenceNumber = UINT_MAX;
  outSequenceNumber = 0;
  state = e_Idle;
  receivedCapabilites = false;

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ")");
}


bool H245NegTerminalCapabilitySet::Start()
{
    SDEFFUNC("H245NegTerminalCapSet::Start");
    SWaitAndSignal wait(mutex);

    if (state == e_InProgress)
    {
        STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() TerminalCapabilitySet already in progress: outSeq=" << outSequenceNumber);
        return true;
    }
    if (state == e_Sent)
    {
        STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() TerminalCapabilitySet already sent: outSeq=" << outSequenceNumber);
        return true;
    }
    
  // Begin the capability exchange procedure
  outSequenceNumber = (outSequenceNumber+1)%256;
  if (!replyTimer.IsRunning())
    IncRef(__FILE__,__LINE__); // reference counting for timer
  replyTimer = endpoint.GetCapabilityExchangeTimeout();
  state = e_InProgress;

  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() Sending TerminalCapabilitySet: outSeq=" << outSequenceNumber);

  H323ControlPDU pdu;
  connection.OnSendCapabilitySet(pdu.BuildTerminalCapabilitySet(connection, outSequenceNumber));
  return connection.WriteControlPDU(pdu);
}


bool H245NegTerminalCapabilitySet::HandleIncoming(const H245_TerminalCapabilitySet & pdu)
{
    SDEFFUNC("H245NegTerminalCapSet::HdlIncom");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received TerminalCapabilitySet:"
        " state=" << state <<
        " pduSeq=" << pdu.m_sequenceNumber <<
        " inSeq=" << inSequenceNumber);

  if (pdu.m_sequenceNumber == inSequenceNumber)
  {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Already had received TerminalCapabilitySet with sequence number "
            << inSequenceNumber << " - ignoring it!");
        return true;
  }

  inSequenceNumber = pdu.m_sequenceNumber;

    if (   !pdu.HasOptionalField(H245_TerminalCapabilitySet::e_multiplexCapability)
        && !pdu.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityTable)
        && !pdu.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityDescriptors) )
    {
        // Received empty TerminalCapabilitySet
        // (H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing)
        STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Received empty TerminalCapabilitySet.");

        H323ControlPDU reply;
        reply.BuildTerminalCapabilitySetAck(inSequenceNumber);
        if (!connection.WriteControlPDU(reply))
            return false;

        if (!connection.OnReceivedEmptyCapabilitySet())
            return false;

        // Reset capapilities of this object.
        // It includes a ResetCapabilities() at the associated connection!
        Reset();
        return true;
    }

  receivedCapabilites = true;

  H323Capabilities remoteCapabilities(connection, pdu);

  const H245_MultiplexCapability * muxCap = NULL;
  if (pdu.HasOptionalField(H245_TerminalCapabilitySet::e_multiplexCapability))
    muxCap = &pdu.m_multiplexCapability;


  H323ControlPDU reject;
  if (connection.OnReceivedCapabilitySet(remoteCapabilities, muxCap,
                    reject.BuildTerminalCapabilitySetReject(inSequenceNumber,
                            H245_TerminalCapabilitySetReject_cause::e_unspecified))) {
    H323ControlPDU ack;
    ack.BuildTerminalCapabilitySetAck(inSequenceNumber);
    return connection.WriteControlPDU(ack);
  }
  STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() No matching capabilities in received TerminalCapabilitySet"
      " -> sending TCS Reject and clear call!");
  connection.WriteControlPDU(reject);
  connection.ClearCall(H323Connection::EndedByCapabilityExchange);
  return true;
}


bool H245NegTerminalCapabilitySet::HandleAck(const H245_TerminalCapabilitySetAck & pdu)
{
    SDEFFUNC("H245NegTerminalCapSet::HandleAck");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received TerminalCapabilitySetAck:"
        " state=" << state <<
        " pduSeq=" << pdu.m_sequenceNumber <<
        " outSeq=" << (unsigned)outSequenceNumber);

  if (state != e_InProgress)
    return true;

  if (pdu.m_sequenceNumber != outSequenceNumber)
    return true;

  if (replyTimer.Stop() == true)
  {
      STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
          "() Timer stop ok.");
      DecRef(__FILE__,__LINE__); // reference counting for timer
  }
  state = e_Sent;
  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() TerminalCapabilitySet sent.");
  return true;
}


bool H245NegTerminalCapabilitySet::HandleReject(const H245_TerminalCapabilitySetReject & pdu,
                                                bool bIsError)
{
    SDEFFUNC("H245NegTerminalCapSet::HdlReject");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received TerminalCapabilitySetReject:"
        " state=" << state <<
        " pduSeq=" << pdu.m_sequenceNumber <<
        " outSeq=" << (unsigned)outSequenceNumber);

  if (state != e_InProgress)
    return true;

  if (pdu.m_sequenceNumber != outSequenceNumber)
    return true;

  state = e_Idle;
  if (replyTimer.Stop() == true)
  {
      STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
          "() Timer stop ok.");
      DecRef(__FILE__,__LINE__); // reference counting for timer
  }

  if (bIsError == false)
      return true;

  return connection.OnControlProtocolError(H323Connection::e_CapabilityExchange,
                                           "Rejected");
}


bool H245NegTerminalCapabilitySet::HandleRelease(const H245_TerminalCapabilitySetRelease & /*pdu*/)
{
    SDEFFUNC("H245NegTerminalCapSet::HdlRelease");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received TerminalCapabilityRelease: state=" << state);

  receivedCapabilites = false;
  return connection.OnControlProtocolError(H323Connection::e_CapabilityExchange,
                                           "Aborted");
}


void H245NegTerminalCapabilitySet::HandleTimeout(STimer &, INT)
{
    SDEFFUNC("H245NegTerminalCapSet::HdlTimeout");

    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Timeout on TerminalCapabilitySet: state=" << state);

  H323ControlPDU reply;
  reply.Build(H245_IndicationMessage::e_terminalCapabilitySetRelease);
  connection.WriteControlPDU(reply);

  connection.OnControlProtocolError(H323Connection::e_CapabilityExchange, "Timeout");

    DecRef(__FILE__,__LINE__); // reference counting for timer
}


bool H245NegTerminalCapabilitySet::SendEmptyTerminalCapabilitySet()
{
    SDEFFUNC("H245NegTermCapSet::SendEmptyTCS");

    // Send empty TerminalCapabilitySet
    // (H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing)

    SWaitAndSignal wait(mutex);

    outSequenceNumber = (outSequenceNumber+1)%256;

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Sending empty TerminalCapabilitySet: outSeq=" << outSequenceNumber);

    H323ControlPDU pdu;
    connection.OnSendCapabilitySet(pdu.BuildEmptyTerminalCapabilitySet(outSequenceNumber));
    return connection.WriteControlPDU(pdu);
}


void H245NegTerminalCapabilitySet::Reset()
{
    SDEFFUNC("H245NegTerminalCapSet::Reset");

    // Reset this object. This is done after receive of an empty  TerminalCapabilitySet.
    // (H.323 Recomm. chap. 8.4.6 - Third party initiated pause and re-routing)

    SWaitAndSignal wait(mutex);

    inSequenceNumber = UINT_MAX;
    outSequenceNumber = 0;
    state = e_Idle;
    receivedCapabilites = false;

    // Reset capabilities at connection
    connection.ResetCapabilities();

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Reset of capability exchange object done.");
}


//#if STRACING
const char * const H245NegTerminalCapabilitySet::StateNames[] = {
  "Idle", "InProgress", "Sent"
};
//#endif


/////////////////////////////////////////////////////////////////////////////

H245NegLogicalChannel::H245NegLogicalChannel(H323EndPoint & end,
                                             H323Connection & conn,
                                             const H323ChannelNumber & chanNum)
  : H245Negotiator(end, conn),
    channelNumber(chanNum)
{
    SDEFFUNC("H245NegLogChan::H245NegLogChan 1");

  channel = NULL;
  state = e_Released;

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ", chanNum=" << chanNum << ")");
}


H245NegLogicalChannel::H245NegLogicalChannel(H323EndPoint & end,
                                             H323Connection & conn,
                                             H323Channel & chan)
  : H245Negotiator(end, conn),
    channelNumber(chan.GetNumber())
{
    SDEFFUNC("H245NegLogChan::H245NegLogChan 2");

  channel = &chan;
  state = e_Established;

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ", chan=" << &chan << ")");
}


H245NegLogicalChannel::~H245NegLogicalChannel()
{
    SDEFFUNC("H245NegLogChan::~H245NegLogChan");

//  delete channel; moved to Delete() because timeout may occur if H323Connection is already deleted!

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "()");
}


void H245NegLogicalChannel::Delete()
{
    SDEFFUNC("H245NegLogicalChannel::Delete");

    // This was in destructor, moved to Delete() because timeout may occur
    // if H323Connection is already deleted! 
    delete channel;
    channel = NULL;

    if (replyTimer.Stop() == true)
    {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
    }

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "()");

//    delete this;
}


bool H245NegLogicalChannel::Open(const H323Capability & capability, unsigned sessionID)
{
    SDEFFUNC("H245NegLogicalChannel::Open");
    SWaitAndSignal wait(mutex);

  if (state != e_Released && state != e_AwaitingRelease) {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Open of channel currently in negotiations: " << channelNumber);
    return false;
  }

  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() Opening channel: " << channelNumber);

  if (!SBAD_READPTR(channel))
  {
    channel->CleanUpOnTermination();
    delete channel;
    channel = NULL;
  }

  state = e_AwaitingEstablishment;

  H323ControlPDU pdu;
  H245_OpenLogicalChannel & open = pdu.BuildOpenLogicalChannel(channelNumber);

  if (!capability.OnSendingPDU(open.m_forwardLogicalChannelParameters.m_dataType)) {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Opening channel: " << channelNumber
        << ", capability.OnSendingPDU() failed");
    return false;
  }
  // The following could be used to set fixed values in OLC for tests
  //H245_AudioCapability & cap = open.m_forwardLogicalChannelParameters.m_dataType;
  //SAsnInteger & value = cap;
  //value = 30;

  channel = capability.CreateChannel(connection, H323Channel::IsTransmitter, sessionID, NULL);
  if (channel == NULL) {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Opening channel: " << channelNumber
        << ", capability.CreateChannel() failed");
    return false;
  }

  if (!channel->SetInitialBandwidth()) {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Opening channel: " << channelNumber << ", Insufficient bandwidth");
    return false;
  }

  channel->SetNumber(channelNumber);

  if (!channel->OnSendingPDU(open)) {
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Opening channel: " << channelNumber
        << ", channel->OnSendingPDU() failed");
    return false;
  }

  if (!replyTimer.IsRunning())
    IncRef(__FILE__,__LINE__); // reference counting for timer
  replyTimer = endpoint.GetLogicalChannelTimeout();

  return connection.WriteControlPDU(pdu);
}


bool H245NegLogicalChannel::Close()
{
    SDEFFUNC("H245NegLogicalChannel::Close");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Closing channel: " << channelNumber << ", state=" << state);

  if (state != e_AwaitingEstablishment && state != e_Established)
    return true;

  if (!replyTimer.IsRunning())
    IncRef(__FILE__,__LINE__); // reference counting for timer
  replyTimer = endpoint.GetLogicalChannelTimeout();

  H323ControlPDU reply;

  if (channelNumber.IsFromRemote()) {
    reply.BuildRequestChannelClose(channelNumber, H245_RequestChannelClose_reason::e_normal);
    state = e_AwaitingResponse;
  }
  else {
    reply.BuildCloseLogicalChannel(channelNumber);
    state = e_AwaitingRelease;
  }

  return connection.WriteControlPDU(reply);
}


bool H245NegLogicalChannel::HandleOpen(const H245_OpenLogicalChannel & pdu)
{
    SDEFFUNC("H245NegLogicalChannel::HandleOpen");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received open channel: " << channelNumber << ", state=" << state);

  if (!SBAD_READPTR(channel))
  {
    channel->CleanUpOnTermination();
    delete channel;
    channel = NULL;
  }

  state = e_AwaitingEstablishment;

  H323ControlPDU reply;
  H245_OpenLogicalChannelAck & ack = reply.BuildOpenLogicalChannelAck(channelNumber);

  unsigned cause;
  if (connection.OnOpenLogicalChannel(pdu, ack, cause))
    channel = connection.CreateLogicalChannel(pdu, cause);

  connection.OnOLCAfterCreation(channel, channelNumber, cause);

  if (!SBAD_READPTR(channel))
  {
    if (channel->GetDirection() == H323Channel::IsBidirectional)
      state = e_AwatingConfirmation;
    else {
      state = e_Established;

      if (!g_pH323Module->m_bUsingSwtcl)
      {
          // Only start a LogicalChannel thread if we not use SwTCL.
          channel->Start();
      }
    }
    channel->SetNumber(channelNumber);
    channel->OnSendOpenAck(pdu, ack);
  }
  else {
    reply.BuildOpenLogicalChannelReject(channelNumber, cause);
    Release();
  }

  return connection.WriteControlPDU(reply);
}


bool H245NegLogicalChannel::HandleOpenAck(const H245_OpenLogicalChannelAck & pdu)
{
    SDEFFUNC("H245NegLogicalChannel::HdlOpenAck");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received open channel ack: " << channelNumber << ", state=" << state);

  switch (state) {
    case e_Released :
      return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                               "Ack unknown channel");
    case e_AwaitingEstablishment :
      state = e_Established;
      if (replyTimer.Stop() == true)
      {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
      }

      if (!channel->OnReceivedAckPDU(pdu))
        return Close();

      if (channel->GetDirection() == H323Channel::IsBidirectional) {
        H323ControlPDU reply;
        reply.BuildOpenLogicalChannelConfirm(channelNumber);
        if (!connection.WriteControlPDU(reply))
          return false;
      }

      if (!g_pH323Module->m_bUsingSwtcl)
      {
          // Only start a LogicalChannel thread if we not use SwTCL.
          channel->Start();
      }

    default :
      break;
  }

  return true;
}


bool H245NegLogicalChannel::HandleOpenConfirm(const H245_OpenLogicalChannelConfirm & /*pdu*/)
{
    SDEFFUNC("H245NegLogicChannel::HdlOpenConf");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received open channel confirm: " << channelNumber << ", state=" << state);

  switch (state) {
    case e_Released :
      return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                               "Confirm unknown channel");
    case e_AwaitingEstablishment :
      return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                               "Confirm established channel");
    case e_AwatingConfirmation :
      if (replyTimer.Stop() == true)
      {
          STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
              "() Timer stop ok.");
          DecRef(__FILE__,__LINE__); // reference counting for timer
      }
      state = e_Established;
      if (!g_pH323Module->m_bUsingSwtcl)
      {
          // Only start a LogicalChannel thread if we not use SwTCL.
          channel->Start();
      }

    default :
      break;
  }

  return true;
}


bool H245NegLogicalChannel::HandleReject(const H245_OpenLogicalChannelReject & /*pdu*/)
{
    SDEFFUNC("H245NegLogicalChannel::HdlReject");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received open channel reject: " << channelNumber << ", state=" << state);

  switch (state) {
    case e_Released :
      return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                               "Reject unknown channel");
    case e_Established :
      Release();
      return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                               "Reject established channel");
    case e_AwaitingEstablishment :
    case e_AwaitingRelease :
      Release();

    default :
      break;
  }

  return true;
}


bool H245NegLogicalChannel::HandleClose(const H245_CloseLogicalChannel & /*pdu*/)
{
    SDEFFUNC("H245NegLogicalChannel::HdlClose");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received close channel: " << channelNumber << ", state=" << state);

  //if (pdu.m_source.GetTag() == H245_CloseLogicalChannel_source::e_user)

  H323ControlPDU reply;
  reply.BuildCloseLogicalChannelAck(channelNumber);
  if (!connection.WriteControlPDU(reply))
    return false;

  Release();
  return true;
}


bool H245NegLogicalChannel::HandleCloseAck(const H245_CloseLogicalChannelAck & /*pdu*/)
{
    SDEFFUNC("H245NegLogicChannel::HdlCloseAck");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received close channel ack: " << channelNumber << ", state=" << state);

  switch (state) {
    case e_Established :
      Release();
      return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                               "Close ack open channel");
    case e_AwaitingRelease :
      Release();

    default :
      break;
  }

  return true;
}


bool H245NegLogicalChannel::HandleRequestClose(const H245_RequestChannelClose & /*pdu*/)
{
    SDEFFUNC("H245NegLogicChannel::HdlReqClose");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received request close channel: " << channelNumber << ", state=" << state);

  if (state != e_Established)
    return true;    // Already closed

  H323ControlPDU reply;
  if (connection.OnClosingLogicalChannel(*channel)) {
    reply.BuildRequestChannelCloseAck(channelNumber);
    Release();
  }
  else
    reply.BuildRequestChannelCloseReject(channelNumber);

  return connection.WriteControlPDU(reply);
}


bool H245NegLogicalChannel::HandleRequestCloseAck(const H245_RequestChannelCloseAck & /*pdu*/)
{
    SDEFFUNC("H245NegLogicChan::HdlReqCloseAck");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received request close ack channel: " << channelNumber << ", state=" << state);

  if (state == e_AwaitingResponse)
    Release();  // Other end says close OK, so do so.

  return true;
}


bool H245NegLogicalChannel::HandleRequestCloseReject(const H245_RequestChannelCloseReject & /*pdu*/)
{
    SDEFFUNC("H245NegLogicChan::HdlReqCloseRej");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received request close reject channel: " << channelNumber << ", state=" << state);

  // Other end refused close, so go back to still having channel open
  if (state == e_AwaitingResponse)
    state = e_Established;

  return true;
}


bool H245NegLogicalChannel::HandleRequestCloseRelease(const H245_RequestChannelCloseRelease & /*pdu*/)
{
    SDEFFUNC("H245NegLogicChan::HdlReqCloseRel");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received request close release channel: " << channelNumber << ", state=" << state);

  // Other end refused close, so go back to still having channel open
  state = e_Established;

  return true;
}


void H245NegLogicalChannel::HandleTimeout(STimer &, INT)
{
    SDEFFUNC("H245NegLogicalChannel::HdlTimeout");

    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Timeout on open channel: " << channelNumber << ", state=" << state);

  if (state == e_AwaitingEstablishment) {
    H323ControlPDU reply;
    reply.BuildCloseLogicalChannel(channelNumber);
    connection.WriteControlPDU(reply);
  }
  else if (state == e_AwaitingResponse) {
    H323ControlPDU reply;
    reply.BuildRequestChannelCloseRelease(channelNumber);
    connection.WriteControlPDU(reply);
  }

  Release();
  connection.OnControlProtocolError(H323Connection::e_LogicalChannel, "Timeout");

    DecRef(__FILE__,__LINE__); // reference counting for timer
}


void H245NegLogicalChannel::Release()
{
    SDEFFUNC("H245NegLogicalChannel::Release");
              
  SWaitAndSignal wait(mutex);

  if (replyTimer.Stop() == true)
  {
      STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
          "() Timer stop ok.");
      DecRef(__FILE__,__LINE__); // reference counting for timer
  }

  if (!SBAD_READPTR(channel))
  {
    channel->CleanUpOnTermination();
//    delete channel;  Should not be deleted here, will be done later in destructor!
//    channel = NULL;
  }

  state = e_Released;
}


H323Channel * H245NegLogicalChannel::GetChannel()
{
  SWaitAndSignal wait(mutex);

  return channel;
}


//#if STRACING
const char * const H245NegLogicalChannel::StateNames[] = {
  "Released",
  "AwaitingEstablishment",
  "Established",
  "AwaitingRelease",
  "AwatingConfirmation",
  "AwaitingResponse"
};
//#endif


/////////////////////////////////////////////////////////////////////////////

H245NegLogicalChannels::H245NegLogicalChannels(H323EndPoint & end,
                                               H323Connection & conn)
  : H245Negotiator(end, conn),
    lastChannelNumber(100, false)
{
    SDEFFUNC("H245NegLogChans::H245NegLogChans");

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ")");
}

H245NegLogicalChannels::~H245NegLogicalChannels()
{
    SDEFFUNC("H245NegLogChans::~H245NegLogChans");

    // clean up channel dictionary
    if ( channels.size() > 0 )
    {
        H245LogicalChannelDict::iterator LCDictIter;
        for ( LCDictIter = channels.begin(); LCDictIter != channels.end(); ++LCDictIter )
        {
            //delete (*LCDictIter).second;
            (*LCDictIter).second->Delete();
            (*LCDictIter).second->DecRef(__FILE__,__LINE__);
        }
    }

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "()");
}


void H245NegLogicalChannels::Add(H323Channel & channel)
{
  SWaitAndSignal wait(mutex);

  H245NegLogicalChannel * negChan = new H245NegLogicalChannel(endpoint, connection, channel);
  if (SBAD_READPTR(negChan))
      return;
  negChan->IncRef(__FILE__,__LINE__);
  channels.insert(pair<const H323ChannelNumber, H245NegLogicalChannel *>(channel.GetNumber(), negChan));
}


bool H245NegLogicalChannels::Open(const H323Capability & capability, unsigned sessionID, H323ChannelNumber & rNumber)
{
    mutex.Wait();

    lastChannelNumber++;

    rNumber = lastChannelNumber;

    H245NegLogicalChannel * negChan = new H245NegLogicalChannel(endpoint, connection, lastChannelNumber);
    if (SBAD_READPTR(negChan))
    {
        mutex.Signal();
        return false;
    }
    negChan->IncRef(__FILE__,__LINE__);
    channels.insert(pair<const H323ChannelNumber, H245NegLogicalChannel *>(lastChannelNumber, negChan));

    mutex.Signal();

    return negChan->Open(capability, sessionID);
}


bool H245NegLogicalChannels::Close(unsigned channelNumber, bool fromRemote)
{
  H245NegLogicalChannel * chan = FindNegLogicalChannel(channelNumber, fromRemote);
  if (!SBAD_READPTR(chan))
  {
      return chan->Close();
  }

  return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                           "Close unknown");
}


bool H245NegLogicalChannels::Erase(H323Channel * channel)
{
    SWaitAndSignal wait(mutex);

    H245LogicalChannelDict::iterator LCDictIter;
    for ( LCDictIter = channels.begin(); LCDictIter != channels.end(); ++LCDictIter )
    {
        STRACE(STrace::Debug, g_pH323Module->m_H323TraceId,
               "(" << this << ") *** H245NegLogicalChannels::Erase(): channel: " << (*LCDictIter).second->GetChannel());

        if ( channel == (*LCDictIter).second->GetChannel() )
        {
//          delete (*LCDictIter).second;
            STRACE(STrace::Debug, g_pH323Module->m_H323TraceId,
                   "(" << this << ") *** H245NegLogicalChannels::Erase(): Calling Delete() for channel: " << (*LCDictIter).second->GetChannel());
            (*LCDictIter).second->Delete();
            (*LCDictIter).second->DecRef(__FILE__,__LINE__);
            channels.erase(LCDictIter);
            return true;
        }
    }
    
    return false;
}


void H245NegLogicalChannels::CloseAllChannels()
{
    SWaitAndSignal wait(mutex);

    // Close all logical channels
    H245LogicalChannelDict::iterator LCDictIter;
    for (LCDictIter = channels.begin(); LCDictIter != channels.end(); ++LCDictIter)
    {
        H323Channel * channel = (*LCDictIter).second->GetChannel();
        STRACE(STrace::Debug, g_pH323Module->m_H323TraceId,
               "(" << this << ") *** H245NegLogicalChannels::CloseAllChannels(): channel: " << channel);
        if (!SBAD_READPTR(channel))
        {
            if (channel->GetDirection() == H323Channel::IsTransmitter)
            {
                // If we use SwTCL the receive and transmit threads are not running.
                // Normally they are closing the channels.
                (*LCDictIter).second->Close();
            }
            channel->CleanUpOnTermination();
        }
    }
}


bool H245NegLogicalChannels::HandleOpen(const H245_OpenLogicalChannel & pdu)
{
    SDEFFUNC("H245NegLogicalChannels::HdlOpen");
    
    H323ChannelNumber chanNum(pdu.m_forwardLogicalChannelNumber, true);
    H245NegLogicalChannel * chan;

    mutex.Wait();

    if (channels.find(chanNum) != channels.end())
        chan = channels[chanNum];
    else
    {
        chan = new H245NegLogicalChannel(endpoint, connection, chanNum);
        if (SBAD_READPTR(chan))
        {
            mutex.Signal();
            return false;
        }
        chan->IncRef(__FILE__,__LINE__);
        channels.insert(pair<const H323ChannelNumber, H245NegLogicalChannel *>(chanNum, chan));
    }

    mutex.Signal();

    if (!chan->HandleOpen(pdu))
        return false;

    if (chan->IsReleased())
    {
        STRACE(STrace::Debug, g_pH323Module->m_H323TraceId,
               "(" << this << ") *** Calling Delete() for released channel: " << chan->GetChannel());
        chan->Delete();
        chan->DecRef(__FILE__,__LINE__);
        channels.erase(chanNum);
    }
    return true;
}


bool H245NegLogicalChannels::HandleOpenAck(const H245_OpenLogicalChannelAck & pdu)
{
  H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, false);
  if (!SBAD_READPTR(chan))
    return chan->HandleOpenAck(pdu);

  return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                           "Ack unknown");
}


bool H245NegLogicalChannels::HandleOpenConfirm(const H245_OpenLogicalChannelConfirm & pdu)
{
  H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, true);
  if (!SBAD_READPTR(chan))
    return chan->HandleOpenConfirm(pdu);

  return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                           "Confirm unknown");
}


bool H245NegLogicalChannels::HandleReject(const H245_OpenLogicalChannelReject & pdu)
{
    SDEFFUNC("H245NegLogicalChannels::HdlReject");

    SWaitAndSignal wait(mutex);

    H323ChannelNumber chanNum(pdu.m_forwardLogicalChannelNumber, false);

    H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, false);
    if (!SBAD_READPTR(chan))
    {
        if (!chan->HandleReject(pdu))
            return false;
        STRACE(STrace::Debug, g_pH323Module->m_H323TraceId,
               "(" << this << ") *** Calling Delete() for channel: " << chan->GetChannel());
        chan->Delete();
        chan->DecRef(__FILE__,__LINE__);
        channels.erase(chanNum);
        return true;
    }
    return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                             "Reject unknown");
}


bool H245NegLogicalChannels::HandleClose(const H245_CloseLogicalChannel & pdu)
{
    SDEFFUNC("H245NegLogicalChannels::HdlClose");

    SWaitAndSignal wait(mutex);

    H323ChannelNumber chanNum(pdu.m_forwardLogicalChannelNumber, true);

    H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, true);
    if (!SBAD_READPTR(chan))
    {
        if (!chan->HandleClose(pdu))
            return false;
//      delete chan;
        STRACE(STrace::Debug, g_pH323Module->m_H323TraceId,
               "(" << this << ") *** Calling Delete() for channel: " << chan->GetChannel());
        chan->Delete();
        chan->DecRef(__FILE__,__LINE__);
        channels.erase(chanNum);
        return true;
    }
    STRACE2_WARN(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() No logical channel found for channel number " << chanNum);
    return true;
}


bool H245NegLogicalChannels::HandleCloseAck(const H245_CloseLogicalChannelAck & pdu)
{
    SWaitAndSignal wait(mutex);

    H323ChannelNumber chanNum(pdu.m_forwardLogicalChannelNumber, false);

    H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, false);
    if (!SBAD_READPTR(chan))
    {
        if (!chan->HandleCloseAck(pdu))
            return false;
        //delete chan;
        STRACE(STrace::Debug, g_pH323Module->m_H323TraceId,
               "(" << this << ") *** Calling Delete() for channel: " << chan->GetChannel());
        chan->Delete();
        chan->DecRef(__FILE__,__LINE__);
        channels.erase(chanNum);
        return true;
    }
    return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                             "Close Ack unknown");
}


bool H245NegLogicalChannels::HandleRequestClose(const H245_RequestChannelClose & pdu)
{
  H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, false);
  if (!SBAD_READPTR(chan))
    return chan->HandleRequestClose(pdu);

  return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                           "Request Close unknown");
}


bool H245NegLogicalChannels::HandleRequestCloseAck(const H245_RequestChannelCloseAck & pdu)
{
  H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, true);
  if (!SBAD_READPTR(chan))
    return chan->HandleRequestCloseAck(pdu);

  return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                           "Request Close Ack unknown");
}


bool H245NegLogicalChannels::HandleRequestCloseReject(const H245_RequestChannelCloseReject & pdu)
{
  H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, true);
  if (!SBAD_READPTR(chan))
    return chan->HandleRequestCloseReject(pdu);

  return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                           "Request Close Reject unknown");
}


bool H245NegLogicalChannels::HandleRequestCloseRelease(const H245_RequestChannelCloseRelease & pdu)
{
  H245NegLogicalChannel * chan = FindNegLogicalChannel(pdu.m_forwardLogicalChannelNumber, false);
  if (!SBAD_READPTR(chan))
    return chan->HandleRequestCloseRelease(pdu);

  return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                           "Request Close Release unknown");
}


H323ChannelNumber H245NegLogicalChannels::GetNextChannelNumber()
{
  lastChannelNumber++;
  return lastChannelNumber;
}


// Swyx version using STL hash_map !
H323Channel * H245NegLogicalChannels::GetChannelAt(int i)
{
    SWaitAndSignal wait(mutex);

    H245LogicalChannelDict::iterator pos;
    int j;

    for (pos = channels.begin(), j = 0; pos != channels.end(); ++pos, j++)
    {
        if (j == i)
        {
            return (*pos).second->GetChannel();
        }
    }
    return NULL;
}


H323Channel * H245NegLogicalChannels::FindChannel(unsigned channelNumber,
                                                  bool fromRemote)
{
  SWaitAndSignal wait(mutex);

  H323ChannelNumber chanNum(channelNumber, fromRemote);

  if (channels.find(chanNum) != channels.end())
    return channels[chanNum]->GetChannel();

  return NULL;
}


H245NegLogicalChannel * H245NegLogicalChannels::FindNegLogicalChannel(unsigned channelNumber,
                                                                      bool fromRemote)
{
  SWaitAndSignal wait(mutex);

  H323ChannelNumber chanNum(channelNumber, fromRemote);

  if (channels.find(chanNum) != channels.end())
    return channels[chanNum];

  return NULL;
}


H323Channel * H245NegLogicalChannels::FindChannelBySession(unsigned rtpSessionId,
                                                           bool fromRemote)
{
    SWaitAndSignal wait(mutex);

    H323Channel::Directions desiredDirection = fromRemote ? H323Channel::IsReceiver : H323Channel::IsTransmitter;
    H245LogicalChannelDict::iterator pos;
    for (pos = channels.begin(); pos != channels.end(); ++pos)
    {
        if (SBAD_READPTR((*pos).second))
            continue;

        H323Channel * channel = (*pos).second->GetChannel();
        if (!SBAD_READPTR(channel) && (channel->GetSessionID() == rtpSessionId) && (channel->GetDirection() == desiredDirection))
            return channel;
    }

    return NULL;
}


bool H245NegLogicalChannels::IsChannelInMap(H323Channel * pChannel)
{
    SDEFFUNC("H245NegLogicalChs::IsChannelInMap");

    SWaitAndSignal wait(mutex);

    H245LogicalChannelDict::iterator pos;
    for (pos = channels.begin(); pos != channels.end(); ++pos)
    {
        if (SBAD_READPTR((*pos).second))
            continue;

        H323Channel * pLogCh = (*pos).second->GetChannel();
        if (SBAD_READPTR(pLogCh))
            continue;

        STRACE2_INFO3(g_pH323Module->m_H323TraceId, this, GetConnID(),
            "(" << pChannel << ") logical channel: " << pLogCh);

        if (pLogCh == pChannel)
            return true;
    }
    return false;
}


/////////////////////////////////////////////////////////////////////////////

H245NegRequestMode::H245NegRequestMode(H323EndPoint & end, H323Connection & conn)
  : H245Negotiator(end, conn)
{
    SDEFFUNC("H245NegReqMode::H245NegReqMode");

  awaitingResponse = false;
  inSequenceNumber = UINT_MAX;
  outSequenceNumber = 0;

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ")");
}


bool H245NegRequestMode::StartRequest(const H323Capability & rCapability)
{
    SDEFFUNC("H245NegRequestMode::StartRequest");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Started request mode: outSeq=" << outSequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

    if (awaitingResponse)
        return false;

    // Initiate a mode request
    outSequenceNumber = (outSequenceNumber+1)%256;
    if (!replyTimer.IsRunning())
        IncRef(__FILE__,__LINE__); // reference counting for timer
    replyTimer = endpoint.GetRequestModeTimeout();
    awaitingResponse = true;

    const H323Capabilities & localCapabilities = connection.GetLocalCapabilities();
    H323Capability * pCapability = localCapabilities.FindCapability(rCapability);
    if (SBAD_READPTR(pCapability))
    {
        STRACE2_ERROR(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Could not find local capability!");
        return false;
    }

    H245_ModeDescription description;
    H245_ArrayOf_ModeDescription descriptions;
    description.SetSize(1);
    pCapability->OnSendingPDU(description[0]);
    descriptions.SetSize(1);
    descriptions[0] = description;

    H323ControlPDU pdu;
    H245_RequestMode & requestMode = pdu.BuildRequestMode(outSequenceNumber);
    requestMode.m_requestedModes = descriptions;
    requestMode.m_requestedModes.SetConstraints(SAsnObject::FixedConstraint, 1, 256);

    return connection.WriteControlPDU(pdu);
}


bool H245NegRequestMode::HandleRequest(const H245_RequestMode & pdu)
{
    SDEFFUNC("H245NegRequestMode::HandleRequest");
    SWaitAndSignal wait(mutex);

    inSequenceNumber = pdu.m_sequenceNumber;

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received request mode: inSeq=" << inSequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

  H323ControlPDU reply_ack;
  H245_RequestModeAck & ack = reply_ack.BuildRequestModeAck(inSequenceNumber,
              H245_RequestModeAck_response::e_willTransmitMostPreferredMode);

  H323ControlPDU reply_reject;
  H245_RequestModeReject & reject = reply_reject.BuildRequestModeReject(inSequenceNumber,
                                        H245_RequestModeReject_cause::e_modeUnavailable);

  int selectedMode = 0;
  if (!connection.OnRequestModeChange(pdu, ack, reject, selectedMode))
    return connection.WriteControlPDU(reply_reject);

  if (selectedMode != 0)
    ack.m_response.SetTag(H245_RequestModeAck_response::e_willTransmitLessPreferredMode);

  if (!connection.WriteControlPDU(reply_ack))
    return false;

  connection.OnModeChanged(pdu.m_requestedModes[selectedMode]);
  return true;
}


bool H245NegRequestMode::HandleAck(const H245_RequestModeAck & pdu)
{
    SDEFFUNC("H245NegRequestMode::HandleAck");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received ack on request mode: outSeq=" << outSequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

  if (awaitingResponse && pdu.m_sequenceNumber == outSequenceNumber) {
    awaitingResponse = false;
    if (replyTimer.Stop() == true)
    {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
    }
    connection.OnAcceptModeChange(pdu);
  }

  return true;
}


bool H245NegRequestMode::HandleReject(const H245_RequestModeReject & pdu)
{
    SDEFFUNC("H245NegRequestMode::HandleReject");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received reject on request mode: outSeq=" << outSequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

  if (awaitingResponse && pdu.m_sequenceNumber == outSequenceNumber) {
    awaitingResponse = false;
    if (replyTimer.Stop() == true)
    {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
    }
    connection.OnRefusedModeChange(&pdu);
  }

  return true;
}


bool H245NegRequestMode::HandleRelease(const H245_RequestModeRelease & /*pdu*/)
{
    SDEFFUNC("H245NegRequestMode::HandleRelease");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Received release on request mode: inSeq=" << inSequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

  return true;
}


void H245NegRequestMode::HandleTimeout(STimer &, INT)
{
    SDEFFUNC("H245NegRequestMode::HandleTimeout");

    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Timeout on request mode: outSeq=" << outSequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

  if (awaitingResponse) {
    awaitingResponse = false;
    H323ControlPDU pdu;
    pdu.Build(H245_IndicationMessage::e_requestModeRelease);
    connection.WriteControlPDU(pdu);
    connection.OnRefusedModeChange(NULL);
  }

  connection.OnControlProtocolError(H323Connection::e_ModeRequest, "Timeout");

    DecRef(__FILE__,__LINE__); // reference counting for timer
}


/////////////////////////////////////////////////////////////////////////////

H245NegRoundTripDelay::H245NegRoundTripDelay(H323EndPoint & end, H323Connection & conn)
  : H245Negotiator(end, conn)
{
    SDEFFUNC("H245NegRTD::H245NegRTD");

  awaitingResponse = false;
  sequenceNumber = 0;
#ifdef ROUND_TRIP_DELAY
  retryCount = 3;
#endif // ROUND_TRIP_DELAY

    STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "(ep=" << &end <<", conn=" << &conn << ")");
}


bool H245NegRoundTripDelay::StartRequest()
{
    SDEFFUNC("H245NegRoundTripDelay::StartReq");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Started round trip delay: seq=" << sequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

#ifndef ROUND_TRIP_DELAY // (Could be deleted later)
  tripStartTime = STimer::Tick();
#endif // ! ROUND_TRIP_DELAY

  if (!replyTimer.IsRunning())
    IncRef(__FILE__,__LINE__); // reference counting for timer
  replyTimer = endpoint.GetRoundTripDelayTimeout();
  sequenceNumber = (sequenceNumber + 1)%256;
  awaitingResponse = true;

  H323ControlPDU pdu;
  pdu.BuildRoundTripDelayRequest(sequenceNumber);
#ifdef ROUND_TRIP_DELAY
  if (!connection.WriteControlPDU(pdu))
    return false;

  tripStartTime = STimer::Tick();
  return true;
#else
  return connection.WriteControlPDU(pdu);
#endif // ROUND_TRIP_DELAY
}


bool H245NegRoundTripDelay::HandleRequest(const H245_RoundTripDelayRequest & pdu)
{
    SDEFFUNC("H245NegRoundTripDelay::HandleReq");
    SWaitAndSignal wait(mutex);

    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Started round trip delay: seq=" << sequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));

  H323ControlPDU reply;
  reply.BuildRoundTripDelayResponse(pdu.m_sequenceNumber);
  return connection.WriteControlPDU(reply);
}


bool H245NegRoundTripDelay::HandleResponse(const H245_RoundTripDelayResponse & pdu)
{
    SDEFFUNC("H245NegRoundTripDelay::HdlRespons");
    SWaitAndSignal wait(mutex);

#ifdef ROUND_TRIP_DELAY
  STimeInterval tripEndTime = STimer::Tick();

  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() Handling round trip delay: seq=" << sequenceNumber
      << (awaitingResponse ? " awaitingResponse" : " idle"));
#else
  STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
      "() Started round trip delay: seq=" << sequenceNumber
      << (awaitingResponse ? " awaitingResponse" : " idle"));
#endif // ROUND_TRIP_DELAY

  if (awaitingResponse && pdu.m_sequenceNumber == sequenceNumber) {
#ifdef ROUND_TRIP_DELAY
    if (replyTimer.Stop() == true)
    {
        STRACE2_INFO3(g_pH323Module->m_H245TraceId, this, GetConnID(),
            "() Timer stop ok.");
        DecRef(__FILE__,__LINE__); // reference counting for timer
    }
    awaitingResponse = false;
    roundTripTime = tripEndTime - tripStartTime;
    retryCount = 3;
#else
    awaitingResponse = false;
    roundTripTime = STimer::Tick() - tripStartTime;
#endif // ROUND_TRIP_DELAY
  }

  return true;
}


void H245NegRoundTripDelay::HandleTimeout(STimer &, INT)
{
    SDEFFUNC("H245NegRoundTripDelay::HdlTimeout");

    mutex.Wait();

#ifdef ROUND_TRIP_DELAY
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Timeout on round trip delay (retry " << retryCount << "): seq=" << sequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));
#else
    STRACE2_INFO(g_pH323Module->m_H245TraceId, this, GetConnID(),
        "() Timeout on round trip delay: seq=" << sequenceNumber
        << (awaitingResponse ? " awaitingResponse" : " idle"));
#endif // ROUND_TRIP_DELAY

#ifdef ROUND_TRIP_DELAY
  if (awaitingResponse && retryCount > 0)
  {
    retryCount--;
    mutex.Signal();
    DecRef(__FILE__,__LINE__); // reference counting for timer
    return;
  }
#endif // ROUND_TRIP_DELAY
  awaitingResponse = false;

  connection.OnControlProtocolError(H323Connection::e_RoundTripDelay, "Timeout");
  mutex.Signal();
  DecRef(__FILE__,__LINE__); // reference counting for timer
}


/////////////////////////////////////////////////////////////////////////////
