/*---------------------------------------------------------------------------

   Swyx Solutions GmbH
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Solutions GmbH are
  Copyright (C) 2000-2005 Swyx Solutions GmbH.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: Q.931
 
    Comments: Q.931 protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/IpPbxSrv.root/H323/source/q931.cpp                   $
    $Author:: Martin.hueser                                                $
      $Date:: 31.10.05 10:18                                               $
   $Modtime:: 26.10.05 15:42                                               $
  $Revision:: 3                                                            $
  $Workfile:: q931.cpp                                                     $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "q931.h"
#include "h323lib.h"

// Module including Swyx H.323 specific stuff, e.g. tracing
extern H323Module *g_pH323Module;

using namespace std;
using namespace Tools;

#define new SNEW


Q931::Q931()
{
  protocolDiscriminator = 8;  // Q931 always has 00001000
  messageType = NationalEscapeMsg;
  fromDestination = false;
  callReference = 0;
  m_bCopy = false;
  informationElements.clear();
}


Q931::~Q931()
{
  if (!m_bCopy)
    ClearIEs();
}

void Q931::ClearIEs()
{
  if ( informationElements.size() > 0 )
  {
    for (InternalInformationElements::iterator pos = informationElements.begin(); pos != informationElements.end(); ++pos)
    {
      delete (*pos).second;
    }
    informationElements.clear();
  }
}

Q931 & Q931::operator=(const Q931 & other)
{
  protocolDiscriminator = other.protocolDiscriminator;
  callReference = other.callReference;
  fromDestination = other.fromDestination;
  messageType = other.messageType;

  ClearIEs();

  informationElements = other.informationElements;
  m_bCopy = true;

  return *this;
}


void Q931::BuildFacility(int callRef, bool fromDest)
{
  messageType = FacilityMsg;
  callReference = callRef;
  fromDestination = fromDest;
  ClearIEs();

  // OpenH323 added this for call forwarding...
//  SByteVector data;
//  SetIE(FacilityIE, data);
}


void Q931::BuildCallProceeding(int callRef)
{
  messageType = CallProceedingMsg;
  callReference = callRef;
  fromDestination = true;
  ClearIEs();
}


void Q931::BuildAlerting(int callRef, const Q931::ProgressIndicator & ProgressInd)
{
    messageType = AlertingMsg;
    callReference = callRef;
    fromDestination = true;
    ClearIEs();

    SetProgressIndicator(ProgressInd);
}

// Bearer capability:
// 
// Octet 3: 1         extension bit
//           00       Coding standard: ITU-T
//                    Information transfer capability:
//             00000    Speech                            (if audio only)
//             01000    Unrestricted digital information  (if audio and video, 64 kbps)
//
// => 1 00 00000 = 0x80
//
// Octet 4: 1         extension bit
//                    Transfer mode:
//           00         Circuit mode (always in H.225!)
//           10         Packet mode  (never in H.225!)
//                    Information transfer rate:
//             00000    Shall be used for packet mode calls (=> never in H.225!)
//             10000    64 kbit/s
//             10001    2 x 64 kbit/s
//             10011    384 kbit/s
//             10101    1536 kbit/s
//             10111    1920 kbit/s
//             11000    Multirate (64 kbit/s base rate) (use octet 4.1 then!)
//
// => 1 00 10000 = 0x90
//
// Octec 5: 1         extension bit
//           01       Layer one identifier
//                    Layer one protocol:
//             00010    Recommendation G.711 u-law
//             00011    Recommendation G.711 A-law
//             00101    H.221 and H.242 to indicate an H.323 videophone call
//
// => 1 01 00010 = 0xa2
// or 1 01 00011 = 0xa3

#define BearerCapabilityLength 3
static const BYTE BearerCapabilityData[BearerCapabilityLength] = { 0x80, 0x90, 0xa2 };

void Q931::BuildSetup(int callRef, const SString & name, bool bSendingComplete)
{
  int i;
  SByteVector tempVec(BearerCapabilityLength);

  callReference = callRef;
  messageType = SetupMsg;
  fromDestination = false;
  ClearIEs();
  SetDisplayName(name);
  for (i = 0; i < BearerCapabilityLength; i++)
  {
      tempVec[i] = BearerCapabilityData[i];
  }
  SetIE(BearerCapabilityIE, tempVec);

  if (bSendingComplete == true)
  {
      SByteVector data;
      SetIE(SendingCompleteIE, data);
  }
}


void Q931::BuildSetupAck(int callRef)
{
  callReference = callRef;
  messageType = SetupAckMsg;
  fromDestination = true;
  ClearIEs();
}


void Q931::BuildProgress(int callRef, bool fromDest, const Q931::ProgressIndicator & ProgressInd, const Q931::Cause & cause)
{
    int i;
    SByteVector tempVec(BearerCapabilityLength);

    messageType = ProgressMsg;
    callReference = callRef;
    fromDestination = fromDest;
    ClearIEs();
    for (i = 0; i < BearerCapabilityLength; i++)
    {
        tempVec[i] = BearerCapabilityData[i];
    }
    SetIE(BearerCapabilityIE, tempVec);

    SetProgressIndicator(ProgressInd);
    SetCause(cause);
}


void Q931::BuildNotify(int callRef, bool fromDest, const Q931::NotificationIndicator & NotificationInd)
{
    messageType = NotifyMsg;
    callReference = callRef;
    fromDestination = fromDest;
    ClearIEs();
    SetNotificationIndicator(NotificationInd);
}

void Q931::BuildConnect(int callRef, const SString & name)
{
  int i;
  SByteVector tempVec(BearerCapabilityLength);

  messageType = ConnectMsg;
  callReference = callRef;
  fromDestination = true;
  ClearIEs();
  SetDisplayName(name);
  for (i = 0; i < BearerCapabilityLength; i++)
  {
      tempVec[i] = BearerCapabilityData[i];
  }
  SetIE(BearerCapabilityIE, tempVec);
}


void Q931::BuildInformation(int callRef, bool fromDest)
{
  messageType = InformationMsg;
  callReference = callRef;
  fromDestination = fromDest;
  ClearIEs();
}


void Q931::BuildStatus(int callRef, bool fromDest)
{
  messageType = StatusMsg;
  callReference = callRef;
  fromDestination = fromDest;
  ClearIEs();
  SByteVector CallStateData(1);
  SByteVector CauseIEData(2);

  // Call State as per Q.931 section 4.5.7
  CallStateData[0] = 0; // Use CCITT Null
  SetIE(CallStateIE, CallStateData);

  // Cause IE as per Q.850
  CauseIEData[0] = 0x00; // ITU-T code, User Location
  CauseIEData[1] = 0x9e; // Last octet, Cause value is response to STATUS ENQUIRY
  SetIE(CauseIE, CauseIEData);
}


void Q931::BuildReleaseComplete(int callRef, bool fromDest, const Q931::Cause & cause)
{
  messageType = ReleaseCompleteMsg;
  callReference = callRef;
  fromDestination = fromDest;
  ClearIEs();

  SetCause(cause);
}


bool Q931::Decode(const SByteVector & data)
{
    SDEFFUNC("Q931::Decode");

  // Clear all existing data before reading new
  ClearIEs();

  if (data.size() < 5) // Packet too short
  {
    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                  "() Error in Q931 decode: Packet length (" << data.size() << ") < 5");
    return false;
  }

  protocolDiscriminator = data[0];

  if (data[1] != 2) // Call reference must be 2 bytes long
  {
    STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                  "() Error in Q931 decode: Length of call reference (0x" << hex
                  << (int)data[1] << dec << ") != 2");
    return false;
  }

  callReference = ((data[2]&0x7f) << 8) | data[3];
  fromDestination = (data[2]&0x80) != 0;

  messageType = (MsgTypes)data[4];

  // Have preamble, start getting the information elements into buffers
  SByteVector::size_type offset = 5;
  while (offset < data.size()) {
    // Get field discriminator
    int discriminator = data[offset++];

    SByteVector * item = new SByteVector;

    // For discriminator with high bit set there is no data
    if ((discriminator&0x80) == 0) {
      SByteVector::size_type len = data[offset++];

      if (discriminator == UserUserIE) {
        // Special case of User-user IE, there is some confusion here as
        // the Q931 documentation claims the length is a single byte,
        // unfortunately all H.323 based apps have a 16 bit length here, so
        // we allow for said longer length. There is presumably an addendum
        // to Q931 which describes this, and provides a means to discriminate
        // between the old 1 byte and the new 2 byte systems. However, at
        // present we assume it is always 2 bytes until we find something that
        // breaks it.
        len <<= 8;
        len |= data[offset++];

        // we also have a protocol discriminator, which we ignore
        offset++;
        len--;
      }

      if (offset + len > data.size())
      {
        STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
                      "() Error in Q931 decode: offset (" << offset
                      << ") + length (" << len << ") > data size (" << data.size() << ")");
        return false;
      }

      if (len > item->size())
          item->resize(len);

      Tools::copy_n(data.begin()+offset, len, item->begin());
      offset += len;
    }

    informationElements.insert(pair<const int, SByteVector *>(discriminator, item));
  }

  return true;
}


bool Q931::Encode(SByteVector & data) const
{
    SDEFFUNC("Q931::Encode");

  SByteVector::size_type totalBytes = 5;
  unsigned discriminator;
  for (discriminator = 0; discriminator < 256; discriminator++) {
    if (informationElements.find(discriminator) != informationElements.end()) {
      if (discriminator < 128)
        totalBytes += informationElements[discriminator]->size() +
                            (discriminator != UserUserIE ? 2 : 4);
      else
        totalBytes++;
    }
  }

  if (totalBytes > data.size())
     data.resize(totalBytes);

  // Put in Q931 header
  SAssert(protocolDiscriminator < 256, SInvalidParameter);
  data[0] = (BYTE)protocolDiscriminator;
  data[1] = 2; // Length of call reference
  data[2] = (BYTE)(callReference >> 8);
  if (fromDestination)
    data[2] |= 0x80;
  data[3] = (BYTE)callReference;
  SAssert(messageType < 256, SInvalidParameter);
  data[4] = (BYTE)messageType;

  // The following assures disciminators are in ascending value order
  // as required by Q931 specification
  int offset = 5;

    // Hack to set SendingComplete IE to the beginning, not after UUIE
    // (for possible compability reasons)
    if (informationElements.find(SendingCompleteIE) != informationElements.end())
    {
        data[offset++] = (BYTE)SendingCompleteIE;
    }

  for (discriminator = 0; discriminator < 256; discriminator++) {
    if (informationElements.find(discriminator) != informationElements.end()) {
      if (discriminator < 128) {
        int len = informationElements[discriminator]->size();

        if (discriminator != UserUserIE) {
          data[offset++] = (BYTE)discriminator;
          data[offset++] = (BYTE)len;
        }
        else {
          len++; // Allow for protocol discriminator
          data[offset++] = (BYTE)discriminator;
          data[offset++] = (BYTE)(len >> 8);
          data[offset++] = (BYTE)len;
          len--; // Then put the length back again
          // We shall assume that the user-user IE is an ITU protocol block (5)
          data[offset++] = 5;
        }

        Tools::copy_n(informationElements[discriminator]->begin(), len, &data[offset]);
        offset += len;
      }
      else
      {
        if (discriminator == SendingCompleteIE) // Skip SendingComplete IE (see above)
            continue;
        data[offset++] = (BYTE)discriminator;
      }
    }
  }

    if (data.size() != offset)
    {
        STRACE2_ERROR(g_pH323Module->m_H225TraceId, this, 0,
            "() Q931 encode: data size (" << data.size() << ") != offset (" << offset << ")");
    }
//  return data.SetSize(offset);
//  data.resize(offset); // necessary?
  return true;
}


static SString Q931IENameString(int number)
{
  switch (number) {
    case Q931::BearerCapabilityIE:
      return "Bearer-Capability";
    case Q931::CauseIE:
      return "Cause";
    case Q931::CallIdentityIE:
      return "Call-Identity";
    case Q931::CallStateIE:
      return "Call-State";
    case Q931::ChannelIdentificationIE:
      return "Channel-Identification";
    case Q931::FacilityIE:
      return "Facility";
    case Q931::ProgressIndicatorIE:
      return "Progress-Indicator";
    case Q931::NetworkSpecificFacilitiesIE:
      return "Network-specific-Facilities";
    case Q931::NotificationIndicatorIE:
      return "Notification-Indicator";
    case Q931::DisplayIE :
      return "Display";
    case Q931::DateTimeIE:
      return "Date-Time";
    case Q931::KeypadFacilityIE:
      return "Keypad-Facility";
    case Q931::SignalIE:
      return "Signal";
    case Q931::InformationRateIE:
      return "Information-Rate";
    case Q931::EndToEndTransitDelayIE:
      return "End-to-end-Transit-Delay";
    case Q931::TransitDelaySelectionAndIndicationIE:
      return "Transit-Delay-Selection-And-Indication";
    case Q931::PacketLayerBinaryParametersIE:
      return "Packet-Layer-Binary-Parameters";
    case Q931::PacketLayerWindowSizeIE:
      return "Packet-Layer-Window-Size";
    case Q931::PacketSizeIE:
      return "Packet-Size";
    case Q931::ClosedUserGroupIE:
      return "Closed-User-Group";
    case Q931::ReverseChargeIndicationIE:
     return "Reverse-Charge-Indication";
    case Q931::ConnectedNumberIE:
      return "Connected-Number";
    case Q931::CallingPartyNumberIE:
      return "Calling-Party-Number";
    case Q931::CallingPartySubaddressIE:
      return "Calling-Party-Subaddress";
    case Q931::CalledPartyNumberIE:
      return "Called-Party-Number";
    case Q931::CalledPartySubaddressIE:
      return "Called-Party-Subaddress";
    case Q931::RedirectingNumberIE:
      return "Redirecting-Number";
    case Q931::TransitNetworkSelectionIE:
      return "Transit-Network-Selection";
    case Q931::RestartIndicatorIE:
      return "Restart-Indicator";
    case Q931::LowLayerCompatibilityIE:
      return "Low-Layer-Compatibility";
    case Q931::HighLayerCompatibilityIE:
      return "High-Layer-Compatibility";
    case Q931::UserUserIE:
      return "User-User";
    case Q931::SendingCompleteIE:
      return "SendingComplete";
  }
  return ssprintf("0x%02x", number);
}


void Q931::PrintOn(ostream & strm) const
{
  long long indent = strm.precision() + 2;
  strm << "{\n" << STrace::MultiLinePrefix
       << setw(indent+24) << "protocolDiscriminator = " << protocolDiscriminator << '\n' << STrace::MultiLinePrefix
       << setw(indent+16) << "callReference = " << callReference << '\n' << STrace::MultiLinePrefix
       << setw(indent+7)  << "from = " << (fromDestination ? "destination" : "originator") << '\n' << STrace::MultiLinePrefix
       << setw(indent+14) << "messageType = " << GetMessageTypeName() << '\n' << STrace::MultiLinePrefix;

  for (InternalInformationElements::iterator pos = informationElements.begin(); pos != informationElements.end(); ++pos)
  {
    strm << setw(indent+4)
         << "IE: " << Q931IENameString((*pos).first) << " = {\n" << STrace::MultiLinePrefix
         << hex << setfill('0')
         << setprecision(indent+2) << *(*pos).second << '\n' << STrace::MultiLinePrefix
         << dec << setfill(' ')
         << setw(indent+2) << "}\n" << STrace::MultiLinePrefix;
  }

  strm << setw(indent-1) << "}";
}


SString Q931::GetMessageTypeName() const
{
  switch (messageType) {
    case NationalEscapeMsg :
      return "Escape";
    case AlertingMsg :
      return "Alerting";
    case CallProceedingMsg :
      return "CallProceeding";
    case ProgressMsg :
      return "Progress";
    case SetupMsg :
      return "Setup";
    case ConnectMsg :
      return "Connect";
    case SetupAckMsg :
      return "SetupAck";
    case ConnectAckMsg :
      return "ConnectAck";
    case ReleaseCompleteMsg :
      return "ReleaseComplete";
    case FacilityMsg :
      return "Facility";
    case NotifyMsg :
      return "Notify";
    case StatusEnquiryMsg :
      return "StatusEnquiry";
    case InformationMsg :
      return "Information";    
    case StatusMsg :
      return "Status";
    default :
      break;
  }
  return ssprintf("<%u>", messageType);
}


unsigned Q931::GenerateCallReference()
{
    static unsigned LastCallReference;
    static SMutex mutex;
    SWaitAndSignal wait(mutex);

    if (LastCallReference == 0)
    {
        SRandom::Seed();
        LastCallReference = SRandom::GetDword();
    }
    else
    {
        LastCallReference++;
    }

    LastCallReference &= 0x7fff;

    if (LastCallReference == 0)
    {
        LastCallReference = 1;
    }

    return LastCallReference;
}


bool Q931::HasIE(InformationElementCodes ie) const
{
    if (informationElements.find(ie) != informationElements.end())
        return true;
    else
        return false;
}


SByteVector Q931::GetIE(InformationElementCodes ie) const
{
  if (informationElements.find(ie) != informationElements.end())
    return *informationElements[ie];

  return SByteVector();
}


void Q931::SetIE(InformationElementCodes ie, const SByteVector & userData)
{
    informationElements.insert(pair<const int, SByteVector *>(ie, new SByteVector(userData)));
}


void Q931::SetProgressIndicator(const Q931::ProgressIndicator & ProgressInd)
{
    if (ProgressInd.bIndicatorIsSet == true)
    {
        SByteVector data(2);
        // Progress indicator as per Q.931 section 4.5.23
        data[0] = 0x80 | (BYTE)(ProgressInd.location & 0x0f); // ITU-T code, location
        data[1] = 0x80 | (BYTE)(ProgressInd.description & 0x7f);
        SetIE(ProgressIndicatorIE, data);
    }
}


bool Q931::GetProgressIndicator(Q931::ProgressIndicator & ProgressInd) const
{
    if (!HasIE(ProgressIndicatorIE))
    {
        ProgressInd.bIndicatorIsSet = false;
        return false;
    }
    SByteVector bytes = GetIE(ProgressIndicatorIE);
    if (bytes.size() != 2)
    {
        ProgressInd.bIndicatorIsSet = false;
        return false;
    }
    ProgressInd.bIndicatorIsSet = true;
    ProgressInd.location = (Q931::ProgressIndicator::Location)(bytes[0] & 0x0f);
    ProgressInd.description = (Q931::ProgressIndicator::ProgressDescription)(bytes[1] & 0x7f);
    return true;
}


SString Q931::GetProgressIndicatorInfo(Q931::ProgressIndicator & ProgressInd) const
{
    SString str;
    if (ProgressInd.bIndicatorIsSet == true)
    {
        switch (ProgressInd.location)
        {
        case ProgressIndicator::e_User:
            str ="location: User, ";
            break;
          case ProgressIndicator::e_PrivateNetworkServingLocalUser:
            str ="location: Private network serving the local user, ";
            break;
          case ProgressIndicator::e_PublicNetworkServingLocalUser:
            str ="location: Public network serving the local user, ";
            break;
          case ProgressIndicator::e_TransitNetwork:
            str ="location: Transit network, ";
            break;
          case ProgressIndicator::e_PublicNetworkServingRemoteUser:
            str ="location: Public network serving the remote user, ";
            break;
          case ProgressIndicator::e_PrivateNetworkServingRemoteUser:
            str ="location: Private network serving the remote user, ";
            break;
          case ProgressIndicator::e_NetworkBeyondInterworkingPoint:
            str ="location: Network beyond the interworking point, ";
            break;
          default:
            str = ssprintf("location: <unknown> (%u), ", ProgressInd.location);
            break;
        };

        switch (ProgressInd.description)
        {
          case ProgressIndicator::e_CallIsNotEndToEndIsdn:
            str += "description: Call is not end-to-end ISDN";
            break;
          case ProgressIndicator::e_DestinationAddressIsNonIsdn:
            str += "description: Destination address is non ISDN";
            break;
          case ProgressIndicator::e_OriginationAddressIsNonIsdn:
            str += "description: Origination address is non ISDN";
            break;
          case ProgressIndicator::e_CallHasReturnedToTheIsdn:
            str += "description: Call has returned to the ISDN";
            break;
          case ProgressIndicator::e_InterworkingHasOccurred:
            str += "description: Interworking has occurred";
            break;
          case ProgressIndicator::e_InBandInformationAvailable:
            str += "description: In-band information or an appropriate pattern is now available";
            break;
          default:
            str += ssprintf("description: <unknown> (%u)", ProgressInd.description);
            break;
        }
    }
    else
    {
        str ="No process indicator present!";
    }
    return str;
}


void Q931::SetNotificationIndicator(const Q931::NotificationIndicator & NotificationInd)
{
    SByteVector data(1);
    // Notification indicator as per Q.931 section 4.5.22
    data[0] = 0x80 | (BYTE)(NotificationInd.description & 0x7f);
    SetIE(NotificationIndicatorIE, data);
}


bool Q931::GetNotificationIndicator(Q931::NotificationIndicator & NotificationInd) const
{
    if (!HasIE(NotificationIndicatorIE))
    {
        return false;
    }
    SByteVector bytes = GetIE(NotificationIndicatorIE);
    if (bytes.size() != 1)
    {
        return false;
    }
    NotificationInd.description = (Q931::NotificationIndicator::NotificationDescription)(bytes[0] & 0x7f);
    return true;
}


SString Q931::GetNotificationIndicatorInfo(Q931::NotificationIndicator & NotificationInd) const
{
    SString str;
    switch (NotificationInd.description)
    {
      // Q.931, table 4-19
      case NotificationIndicator::e_UserSuspended:
        str = "User suspended";
        break;
      case NotificationIndicator::e_UserResumed:
        str = "User resumed";
        break;
      case NotificationIndicator::e_BearerServiceChange:
        str = "Bearer service change";
        break;

      // Swyx proprietary values
      case NotificationIndicator::e_SwyxStopDtmfDetection:
        str = "Stop DTMF detection (proprietary)";
        break;
      case NotificationIndicator::e_SwyxStartDtmfDetection:
        str = "Start DTMF detection (proprietary)";
        break;

      default:
        str = ssprintf("description: <unknown> (%u)", NotificationInd.description);
        break;
    }
    return str;
}


void Q931::SetSignalInfo(SignalInfo value)
{
    SByteVector data(1);
    data[0] = (BYTE)value;
    SetIE(SignalIE, data);
}


Q931::SignalInfo Q931::GetSignalInfo() const
{
    if (!HasIE(SignalIE))
        return SignalErrorInIE;

    SByteVector data = GetIE(SignalIE);
    if (data.empty())
        return SignalErrorInIE;

    return (SignalInfo)data[0];
}


SString Q931::GetSignalInfoName(SignalInfo value) const
{
    switch (value)
    {
        case SignalDialToneOn:
            return "Dial tone on";
        case SignalRingBackToneOn:
            return "Ring back tone on";
        case SignalInterceptToneOn:
            return "Intercept tone on";
        case SignalNetworkCongestionToneOn:
            return "Network congestion tone on";
        case SignalBusyToneOn:
            return "Network congestion tone on";
        case SignalConfirmToneOn:
            return "Confirm tone on";
        case SignalAnswerToneOn:
            return "Answer tone on";
        case SignalCallWaitingTone:
            return "Call waiting tone";
        case SignalOffhookWarningTone:
            return "Off-hook warning tone";
        case SignalPreemptionToneOn:
            return "Preemption tone on";
        case SignalTonesOff:
            return "Preemption tone on";
        case SignalAlertingPattern0:
            return "Alerting on - pattern 0";
        case SignalAlertingPattern1:
            return "Alerting on - pattern 1";
        case SignalAlertingPattern2:
            return "Alerting on - pattern 2";
        case SignalAlertingPattern3:
            return "Alerting on - pattern 3";
        case SignalAlertingPattern4:
            return "Alerting on - pattern 4";
        case SignalAlertingPattern5:
            return "Alerting on - pattern 5";
        case SignalAlertingPattern6:
            return "Alerting on - pattern 6";
        case SignalAlertingPattern7:
            return "Alerting on - pattern 7";
        case SignalAlertingOff:
            return "Alerting off";
        default :
          break;
    }   
    return ssprintf("unknown <%u>", value);
}


void Q931::SetDisplayName(const SString & name)
{
    SByteVector tempVec;
    tempVec.clear();
    for (SString::const_iterator pos = name.begin(); pos != name.end(); ++pos)
    {
        tempVec.push_back((BYTE)(*pos));
    }
    SetIE(DisplayIE, tempVec);
}


SString Q931::GetDisplayName() const
{
  if (!HasIE(Q931::DisplayIE))
    return SString();

  SByteVector display = GetIE(Q931::DisplayIE);
  if (display.empty())
    return SString();

  return SString((const char *)&(*(display.begin())), display.size());
}


void Q931::SetKeypadFacility(const SString & info)
{
    SByteVector bytes;
    std::copy(info.begin(), info.end(), std::back_inserter(bytes));
    SetIE(KeypadFacilityIE, bytes);
}


SString Q931::GetKeypadFacility() const
{
    if (!HasIE(KeypadFacilityIE))
        return SString();

    SByteVector bytes = GetIE(KeypadFacilityIE);
    if (bytes.empty())
        return SString();

    return SString((const char *)&(*(bytes.begin())), bytes.size());
}


void Q931::SetCallingPartyNumber(const SString & number,
                                 unsigned plan,
                                 unsigned type,
                                 int presentation,
                                 int screening)
{
  int len = number.size();
  SByteVector bytes;
  if (presentation == -1 || screening == -1) {
    bytes.resize(len+1);
    bytes[0] = (BYTE)(0x80|((type&7)<<4)|(plan&15));
    std::copy(number.begin(), number.end(), bytes.begin()+1);
  }
  else {
    bytes.resize(len+2);
    bytes[0] = (BYTE)(((type&7)<<4)|(plan&15));
    bytes[1] = (BYTE)(0x80|((presentation&3)<<5)|(screening&3));
    std::copy(number.begin(), number.end(), bytes.begin()+2);
  }
  SetIE(CallingPartyNumberIE, bytes);
}


bool Q931::GetCallingPartyNumber(SString  & number,
                                 unsigned * plan,
                                 unsigned * type,
                                 unsigned * presentation,
                                 unsigned * screening,
                                 unsigned   defPresentation,
                                 unsigned   defScreening) const
{
  number = SString();

  if (!HasIE(CallingPartyNumberIE))
    return false;

  SByteVector bytes = GetIE(CallingPartyNumberIE);
  if (bytes.empty())
    return false;

  if (plan != NULL)
    *plan = bytes[0]&15;

  if (type != NULL)
    *type = (bytes[0]>>4)&7;

  int offset;
  if (bytes[0] & 0x80) {  // Octet 3a not provided, set defaults
    if (presentation != NULL)
      *presentation = defPresentation;

    if (screening != NULL)
      *screening = defScreening;

    offset = 1;
  }
  else {
    if (presentation != NULL)
      *presentation = (bytes[1]>>5)&3;

    if (screening != NULL)
      *screening = bytes[1]&3;

    offset = 2;
  }

  int len = bytes.size()-offset;
  number.assign((const char *)&bytes[offset], len);  

  return !number.empty();
}


void Q931::SetCalledPartyNumber(const SString & number, unsigned plan, unsigned type)
{
  int len = number.size();
  SByteVector bytes(len+1);
  bytes[0] = (BYTE)(0x80|((type&7)<<4)|(plan&15));
  std::copy(number.begin(), number.end(), bytes.begin()+1);
  SetIE(CalledPartyNumberIE, bytes);
}


bool Q931::GetCalledPartyNumber(SString & number, unsigned * plan, unsigned * type) const
{
  number = SString();

  if (!HasIE(CalledPartyNumberIE))
    return false;

  SByteVector bytes = GetIE(CalledPartyNumberIE);
  if (bytes.empty())
      return false;
  
  if (plan != NULL)
      *plan = bytes[0]&15;
  
  if (type != NULL)
      *type = (bytes[0]>>4)&7;
  
  int len = bytes.size()-1;
  number.assign((const char *)&bytes[1], len);
  
  return !number.empty();
}

void Q931::SetCause(const Q931::Cause & cause)
{
    if (cause.cause != Q931::Cause::e_UnknownCause)
    {
        // Cause IE as per Q.850
        SByteVector CauseIEData(2);
//hk: I make it just as needed for Siemens-phones
//        CauseIEData[0] = 0xc0; // No Octet 3a, coding: national standard, location: user

        CauseIEData[0] = 0x80; // this is the way siemens does it 
                               // .... 0000 cause location: user (0)
                               // .00. .... coding standard ITU-T standardized coding (0x00)
                               // 1... .... extension locator
        CauseIEData[1] = 0x80 | (BYTE)(cause.cause & 0x7f);
        SetIE(CauseIE, CauseIEData);
    }
}


bool Q931::GetCause(Cause & cause) const
{
    cause.cause = Q931::Cause::e_UnknownCause;

    if (!HasIE(CauseIE))
        return false;

    SByteVector bytes = GetIE(CauseIE);
    if (bytes.empty())
        return false;

    if (bytes[0] & 0x80)
    {
        // Octet 3a not present
        cause.cause = (Q931::Cause::Causes)(bytes[1] & 0x7f);
    }
    else
    {
        // Octet 3a is present
        cause.cause = (Q931::Cause::Causes)(bytes[2] & 0x7f);
    }
    return true;
}


Tools::SString Q931::Cause::GetCauseString() const
{
    // ETSI 300 102-1, table 4.13
    switch (cause)
    {
        case Q931::Cause::e_UnknownCause:
            return "Unknown cause (0x0)";

        // class 0        
        case Q931::Cause::e_UnallocatedNumber:
            return "Unallocated (unassigned) number";
        case Q931::Cause::e_NoRouteToSpecifiedTransitNetwork:
            return "No route to specified transit network";
        case Q931::Cause::e_NoRouteToDestination:
            return "No route to destination";
        case Q931::Cause::e_ChannelUnacceptable:
            return "Channel unacceptable";
        case Q931::Cause::e_CallAwarded:
            return "Call awarded and being delivered in an established channel";

        // class 1
        case Q931::Cause::e_NormalCallClearing:
            return "Normal call clearing";
        case Q931::Cause::e_UserBusy:
            return "User busy";
        case Q931::Cause::e_NoUserResponding:
            return "No user responding";
        case Q931::Cause::e_NoAnswerFromUser:
            return "No answer from user (user alerted)";
        case Q931::Cause::e_CallRejected:
            return "Call rejected";
        case Q931::Cause::e_NumberChanged:
            return "Number changed";
        case Q931::Cause::e_NonSelectedUserClearing:
            return "Non-selected user clearing";
        case Q931::Cause::e_DestinationOutOfOrder:
            return "Destination out of order";
        case Q931::Cause::e_InvalidNumberFormat:
            return "Invalid number format";
        case Q931::Cause::e_FacilityRejected:
            return "Facility rejected";
        case Q931::Cause::e_ResponsetoStatusEnquiry:
            return "Response to STATUS ENQUIRY";
        case Q931::Cause::e_Normal:
            return "Normal, unspecified";

        // class 2
        case Q931::Cause::e_NoChannelAvailable:
            return "No circuit/channel available";
        case Q931::Cause::e_NetworkOutOfOrder:
            return "Network out of order";
        case Q931::Cause::e_TemporaryFailure:
            return "Temporary failure";
        case Q931::Cause::e_NetworkCongestion:
            return "Network congestion";
        case Q931::Cause::e_AccessInformationDiscarded:
            return "Access information discarded";
        case Q931::Cause::e_RequestedChannelNotAvailable:
            return "Requested circuit/channel not available";
        case Q931::Cause::e_ResourcesUnavailable:
            return "Resources unavailable, unspecified";

        // class 3
        case Q931::Cause::e_QualityOfServiceUnavailable:
            return "Quality of service unavailable";
        case Q931::Cause::e_RequestedFacilityNotSubscribed:
            return "Requested facility not subscribed";
        case Q931::Cause::e_BearerCapabilityNotAuthorized:
            return "Bearer capability not authorized";
        case Q931::Cause::e_BearerCapabilityNotPresentlyAvailable:
            return "Bearer capability not presently available";
        case Q931::Cause::e_ServiceOrOptionNotAvailable:
            return "Service or option not available, unspecified";

        // class 4
        case Q931::Cause::e_BearerCapabilityNotImplemented:
            return "Bearer capability not implemented";
        case Q931::Cause::e_ChannelTypeNotImplemented:
            return "Channel type not implemented";
        case Q931::Cause::e_RequestedFacilityNotImplemented:
            return "Requested facility not implemented";
        case Q931::Cause::e_OnlyRestrictedDigitalInformationBearerCapabilityIsAvailable:
            return "Only restricted digital information bearer capability is available";
        case Q931::Cause::e_ServiceOrOptionNotImplemented:
            return "Service or option not implemented, unspecified";

            // class 5
        case Q931::Cause::e_InvalidCallReferenceValue:
            return "Invalid call reference value";
        case Q931::Cause::e_IdentifiedChannelDoesNotExist:
            return "Identified channel does not exist";
        case Q931::Cause::e_SuspendedCallExists:
            return "A suspended call exists, but this call identity does not";
        case Q931::Cause::e_CallIdentityInUse:
            return "Call identity in use";
        case Q931::Cause::e_NoCallSuspended:
            return "No call suspended";
        case Q931::Cause::e_CallHavingTheRequestedCallIdentityHasBeenCleared:
            return "Call having the requested call identity has been cleared";
        case Q931::Cause::e_IncompatibleDestination:
            return "Incompatible destination";
        case Q931::Cause::e_InvalidTransitNetworkSelection:
            return "Invalid transit network selection";
        case Q931::Cause::e_InvalidMessage:
            return "Invalid message, unspecified";

            // class 6
        case Q931::Cause::e_MandatoryInformationElementIsMissing:
            return "Mandatory information element is missing";
        case Q931::Cause::e_MessageTypeNonExistentOrNotImplemented:
            return "Message type non-existent or not implemented";
        case Q931::Cause::e_MessageNotCompatibleWithCallStateOrNonExistent:
            return "Message not compatible with call state or message type non-existent or not implemented";
        case Q931::Cause::e_InformationElementNonExistentOrNotImplemented:
            return "Information element non-existent or not implemented";
        case Q931::Cause::e_InvalidInformationElementContents:
            return "Invalid information element contents";
        case Q931::Cause::e_MessageNotCompatibleWithCallState:
            return "Message not compatible with call state";
        case Q931::Cause::e_RecoveryOnTimerExpiry:
            return "Recovery on timer expiry";
        case Q931::Cause::e_ProtocolError:
            return "Protocol error, unspecified";

            // class 7
        case Q931::Cause::e_Interworking:
            return "Interworking, unspecified";
        
        default :
          break;
    }   
    return ssprintf("<unknown> %u = 0x%x", cause, cause);
}


void Q931::SetFacilityIE(const Tools::SByteVector & data)
{
    // No check for empty Facility info element because it may e.g. signal
    // call redirection (see H.225.0 chapter '7.2.2.14 Facility')!

    SetIE(FacilityIE, data);    
}


bool Q931::GetFacilityIE(Tools::SByteVector & data) const
{
    if (!HasIE(FacilityIE))
        return false;

    data = GetIE(FacilityIE);
    
    // No check for empty Facility info element because it may e.g. signal
    // call redirection (see H.225.0 chapter '7.2.2.14 Facility')!

    return true;
}


/////////////////////////////////////////////////////////////////////////////
