/*---------------------------------------------------------------------------

   Swyx Communications AG
   Joseph-von-Fraunhofer-Str. 13a
   44227 Dortmund

  ---------------------------------------------------------------------------

  The contents of this file are subject to the Mozilla Public License
  Version 1.0 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Code is Open H323 Library.

  The Initial Developer of the Original Code is Equivalence Pty. Ltd.

  Portions of this code were written with the assistance of funding from
  Vovida Networks, Inc. http://www.vovida.com.

  Portions created by Swyx Communications AG are
  Copyright (C) 2000-2002 Swyx Communications AG.
  All Rights Reserved.

  Contributor(s): ______________________________________.

  ---------------------------------------------------------------------------

     Project: H.323 Library

      Module: RTP
 
    Comments: RTP protocol handler 

  --------------------------------------------------------------------------

   $Archive:: /Server/SwyxSrv.root/H323/source/rtp.cpp                     $
    $Author:: Martin.hueser                                                $
      $Date:: 6.08.04 16:30                                                $
   $Modtime:: 3.08.04 9:45                                                 $
  $Revision:: 27                                                           $
  $Workfile:: rtp.cpp                                                      $
$NoKeywords: $
----------------------------------------------------------------------------*/


#include <stlib.h>
#include "../include/rtp.h" // prevent conflicts with platform SDK
#include "h323.h"
#include "h225.h"
#include "h245.h"

using namespace std;
using namespace Tools;

#define new SNEW

#define UDP_BUFFER_SIZE 32768



/////////////////////////////////////////////////////////////////////////////

RTP_DataFrame::RTP_DataFrame(int sz)
  : SSimpleArray<BYTE>(12+sz)
{
  payloadSize = sz;
  m_pArray[0] = '\x80';
}


void RTP_DataFrame::SetExtension(bool ext)
{
  if (ext)
    m_pArray[0] |= 0x10;
  else
    m_pArray[0] &= 0xef;
}


void RTP_DataFrame::SetMarker(bool m)
{
  if (m)
    m_pArray[1] |= 0x80;
  else
    m_pArray[1] &= 0x7f;
}


void RTP_DataFrame::SetPayloadType(unsigned t)
{
  SAssert(t <= 0x7f, SInvalidParameter);

  m_pArray[1] &= 0x80;
  m_pArray[1] |= t;
}


DWORD RTP_DataFrame::GetContribSource(int idx) const
{
  SAssert(idx < GetContribSrcCount(), SInvalidParameter);
  return ((SUInt32b *)&m_pArray[12])[idx];
}


void RTP_DataFrame::SetContribSource(int idx, DWORD src)
{
  SAssert(idx <= 15, SInvalidParameter);

  if (idx >= GetContribSrcCount()) {
    BYTE * oldPayload = GetPayloadPtr();
    m_pArray[0] &= 0xf0;
    m_pArray[0] |= idx+1;
    SetSize(GetHeaderSize()+payloadSize);
    memmove(GetPayloadPtr(), oldPayload, payloadSize);
  }

  ((SUInt32b *)&m_pArray[12])[idx] = src;
}


bool RTP_DataFrame::SetPayloadSize(int sz)
{
  payloadSize = sz;

  if ( (GetHeaderSize()+payloadSize) > GetSize() )
     SetSize(GetHeaderSize()+payloadSize);

  return true;
}


/////////////////////////////////////////////////////////////////////////////

RTP_ControlFrame::RTP_ControlFrame(int sz)
{
  SetPayloadSize(sz);
  m_pArray[0] = '\x80';
}


void RTP_ControlFrame::SetPayloadType(unsigned t)
{
  SAssert(t < 256, SInvalidParameter);
  m_pArray[1] = (BYTE)t;
}


void RTP_ControlFrame::SetPayloadSize(int sz)
{
  sz = (sz+3)/4;
  SAssert(sz <= 0xffff, SInvalidParameter);

  if ((4*(sz+1)) > GetSize())
    SetSize(4*(sz+1));

  *(SUInt16b *)&m_pArray[2] = (WORD)sz;
}


/////////////////////////////////////////////////////////////////////////////

RTP_Session::RTP_Session(const H323Connection & conn, unsigned id)
  : connection(conn)
{
  SAssert(id >= 0 && id < 256, SInvalidParameter);
  sessionID = (BYTE)id;

  referenceCount = 1;
  shutdownRead = false;
  shutdownWrite = false;

  expectedSequenceNumber = 0;
  lastSequenceNumber = (WORD)rand();

  syncSourceOut = rand()*rand();
  senderReportInterval = 250;  // Number of data packets between reports

  packetsSent = 0;
  octetsSent = 0;
  packetsReceived = 0;
  octetsReceived = 0;
  packetsLost = 0;
  jitterLevel = 0;
}


void RTP_Session::OnSendRasInfo(H225_RTPSession & info)
{
  info.m_sessionId = sessionID;
  info.m_ssrc = syncSourceOut;
}


#if STRACING
static void TraceReceiverReport(const char * prefixMessage,
                                const RTP_ControlFrame::ReceptionReport * reports,
                                int count)
{
  for (int i = 0; i < count; i++)
  {
    STRACE_INFO(g_pH323Module->m_RtpTraceId, prefixMessage << "ReceiverReport " << i << ":"
              " ssrc=" << reports[i].ssrc
           << " fraction=" << (unsigned)reports[i].fraction
           << " lost=" << reports[i].GetLostPackets()
           << " last_seq=" << reports[i].last_seq
           << " jitter=" << reports[i].jitter
           << " lsr=" << reports[i].lsr
           << " dlsr=" << reports[i].dlsr);
  }
}


static void TraceSenderReport(const char * prefixMessage,
                              const RTP_ControlFrame::SenderReport & sender,
                              const RTP_ControlFrame::ReceptionReport * reports,
                              int count)
{
  STRACE_INFO(g_pH323Module->m_RtpTraceId, prefixMessage << "SenderReport: "
              " ssrc=" << sender.ssrc
           << " ntp=" << sender.ntp_sec
           << " rtp=" << sender.rtp_ts
           << " psent=" << sender.psent
           << " osent=" << sender.osent
           << " rr_cnt=" << count);
  TraceReceiverReport(prefixMessage, reports, count);
}
#endif


RTP_Session::SendReceiveStatus RTP_Session::OnSendData(RTP_DataFrame & frame)
{
  frame.SetSequenceNumber(++lastSequenceNumber);
  frame.SetSyncSource(syncSourceOut);

  octetsSent += frame.GetPayloadSize();
  packetsSent++;
  if (packetsSent%senderReportInterval != 0)
    return e_ProcessPacket;

  RTP_ControlFrame report;
  report.SetPayloadType(RTP_ControlFrame::e_SenderReport);
  report.SetPayloadSize(sizeof(RTP_ControlFrame::SenderReport));

  RTP_ControlFrame::SenderReport * sender =
                            (RTP_ControlFrame::SenderReport *)report.GetPayloadPtr();
  sender->ssrc = syncSourceOut;
  sender->ntp_sec = ::time(NULL)+(70*365+8)*24*60*60U;
  sender->ntp_frac = rand();
  sender->rtp_ts = lastSequenceNumber;
  sender->psent = packetsSent;
  sender->osent = octetsSent;

#if STRACING
    TraceSenderReport("Sent", *sender, NULL, 0);
#endif

  WriteControl(report);
  return e_ProcessPacket;
}


RTP_Session::SendReceiveStatus RTP_Session::OnReceiveData(const RTP_DataFrame & frame)
{
    SDEFFUNC("RTP_Session::OnReceiveData");

  // Check that the PDU is the right version
  if (frame.GetVersion() != RTP_DataFrame::ProtocolVersion)
    return e_IgnorePacket; // Non fatal error, just ignore

  // Check for if a control packet rather than data packet.
  if (frame.GetPayloadType() > RTP_DataFrame::MaxPayloadType)
    return e_IgnorePacket; // Non fatal error, just ignore

  // Check packet sequence numbers
  if (expectedSequenceNumber == 0) {
    expectedSequenceNumber = (WORD)(frame.GetSequenceNumber() + 1);
    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() First data:"
        " ver=" << frame.GetVersion()
        << " pt=" << frame.GetPayloadType()
        << " psz=" << frame.GetPayloadSize()
        << " m=" << frame.GetMarker()
        << " x=" << frame.GetExtension()
        << " seq=" << frame.GetSequenceNumber()
        << " ts=" << frame.GetTimestamp()
        << " src=" << frame.GetSyncSource()
        << " ccnt=" << frame.GetContribSrcCount());
  }
  else {
    WORD sequenceNumber = frame.GetSequenceNumber();
    if (sequenceNumber == expectedSequenceNumber)
      expectedSequenceNumber++;
    else if (sequenceNumber < expectedSequenceNumber) {
      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
          "() Out of order packet, received "
          << sequenceNumber << " expected " << expectedSequenceNumber);
      // This implementation simply ignores out of order packets
      return e_IgnorePacket; // Non fatal error, just ignore
    }
    else {
      unsigned dropped = sequenceNumber - expectedSequenceNumber;
      packetsLost += dropped;
      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
          "() " << dropped << " packets dropped.");
      expectedSequenceNumber = (WORD)(sequenceNumber + 1);
    }
  }

  packetsReceived++;
  octetsReceived += frame.GetPayloadSize();

  return e_ProcessPacket;
}


RTP_Session::SendReceiveStatus RTP_Session::OnReceiveControl(const RTP_ControlFrame & frame)
{
    SDEFFUNC("RTP_Session::OnReceiveControl");

  BYTE * payload = frame.GetPayloadPtr();
  switch (frame.GetPayloadType()) {
    case RTP_ControlFrame::e_SenderReport :
      OnRxSenderReport(*(const RTP_ControlFrame::SenderReport *)payload,
                       (const RTP_ControlFrame::ReceptionReport *)(payload+sizeof(RTP_ControlFrame::SenderReport)),
                       frame.GetCount());
      break;

    case RTP_ControlFrame::e_ReceiverReport :
      if (frame.GetPayloadSize() < 4)
      {
        STRACE2_WARN(g_pH323Module->m_RtpTraceId, this, 0,
            "() ReceiverReport packet truncated");
      }
      else
        OnRxReceiverReport(*(const SUInt32b *)payload,
                           (const RTP_ControlFrame::ReceptionReport *)(payload+sizeof(SUInt32b)),
                           frame.GetCount());
      break;

    case RTP_ControlFrame::e_SourceDescription :
      OnRxSourceDescription(*(const RTP_ControlFrame::SourceDescription *)payload);
      break;

    case RTP_ControlFrame::e_Goodbye :
    {
      SString str;
      int count = frame.GetCount()*4;
      if (frame.GetPayloadSize() > count)
        str = SString((const char *)(payload+count+4), *payload);
      OnRxGoodbye((const SUInt32b *)payload, count, str);
      break;
    }

    case RTP_ControlFrame::e_ApplDefined :
    {
      SString str((const char *)(payload+4), 4);
      OnRxApplDefined(str, frame.GetCount(), *(const SUInt32b *)payload,
                      payload+8, frame.GetPayloadSize()-8);
      break;
    }

    default :
      STRACE2_WARN(g_pH323Module->m_RtpTraceId, this, 0,
          "() Unknown control payload type: " << frame.GetPayloadType());
  }

  return e_ProcessPacket;
}


void RTP_Session::OnRxSenderReport(const RTP_ControlFrame::SenderReport & sender,
                                   const RTP_ControlFrame::ReceptionReport * reports,
                                   int count)
{
#if STRACING
  TraceSenderReport("OnRx", sender, reports, count);
#endif
}


void RTP_Session::OnRxReceiverReport(DWORD src,
                                     const RTP_ControlFrame::ReceptionReport * reports,
                                     int count)
{
    SDEFFUNC("RTP_Session::OnRxReceiverReport");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() ssrc=" << src << " rr_cnt=" << count);
#if STRACING
  TraceReceiverReport("OnRx", reports, count);
#endif
}


void RTP_Session::OnRxSourceDescription(const RTP_ControlFrame::SourceDescription & description)
{
    SDEFFUNC("RTP_Session::OnRxSourceDescriptio");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() " << description.src);
}


void RTP_Session::OnRxGoodbye(const SUInt32b * /*ssrc*/, int count,
                              const SString & reason)
{
    SDEFFUNC("RTP_Session::OnRxGoodbye");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() \"" << reason << "\" src_cnt=" << count);
}


void RTP_Session::OnRxApplDefined(const SString & type,
                                  unsigned subtype, DWORD src,
                                  const BYTE * /*data*/, int size)
{
    SDEFFUNC("RTP_Session::OnRxApplDefined");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() \"" << type << "\"-" << subtype
        << " " << src << " [" << size << ']');
}


/////////////////////////////////////////////////////////////////////////////

static void SetMinBufferSize(SUDPSocket & sock, int buftype)
{
  int sz = 0;
  if (sock.GetOption(buftype, sz)) {
    if (sz >= UDP_BUFFER_SIZE)
      return;
  }

  if (!sock.SetOption(buftype, UDP_BUFFER_SIZE))
  {
    STRACE_ERROR(g_pH323Module->m_RtpTraceId, "SetOption(" << buftype << ") failed: " << sock.GetErrorText());
  }
}


RTP_UDP::RTP_UDP(const H323Connection & conn, unsigned id, WORD portBase)
  : RTP_Session(conn, id), remoteDataPort(0), remoteControlPort(0)
{
    SDEFFUNC("RTP_UDP::RTP_UDP");

    if (!g_pH323Module->m_bUsingSwtcl)
    {
        const H323Transport & transport = connection.GetControlChannel();
        SIPSocket::Address localAddress;
        WORD dummy;
        transport.GetLocalAddress().GetIpAndPort(localAddress, dummy);

        // Do not open RTP sockets if using SwTCL.
        WORD nextPort = (WORD)(portBase&0xfffe);
        while (!dataSocket.Listen(localAddress, 1, nextPort) ||
               !controlSocket.Listen(localAddress, 1, (WORD)(nextPort+1))) {
          dataSocket.Close();
          controlSocket.Close();
          nextPort += 2;
        }
    }

#ifndef __BEOS__
  SetMinBufferSize(dataSocket,    SO_RCVBUF);
  SetMinBufferSize(dataSocket,    SO_SNDBUF);
  SetMinBufferSize(controlSocket, SO_RCVBUF);
  SetMinBufferSize(controlSocket, SO_SNDBUF);
#endif

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() UDP Session " << id
        << " created: data=" << dataSocket.GetPort()
        << " control=" << controlSocket.GetPort()
        << " ssrc=" << syncSourceOut);
}


RTP_UDP::~RTP_UDP()
{
  Close(true);
  Close(false);
}


void RTP_UDP::Close(bool reading)
{
    SDEFFUNC("RTP_UDP::Close");

  if (reading) {
    if (!shutdownRead) {
      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
          "() Shutdown read");
      shutdownRead = true;
      SIPSocket::Address addr;
      controlSocket.GetLocalAddress(addr);
      dataSocket.WriteTo("", 1, addr, controlSocket.GetPort());
    }
  }
  else
    shutdownWrite = true;

  if (shutdownRead && shutdownWrite) {
    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "() Closing sockets "
        << dataSocket.GetHandle() << ' ' << controlSocket.GetHandle());
    dataSocket.Close();
    controlSocket.Close();
  }
}


bool RTP_UDP::OnSendingPDU(H245_H2250LogicalChannelParameters & param,
                           bool includeMediaChannel,
                           H323Codec::CodecType type) const
{
    SDEFFUNC("RTP_UDP::OnSendingPDU");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "(," << includeMediaChannel << "," << type << ")");

  param.m_sessionID = GetSessionID();

  param.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaGuaranteedDelivery);
  param.m_mediaGuaranteedDelivery = false;

  param.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_silenceSuppression);
  param.m_silenceSuppression = false;

  const H323Transport & transport = connection.GetControlChannel();

  if (g_pH323Module->m_bUsingSwtcl)
  {
      Tools::SIpAddress SourceMediaTSAP        = connection.GetSourceMediaTSAP(type);
      Tools::SIpAddress SourceMediaControlTSAP = connection.GetSourceMediaControlTSAP(type);

      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
          "() (Using SwTCL) RTP: " << SourceMediaTSAP << ", RTCP: " << SourceMediaControlTSAP);

      // unicast must have mediaControlChannel
      param.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel);
      transport.SetUpTransportPDU(param.m_mediaControlChannel, SourceMediaControlTSAP);

      if (includeMediaChannel) {
        // set mediaChannel
        param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);
        transport.SetUpTransportPDU(param.m_mediaChannel, SourceMediaTSAP);
      }
  }
  else
  {
      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
          "() Not using SwTCL"); 

      // unicast must have mediaControlChannel
      param.IncludeOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel);
      transport.SetUpTransportPDU(param.m_mediaControlChannel, controlSocket.GetPort());

      if (includeMediaChannel) {
        // set mediaChannel
        param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);
        transport.SetUpTransportPDU(param.m_mediaChannel, dataSocket.GetPort());
      }
  }
  return true;
}


void RTP_UDP::OnSendingAckPDU(H245_H2250LogicalChannelAckParameters & param,
                              H323Codec::CodecType type) const
{
    SDEFFUNC("RTP_UDP::OnSendingAckPDU");

    STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
        "(" << "," << type << ")");

  const H323Transport & transport = connection.GetControlChannel();

  if (g_pH323Module->m_bUsingSwtcl)
  {
      Tools::SIpAddress SourceMediaTSAP        = connection.GetSourceMediaTSAP(type);
      Tools::SIpAddress SourceMediaControlTSAP = connection.GetSourceMediaControlTSAP(type);

      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
          "() (Using SwTCL) RTP: " << SourceMediaTSAP << ", RTCP: " << SourceMediaControlTSAP);

      // set mediaControlChannel
      param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel);
      transport.SetUpTransportPDU(param.m_mediaControlChannel, SourceMediaControlTSAP);

      // set mediaChannel
      param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);
      transport.SetUpTransportPDU(param.m_mediaChannel, SourceMediaTSAP);
  }
  else
  {
      STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
          "() Not using SwTCL"); 

      // set mediaControlChannel
      param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel);
      transport.SetUpTransportPDU(param.m_mediaControlChannel, controlSocket.GetPort());

      // set mediaChannel
      param.IncludeOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel);
      transport.SetUpTransportPDU(param.m_mediaChannel, dataSocket.GetPort());
  }
}


static bool ExtractTransport(const H245_TransportAddress & pdu,
                             SIPSocket::Address & ipnum, WORD & port,
                             unsigned & errorCode)
{
  if (pdu.GetTag() != H245_TransportAddress::e_unicastAddress) {
    STRACE_ERROR(g_pH323Module->m_RtpTraceId, "UDP\tOnly unicast supported at this time");
    errorCode = H245_OpenLogicalChannelReject_cause::e_multicastChannelNotAllowed;
    return false;
  }

  const H245_UnicastAddress & unicast = pdu;
  if (unicast.GetTag() != H245_UnicastAddress::e_iPAddress) {
    STRACE_ERROR(g_pH323Module->m_RtpTraceId, "UDP\tLogic error, UDP must be IP unicast");
    errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
    return false;
  }

  const H245_UnicastAddress_iPAddress & addr = unicast;

  ipnum = SIPSocket::Address((DWORD)addr.m_network);

  port = (WORD)(unsigned)addr.m_tsapIdentifier;
  return true;
}


bool RTP_UDP::OnReceivedPDU(const H245_H2250LogicalChannelParameters & param,
                            unsigned & errorCode)
{
    SDEFFUNC("RTP_UDP::OnReceivedPDU");

    STRACE2_INFO3(g_pH323Module->m_RtpTraceId, this, 0,
        "()");

  if (param.m_sessionID != GetSessionID()) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() UDP\tOpen for invalid session: " << param.m_sessionID);
    errorCode = H245_OpenLogicalChannelReject_cause::e_invalidSessionID;
    return false;
  }

  if (!param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel)) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() UDP\tNo mediaControlChannel specified");
    errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
    return false;
  }

  if (!ExtractTransport(param.m_mediaControlChannel, remoteAddress, remoteControlPort, errorCode))
    return false;

  if (param.HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaChannel))
    return ExtractTransport(param.m_mediaChannel, remoteAddress, remoteDataPort, errorCode);

  if (!g_pH323Module->m_bUsingSwtcl)
  {
    // Do not choose RTP port based on RTCP port-1 if using SwTCL.
    if (remoteDataPort == 0)
      remoteDataPort = (WORD)(remoteControlPort - 1);
  }
  return true;
}


bool RTP_UDP::OnReceivedAckPDU(const H245_H2250LogicalChannelAckParameters & param)
{
    SDEFFUNC("RTP_UDP::OnReceivedAckPDU");

    STRACE2_INFO3(g_pH323Module->m_RtpTraceId, this, 0,
        "()");

  if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_sessionID)) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() UDP\tNo session specified");
    //return false;  // sessionID is optional !
  }
  else
  {
      if (param.m_sessionID != GetSessionID())
      {
          STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
              "() UDP\tAck for invalid session: " << param.m_sessionID);
      }
  }

  if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaControlChannel)) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() UDP\tNo mediaControlChannel specified");
    return false;
  }

  unsigned errorCode;
  if (!ExtractTransport(param.m_mediaControlChannel, remoteAddress, remoteControlPort, errorCode))
    return false;

  if (!param.HasOptionalField(H245_H2250LogicalChannelAckParameters::e_mediaChannel)) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() UDP\tNo mediaChannel specified");
    return false;
  }

  if (!ExtractTransport(param.m_mediaChannel, remoteAddress, remoteDataPort, errorCode))
    return false;

  return true;
}


void RTP_UDP::OnSendRasInfo(H225_RTPSession & info)
{
  RTP_Session::OnSendRasInfo(info);

  const H323Transport & transport = connection.GetControlChannel();

  transport.SetUpTransportPDU(info.m_rtpAddress.m_recvAddress, dataSocket.GetPort() ? true : false);
  H323TransportAddress(remoteAddress, remoteDataPort).SetPDU(info.m_rtpAddress.m_sendAddress);

  transport.SetUpTransportPDU(info.m_rtcpAddress.m_recvAddress, controlSocket.GetPort() ? true : false);
  H323TransportAddress(remoteAddress, remoteDataPort).SetPDU(info.m_rtcpAddress.m_sendAddress);
}


bool RTP_UDP::ReadData(RTP_DataFrame & frame)
{
    SDEFFUNC("RTP_UDP::ReadData");

  for (;;) {
    int selectStatus = SSocket::Select(dataSocket, controlSocket);

    if (shutdownRead)
      return false;

    switch (selectStatus) {
      case -2 :
        if (ReadControlPDU() == e_AbortTransport)
          return false;
        break;

      case -3 :
        if (ReadControlPDU() == e_AbortTransport)
          return false;
        // Then do -1 case

      case -1 :
        switch (ReadDataPDU(frame)) {
          case e_ProcessPacket :
            return true;
          case e_IgnorePacket :
            break;
          case e_AbortTransport :
            return false;
        }
        break;

      case SSocket::Interrupted:
        STRACE2_INFO(g_pH323Module->m_RtpTraceId, this, 0,
            "() RTP\tInterrupted");
        return false;

      default :
        STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
            "() Select error: " << SChannel::GetErrorText((SChannel::Errors)selectStatus));
        return false;
    }
  }
}


RTP_Session::SendReceiveStatus RTP_UDP::ReadDataPDU(RTP_DataFrame & frame)
{
    SDEFFUNC("RTP_UDP::ReadDataPDU");

  SIPSocket::Address addr;
  WORD port;

  if (!dataSocket.ReadFrom(frame.GetPointer(), frame.GetSize(), addr, port)) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() Data read error: " << dataSocket.GetErrorText());
    return e_AbortTransport;
  }

  if (addr != remoteAddress) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() Data PDU not from correct host or port: is "
        << addr << ':' << port << ", should be "
        << remoteAddress << ':' << remoteDataPort);
    return e_IgnorePacket;
  }

  // Check recieved PDU is big enough
  int pduSize = dataSocket.GetLastReadCount();
  if (pduSize < RTP_DataFrame::MinHeaderSize || pduSize < frame.GetHeaderSize()) {
    STRACE2_WARN(g_pH323Module->m_RtpTraceId, this, 0,
        "() Received data packet too small: " << pduSize << " bytes");
    return e_IgnorePacket;
  }

  frame.SetPayloadSize(pduSize - frame.GetHeaderSize());
  return OnReceiveData(frame);
}


RTP_Session::SendReceiveStatus RTP_UDP::ReadControlPDU()
{
    SDEFFUNC("RTP_UDP::ReadControlPDU");

  RTP_ControlFrame frame(2048);
  SIPSocket::Address addr;
  WORD port;

  if (!controlSocket.ReadFrom(frame.GetPointer(), frame.GetSize(), addr, port)) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() Control read error: " << controlSocket.GetErrorText());
    return e_AbortTransport;
  }

  if (addr != remoteAddress) {
    STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
        "() Control PDU not from correct host or port: is "
        << addr << ':' << port << ", should be "
        << remoteAddress << ':' << remoteControlPort);
    return e_IgnorePacket;
  }

  int pduSize = controlSocket.GetLastReadCount();
  if (pduSize < 4 || pduSize < 4+frame.GetPayloadSize()) {
    STRACE2_WARN(g_pH323Module->m_RtpTraceId, this, 0,
        "() Received control packet too small: " << pduSize << " bytes");
    return e_IgnorePacket;
  }

  return OnReceiveControl(frame);
}


bool RTP_UDP::WriteData(RTP_DataFrame & frame)
{
    SDEFFUNC("RTP_UDP::WriteData");

  if (shutdownWrite)
    return false;

  switch (OnSendData(frame)) {
    case e_ProcessPacket :
      break;
    case e_IgnorePacket :
      return true;
    case e_AbortTransport :
      return false;
  }

  if (dataSocket.WriteTo(frame.GetPointer(),
                         frame.GetHeaderSize()+frame.GetPayloadSize(),
                         remoteAddress, remoteDataPort))
    return true;

  STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
      "() Write error on data port: " << dataSocket.GetErrorText());
  return false;
}


bool RTP_UDP::WriteControl(RTP_ControlFrame & frame)
{
    SDEFFUNC("RTP_UDP::WriteControl");

  if (controlSocket.WriteTo(frame.GetPointer(), frame.GetPayloadSize()+4,
                            remoteAddress, remoteControlPort))
    return true;

  STRACE2_ERROR(g_pH323Module->m_RtpTraceId, this, 0,
      "() Write error on control port: " << controlSocket.GetErrorText());
  return false;
}


Tools::SIpAddress RTP_UDP::GetRemoteRtpTsap() const
{
    Tools::SIpAddress RtpTsap;

    RtpTsap.m_Address = remoteAddress;
    RtpTsap.m_Port    = remoteDataPort;

    return RtpTsap;
}


Tools::SIpAddress RTP_UDP::GetRemoteRtcpTsap() const
{
    Tools::SIpAddress RtcpTsap;

    RtcpTsap.m_Address = remoteAddress;
    RtcpTsap.m_Port    = remoteControlPort;

    return RtcpTsap;
}

/////////////////////////////////////////////////////////////////////////////
